#!/opt/edge/venv/bin/python3.12

from edge import eventlog
from edge.support import cgi
from edge.webserver.sysconf.api_iit_ca import IitCaApi


class _HandlerFactory:
    """
    GET      /api/v2/sys/conf/iit-ca   /<uiref>
    PUT      /api/v2/sys/conf/iit-ca   /<uiref>
    """

    @classmethod
    def create(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        method_to_handlers_map: cgi.CgiHandlerFactoryDict = {
            'GET': [
                cls._iit_ca__get_uiref,
                cls._iit_ca__get_uiref_factory_limits,
                cls._iit_ca__get_uiref_factory
            ],
            'PUT': [
                cls._iit_ca__put_uiref
            ],
            'POST': [
                cls._iit_ca__post_units_uiref
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _iit_ca__get_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            uiconf_pub = IitCaApi.load_service(request.path_info[0])
            return cgi.CgiResponse(body=uiconf_pub)

        return handler

    @classmethod
    def _iit_ca__get_uiref_factory_limits(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        if request.path_info[1] != 'factory':
            raise cgi.CgiError404

        roles = []
        for name_value in request.query_params:
            name, value = name_value.split('=', 1)
            if name == 'role':
                roles.append(value)

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            reply = [
                {
                    "type": "iit-ca~end-user",
                    "role": "unit"
                }
            ]
            return cgi.CgiResponse(body=reply)

        return handler

    @classmethod
    def _iit_ca__get_uiref_factory(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 3:
            raise cgi.CgiError404

        if request.path_info[1] != 'factory':
            raise cgi.CgiError404

        if request.path_info[2] != 'iit-ca~end-user':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            factory_limits = IitCaApi.load_service_factory(request.path_info[0])
            return cgi.CgiResponse(body=factory_limits)

        return handler

    @classmethod
    def _iit_ca__put_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            new_uiconf = IitCaApi.save_service(request.path_info[0], request.body_jnode)
            return cgi.CgiResponse(body=new_uiconf)

        return handler

    @classmethod
    def _iit_ca__post_units_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        if request.path_info[1] != 'units':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            reply = IitCaApi.make_end_user(request.path_info[0], request.body_jnode)
            return cgi.CgiResponse(body=reply)

        return handler


if __name__ == '__main__':
    eventlog.open_log('cgi-iit-ca', 'debug')
    try:
        cgi.cgi_handle(_HandlerFactory.create)
    finally:
        eventlog.close_log()
