#!/opt/edge/venv/bin/python3.12

from edge import eventlog
from edge.support import cgi
from edge.webserver.sysconf.api_location import LocationApi


class _HandlerFactory:
    """
    GET   /api/v2/sys/conf/location    /<uiref>/rotation
    GET   /api/v2/sys/conf/location    /<uiref>
    PUT   /api/v2/sys/conf/location    /<uiref>
    """

    @classmethod
    def create(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        method_to_handlers_map: cgi.CgiHandlerFactoryDict = {
            'GET': [
                cls._location__get_uiref,
                cls._location__get_uiref_rotation
            ],
            'PUT': [
                cls._location__put_uiref
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    @classmethod
    def _location__get_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            uiconf_pub = LocationApi.load_unit(request.path_info[0])
            return cgi.CgiResponse(body=uiconf_pub)

        return handler

    @classmethod
    def _location__put_uiref(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            uiconf_pub = LocationApi.save_unit(request.path_info[0], request.body_jnode)
            return cgi.CgiResponse(body=uiconf_pub)

        return handler

    @classmethod
    def _location__get_uiref_rotation(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[1] != 'rotation':
            raise cgi.CgiError404

        request.assert_query_present()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            cam_rotation = LocationApi.estimate_rotation(request.path_info[0], request.query_params)
            return cgi.CgiResponse.create200(cam_rotation, envelope=False)

        return handler


if __name__ == '__main__':
    eventlog.open_log('cgi-location', 'debug')
    try:
        cgi.cgi_handle(_HandlerFactory.create)
    finally:
        eventlog.close_log()
