from typing import Final, Optional

from PIL import ImageFont, Image, ImageDraw

from edge import eventlog
from edge.image.helpers import ImHelpers, LineDrawer, MarkingsDrawer
from edge.image.improc import ImProc
from edge.image.profiles import FONT_PATH
from edge.image.types.footer import ImFooterVars
from edge.image.types.builder import SceneImConf
from edge.services.types.cam import ICamService
from edge.types.blob import MemBlob, JsonBlobMeta
from edge.types.image import ImRgba, ImTrans
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF, RectI
from edge.types.track import ITrack
from speedcontrol.events.position import Position


def _add_margin(rect: RectF, margin: float = 5) -> RectF:
    return RectF(rect.fl - margin, rect.ft - margin, rect.fr + margin, rect.fb + margin)


class SceneImProfile:
    def __init__(self, conf: SceneImConf, cam_service: Optional[ICamService] = None):
        self._conf: Final[SceneImConf] = conf
        self._rgba_scaled_: Final[ImRgba] = ImRgba()
        self._cam_service: Optional[ICamService] = cam_service

        self._font = ImageFont.truetype(FONT_PATH, conf.footer.font_size, encoding='unic')

    @property
    def dh(self) -> int:
        return self._conf.footer.height

    def _preprocess(self,
                    scene_dh_data: ImRgba,
                    vehicle_bbox: Optional[RectF],
                    improc: Optional[ImProc]) -> tuple[ImRgba, ImTrans | None]:

        if scene_dh_data.w != self._conf.width:
            target_w: float | int = min(self._conf.width, scene_dh_data.w)
            target_h: float = target_w / scene_dh_data.w * scene_dh_data.h
            scaled_im: Image.Image = ImHelpers.mk_resize_image(scene_dh_data, target_w, target_h)

            dh_data: ImRgba = self._rgba_scaled_
            dh_data.reset(scaled_im.size[0], scaled_im.size[1], self._conf.footer.height, scene_dh_data.ts)
            dh_trans = ImHelpers.mk_resize_trans(scene_dh_data, target_w, target_h)

            if improc is None:
                dh_data.with_image(lambda im: im.paste(scaled_im, (0, 0)))

            elif vehicle_bbox is None:
                improc.improve_contrast_rgba(scaled_im, dh_data)

            else:
                improc.improve_contrast_rgba_roi(scaled_im, dh_data, dh_trans.map_rectf(vehicle_bbox))
                dh_data.with_image(lambda im: ImHelpers.draw_rect(im, vehicle_bbox, dh_trans))

        else:
            dh_data = scene_dh_data
            dh_trans = None

            if improc is not None:
                improc.improve_contrast_rgba_inplace(dh_data)

        return dh_data, dh_trans
    def draw_vehicle_line(self, im: Image.Image, vehicle_pos: Position, source_w: int, source_h: int) -> None:
        im_draw = ImageDraw.Draw(im)
        im_draw_w, im_draw_h = im.size

        w_ratio =  im_draw_w / source_w

        p0_x, p0_y = self._cam_service.mk_image_point(vehicle_pos.left / 100.0 - 1, 0, vehicle_pos.forward / 100.0)
        p1_x, p1_y = self._cam_service.mk_image_point(vehicle_pos.left / 100.0  + 1, 0, vehicle_pos.forward / 100.0)

        p0_x = int(p0_x * w_ratio)
        p0_y = int(p0_y * w_ratio)
        p1_x = int(p1_x * w_ratio)
        p1_y = int(p1_y * w_ratio)

        im_draw.line((p0_x, p0_y, p1_x, p1_y), fill='cyan', width=2)

    def process(self,
                scene_dh_data: ImRgba,
                plate_bbox: RectF,
                vehicle_bbox: Optional[RectF],
                pixelate_bboxes: Optional[list[RectF]],
                draw_bboxes: Optional[list[RectF]],
                footer_vars: ImFooterVars,
                track: Optional[ITrack],
                plate_quad: QuadI,
                improc: Optional[ImProc],
                vehicle_pos: Optional[Position] = None) -> list[MemBlob]:

        dh_data, dh_trans = self._preprocess(scene_dh_data, vehicle_bbox, improc)

        # render footer
        footer_conf = self._conf.footer.eval(footer_vars)
        dh_data.with_footer_image(
            lambda im: ImHelpers.draw_footer(im, footer_conf, self._font, self._conf.footer.font_size))

        # render annotations
        def handler(im: Image.Image) -> bytes:
            if self._cam_service is not None:
                im_draw = ImageDraw.Draw(im)
                im_draw_w, im_draw_h = im.size

                line_drawer = LineDrawer(im_draw, im_draw_w, im_draw_h, scene_dh_data.w, scene_dh_data.h)
                markings_drawer = MarkingsDrawer(self._cam_service, self._conf.markings,
                                                 scene_dh_data.w, scene_dh_data.h)

                markings_drawer.draw_markings(line_drawer)
                if self._conf.markings.vehilce:
                    self.draw_vehicle_line(im, vehicle_pos, scene_dh_data.w, scene_dh_data.h)

            if self._conf.anno.plate:
                ImHelpers.draw_poly(im, plate_quad, rect_trans=dh_trans)

            if self._conf.anno.vehicle and vehicle_bbox:
                ImHelpers.draw_rect(im, vehicle_bbox, dh_trans, line_w=2)

            skip_zones: list[RectF] = []

            for pixelate_bbox in pixelate_bboxes or []:
                if dh_trans:
                    pixelate_bbox = dh_trans.map_rectf(pixelate_bbox)
                else:
                    pixelate_bbox = pixelate_bbox

                ImHelpers.pixelate_roi(im, RectI(*pixelate_bbox.ltrb))
                ImHelpers.draw_rect(im, pixelate_bbox, color='yellow')
                skip_zones.append(_add_margin(pixelate_bbox, 5))

            for draw_bbox in draw_bboxes or []:
                if dh_trans:
                    draw_bbox = dh_trans.map_rectf(draw_bbox)
                else:
                    draw_bbox = draw_bbox

                ImHelpers.draw_rect(im, draw_bbox, None)
                skip_zones.append(_add_margin(draw_bbox, 5))

            if self._conf.anno.track and track:
                if dh_trans:
                    trans_plate_bbox = dh_trans.map_rectf(plate_bbox)
                    trans_plate_trace = dh_trans.map_track(track.vrp_center_trace)
                else:
                    trans_plate_bbox = plate_bbox
                    trans_plate_trace = track.vrp_center_trace

                skip_zones.append(_add_margin(trans_plate_bbox, 0.5 * trans_plate_bbox.fh))

                try:
                    ImHelpers.draw_track(im, trans_plate_trace, skip_zones)
                except ValueError as x:
                    eventlog.log_w(f'fail: draw_track <{footer_vars.get_value("plate")}> ({x})')

            return ImHelpers.mk_jpeg(im)

        dh_data.with_header_image(handler)

        blob_data = dh_data.with_image(lambda im: ImHelpers.mk_jpeg(im))
        blob_meta = JsonBlobMeta({
            'trans': ImTrans().serialize(),
            'size': [dh_data.w, dh_data.h + dh_data.dh, len(blob_data)]
        })

        return [MemBlob('scene', scene_dh_data.ts, blob_data, blob_meta)]
