from typing import Final, List, Optional

from edge.image.helpers import ImHelpers
from edge.types.blob import MemBlob
from edge.types.image import ImRgba
from edge.image.types.builder import ThumbImConf
from edge.support.bbox import mk_mmr_bbox
from edge.types.shape.rect import RectF


class ThumbImProfile:
    def __init__(self, conf: ThumbImConf):
        self._w: Final[float] = float(conf.width)
        self._h: Final[float] = float(conf.height)
        self._wh_ratio: Final[float] = self._w / self._h

    def process(self, scene_dh_data: ImRgba, plate_bbox: RectF, vehicle_bbox: Optional[RectF]) -> List[MemBlob]:
        min_visible_factor = 0.0  # assume that all mmr bboxes are suitable
        try:
            crop_rect = mk_mmr_bbox(plate_bbox, scene_dh_data.wh, min_visible_factor)
        except ValueError:
            crop_rect = self._get_stub_bbox(plate_bbox, *scene_dh_data.wh)
            crop_rect.strip_invisible(*scene_dh_data.wh)

        blob_data = ImHelpers.mk_crop_resize_jpeg(
            scene_dh_data, crop_rect, self._w, self._h)
        return [MemBlob('thumb', scene_dh_data.ts, blob_data)]

    def _get_stub_bbox(self, plate_bbox: RectF, scene_iw: int, scene_ih: int) -> RectF:
        # full scene-h
        car_fw = scene_ih * self._wh_ratio
        if car_fw < scene_iw:
            car_center_fl = car_fw / 2.0
            car_center_fl_min = 0.0 + car_center_fl
            car_center_fl_max = scene_iw - car_center_fl

            plate_center_fl = (plate_bbox.fl + plate_bbox.fr) / 2.0

            if plate_center_fl < car_center_fl_min:
                car_fl = car_center_fl_min - car_center_fl
            elif plate_center_fl > car_center_fl_max:
                car_fl = car_center_fl_max - car_center_fl
            else:
                car_fl = plate_center_fl - car_center_fl

            return RectF(car_fl, 0.0, car_fl + car_fw, float(scene_ih))

        # full scene-w
        car_fh: float = scene_iw / self._wh_ratio
        if car_fh < scene_ih:
            car_center_ft: float = car_fh / 2.0
            car_center_ft_min: float = 0.0 + car_center_ft
            car_center_ft_max: float = scene_ih - car_center_ft

            plate_center_ft = (plate_bbox.ft + plate_bbox.fb) / 2.0

            if plate_center_ft < car_center_ft_min:
                car_ft = car_center_ft_min - car_center_ft
            elif plate_center_ft > car_center_ft_max:
                car_ft = car_center_ft_max - car_center_ft
            else:
                car_ft = plate_center_ft - car_center_ft

            return RectF(0.0, car_ft, float(scene_iw), car_ft + car_fh)

        raise NotImplementedError
