import hashlib
import os
import subprocess

from typing import Callable, TypedDict

try:
    from typing import NotRequired
except ImportError:
    from typing_extensions import NotRequired

from edge import eventlog
from edge.support import fs, net
from edge.types.geoposition import Geoposition
from edge.types.origin import Origin
from edge.types.location import Location
from edge.services.types.gnss import IGnssService
from edge.services.types.about import AboutServiceConf, IAboutService
from edge.webserver.sysconf.types.defs import CTRL_UICONF_PATH

SW_DIRPATH = '/opt/edge/share/dpkg/info'


class SwFileInfo(TypedDict):
    path: str
    md5: str


SwPackageAuthor = str
SwPackageName = str
SwPackageVer = str


class SwPackageInfo(TypedDict):
    package: tuple[SwPackageAuthor, SwPackageName, SwPackageVer]
    files: NotRequired[list[SwFileInfo]]


class AboutError(RuntimeError):
    pass


def get_device_hw_vendor_model_serial() -> tuple[str, str, str] | None:
    jnode: dict = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    hw_jnode: list[str] | None = jnode.get('hw')
    if type(hw_jnode) is list and len(hw_jnode) == 3:
        return hw_jnode


def get_device_hw_model() -> str | None:
    about_hw = get_device_hw_vendor_model_serial()
    if about_hw is not None:
        return about_hw[1]


def get_device_hw_serial() -> str | None:
    vendor_model_serial: tuple[str, str, str] | None = get_device_hw_vendor_model_serial()
    if vendor_model_serial is not None:
        return vendor_model_serial[2]


def get_device_sw_name() -> dict | None:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('sw-name')


def get_device_sw_lang() -> list[str] | None:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('sw-lang')


def _attach_files_md5(file_infos: list[SwFileInfo]) -> list[SwFileInfo]:
    """
    :param file_infos:
        *.sw/#/files
    """
    blocksize = 65536

    for file_info in file_infos:
        hasher = hashlib.md5()
        file_path = file_info['path']
        with open(file_path, 'rb') as bin_file:
            bin_data = bin_file.read(blocksize)
            while len(bin_data) > 0:
                hasher.update(bin_data)
                bin_data = bin_file.read(blocksize)

        file_info['md5'] = hasher.hexdigest()

    return file_infos


def _get_device_sw_ver__etc() -> str | None:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    if about_jnode:
        return about_jnode.get('sw-ver')


def _get_device_sw_ver__opt() -> str:
    with open('/opt/edge/semver', 'r') as text_file:
        return text_file.readline()


def get_device_sw_packages(sw_files_dir: str) -> list[SwPackageInfo]:
    """
    :param sw_files_dir:
        /opt/edge/share/dpkg/info/
    """
    sw_version: str = _get_device_sw_ver__etc() or _get_device_sw_ver__opt()
    sw_packages: list[SwPackageInfo] = [{
        'package': ('VIT', 'edge', sw_version)
    }]

    try:
        sw_file_names: list[str] = os.listdir(sw_files_dir)
    except os.error as x:
        eventlog.log_d('fail: os.listdir <{}>'.format(x))
        return sw_packages

    # The default format string is “${binary:Package}\t${Version}\n”.
    try:
        cli_params = ['dpkg-query', '--show'] + [name[:-3] for name in sw_file_names if name.endswith('.sw')]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True, encoding='ascii',
                                    check=True, timeout=10.0)
    except subprocess.TimeoutExpired as x:
        eventlog.log_e('fail: <{}> (timeout:{})'.format(x.cmd, x.timeout))
        raise AboutError

    except subprocess.CalledProcessError as x:
        eventlog.log_e('fail: <{}> (returncode:{})'.format(x.cmd, x.returncode))
        raise AboutError

    for line in cli_result.stdout.split('\n'):
        if not line:
            continue

        deb_name_arch, deb_version = line.split(None, 1)
        deb_name = deb_name_arch.split(':', 1)[0]

        sw_path = fs.join_path(sw_files_dir, deb_name + '.sw')
        sw_jnode = fs.load_json3(sw_path, default=dict())
        sw_jnode_files: list[SwFileInfo] = sw_jnode.get('files', [])

        if sw_jnode_files:
            sw_packages.append({
                'package': ('VIT', deb_name, deb_version),
                'files': _attach_files_md5(sw_jnode_files)
            })
        else:
            sw_packages.append({
                'package': ('VIT', deb_name, deb_version)
            })

    return sw_packages


def get_device_sw_md5() -> str | None:
    about_jnode = fs.load_json3(CTRL_UICONF_PATH, default=dict())
    return about_jnode.get('sw-md5')


class _Service(IAboutService):
    def __init__(self, args: AboutServiceConf, geolocation_service: IGnssService) -> None:
        self._gpsd = geolocation_service

        def get_hw_info() -> tuple[str, str, str]:
            jnode = fs.load_json3(args.server_about_path, default=dict())
            hw_jnode = jnode.get('hw') or list()
            return hw_jnode if len(hw_jnode) == 3 else ['', '', '']

        def mk_origin_factory() -> Callable[[Geoposition | None, str], Origin]:
            jnode = fs.load_json3(args.origin_about_path, default=dict())
            o_jnode = jnode.get('origin') or dict()

            uuid = o_jnode['uuid']
            ctime = o_jnode['time']
            place = o_jnode.get('place') or ''
            name = o_jnode.get('name') or ''

            return lambda geolocation, lane: Origin(
                uuid, ctime, name, Location(place=place, lane=lane), geolocation)

        self._hw_info = get_hw_info()
        self._mk_origin = mk_origin_factory()
        self._sw_info = get_device_sw_packages(SW_DIRPATH)
        self._sw_name = get_device_sw_name() or 'EDGE'
        self._base_url = None

    @property
    def hw_info(self) -> tuple[str, str, str]:
        return self._hw_info

    @property
    def sw_info(self) -> list[dict]:
        return self._sw_info

    @property
    def sw_name(self) -> str:
        return self._sw_name

    @property
    def base_url(self) -> str:
        if self._base_url is None:
            try:
                endpoint = net.get_http_api_endpoint()
                self._base_url = f'{endpoint}/api/v2/origins/event/dist'
            except net.NetError:
                eventlog.log_w('fail: define <base_url>')

        return self._base_url

    def get_origin(self, ts: int, lane: str = '') -> Origin:
        geoposition = self._gpsd.get_geoposition(ts)
        return self._mk_origin(geoposition, lane)

    @property
    def hw_vendor(self) -> str:
        return self._hw_info[0]

    @property
    def hw_model(self) -> str:
        return self._hw_info[1]

    @property
    def hw_serial(self) -> str:
        return self._hw_info[2]


def init_service(args: AboutServiceConf, gpsd_service: IGnssService) -> IAboutService:
    return _Service(args, gpsd_service)


# noinspection PyUnusedLocal
def fini_service(service: IAboutService):
    service.destroy()
