import sortedcontainers
import time
import ujson

from abc import abstractmethod, ABCMeta
from twisted.internet.protocol import ReconnectingClientFactory, connectionDone, Protocol
from twisted.protocols.basic import LineOnlyReceiver
from typing import Final, Callable, Dict, Optional

from astral.location import Location, LocationInfo

from edge import eventlog
from edge.services import tz
from edge.services.types.gnss import GnssServiceConf, IGnssService, SkyValidatorConf, ChangelogConf
from edge.support import isotime
from edge.types.geoposition import Geoposition

_TPV_MODE_NONE = 0
_TPV_MODE_NO_FIX = 1
_TPV_MODE_2D_FIX = 2
_TPV_MODE_3D_FIX = 3


class _TpvStats:
    def __init__(self):
        self.mts: int | None = None

        self.tpv_ts: int | None = None
        self.tpv_lat: float | None = None
        self.tpv_lon: float | None = None
        self.tpv_alt: float | None = None

    def inc(self,
            tpv_ts: int | None,
            tpv_lat: float | None,
            tpv_lon: float | None,
            tpv_alt: float | None) -> None:

        self.mts = int(time.time())

        self.tpv_ts = tpv_ts
        self.tpv_lat = tpv_lat
        self.tpv_lon = tpv_lon
        self.tpv_alt = tpv_alt


class _GpsdStats:
    MODE_0D_FIX = 0
    MODE_1D_FIX = 1
    MODE_2D_FIX = 2
    MODE_3D_FIX = 3

    def __init__(self):
        self.mode: int = 0
        self.mode_mts: int | None = None

        self.changelog_size = int()
        self.tpv = _TpvStats()

    def inc_mode(self, mode):
        if self.mode != mode:
            self.mode = mode
            self.mode_mts = int(time.time())

    def inc_changelog(self, record_count: int) -> None:
        self.changelog_size += record_count


class _Changelog:
    def __init__(self, conf: ChangelogConf) -> None:
        self._conf: Final[ChangelogConf] = conf
        self._records = sortedcontainers.SortedDict[int, Geoposition]()

    def add_record(self, tpv_time_usec: int, geoposition: Geoposition) -> None:
        self._records[tpv_time_usec] = geoposition

        while len(self._records) > self._conf.max_depth_records:
            self._records.popitem(index=0)

        while True:
            max_ts = self._records.peekitem()[0]
            min_ts = self._records.peekitem(0)[0]
            if max_ts - min_ts > self._conf.max_depth_usec:
                self._records.popitem(index=0)
            else:
                break

    def lookup_nearest(self, lookup_ts: int) -> Geoposition | None:
        entry_count = len(self._records)
        if not entry_count:
            return None

        # insert lookup_ts
        bisect_index = self._records.bisect_right(lookup_ts)
        if 0 < bisect_index < entry_count:
            # exact-match with lookup_ts
            less_equal_ts_dist = lookup_ts - self._records.iloc[bisect_index - 1]
            if less_equal_ts_dist == 0:
                return self._records.peekitem(bisect_index - 1)[1]

            # greater_ts is better
            greater_ts_dist = self._records.iloc[bisect_index] - lookup_ts
            if less_equal_ts_dist < greater_ts_dist and less_equal_ts_dist <= self._conf.max_lookup_dist_usec:
                return self._records.peekitem(bisect_index - 1)[1]

            # less_equal_ts is better
            return self._records.peekitem(bisect_index)[1] \
                if greater_ts_dist <= self._conf.max_lookup_dist_usec else None

        # push-back lookup_ts
        if bisect_index == entry_count:
            less_equal_ts_dist = lookup_ts - self._records.iloc[entry_count - 1]
            return self._records.peekitem(bisect_index - 1)[1] \
                if less_equal_ts_dist <= self._conf.max_lookup_dist_usec else None

        # push-front lookup_ts
        greater_ts_dist = self._records.iloc[bisect_index] - lookup_ts
        if greater_ts_dist <= self._conf.max_lookup_dist_usec:
            return self._records.peekitem(bisect_index)[1]
        else:
            eventlog.log_w('fail: target-ts:{}usec, min-ts:{}usec, dt:{}sec'.format(
                lookup_ts, self._records.peekitem(0)[0],
                (self._records.peekitem(0)[0] - lookup_ts) / 1e6))
            return None

    def get_count(self) -> int:
        return len(self._records)


class _IValidator(metaclass=ABCMeta):
    def setup(self, sky: dict) -> None:
        pass

    @abstractmethod
    def validate(self, time_usec: int) -> bool:
        pass


class _TrueValidator(_IValidator):
    def validate(self, time_usec: int) -> bool:
        return True


class _SkyValidator(_IValidator):
    def __init__(self, conf: SkyValidatorConf) -> None:
        self._conf: Final[SkyValidatorConf] = conf
        self._max_valid_time_usec: int = 0

    def setup(self, sky: dict) -> None:
        raw_sky_time: str | None = sky.get('time')
        pdop: float | None = sky.get('pdop')
        if pdop is not None and pdop <= self._conf.max_pdop:
            sky_time_usec = isotime.isostr_to_timestamp_usec(raw_sky_time) if raw_sky_time else int(time.time() * 1e6)
            self._max_valid_time_usec = sky_time_usec + self._conf.timeout_usec
        else:
            self._max_valid_time_usec = 0

    def validate(self, time_usec: int) -> bool:
        valid = time_usec <= self._max_valid_time_usec
        if not valid:
            eventlog.log_w(f'fail: time:{time_usec}, max-valid-time:{self._max_valid_time_usec}')
        return valid


class _GpsdProtocol(LineOnlyReceiver):
    delimiter = b'\n'

    def __init__(self, changelog: _Changelog, stats: _GpsdStats, conf: SkyValidatorConf | None) -> None:
        self._changelog: Final[_Changelog] = changelog
        self._stats: Final[_GpsdStats] = stats

        self._sky_validator: Final[_IValidator] = _SkyValidator(conf) if conf else _TrueValidator()

    def connectionMade(self):
        # noinspection PyBroadException
        try:
            self.sendLine(b'?WATCH={"enable":true,"json":true}')
        except Exception:
            eventlog.log_x()

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        pass

    def lineReceived(self, raw_reply: str) -> None:
        try:
            self._handle(raw_reply)

        except ValueError:
            eventlog.log_x()

    def _handle(self, raw_reply: str) -> None:
        reply_jnode = ujson.loads(raw_reply)
        if reply_jnode['class'] == 'TPV':
            # Time/date stamp in ISO8601 format, UTC.
            # May have a fractional part of up to .001sec precision. May be absent if mode is not 2 or 3.
            raw_tpv_time: str | None = reply_jnode.get('time')
            tpv_time_usec = None
            tpv_time_sec = None
            if raw_tpv_time:
                tpv_time_usec = isotime.isostr_to_timestamp_usec(raw_tpv_time)
                tpv_time_sec = int(tpv_time_usec / 1e6)

            # Latitude in degrees: +/- signifies North/South. Present when mode is 2 or 3.
            tpv_lat = reply_jnode.get('lat')

            # Longitude in degrees: +/- signifies East/West. Present when mode is 2 or 3.
            tpv_lon = reply_jnode.get('lon')

            # Altitude in meters. Present if mode is 3.
            tpv_alt = reply_jnode.get('alt')

            # Speed over ground, meters per second.
            tpv_speed = reply_jnode.get('speed')

            self._stats.tpv.inc(tpv_time_sec, tpv_lat, tpv_lon, tpv_alt)

            # NMEA mode: %d, 0=no mode value yet seen, 1=no fix, 2=2D, 3=3D.
            tpv_mode = reply_jnode['mode']
            self._stats.inc_mode(tpv_mode)

            if tpv_time_usec is not None and tpv_mode in [_TPV_MODE_2D_FIX, _TPV_MODE_3D_FIX]:
                if self._sky_validator.validate(tpv_time_usec):
                    self._changelog.add_record(tpv_time_usec, Geoposition(tpv_lat, tpv_lon, tpv_alt, tpv_speed))
                    self._stats.inc_changelog(self._changelog.get_count())

        elif reply_jnode['class'] == 'SKY':
            self._sky_validator.setup(reply_jnode)


class _GpsdClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, changelog: _Changelog, stats: _GpsdStats, conf: SkyValidatorConf | None) -> None:
        self._changelog: Final[_Changelog] = changelog
        self._stats: Final[_GpsdStats] = stats
        self._conf: Final[Optional[SkyValidatorConf]] = conf

    def buildProtocol(self, addr) -> Protocol:
        self.resetDelay()
        protocol = _GpsdProtocol(self._changelog, self._stats, self._conf)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason) -> None:
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason) -> None:
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


class _GpsdGnssService(IGnssService):
    def __init__(self, twisted_reactor, tz_service: tz.ITzService, conf: GnssServiceConf) -> None:
        self._tz_service = tz_service
        self._azimuth: Final[float | None] = conf.azimuth
        self._stats = _GpsdStats()

        self._changelog = _Changelog(conf.gpsd_changelog)

        gpsd_host, gpsd_port = conf.gpsd_endpoint
        client_factory = _GpsdClientFactory(self._changelog, self._stats, conf.gpsd_sky_validator)
        twisted_reactor.connectTCP(gpsd_host, gpsd_port, client_factory)

    def get_geoposition(self, ts: int) -> Geoposition | None:
        geoposition: Geoposition | None = self._changelog.lookup_nearest(ts)
        if geoposition is not None:
            geoposition.azimuth = self._azimuth

        return geoposition

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        stat_jnode = {
            'last-geoposition': {
                'azimuth': self._azimuth,
            }
        }

        if self._stats.mode_mts is not None:
            stat_jnode['last-status'] = {
                'mtime': [
                    self._tz_service.map_ts_sec_to_isostr(self._stats.mode_mts),
                    int(self._stats.mode_mts)
                ],
                'fix-mode': self._stats.mode
            }

        if self._stats.tpv.mts is not None:
            geo_jnode = {
                'mtime': [
                    self._tz_service.map_ts_sec_to_isostr(self._stats.tpv.mts),
                    int(self._stats.tpv.mts)
                ]
            }

            if self._stats.tpv.tpv_ts is not None:
                geo_jnode['time'] = [
                    self._tz_service.map_ts_sec_to_isostr(self._stats.tpv.tpv_ts),
                    int(self._stats.tpv.tpv_ts)
                ]

            if None not in [self._stats.tpv.tpv_lat, self._stats.tpv.tpv_lon]:
                geo = Geoposition(self._stats.tpv.tpv_lat, self._stats.tpv.tpv_lon)

                location_info = LocationInfo()
                location_info.longitude = geo.longitude
                location_info.latitude = geo.latitude
                location_info.timezone = 'UTC'

                astral_loc = Location(location_info)
                # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
                coordinates = '{}{:.6f}{}{:.6f}'.format(
                    'N' if geo.latitude >= 0 else 'S', abs(geo.latitude),
                    'E' if geo.longitude >= 0 else 'W', abs(geo.longitude))

                geo_jnode['latitude'] = geo.latitude
                geo_jnode['longitude'] = geo.longitude
                geo_jnode['coordinates'] = coordinates
                geo_jnode['sunset'] = self._tz_service.map_datetime_to_isostr(astral_loc.sunset())
                geo_jnode['sunrise'] = self._tz_service.map_datetime_to_isostr(astral_loc.sunrise())

            # add altitude
            if self._stats.tpv.tpv_alt is not None:
                geo_jnode['altitude'] = self._stats.tpv.tpv_alt

            stat_jnode['last-geoposition'] = geo_jnode

        json_handler({
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': stat_jnode
        })


class _StaticGnssService(IGnssService):
    def __init__(self, geoposition: Geoposition | None, tz_service: tz.ITzService) -> None:
        # http://catb.org/gpsd/gpsd_json.html
        self._geoposition = geoposition

        if geoposition:
            location_info = LocationInfo()
            location_info.longitude = geoposition.longitude
            location_info.latitude = geoposition.latitude
            location_info.timezone = 'UTC'

            self._astral_loc = Location(location_info)
        else:
            self._astral_loc = None

        self._tz_service = tz_service

    def get_geoposition(self, ts: int) -> Geoposition | None:
        return self._geoposition

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        if self._geoposition is None:
            return

        # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
        coordinates = '{}{:.6f}{}{:.6f}'.format(
            'N' if self._geoposition.latitude >= 0 else 'S', abs(self._geoposition.latitude),
            'E' if self._geoposition.longitude >= 0 else 'W', abs(self._geoposition.longitude))

        jnode = {
            'type': 'geolocation',
            'name': 'geolocation.0',
            'stat': {
                'last-geoposition': {
                    'mtime': self._tz_service.now_isostr_tssec(),
                    'coordinates': coordinates,
                    'latitude': self._geoposition.latitude,
                    'longitude': self._geoposition.longitude,
                    'sunset': self._tz_service.map_datetime_to_isostr(self._astral_loc.sunset()),
                    'sunrise': self._tz_service.map_datetime_to_isostr(self._astral_loc.sunrise())
                }
            }
        }

        if self._geoposition.altitude is not None:
            jnode['stat']['last-geoposition']['altitude'] = self._geoposition.altitude

        if self._geoposition.azimuth is not None:
            jnode['stat']['last-geoposition']['azimuth'] = self._geoposition.azimuth

        json_handler(jnode)


def init_service(twisted_reactor, conf: GnssServiceConf, tz_service: tz.ITzService) -> IGnssService:
    if conf.provider == conf.PROVIDER_STATIC:
        return _StaticGnssService(conf.static_geoposition, tz_service)
    elif conf.provider == conf.PROVIDER_GPSD:
        return _GpsdGnssService(twisted_reactor, tz_service, conf)
    else:
        eventlog.log_e('fail: unsupported (provider:{})'.format(conf.provider))
        raise NotImplementedError
