from collections.abc import Iterator
from typing import Final

from edge.services.types.lane import LaneServiceConf, ILaneService
from edge.support.mask import Mask, MaskConf
from edge.types.shape.point import PointI, PointF
from edge.types.track import LaneTrackItem


class _LaneService(ILaneService):
    def __init__(self, conf: LaneServiceConf) -> None:
        self._mask_w: Final[int] = conf.mask_w
        self._mask_h: Final[int] = conf.mask_h

        lane_to_mask_map: list[tuple[str, Mask]] = []
        for lane_ref, lane_mask_region in conf.lanes:
            lane_mask_conf = MaskConf(lane_mask_region, (conf.mask_w, conf.mask_h))
            lane_mask = Mask(lane_mask_conf)

            lane_to_mask_map.append((lane_ref, lane_mask))

        self._lane_to_mask_map: Final[list[tuple[str, Mask]]] = lane_to_mask_map

    def map_point_abs(self, abs_point: PointI) -> set[str]:
        """
        :param abs_point:
            Visible coordinates are in range [0, mask_w) and [0, mask_h), px
        """
        lanes: set[str] = set()
        for lane_ref, lane_mask in self._lane_to_mask_map:
            if lane_mask.check_point_abs(abs_point):
                lanes.add(lane_ref)

        return lanes

    def map_point_rel(self, rel_point: PointF) -> set[str]:
        """
        :param rel_point:
            Visible coordinates are in range [0.0, 1.0) and [0.0, 1.0)
        """
        abs_point = PointI(
            rel_point.fl * self._mask_w,
            rel_point.ft * self._mask_h)
        return self.map_point_abs(abs_point)

    def map_points_abs(self, point_provider: Iterator[tuple[int, PointI]]) -> tuple[list[LaneTrackItem], set[str]]:
        lanes: set[str] = set()
        items: list[LaneTrackItem] = []
        for item in point_provider:
            item_ts, item_point = item

            lane_set = self.map_point_abs(item_point)
            lane: str = lane_set.pop() if len(lane_set) == 1 else None
            if lane is not None:
                items.append(LaneTrackItem(item_ts, lane))
                lanes.add(lane)

        return items, lanes

    def map_points_rel(self, point_provider: Iterator[tuple[int, PointF]]) -> tuple[list[LaneTrackItem], set[str]]:
        lanes: set[str] = set()
        items: list[LaneTrackItem] = []
        for item in point_provider:
            item_ts, item_point = item

            lane_set = self.map_point_rel(item_point)
            lane: str = lane_set.pop() if len(lane_set) == 1 else None
            if lane is not None:
                items.append(LaneTrackItem(item_ts, lane))
                lanes.add(lane)

        return items, lanes


class _StubLaneService(ILaneService):
    def map_point_abs(self, abs_point: PointI) -> set[str]:
        return set()

    def map_point_rel(self, rel_point: PointF) -> set[str]:
        return set()

    def map_points_abs(self, point_provider: Iterator[tuple[int, PointI]]) -> tuple[list[LaneTrackItem], set[str]]:
        return [], set()

    def map_points_rel(self, point_provider: Iterator[tuple[int, PointF]]) -> tuple[list[LaneTrackItem], set[str]]:
        return [], set()


def init_service(conf: LaneServiceConf | None) -> ILaneService:
    return _StubLaneService() if conf is None else _LaneService(conf)
