import time
import os

from typing import Final

from edge import eventlog
from edge.services.types.stats import IStatsService, IStatsProvider, StatsServiceConf
from edge.support import fs


class StatsService(IStatsService):
    def __init__(self, twisted_reactor, conf: StatsServiceConf) -> None:
        self._providers: Final[list[IStatsProvider]] = []
        self._conf: Final[StatsServiceConf] = conf

        if conf.interval_sec > 0:
            twisted_reactor.callLater(0, self._tick, twisted_reactor)

    def add_providers(self, providers: list[IStatsProvider]) -> None:
        self._providers.extend(providers)

    def _tick(self, twisted_reactor) -> None:
        # noinspection PyBroadException
        try:
            unixtime_utc: float = time.time()
            self._collect(unixtime_utc)
        except Exception:
            eventlog.log_x()

        twisted_reactor.callLater(self._conf.interval_sec, self._tick, twisted_reactor)

    def _collect(self, unixtime_utc: float) -> None:
        # noinspection PyBroadException
        try:
            units = self._collect_units(unixtime_utc)

            try:
                fs.write_json_file({'units': units}, self._conf.filepath_tmp, True)
                os.rename(self._conf.filepath_tmp, self._conf.filepath)
            except IOError as x:
                eventlog.log_w(f'fail: write <{self._conf.filepath}> ({x})')

        except Exception:
            eventlog.log_x()

    def _collect_units(self, unixtime_utc: float) -> list[dict]:
        units = [
            {
                'type': self._conf.type,
                'name': self._conf.name,
                'stat': {
                    'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(unixtime_utc)), int(round(unixtime_utc))],
                    'tout-sec': self._conf.interval_sec
                }
            }
        ]

        list(map(lambda provider: provider.collect_stats(units.append), self._providers))
        return units


def init_service(twisted_reactor, conf: StatsServiceConf) -> IStatsService:
    return StatsService(twisted_reactor, conf)
