from typing import Optional, Dict

from edge import eventlog
from edge.services.tl.trafficlight import ITrafficLight
from edge.services.tl.trafficlight.tl_type_r import TrafficLightR
from edge.services.tl.trafficlight.tl_type_rag import TrafficLightRag
from edge.services.tl.tl_signal import Signal
from edge.types.exceptions import JnodeError
from edge.types.timespan import TimeSpan, decode__timespans


def _maybe_max(left: Optional[int | float], rigth: Optional[int | float]) -> Optional[int | float]:
    if None not in [left, rigth]:
        return max(left, rigth)

    elif left is not None:
        return left

    elif rigth is not None:
        return rigth


def _maybe_min(left: Optional[int | float], rigth: Optional[int | float]) -> Optional[int | float]:
    if None not in [left, rigth]:
        return min(left, rigth)

    elif left is not None:
        return left

    elif rigth is not None:
        return rigth


def _decode__halt_signal_rag(signal_r: Signal,
                             signal_a: Signal,
                             signal_g: Signal,
                             accuracy_usec: int = 0) -> Signal:

    signal_h: Signal = Signal()

    # Add first span to the halt signal if
    #   - min_ts_a < min_ts_g
    #   - or min_ts_r < min_ts_g
    min_ts_ar: int | None = _maybe_min(signal_a.get_min_ts(), signal_r.get_min_ts())
    if min_ts_ar is not None and len(signal_g) and min_ts_ar < signal_g.get_min_ts():
        max_ts_ar: int | None = _maybe_max(
            signal_a.get_max_ts(max_ts=signal_g.get_min_ts()),
            signal_r.get_max_ts(max_ts=signal_g.get_min_ts()))

        if max_ts_ar is not None:
            signal_h.append(TimeSpan(min_ts_ar, max_ts_ar))

    # all spaces between green are halt intervals
    for index in range(1, len(signal_g)):
        prev_max_ts_g = signal_g[index - 1].max_ts
        curr_min_ts_g = signal_g[index - 0].min_ts

        min_ts_ar: int | None = _maybe_min(
            signal_a.get_min_ts(min_ts=prev_max_ts_g, max_ts=curr_min_ts_g),
            signal_r.get_min_ts(min_ts=prev_max_ts_g, max_ts=curr_min_ts_g))

        max_ts_ar: int | None = _maybe_max(
            signal_a.get_max_ts(max_ts=curr_min_ts_g),
            signal_r.get_max_ts(max_ts=curr_min_ts_g))

        if None not in [min_ts_ar, max_ts_ar]:
            signal_h.append(TimeSpan(min_ts_ar, max_ts_ar))

    # add one more span to the halt signal if:
    #   - max_ts_a > max_ts_g
    #   - or max_ts_r > max_ts_g.
    max_ts_ar: int | None = _maybe_max(signal_a.get_max_ts(), signal_r.get_max_ts())
    if max_ts_ar is not None and len(signal_g) and signal_g.get_max_ts() < max_ts_ar:
        min_ts_ar: int | None = _maybe_min(
            signal_a.get_min_ts(min_ts=signal_g.get_max_ts()),
            signal_r.get_min_ts(min_ts=signal_g.get_max_ts()))

        if min_ts_ar is not None:
            signal_h.append(TimeSpan(min_ts_ar, max_ts_ar))

    signal_h.trim(accuracy_usec)
    return signal_h


def _decode__halt_signal_r(signal_r: Signal, accuracy_usec: int = 0) -> Signal:
    from copy import deepcopy
    signal_h = deepcopy(signal_r)
    signal_h.trim(accuracy_usec)
    return Signal(signal_h)


def decode__traffic_light(jnode: Dict, ref: int, red_only: bool, accuracy_usec: int) -> ITrafficLight | None:
    """
    :raise JnodeError:
    """
    tl_jnode_list = jnode.get('tl')
    if type(tl_jnode_list) != list:
        eventlog.log_e_param('#/tl')
        return None

    for tl_jnode in tl_jnode_list:
        if ref != tl_jnode['id']:
            continue

        ts_jnode = tl_jnode.get('ts')
        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_span = TimeSpan(*ts_jnode)

        if red_only:
            signal_r = Signal(decode__timespans(tl_jnode.get('red') or []))
            signal_h = _decode__halt_signal_r(signal_r, accuracy_usec)

            return TrafficLightR(ref, time_span, signal_h, signal_r)

        else:
            signal_r = Signal(decode__timespans(tl_jnode.get('red') or []))
            signal_a = Signal(decode__timespans(tl_jnode.get('amber') or []))
            signal_g = Signal(decode__timespans(tl_jnode.get('green') or []))
            signal_h = _decode__halt_signal_rag(signal_r, signal_a, signal_g, accuracy_usec)

            return TrafficLightRag(ref, time_span, signal_h, signal_r, signal_a, signal_g)
