from abc import ABCMeta, abstractmethod
from typing import Final, Self, TextIO

from edge.types.exceptions import *

CliValues = list[str]
CliParams = dict[str, CliValues]


def cli_decode_params(cli_args: list[str]) -> CliParams:
    param_to_values_map: CliParams = CliParams()

    for cli_arg in cli_args:
        param_value: list[str] = cli_arg.split('=', 1)
        param_value.append('')  # ensure 2 elements

        param: str = param_value[0].strip()
        if param not in param_to_values_map:
            param_to_values_map[param] = []

        value: str = param_value[1].strip()
        if value:
            param_to_values_map[param].append(value)

    return param_to_values_map


class CliRequest:
    def __init__(self, cli_path: str, cli_args: list[str]) -> None:
        self.path: Final[str] = cli_path        # argv[0]
        self.args: Final[list[str]] = cli_args  # argv[1:]

    @classmethod
    def create(cls) -> Self:
        import sys
        return cls.create2(sys.argv)

    @classmethod
    def create2(cls, argv: list[str]) -> Self:
        return cls(argv[0], argv[1:])

    @property
    def log_name(self) -> str:
        from os import path
        return path.splitext(path.basename(self.path))[0]

    @property
    def log_level(self) -> str:
        return 'debug'

    def assert_args(self, count: int) -> None:
        if len(self.args) != count:
            raise NotFoundError

    def assert_min_args(self, count: int) -> None:
        if len(self.args) < count:
            raise NotFoundError

    def assert_max_args(self, count: int) -> None:
        if len(self.args) > count:
            raise NotFoundError

    def assert_arg(self, index: int, value: str) -> None:
        if self.args[index] != value:
            raise NotFoundError

    @staticmethod
    def assert_env(name: str) -> None:
        import os
        if name not in os.environ:
            raise NotFoundError


class CliResponse:
    EX_OK = 0         # sysexists.h / successful termination
    EX_USAGE = 64     # sysexists.h / command line usage error
    EX_SOFTWARE = 70  # sysexists.h / internal software error

    def __init__(self, cli_status: int, cli_stdout: str = '', cli_stderr: str = ''):
        self._status = cli_status
        self._stdout = cli_stdout
        self._stderr = cli_stderr

    @classmethod
    def done(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_OK, cli_stdout, cli_stderr)

    @classmethod
    def fail_external(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_USAGE, cli_stdout, cli_stderr)

    @classmethod
    def fail_internal(cls, cli_stdout: str = '', cli_stderr: str = '') -> Self:
        return cls(cls.EX_SOFTWARE, cli_stdout, cli_stderr)

    def render(self) -> int:
        import sys
        return self.render2(sys.stdout, sys.stderr)

    def render2(self, stdout_stream: TextIO, stderr_stream: TextIO) -> int:
        if self._stdout:
            stdout_stream.write(self._stdout)

        if self._stderr:
            stderr_stream.write(self._stderr)

        return self._status


class CliBoolResponse(CliResponse):
    def __init__(self, cli_status: int, cli_stdout: str = '', cli_stderr: str = ''):
        super().__init__(cli_status, cli_stdout, cli_stderr)

    @classmethod
    def done(cls, cli_stdout: str = 'DONE\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_OK, cli_stdout, cli_stderr)

    @classmethod
    def fail_external(cls, cli_stdout: str = 'FAIL\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_USAGE, cli_stdout, cli_stderr)

    @classmethod
    def fail_internal(cls, cli_stdout: str = 'FAIL\n', cli_stderr: str = '') -> Self:
        return cls(cls.EX_SOFTWARE, cli_stdout, cli_stderr)


class ICliHandler(metaclass=ABCMeta):
    @abstractmethod
    def usage(self) -> str:
        pass

    @abstractmethod
    def handle(self, request: CliRequest) -> CliResponse:
        pass
