import ujson
import psycopg2
import psycopg2.pool
import psycopg2.extras

from typing import Self

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import GenericError
from edge.webserver.sysconf.uiconf.db import DbUiconf

psycopg2.extras.register_default_json(globally=True, loads=ujson.loads)
psycopg2.extras.register_default_jsonb(globally=True, loads=ujson.loads)


class DbError(GenericError):
    pass


class DbHintError(DbError):
    def __init__(self, message: str, hint: str):
        super().__init__(message)
        self.hint = hint


class DbQueryError(DbHintError):
    pass


class DbNotFoundError(DbHintError):
    pass


class DbInternalError(DbHintError):
    pass


_x_map = {
    '400': DbQueryError,
    '404': DbNotFoundError,
    '500': DbInternalError
}


def trap_hint_error(x: psycopg2.Error) -> None:
    """
    :raise DbHintError:
    """
    error = str(x.pgerror)
    for line in error.splitlines():
        if line.startswith('HINT'):
            hint = line.split(' ')[-1]
            raise _x_map.get(hint, DbHintError)(str(x), hint)


class DbJsonArg(psycopg2.extras.Json):
    def dumps(self, jnode: dict | list) -> str:
        return ujson.dumps(jnode)


class DbArgs:
    def __init__(self, db_user: str, db_pass: str, db_name: str, db_host: str = 'localhost', db_port: int = 5433):
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port

    @classmethod
    def deserialize(cls, db_jnode: dict) -> Self:
        """
        :raise: KeyError
        """
        db_name = db_jnode['name']
        db_user = db_jnode['user']
        db_pass = db_jnode['pass']
        db_host = db_jnode['host']
        db_port = db_jnode['port']
        return cls(db_user, db_pass, db_name, db_host, db_port)

    @classmethod
    def deserialize2(cls, filepath: str = DbUiconf.CONF_PATH) -> Self:
        """
        :raise: KeyError
        :raise: ValueError
        """
        db_jnode = fs.load_json3(filepath, default=None)
        if db_jnode:
            db_args = cls.deserialize(db_jnode['db'])
            return db_args

        eventlog.log_e(f'fail: invalid {filepath}')
        raise ValueError


def with_db(db_args: DbArgs, handler):
    """
    :type db_args: DbArgs
    :type handler: (Any) -> Any
    :rtype: Any
    :raises: DbError
    """
    try:
        db_link = psycopg2.connect(
            database=db_args.db_name, user=db_args.db_user, password=db_args.db_pass,
            host=db_args.db_host, port=db_args.db_port)
        try:
            return handler(db_link)
        finally:
            db_link.close()

    except psycopg2.Error as x:
        eventlog.log_e(f'fail: {x}')
        raise DbError


def edit_term(term_uuid: str, term_name: str, term_place: str, manual: bool = True):
    dbapi_query_jnode = {
        'uuid': term_uuid,
        'name': term_name,
        'param': {
            'public': {
                'uuid': term_uuid,
                'name': term_name,
                'location': {
                    'place': term_place,
                }
            },
            'manual': manual
        }
    }

    db_args = DbArgs.deserialize2()

    def db_handler(db_link) -> None:
        """
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "select errcode, errmsg, result from api_origin_set(%s)",
                [DbJsonArg(dbapi_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_d(f'fail: api_origin_set ({errmsg})')
                raise DbError

    try:
        return with_db(db_args, db_handler)

    except DbError:
        eventlog.log_e('fail: db')
        raise DbError


def get_origins() -> list | None:
    dbapi_query_jnode = {
        'format': 'param-public',
        'object': 'origin'
    }

    db_args = DbArgs.deserialize2()

    def db_handler(db_link) -> list[dict]:
        """
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "select errcode, errmsg, result from api_discovery(%s)",
                [DbJsonArg(dbapi_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_e(f'fail: api_discovery ({errmsg})')
                raise DbError

            return api_reply_jnode

    try:
        return with_db(db_args, db_handler)

    except DbError:
        eventlog.log_e('fail: db')
        raise DbError
