import os
import socket
import stat
import subprocess
import ujson

from typing import Dict, List, Optional, Tuple

from edge import eventlog
from edge.support.networking import map_ifmac_to_ifname

UI_NETWORK_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/network.json'
MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'


class NetError(Exception):
    pass


class NetToutError(NetError):
    pass


def get_es_uds_endpoint() -> str:
    return '/var/spool/edge/system/es-es-uds'


def get_ra_uds_endpoint(ui_pid: str) -> str:
    return f'/var/spool/edge/pipeline-{ui_pid}/ra0-ra0-uds'
    # return '/var/spool/edge/pipeline-{}/ra_uds'.format(ui_pid)


def get_lr_uds_endpoint(ui_pid: str) -> str:
    return f'/var/spool/edge/pipeline-{ui_pid}/lr0-lr0-uds'


def query_uds(uds_endpoint: str, query_jnode: Dict, tout_sec: int = 3) -> Dict:
    """
    :raise NetError:
    """
    try:
        file_mode = os.stat(uds_endpoint).st_mode
        if not stat.S_ISSOCK(file_mode):
            eventlog.log_e(f'fail: file {uds_endpoint} is not a socket')
            raise NetError

    except OSError as x:
        eventlog.log_e(f'fail: stat {x.filename} ({x.strerror})')
        raise NetError

    link = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        link.settimeout(tout_sec or 1)
        link.connect(uds_endpoint)

        query_utf8 = '{}\0'.format(ujson.dumps(query_jnode))
        query_bytes = bytes(query_utf8, 'utf-8')

        link.settimeout(tout_sec or 1)
        link.send(query_bytes)

        reply_utf8 = str(link.recv(1024), 'utf-8')
        reply_utf8 = reply_utf8.split('\0')[0]  # strip '\0' in the end
        reply_jnode = ujson.loads(reply_utf8)
        return reply_jnode

    except socket.timeout:
        if tout_sec:
            eventlog.log_w('fail: timeout')
            raise NetToutError
        else:
            eventlog.log_e('fail: timeout')
            raise NetError

    except socket.error as x:
        eventlog.log_e(f'fail: {x}')
        raise NetError

    finally:
        link.close()


def query_ra_alignment(ui_pid: str, alignment: bool, min_rcs: int,
                       speed_filter: int = 0, tout_sec: int = 3) -> None:
    """
    :raise NetError
    """
    endpoint = get_ra_uds_endpoint(ui_pid)
    query_jnode = {
        'method': 'set',
        'action': 'alignment',
        'args': {
            'sensor': {
                'mode': 'alignment' if alignment else 'monitoring',
                'mode-alignment': {
                    "speed-filter": speed_filter
                },
                'min-rcs': min_rcs
            }
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status != 1:
        eventlog.log_e(f'fail: param reply#/status has invalid value <{status}>')
        raise NetError


def query_es_test(session_ref: str, test_ref: Tuple[str, str], origin_uuid: str,
                  bbox4_jnode: Optional[List[int]] = None, tout_sec: int = 3) -> None:
    """
    :raise NetError
    """
    es_endpoint = get_es_uds_endpoint()
    jnode = {
        'action': 'store-conf',
        'args': {
            'tout': tout_sec,
            'session-ref': session_ref,
            'test-ref': test_ref,
            'origin': origin_uuid
        }
    }

    if bbox4_jnode:
        jnode['args']['bbox4'] = bbox4_jnode

    reply_jnode = query_uds(es_endpoint, jnode, tout_sec)
    code = reply_jnode.get('code')
    if code != 200:
        eventlog.log_e('fail: param reply#/code has invalid value <{}>'.format(code))
        raise NetError


def query_es_config(endpoints: List[str], tout_sec: int = 3) -> None:
    """
    :raise NetError
    """

    # The eventstore.exe does not support the SIGHUP signal.
    # The socket request is used to notify eventstore.exe of changes to its main.conf.
    # The socket file does not exist if eventstore.exe is not running.
    # The eventstore.exe will read endpoints from main.conf on startup.
    # So, there is no need to fail if the socket file does not exist.
    es_endpoint = get_es_uds_endpoint()
    try:
        file_mode = os.stat(es_endpoint).st_mode
        if not stat.S_ISSOCK(file_mode):
            eventlog.log_d(f'fail: file {es_endpoint} is not a socket')
            return

    except OSError:
        return

    jnode = {
        'action': 'react-conf',
        'args': endpoints
    }

    reply_jnode = query_uds(es_endpoint, jnode, tout_sec)
    code = reply_jnode.get('code')
    if code != 200:
        eventlog.log_e(f'fail: param reply#/code has invalid value <{code}>')
        raise NetError


def query_gps_count() -> int:
    from gpsdclient import GPSDClient
    with GPSDClient(host='localhost', port='2947', timeout=2) as client:
        try:
            for result in client.dict_stream(convert_datetime=False, filter='DEVICES'):
                return len(result.get('devices', []))
        except TimeoutError:
            return 0

    return 0


def query_lr_add_alarm(ui_pid: str, alarm_name: str, tout_sec: int = 3):
    """
    :raise NetError
    """
    endpoint = get_lr_uds_endpoint(ui_pid)
    query_jnode = {
        'action': 'add-alarm',
        'params': {
            'name': alarm_name
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status != 1:
        eventlog.log_e(f'fail: add-alarm (status:{status})')
        raise NetError


def query_lr_get_alarm(ui_pid: str, alarm_name: str, tout_sec: int = 3) -> bool:
    """
    :raise NetError
    """
    endpoint = get_lr_uds_endpoint(ui_pid)
    query_jnode = {
        'action': 'get-alarm',
        'params': {
            'name': alarm_name
        }
    }

    reply_jnode = query_uds(endpoint, query_jnode, tout_sec)
    status = reply_jnode.get('status')
    if status == 1:
        return True
    elif status == 2:
        eventlog.log_e('fail: get-alarm (status:fail)')
        return False
    else:
        eventlog.log_e(f'fail: get-alarm (status:{status})')
        raise NetError


def _get_http_api_port() -> int:
    # TODO: parse /etc/nginx/sites-enabled/edge-site
    # https://github.com/fatiherikli/nginxparser
    # https://github.com/peakwinter/python-nginx
    return 8081


def get_http_api_endpoint() -> Optional[str]:
    """
    :returns: uri
    """
    if os.path.exists(UI_NETWORK_JNODE_FILEPATH):
        with open(UI_NETWORK_JNODE_FILEPATH) as json_file:
            ui_network_jnode = ujson.load(json_file)
            if_mac = ui_network_jnode['ro']['mac']

    elif os.path.exists(MANAGEMENT_MAC_FILEPATH):
        with open(MANAGEMENT_MAC_FILEPATH, 'r') as text_file:
            if_mac = text_file.readlines()[0].strip()

    else:
        eventlog.log_e('fail: lookup management mac')
        raise NetError

    try:
        if_name = map_ifmac_to_ifname(if_mac)
        if_ip = subprocess.getoutput(
            f"ip addr show {if_name} | grep -v '127.0.0.1' | grep -w inet | awk '{{print $2}}'").split('/')[0]

        return f'http://{if_ip}:{_get_http_api_port()}'

    except Exception:
        eventlog.log_e('fail: define uri-host')
        raise NetError
