import shlex
import subprocess

from edge.types.exceptions import ExternalError, InternalError, ToutError


def run_silent(esc_cmd: str) -> int:
    """
    :param esc_cmd:
        Escaped shell-command. Use pipes.quote() to escape input form external sources
    """
    return subprocess.run(esc_cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT).returncode


def quote(data: str) -> str:
    return shlex.quote(data)


def exec_cli(cli_binary: str, cli_params: list[str] | tuple[str],
             tout_sec: int | None = 10,
             encoding: str = 'ascii',
             env: dict[str, str] | None = None) -> str:
    """
    :param cli_binary:
    :param cli_params:
    :param tout_sec:
        If the timeout expires, the child process will be killed and waited for.
    :param encoding:
        'ascii' | 'utf-8'
    :param env:
        These are used instead of the default behavior of inheriting the current process’ environment

    :returns: stdout

    :raises ToutError:
    :raises InternalError:
    :raises ExternalError:
    """
    import subprocess
    from edge import eventlog

    eventlog.log_d(f'exec: <{cli_binary}>')
    try:
        cli_args: list[str] | str = [cli_binary] + list(cli_params)
        eventlog.log_d(f'exec-args: <{cli_args}>')
        cli_result: subprocess.CompletedProcess[str] = subprocess.run(
            cli_args, capture_output=True, text=True, encoding=encoding,
            check=True, timeout=tout_sec, env=env)

        return cli_result.stdout

    except subprocess.TimeoutExpired as x:
        eventlog.log_e(f'fail: <{x.cmd}> (timeout:{x.timeout})')
        raise ToutError

    except subprocess.CalledProcessError as x:
        if x.returncode == 64:  # EX_USAGE
            eventlog.log_e(f'fail: ExternalError <{x.cmd}>')
            raise ExternalError

        elif x.returncode == 70:  # EX_SOFTWARE
            eventlog.log_e(f'fail: InternalError <{x.cmd}> ({x.stderr})')
            raise InternalError

        else:
            eventlog.log_e(f'fail: <{x.cmd}> (returncode:{x.returncode})')
            raise InternalError
