from typing import Dict, Final, Optional, List, Tuple

from edge.types.dir import Dir
from edge.types.jnode.imageref import decode__image_ref
from edge.types.track import LrTrackItem
from edge.types.jnode.track import decode__lr_track_item

from edge.types.jnode.trackset.vrp import decode__track_vrp__lr
from edge.types.trackset.lr import LrTrack


def decode__lr_track_stat(stat: int) -> Tuple[bool, bool]:
    """
    It is possible to have a track which is detected and lost at the same time.

    :param stat:
        kPLATE_DETECTED (0x01).
        kPLATE_TRACKING (0x02).
        kPLATE_LOST_VALID (0x04).
        kPLATE_LOST_INVALID (0x8).
    """
    detected_bit: bool = bool(stat & 0x01)
    lost_bit: bool = bool(stat & 0x04)

    detected_not_lost: bool = detected_bit and not lost_bit
    lost_not_detected: bool = lost_bit and not detected_bit

    stat_detected: Final[bool] = detected_not_lost or lost_not_detected
    stat_lost: Final[bool] = lost_bit

    return stat_detected, stat_lost


def decode__lr_track_dir(value: int) -> Dir:
    """
    :param value:
        1 - unknown
        11 - approaching
        12 - receding
    """
    if value == 1:
        return Dir.CODE_U
    elif value == 11:
        return Dir.CODE_A
    elif value == 12:
        return Dir.CODE_R
    else:
        raise NotImplementedError()


def decode__lr_track(jnode: Dict, lr_alarms_jnode: Optional[List]) -> LrTrack:
    """
    :param jnode:
        TrTrack#/Lr
    :param lr_alarms_jnode:
    """
    vrp_dir = decode__lr_track_dir(jnode['Dir'])

    vrp = decode__track_vrp__lr(jnode['Plate'])
    vrp_trace: List[LrTrackItem] = [decode__lr_track_item(x) for x in jnode['PlateTrace']]

    # todo: find error in tracker/lpr and delete this workaround
    vrp_trace.sort(key=lambda item: item.ts, reverse=False)

    stat_detected, stat_lost = decode__lr_track_stat(jnode['Stat'])

    imref = decode__image_ref(jnode['Image'])

    vehicle_type_jnode = jnode.get('Vehicle', {}).get('type') or None
    vehicle_seats_jnode = jnode.get('Vehicle', {}).get('seats') or None
    spot_speed = jnode.get('SpotSpeed') or None

    return LrTrack(imref, vrp, vrp_trace, vrp_dir, lr_alarms_jnode, stat_detected, stat_lost,
                   vehicle_type_jnode, vehicle_seats_jnode, spot_speed)
