from typing import Final

from edge.support.speed import map_speed_uint


class SpotSpeed:
    def __init__(self, source: str, measured: int):
        self.source: Final[str] = source
        self.measured: Final[int] = measured


class CameraSpotSpeed(SpotSpeed):
    def __init__(self, measured: int) -> None:
        super().__init__('camera', measured)


class RadarSpotSpeed(SpotSpeed):
    def __init__(self, measured: int) -> None:
        super().__init__('radar', measured)


class WimSpotSpeed(SpotSpeed):
    def __init__(self, measured: int) -> None:
        super().__init__('wim', measured)


class SpotSpeeds:
    def __init__(self,
                 main_speed: SpotSpeed | None = None,
                 control_speed: SpotSpeed | None = None,
                 ra_speed_2: int | None = None,
                 ra_speed_n: int | None = None,
                 cam_speed_2: int | None = None,
                 cam_speed_n: int | None = None,
                 wim_speed: int | None = None) -> None:
        """
        :param main_speed:
            The value provided by the main measurement channel (e.g. radar)
        :param control_speed:
            The value provided by the control measurement channel (e.g. camera)
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param ra_speed_2:
            The value provided by the radar using 2 track points
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param ra_speed_n:
            The value provided by the radar using all track points
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param cam_speed_2:
            The value provided by the camera using 2 track points
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param cam_speed_n:
            The value provided by the camera using all track points
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param wim_speed:
            The value provided by the Waga system.
            The value is always positive. System does not provide the direction of the vehicle movement.

            It is possible to request the lanes info from the Waga system.
            Each info structure contains the direction of the permitted vehicle movement.
        """

        self.main: Final[SpotSpeed | None] = main_speed
        self.control: Final[SpotSpeed | None] = control_speed

        self.ra_2: Final[int | None] = ra_speed_2
        self.ra_n: Final[int | None] = ra_speed_n
        self.cam_2: Final[int | None] = cam_speed_2
        self.cam_n: Final[int | None] = cam_speed_n
        self.wim: Final[int | None] = wim_speed

    @staticmethod
    def assert_overspeed(main_speed: int, max_speed: float) -> bool:
        """
        :param main_speed:
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param max_speed:
            Maximum allowed speed in km/h.
            Value is always positive.
        """
        return map_speed_uint(main_speed) > max_speed
