from collections.abc import Iterable
from sortedcontainers import SortedDict
from typing import Callable, Final, Self

from edge.types.dir import Dir
from edge.types.track import TrTrackItem, LrTrackItem, RaTrackItem, LaneTrackItem, DirTrack


class TrTrack(SortedDict[int, TrTrackItem]):
    def __init__(self, items: SortedDict[int, TrTrackItem], items_dir: Dir) -> None:
        super().__init__(items)

        self.dir: Final[Dir] = items_dir

    @property
    def min_ts(self) -> int:
        return self.keys()[0]

    @property
    def max_ts(self) -> int:
        return self.keys()[-1]

    @classmethod
    def create(cls,
               lr_track: Iterable[LrTrackItem] | None = None,
               ra_track: Iterable[RaTrackItem] | None = None,
               dir_track: DirTrack | None = None,
               lane_track: Iterable[LaneTrackItem] | None = None) -> Self:

        ts_to_tr_item_map: SortedDict[int, TrTrackItem] = SortedDict()

        for lr_item in (lr_track or ()):
            tr_item = ts_to_tr_item_map.setdefault(lr_item.ts, TrTrackItem(lr_item.ts))
            tr_item.lr_item = lr_item

        for ra_item in (ra_track or ()):
            tr_item = ts_to_tr_item_map.setdefault(ra_item.frame_ts, TrTrackItem(ra_item.frame_ts))
            tr_item.ra_item = ra_item

        for dir_item in (dir_track or ()):
            tr_item = ts_to_tr_item_map.setdefault(dir_item.ts, TrTrackItem(dir_item.ts))
            tr_item.dir_item = dir_item

        for lane_item in (lane_track or ()):
            tr_item = ts_to_tr_item_map.setdefault(lane_item.ts, TrTrackItem(lane_item.ts))
            tr_item.lane_item = lane_item

        items_dir: Dir = Dir.CODE_U if dir_track is None else dir_track.dir
        return cls(ts_to_tr_item_map, items_dir)

    def for_items_asc(self, min_ts: int, max_ts: int, handler: Callable[[TrTrackItem], None]) -> None:
        """
        Calls a handler for each suitable item until the StopIteration exception raised or no more items are available.
        """
        for item_ts in self.irange(min_ts, max_ts):
            try:
                handler(self[item_ts])
            except StopIteration:
                break

    def for_items_desc(self, min_ts: int, max_ts: int, handler: Callable[[TrTrackItem], None]) -> None:
        """
        Calls a handler for each suitable item until the StopIteration exception raised or no more items are available.
        """
        for item_ts in self.irange(min_ts, max_ts, reverse=True):
            try:
                handler(self[item_ts])
            except StopIteration:
                break

    def find_item_asc(self, min_ts: int, max_ts: int, handler: Callable[[TrTrackItem], bool]) -> TrTrackItem | None:
        for item_ts in self.irange(min_ts, max_ts):
            try:
                item = self[item_ts]
                if handler(item):
                    return item

            except StopIteration:
                break

    def find_item_desc(self, min_ts: int, max_ts: int, handler: Callable[[TrTrackItem], bool]) -> TrTrackItem | None:
        for item_ts in self.irange(min_ts, max_ts, reverse=True):
            try:
                item = self[item_ts]
                if handler(item):
                    return item

            except StopIteration:
                break

    def get_item_dir(self, item_ts: int) -> Dir:
        try:
            tr_item = self[item_ts]
            if tr_item.dir_item is not None:
                return tr_item.dir_item.dir

        except KeyError:
            pass

        return Dir.CODE_U

    def get_item_lane(self, item_ts: int) -> str | None:
        try:
            tr_item = self[item_ts]
            if tr_item.lane_item is not None:
                return tr_item.lane_item.lane

        except KeyError:
            pass
