from edge import eventlog
from edge.types.exceptions import *
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.systest import SystestUimod
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class ControllerApi:
    @classmethod
    def about(cls) -> dict:
        from socket import gethostname
        from edge.services import about

        sw_packages: list[about.SwPackageInfo] = about.get_device_sw_packages(about.SW_DIRPATH)

        ctrl_info = {
            'hostname': gethostname(),
            'sw-name': about.get_device_sw_name() or 'EDGE',
            'sw-lang': about.get_device_sw_lang(),
            'sw-ver': sw_packages[0]['package'][2],
            'sw': sw_packages
        }

        sw_md5: str | None = about.get_device_sw_md5()
        if sw_md5 is None:
            if len(sw_packages) >= 2:
                sw_md5 = sw_packages[1]['files'][0]['md5']

        if sw_md5:
            ctrl_info['sw-md5'] = sw_md5

        hw_version = about.get_device_hw_vendor_model_serial()
        if hw_version:
            ctrl_info['hw'] = hw_version

        return {
            'device': ctrl_info
        }

    # noinspection PyUnusedLocal
    @classmethod
    def create_subunit(cls, ctrl_uref: str, factory_conf: dict) -> dict:
        unit_utype = factory_conf.get('type')
        if unit_utype == UTYPE_EVENTSTORE:
            from edge.webserver.sysconf.api_eventstore import EventstoreApi
            return EventstoreApi.make_unit(factory_conf)

        elif unit_utype == UTYPE_TERMINAL:
            # Only terminals with an RTSP camera are allowed.
            if factory_conf.get('rtsp-camera') is None:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            from edge.webserver.sysconf.api_license import LicenseApi
            LicenseApi.test_lpr_license_available()

            from edge.webserver.sysconf.api_terminal import TerminalApi
            factory_conf['factory'] = 'rtsp-terminal-info.json'
            return TerminalApi.create_term(ctrl_uref, factory_conf)

        else:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

    # noinspection PyUnusedLocal
    @classmethod
    def reboot(cls, ctrl_uiref: str) -> None:
        import os
        os.system('shutdown -r now > /dev/null 2>&1')
        # os.system('shutdown -r +1 > /dev/null 2>&1')

    @classmethod
    def _set_systest(cls, value: bool) -> None:
        from edge.webserver.sysconf.dgconf.cron import CronDgconf
        from edge.webserver.sysconf.uiconf.controller import CtrlUiconf
        from edge.webserver.sysconf import edge_creator

        CtrlUiconf.set_systest(value)
        CronDgconf.enable() if value else CronDgconf.disable()

        root_uinfo = rootinfo.decode_info().root
        urefs_l1 = unet.find_urefs_r(root_uinfo, utypes=[UTYPE_TRAFFIC_RULES, UTYPE_FUSION])

        ctrl_umod = Umod('0', UTYPE_CONTROLLER, '0')
        edge_creator.generate(ctrl_umod.create_backup_dirpath(), urefs_l1, forced=True)

        # ph.change_eventstore_systest(True)  # TODO change to generate
        # # TODO: after changing to generate this will happen automatic
        # dg.LifeCycle.sigkill_dgmod(dg.EDGE_TYPE_SPEEDCONTROL)
        # dg.LifeCycle.sigkill_dgmod(dg.EDGE_TYPE_TRACKER)
        dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_EVENTSTORE)

    @classmethod
    def start_systest(cls) -> None:
        cls._set_systest(True)

    @classmethod
    def finish_systest(cls) -> None:
        cls._set_systest(False)

    @classmethod
    def install_systest_crt(cls, crt_jnode: dict) -> None:
        SystestUimod.install_crt(crt_jnode)
        dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_SPEEDCONTROL)

    @classmethod
    def uninstall_systest_crt(cls) -> None:
        SystestUimod.uninstall_crt()
        dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_SPEEDCONTROL)

    @classmethod
    def install_systest_report(cls, report_ref: str, report_path: str) -> None:
        SystestUimod.install_report(report_ref, report_path)

    @classmethod
    def uninstall_systest_report(cls, report_ref: str) -> None:
        SystestUimod.uninstall_report(report_ref)
