import os

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgstats.blobstore import BsDgstats
from edge.webserver.sysconf.types.defs import CTRL_ARCH_DIR
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.eventstore import EsUimod
from edge.webserver.sysconf.uimod.uref import Uref2
from edge.webserver.sysconf.uimod.urole import *
from edge.webserver.sysconf.uimod.utype import *


def _get_disk_size() -> float:
    result = os.statvfs(CTRL_ARCH_DIR)
    block_size = result.f_frsize
    total_blocks = result.f_blocks

    mega = 1024 * 1024
    total_size = total_blocks * block_size / mega

    return total_size


# noinspection PyClassHasNoInit
class EventstoreApi:
    @classmethod
    def _get_limits(cls) -> dict:
        total_size = _get_disk_size()

        limits = {
            'cleanup': {
                'min-df': {
                    'min': 0,
                    'max': total_size
                },
                'max-du': {
                    'min': 0,
                    'max': total_size
                },
                'max-abs-time': {
                    'min': 0,
                    'max': 4294967295
                },
                'max-rel-time': {
                    'min': 0,
                    'max': 4294967295
                }
            }
        }
        return limits

    @classmethod
    def load_unit(cls, uref: str) -> dict:
        unit_umod = EsUimod.create(uref)
        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uref.split('.')[-1], stats_tz)

        reply = {
            'type': UTYPE_EVENTSTORE,
            'conf': unit_uconf['rw']['args'],
            'limits': cls._get_limits()
        }

        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def save_unit(cls, uref: str, new_uconf: dict) -> dict:
        unit_umod = EsUimod.create(uref)
        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        new_uconf_conf = new_uconf['conf']
        fs.validate_json({'args': new_uconf_conf}, unit_umod.schema_path)

        unit_uconf_pub: dict = unit_uconf['rw']
        if unit_uconf_pub['args'] != new_uconf_conf:
            if new_uconf_conf['name'] != unit_uconf_pub['args']['name']:
                rootinfo.set_unit_name(uref, new_uconf_conf['name'])

            unit_uconf_pub['args'] = new_uconf_conf

            # uimod.save_conf(uiconf)
            # bs.bs65535_reappend_chan(uiconf)
            urefs_l0 = [unit_umod.ref]
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0, {unit_umod.ref: unit_uconf})
            dg.CTRL_BS_DGMOD.sighup()

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uref.split('.')[-1], stats_tz)

        reply = {
            'type': UTYPE_EVENTSTORE,
            'conf': unit_uconf_pub['args'],
            'limits': cls._get_limits()
        }

        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def make_unit(cls, factory_conf: dict) -> dict:
        root_unet = rootinfo.decode_info()
        ctrl_uinfo = unet.find_unit_r(root_unet.root, utype=UTYPE_CONTROLLER)

        uref_gen = rootinfo.UrefGenerator(ctrl_uinfo)
        unit_uref: Uref2 = uref_gen.create_unit_uref2(UTYPE_EVENTSTORE)

        # add store_jnode to /etc/edge/setupstore/admin/action/eventstore/<chan-ref>/0.json
        unit_umod = EsUimod.create(unit_uref.ref)
        unit_uconf = unit_umod.install(factory_conf['name'], unit_uref.chan)

        # bs.bs65535_append_chan(uiconf)
        urefs_l0 = [unit_umod.ref]
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l0, {unit_umod.ref: unit_uconf})
        dg.CTRL_BS_DGMOD.sigusr1()

        ctrl_subunits: list = ctrl_uinfo['units']
        ctrl_subunits.append({
            'internal': False,
            'role': UROLE_ARCHIVE,
            'type': UTYPE_EVENTSTORE,
            'name': factory_conf['name'],
            'ref': unit_uref.ref
        })

        rootinfo.encode_info(root_unet)

        return {
            'info': [root_unet.root],
            'created': [unit_uref.chan]
        }

    @classmethod
    def delete_unit(cls, uref: str) -> dict:
        root_info, urefs_l1 = rootinfo.drop_unit_r(uref)

        unit_umod = EsUimod.create(uref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l1)
        dg.CTRL_BS_DGMOD.sigusr1()
        unit_umod.uninstall()

        return {
            'info': [root_info]
        }
