from gpsdclient import GPSDClient
import time
import os

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def _encode_rt_conf(new_jnode: dict) -> dict:
    az = new_jnode.get('azimuth', None)
    provider_jnode = new_jnode['providers'][0]

    if provider_jnode['provider'] == 'manual':
        jnode = {
            'azimuth': az,
            'source': ['static', provider_jnode['latitude'], provider_jnode['longitude']]
        }
    elif provider_jnode['provider'] == 'controller':
        jnode = {
            'azimuth': az,
            'source': ['gpsd', 'localhost', 2947]
        }
    else:
        jnode = dict()

    return jnode


def _test_uiconf_conf_changed(uiconf_conf: dict, new_uiconf_conf: dict) -> bool:
    if uiconf_conf.get('azimuth') != new_uiconf_conf.get('azimuth'):
        return True

    provider_info = uiconf_conf['providers'][0]
    new_provider_info = new_uiconf_conf['providers'][0]

    if provider_info['provider'] != new_provider_info['provider']:
        return True

    # old_mode == new_mode
    if provider_info['provider'] == 'manual':
        if provider_info['latitude'] != new_provider_info['latitude']:
            return True

        if provider_info['longitude'] != new_provider_info['longitude']:
            return True

    return False


def _test_gps_device_configured() -> bool:
    _GPSD_CONF_PATH = '/etc/default/gpsd'

    # gps_present = net.query_gps_count() > 0
    try:
        if os.path.isfile(_GPSD_CONF_PATH):
            with open(_GPSD_CONF_PATH, 'r') as gpsd_conf:
                gpsd_lines = gpsd_conf.readlines()

            for line in gpsd_lines:
                if line.startswith('DEVICES="/'):
                    return True

    except Exception as x:
        eventlog.log_e('fail: {}'.format(str(x)))

    return False


def _get_gps_device_lat_long2(
        host: str = 'localhost',
        port: str = '2947',
        tout_sec: float = 2.0) -> tuple[float, float] | tuple[None, None]:

    call_ts = time.time()
    with GPSDClient(host=host, port=port, timeout=tout_sec) as client:
        try:
            for result in client.dict_stream(convert_datetime=False, filter=['DEVICES', 'TPV']):
                if (time.time() - call_ts) > tout_sec:
                    eventlog.log_d('fail: timeout')
                    break

                msg_class = result.get('class')
                if msg_class == 'DEVICES':
                    if len(result.get('devices', [])) == 0:
                        eventlog.log_w('fail: device-count:0')
                        return None, None

                elif msg_class == 'TPV':
                    lat_long = result.get('lat'), result.get('lon')
                    if None not in lat_long:
                        return lat_long

        except TimeoutError:
            eventlog.log_w('fail: timeout')
            pass

    return None, None


def _add_limits_stats(uiconf_conf: dict) -> dict:
    if _test_gps_device_configured():
        unit_ext_jnode = {
            'type': UTYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual', 'controller']
            }
        }

    else:
        unit_ext_jnode = {
            'type': UTYPE_GNSS,
            'conf': uiconf_conf,
            'limits': {
                'providers': ['manual']
            }
        }

    uiconf_providers = uiconf_conf['providers']
    if len(uiconf_providers) != 1:
        raise NotImplementedError

    for provider in uiconf_providers:
        if provider['provider'] == 'manual':
            unit_ext_jnode['stats'] = {
                'provider': 'manual',
                'latitude': uiconf_providers[0]['latitude'],
                'longitude': uiconf_providers[0]['longitude']
            }

        elif provider['provider'] == 'controller':
            lat, lon = _get_gps_device_lat_long2()
            if None not in [lat, lon]:
                unit_ext_jnode['stats'] = {
                    'provider': 'controller',
                    'latitude': lat,
                    'longitude': lon
                }

    return unit_ext_jnode


# noinspection PyClassHasNoInit
class ApiGnss:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_GNSS)
        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        return _add_limits_stats(unit_uconf['rw']['conf'])

    @staticmethod
    def save_unit(uref: str, new_uconf: dict) -> dict:
        umod = Umod.create(uref)
        umod.assert_type(UTYPE_GNSS)

        new_conf = new_uconf.get('conf')
        fs.validate_json(new_conf, umod.schema_path)

        uconf = umod.load_conf()
        uconf_pub = uconf['rw']
        if _test_uiconf_conf_changed(uconf_pub['conf'], new_conf):
            uconf_pub['conf'] = new_conf
            uconf['rt'] = _encode_rt_conf(new_conf)
            urefs_l0_l1 = rootinfo.get_linked_urefs(umod.ref)
            creator.generate(umod.create_backup_dirpath(), urefs_l0_l1, {umod.ref: uconf})

        return _add_limits_stats(new_conf)
