# coding=utf-8

import os
import datetime
import ujson

from edge import eventlog
from edge.support import fs, sh
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator, file_lock
from edge.webserver.sysconf.uiconf.iit_ca import IitCaUiconf, IitCaEndUserUiconf
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.sign import SignUimod
from edge.webserver.sysconf.uimod.umod2 import Umod2
from edge.webserver.sysconf.uimod.uref import Uref2
from edge.webserver.sysconf.uimod.urole import *
from edge.webserver.sysconf.uimod.utype import *


def _iit_cli(*args) -> str:
    """
    :rtype: str
    """
    return sh.exec_cli('/opt/edge/bin/crypto-iit-cli', list(args), tout_sec=10, encoding='utf-8')


def _list_media(cxt=None) -> list[dict]:
    """
    :rtype: list[dict]
    """
    import EUSignCP

    iit_api = cxt
    if iit_api is None:
        EUSignCP.EULoad()
        iit_api: EUSignCP.EU_INTERFACE = EUSignCP.EUGetInterface()
        # ctx.SetUIMode(1)
        iit_api.Initialize()

    media_list = []

    type_index = 1
    type_desc = list()
    while iit_api.EnumKeyMediaTypes(type_index, type_desc):
        if type_desc[0] != 'е.ключ ІІТ Алмаз-1К':
            type_index += 1
            continue

        dev_index = 0
        dev_desc = list()
        while iit_api.EnumKeyMediaDevices(type_index, dev_index, dev_desc):
            media_list.append({
                'model': type_desc[0],
                'serial': dev_desc[0]
            })

            dev_index += 1

        type_index += 1

    return media_list


def _edit_iit_conf(new_uiconf_conf: dict, save_loaded_cert: bool) -> list[dict]:
    import EUSignCP

    EUSignCP.EULoad()
    ctx: EUSignCP.EU_INTERFACE = EUSignCP.EUGetInterface()
    # ctx.SetUIMode(1)
    ctx.Initialize()

    mode_config = {
        'bOfflineMode': False
    }
    ctx.SetModeSettings(mode_config)

    jnode_fs = new_uiconf_conf['crl']

    file_store_settings = dict()
    try:
        ctx.GetFileStoreSettings(file_store_settings)
    except RuntimeError:
        eventlog.log_d('fail: GetFileStoreSettings(). Creating settings')

    file_store_settings['szPath'] = '/data/certificates'
    file_store_settings['bCheckCRLs'] = jnode_fs['enable']
    file_store_settings['bAutoRefresh'] = False
    file_store_settings['bOwnCRLsOnly'] = jnode_fs['only-own-crls']
    file_store_settings['bFullAndDeltaCRLs'] = jnode_fs['full-and-delta-crls']
    file_store_settings['bAutoDownloadCRLs'] = True
    file_store_settings['bSaveLoadedCerts'] = save_loaded_cert
    file_store_settings['dwExpireTime'] = 3600

    ctx.SetFileStoreSettings(file_store_settings)

    # print 'file_store_settings', file_store_settings

    jnode_proxy = new_uiconf_conf['proxy']
    jnode_proxy_auth_type = jnode_proxy['auth-type']
    proxy_setting = dict()
    try:
        ctx.GetProxySettings(proxy_setting)
    except RuntimeError:
        eventlog.log_i('fail: GetProxySettings(). Creating settings')

    proxy_setting['bUseProxy'] = jnode_proxy.get('enable', False)
    proxy_setting['szAddress'] = '{}'.format(jnode_proxy['host'])
    proxy_setting['szPort'] = '{}'.format(jnode_proxy['port'])
    if jnode_proxy_auth_type == 'anonymous':
        proxy_setting['bAnonymous'] = True
        proxy_setting['szUser'] = ''
        proxy_setting['szPassword'] = ''
        proxy_setting['bSavePassword'] = False

    elif jnode_proxy_auth_type == 'credentials':
        proxy_setting['bAnonymous'] = False
        proxy_setting['szUser'] = '{}'.format(jnode_proxy['auth-user'])
        proxy_setting['szPassword'] = '{}'.format(jnode_proxy['auth-pass'])
        proxy_setting['bSavePassword'] = True

    ctx.SetProxySettings(proxy_setting)

    # print 'proxy_setting', proxy_setting

    jnode_cmp = new_uiconf_conf['cmp']
    cmp_settings = dict()
    try:
        ctx.GetCMPSettings(cmp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetCMPSettings(). Creating settings')

    cmp_settings['szPort'] = '{}'.format(jnode_cmp['port'])
    cmp_settings['szAddress'] = '{}'.format(jnode_cmp['host'])
    # cmp_settings['szCommonName'] = u'name'
    cmp_settings['bUseCMP'] = True
    ctx.SetCMPSettings(cmp_settings)

    # print 'cmp_settings', cmp_settings

    jnode_oscp = new_uiconf_conf['ocsp']
    oscp_settings = dict()
    try:
        ctx.GetOCSPSettings(oscp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetOCSPSettings(). Creating settings')
    oscp_settings['bUseOCSP'] = jnode_oscp.get('enable', False)
    oscp_settings['szAddress'] = '{}'.format(jnode_oscp['host'])
    oscp_settings['szPort'] = '{}'.format(jnode_oscp['port'])
    oscp_settings['bBeforeStore'] = False
    ctx.SetOCSPSettings(oscp_settings)
    # print 'oscp_settig', oscp_settig

    jnode_tsp = new_uiconf_conf['tsp']
    tsp_settings = dict()
    try:
        ctx.GetTSPSettings(tsp_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetTSPSettings(). Creating settings')
    tsp_settings['szAddress'] = '{}'.format(jnode_tsp['host'])
    tsp_settings['szPort'] = '{}'.format(jnode_tsp['port'])
    tsp_settings['bGetStamps'] = True
    ctx.SetTSPSettings(tsp_settings)
    # print 'tsp_settings', tsp_settings

    # IIT library doesn't work if LDAP settings are not set
    ldap_settings = dict()
    try:
        ctx.GetLDAPSettings(ldap_settings)
    except RuntimeError:
        eventlog.log_i('fail: GetLDAPSettings(). Creating stub settings')
    ldap_settings['bUseLDAP'] = False
    ldap_settings['szAddress'] = ''
    ldap_settings['szPort'] = ''
    ldap_settings['bAnonymous'] = True
    ldap_settings['szUser'] = ''
    ldap_settings['szPassword'] = ''
    ctx.SetLDAPSettings(ldap_settings)

    keys = _list_media(ctx)
    return keys
    # ctx.GetPrivateKeyMedia()

    # cmp_settings2 = dict()
    # ctx.GetCMPSettings(cmp_settings2)
    # print cmp_settings2


class IitCaApi:
    @staticmethod
    def create_service_from_sign(ui_sign_jnode: dict) -> tuple[str, dict]:
        root_unet = rootinfo.decode_info()
        root_info, root_links = root_unet.root, root_unet.links
        controller_info = unet.find_unit_r(root_info, utype=UTYPE_CONTROLLER)
        if controller_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UTYPE_TERMINAL))
            raise NotFoundError

        tr_uirefs = unet.find_urefs_r(root_info, utypes=[UTYPE_TRAFFIC_RULES])

        iit_ca_unit_name = 'ІІТ ЦСК'
        # parse old conf
        ui_sign_conf = ui_sign_jnode['rw']['conf']
        ui_sign_conf_key = ui_sign_conf['key']
        media_type = ui_sign_conf_key['model']
        media_serial = ui_sign_conf_key['serial']
        media_password = ui_sign_conf_key['password']
        del ui_sign_conf['key']
        sign_active = ui_sign_conf['active']
        del ui_sign_conf['active']
        ui_sign_conf['name'] = iit_ca_unit_name

        uref_gen = rootinfo.UrefGenerator(controller_info)
        iit_ca_uiref: Uref2 = uref_gen.create_unit_uref2(UTYPE_IIT_CA)

        iit_ca_uimod = Umod2(iit_ca_uiref.ref)
        iit_ca_uiconf = IitCaUiconf.create({'ref': iit_ca_uiref.ref})
        iit_ca_uiconf.set_pub_conf(ui_sign_conf)

        iit_ca_uimod.install(iit_ca_uiconf, iit_ca_uiref.ref)
        iit_ca_units_info = []

        # add enduser
        if 0 not in [len(media_type), len(media_serial), len(media_password)]:
            end_user_uimod = iit_ca_uimod.child(UTYPE_IIT_CA_END_USER, child_chan=0)
            enduser_factory = {
                'conf': {
                    'name': 'ІІТ Користувач ЦСК',
                    'active': sign_active,
                    'media': {
                        'model': media_type,
                        'serial': media_serial,
                        'password': media_password
                    },
                    'parent-ref': iit_ca_uiref.ref,
                    'zip-name': None
                }
            }
            end_user_uiconf = IitCaEndUserUiconf.create(enduser_factory)

            end_user_uimod.install(end_user_uiconf, iit_ca_uiref.ref)

            # install uimod-info
            iit_ca_units_info.append({
                'ref': end_user_uimod.ref,
                'type': end_user_uimod.type,
                'name': end_user_uiconf.pub_name,
                'role': UROLE_SERVICE,
                'internal': True,
                'factory': [{
                    'type': UTYPE_IIT_CA_END_USER,
                    'role': UROLE_UNIT
                }]
            })

            # add links
            for tr_uiref in tr_uirefs:
                root_links.append((tr_uiref, end_user_uimod.ref))

        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': iit_ca_uimod.ref,
            'type': iit_ca_uimod.type,
            'name': iit_ca_unit_name,
            'role': UROLE_UNIT,
            'internal': True,
            'units': iit_ca_units_info
        })

        # add links
        for tr_uiref in tr_uirefs:
            root_links.append((tr_uiref, iit_ca_uimod.ref))

        rootinfo.encode_info(root_unet)
        return iit_ca_uiref.chan, root_info

    @staticmethod
    def make_service(ctrl_uref: str) -> None:
        root_unet = rootinfo.decode_info()
        ctrl_uinfo = unet.find_unit_r(root_unet.root, uref=ctrl_uref)
        if ctrl_uinfo is None:
            eventlog.log_d(f'fail: NotFoundError ({UTYPE_TERMINAL})')
            raise NotFoundError

        uref_gen = rootinfo.UrefGenerator(ctrl_uinfo)
        unit_uref: Uref2 = uref_gen.create_unit_uref2(UTYPE_IIT_CA)
        unit_uconf = IitCaUiconf.create({'ref': unit_uref.ref})

        unit_umod = Umod2(unit_uref.ref)
        unit_umod.install(unit_uconf, unit_uref.ref)

        _edit_iit_conf(unit_uconf.pub_conf, unit_uconf.save_load_certs)

        if ctrl_uinfo.get('units') is None:
            ctrl_uinfo['units'] = []

        ctrl_uinfo['units'].append({
            'ref': unit_umod.ref,
            'type': unit_umod.type,
            'name': unit_uconf.pub_name,
            'role': UROLE_SERVICE,
            'internal': True,
            'units': [],
            'factory': [{
                'type': UTYPE_IIT_CA_END_USER,
                'role': UROLE_UNIT
            }]
        })

        # add links
        urefs_rules = unet.find_urefs_r(root_unet.root, utypes=[UTYPE_TRAFFIC_RULES])
        for uref_rules in urefs_rules:
            root_unet.links.append((uref_rules, unit_umod.ref))

        rootinfo.encode_info(root_unet)

    @staticmethod
    def drop_service(uref: str) -> dict:
        unit_umod_ca = Umod2(uref)
        try:
            unit_umod_ca.assert_type(UTYPE_IIT_CA)
            unit_umod_ca.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        root_unet = rootinfo.decode_info()
        root_uinfo = root_unet.root

        unit_uinfo_ca: unet.UnitInfo = unet.find_unit_r(root_uinfo, uref=uref)
        unit_urefs_enduser = unet.find_urefs_r(unit_uinfo_ca, [UTYPE_IIT_CA_END_USER])

        for enduser_uref in unit_urefs_enduser:
            IitCaApi.drop_end_user(enduser_uref)

        unit_umod_ca.uninstall()

        root_uinfo, _ = rootinfo.drop_unit_r(unit_umod_ca.ref)

        return {
            'info': [root_uinfo]
        }

    @staticmethod
    def load_service(uref: str) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        unit_uconf = IitCaUiconf(unit_umod.uiconf)
        stats_jnode = {
            'end-users': []
        }

        def handle_enduser(user_info: unet.UnitInfo) -> None:
            stats_jnode['end-users'].append({'ref': user_info['ref']})

        root_uinfo = rootinfo.decode_info().root
        unit_uinfo = unet.find_unit_r(root_uinfo, uref=unit_umod.ref)
        unet.map_units_r(unit_uinfo, handle_enduser, [UTYPE_IIT_CA_END_USER])

        return {
            'type': UTYPE_IIT_CA,
            'conf': unit_uconf.pub_conf,
            'stats': stats_jnode
        }

    @staticmethod
    def load_service_factory(uref: str) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        return {
            'limits': {
                'media': _list_media()
            }
        }

    @staticmethod
    def save_service(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        unit_uconf = IitCaUiconf(unit_umod.uiconf)

        new_uconf_conf = new_uconf.get('conf')
        if new_uconf_conf is None:
            eventlog.log_d('ParamsError')
            raise ParamsError

        new_proxy_pass = new_uconf_conf['proxy'].get('auth-pass')

        if new_proxy_pass is None:
            new_uconf_conf['proxy']['auth-pass'] = unit_uconf.proxy_pass

        name_changed = unit_uconf.set_pub_name(new_uconf_conf['name'])
        conf_changed = unit_uconf.set_pub_conf(new_uconf_conf) or name_changed
        if conf_changed:
            unit_uconf.validate()
            unit_umod.save_uiconf(unit_uconf)

        # edit iit config
        _edit_iit_conf(new_uconf_conf, unit_uconf.save_load_certs)

        # generate
        urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf.raw})

        stats_end_users = list[unet.UnitInfo]()

        def handle_end_user(user_info: unet.UnitInfo) -> None:
            stats_end_users.append({'ref': user_info['ref']})

        root_uinfo = rootinfo.decode_info().root
        unit_uinfo = unet.find_unit_r(root_uinfo, uref=unit_umod.ref)
        unet.map_units_r(unit_uinfo, handle_end_user, [UTYPE_IIT_CA_END_USER])

        return {
            'type': UTYPE_IIT_CA,
            'conf': new_uconf_conf,
            'stats': {
                'end-users': stats_end_users
            }
        }

    @staticmethod
    def load_end_user(uref: str) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA_END_USER)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_d('fail: NotFoundError')
            raise

        unit_uconf = IitCaEndUserUiconf(unit_umod.uiconf)

        crt_stats = unit_uconf.crt_stats
        if crt_stats is None:
            try:
                crt_stats = ujson.loads(_iit_cli('sign', 'crt-info', unit_uconf.serial, unit_uconf.password))
                unit_uconf.set_crt_stats(crt_stats)
                unit_umod.save_uiconf(unit_uconf)
            except InternalError:
                pass

        reply = {
            'type': unit_umod.type,
            'conf': unit_uconf.pub_conf,
            'stats': {
                'media': {
                    'model': unit_uconf.model,
                    'serial': unit_uconf.serial
                }
            }
        }

        csr_name = unit_uconf.csr_name
        if csr_name:
            reply['stats']['csr'] = csr_name

        crt_stats = unit_uconf.crt_stats
        if crt_stats:
            reply['stats']['crt'] = crt_stats

        return reply

    @staticmethod
    def save_end_user(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA_END_USER)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        unit_uconf = IitCaEndUserUiconf(unit_umod.uiconf)

        new_uconf_conf = new_uconf.get('conf')
        name_changed = unit_uconf.set_pub_name(new_uconf_conf['name'])
        conf_changed = unit_uconf.set_pub_conf(new_uconf_conf) or name_changed
        if conf_changed:
            unit_uconf.validate()
            unit_umod.save_uiconf(unit_uconf)

        urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf.raw})

        reply = {
            'type': unit_umod.type,
            'conf': unit_uconf.pub_conf,
            'stats': {
                'media': {
                    'model': unit_uconf.model,
                    'serial': unit_uconf.serial
                }
            }
        }

        csr_name = unit_uconf.csr_name
        if csr_name:
            reply['stats']['csr'] = csr_name

        crt_stats = unit_uconf.crt_stats
        if crt_stats:
            reply['stats']['crt'] = crt_stats

        return reply

    @staticmethod
    def make_end_user(parent_uref: str, factory_conf: dict) -> dict:
        factory_conf_conf = factory_conf['conf']

        format_media = factory_conf_conf.get('format-media')
        if type(format_media) not in (bool,):
            eventlog.log_d('fail: ParamsError ({}) <format-media> is not provided'.format(UTYPE_IIT_CA_END_USER))
            raise ParamsError

        root_unet = rootinfo.decode_info()
        root_info, root_links = root_unet.root, root_unet.links
        iit_ca_uimod = Umod2(parent_uref)

        iit_ca_info = unet.find_unit_r(root_info, uref=parent_uref)
        if iit_ca_info is None:
            eventlog.log_d('fail: NotFoundError ({})'.format(UTYPE_IIT_CA))
            raise NotFoundError

        end_user_info = unet.find_unit_r(iit_ca_info, utype=UTYPE_IIT_CA_END_USER)
        if end_user_info is not None:
            eventlog.log_d('fail: ParamsError ({}) already exist'.format(UTYPE_IIT_CA_END_USER))
            raise ParamsError

        from edge.services import about
        hw_serial = about.get_device_hw_serial()
        if hw_serial is None:
            eventlog.log_e('fail: InternalError (about.json#/hw is not provided)')
            raise InternalError

        uref_gen = rootinfo.UrefGenerator(iit_ca_info)
        uiref: Uref2 = uref_gen.create_unit_uref2(UTYPE_IIT_CA_END_USER)
        end_user_uimod = iit_ca_uimod.child(UTYPE_IIT_CA_END_USER, uiref.chan)

        factory_conf_conf['parent-ref'] = iit_ca_uimod.ref

        if format_media:
            # u'EU-15-1219-ЕП-120520.p10'
            # u'EU-KEP-15-1219-ЕП-120520.p10'
            dt = datetime.datetime.now()

            csr_ua_ds_name = 'EU-{}-ЕП-{:02}{:02}{}.p10'.format(hw_serial, dt.day, dt.month, dt.year)
            csr_ua_kep_name = 'EU-KEP-{}-ЕП-{:02}{:02}{}.p10'.format(hw_serial, dt.day, dt.month, dt.year)

            zip_file_name = 'EU-{}-ЕП-{:02}{:02}{}.zip'.format(hw_serial, dt.day, dt.month, dt.year)
            csr_ua_zip_path = '/etc/edge/setupstore/admin/actions/{}/{}'.format(end_user_uimod.ref, zip_file_name)

            factory_conf_conf['zip-name'] = zip_file_name
            end_user_uiconf = IitCaEndUserUiconf.create(factory_conf)

            fs.silent_mkdir_r(os.path.dirname(csr_ua_zip_path))
            _iit_cli('key', 'create', end_user_uiconf.serial, end_user_uiconf.password,
                     csr_ua_ds_name, csr_ua_kep_name, csr_ua_zip_path)
        else:
            # check pass
            end_user_uiconf = IitCaEndUserUiconf.create(factory_conf)
            reply = _iit_cli('key', 'test-access', end_user_uiconf.serial, end_user_uiconf.password)
            if 'PASSED' not in reply:
                eventlog.log_d('fail: ParamsError ({}) media test-access failed'.format(UTYPE_IIT_CA_END_USER))
                raise ParamsError

        end_user_uimod.install(end_user_uiconf, iit_ca_uimod.ref)

        iit_ca_info['units'].append({
            'ref': end_user_uimod.ref,
            'type': end_user_uimod.type,
            'name': end_user_uiconf.pub_name,
            'role': UROLE_UNIT,
            'internal': True,
            'units': []
        })

        tr_uirefs = unet.find_urefs_r(root_info, utypes=[UTYPE_TRAFFIC_RULES])
        for tr_uiref in tr_uirefs:
            root_links.append((tr_uiref, end_user_uimod.ref))

        rootinfo.encode_info(root_unet)

        reply = {
            'info': [root_unet.root],
            'created': [end_user_uimod.ref]
        }
        return reply

    @staticmethod
    def drop_end_user(uref: str) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA_END_USER)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        unit_uconf = IitCaEndUserUiconf(unit_umod.uiconf)
        csr_file_path = '/etc/edge/setupstore/admin/actions/{}/{}'.format(unit_umod.ref, unit_uconf.csr_name)
        fs.silent_rm_r(csr_file_path)

        root_info, urefs_l1 = rootinfo.drop_unit_r(unit_umod.ref)
        unit_umod.uninstall()

        creator.generate(unit_umod.create_backup_dirpath(), urefs_l1, forced=True)

        return {
            'info': [root_info]
        }

    @staticmethod
    def testpack_end_user(uref: str) -> dict:
        unit_umod = Umod2(uref)
        try:
            unit_umod.assert_type(UTYPE_IIT_CA_END_USER)
            unit_umod.assert_ref()
        except NotFoundError:
            eventlog.log_e('fail: NotFoundError')
            raise

        unit_uconf = IitCaEndUserUiconf(unit_umod.uiconf)

        media_serial = unit_uconf.serial
        media_password = unit_uconf.password

        lock = file_lock.UimodsLocker({unit_umod})
        try:
            return ujson.loads(_iit_cli('key', 'test-all', media_serial, media_password))
        finally:
            lock.release()


if __name__ == '__main__':
    eventlog.open_log('api-iit-ca', 'debug')
    try:
        sign_uiconf = SignUimod.load_conf(default=None)
        IitCaApi.create_service_from_sign(sign_uiconf)
    finally:
        eventlog.close_log()
