import subprocess
import ujson

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def get_stats_limits(host: str, port: int) -> dict | None:
    try:
        shell_params = f'env RPC_TARGET={host}:{port} /opt/edge/bin/ipcom-cli get-info'
        shell_result = subprocess.run(shell_params, capture_output=True, text=True, encoding='ascii',
                                      shell=True, check=True, timeout=10.0)

        return ujson.loads(shell_result.stdout)

    except subprocess.TimeoutExpired as x:
        eventlog.log_e(f'fail: <{x.cmd}> (timeout:{x.timeout})')

    except subprocess.CalledProcessError as x:
        eventlog.log_e(f'fail: <{x.cmd}> (returncode:{x.returncode})')


# noinspection PyClassHasNoInit
class IpcomWagaCameraApi:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_IPCOM_WAGA)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_uconf_pub = unit_uconf['rw']
        unit_uconf_server = unit_uconf_pub['conf']['server']
        stats_jnode = get_stats_limits(unit_uconf_server['host'], unit_uconf_server['port'])

        reply = {
            'type': UTYPE_IPCOM_WAGA,
            'conf': unit_uconf_pub['conf']
            # "stats": {
            #     "uuid": "pidhorivka-uuid",
            #     "name": "WIM7",
            #     "vendor": "camea",
            #     "version": "0.0.2",
            #     "road": "M-22 Poltava-Oleksandria, km  74+810",
            #     "latitude": 49.1795,
            #     "longitude": 33.7992,
            #     "last-event": [
            #         1929233,
            #         "2021-11-08T15:40:00.075Z"
            #     ]
            # },
            # "limits": {
            #     "lanes": [
            #         {
            #             "name": "PI-KR-W1",
            #             "desc": "PP: Poltava - Oleksandria"
            #         },
            #         {
            #             "name": "PI-PL-W1",
            #             "desc": "ZP: Oleksandria - Polatava"
            #         }
            #     ]
            # }
        }
        if stats_jnode:
            reply.update(stats_jnode)

        return reply

    @staticmethod
    def save_unit(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_IPCOM_WAGA)

        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')
        fs.validate_json(new_uconf, unit_umod.schema_path)
        new_server_jnode: dict = new_uconf_conf['server']
        unit_stats: dict | None = get_stats_limits(new_server_jnode['host'], new_server_jnode['port'])

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['conf'] != new_uconf_conf:
            unit_uconf['rw']['conf'] = new_uconf_conf
            if unit_stats:
                unit_uconf['ro']['vendor'] = unit_stats['stats']['vendor']

            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        reply = {
            'type': UTYPE_IPCOM_WAGA,
            'conf': new_uconf_conf,
            # "stats": {
            #     "uuid": "pidhorivka-uuid",
            #     "name": "WIM7",
            #     "vendor": "camea",
            #     "version": "0.0.2",
            #     "road": "M-22 Poltava-Oleksandria, km  74+810",
            #     "latitude": 49.1795,
            #     "longitude": 33.7992,
            #     "last-event": [
            #         1929233,
            #         "2021-11-08T15:40:00.075Z"
            #     ]
            # },
            # "limits": {
            #     "lanes": [
            #         {
            #             "name": "PI-KR-W1",
            #             "desc": "PP: Poltava - Oleksandria"
            #         },
            #         {
            #             "name": "PI-PL-W1",
            #             "desc": "ZP: Oleksandria - Polatava"
            #         }
            #     ]
            # }
        }

        if unit_stats:
            reply.update(unit_stats)

        return reply
