from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def get_limits() -> dict:

    jnode = {
        'acquisition': {
            'accumulations': [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024],
            'oversampling': [1, 2, 4, 8],
            'point-count': {
                'min': 2,
                'max': 64
            },
            'threshold-offset': {
                'min': -5.0,
                'max': 100.0
            },
            'smoothing': {
                'min': -17,
                'max': 16
            }
        }
    }

    return jnode


class LeddarApi:
    def __init__(self):
        pass

    @staticmethod
    def make_service(term_uref: str, factory_conf: dict) -> tuple[str, dict]:
        if not unet.find_unit_r(rootinfo.decode_info().root, uref=term_uref):
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_utype = factory_conf.get('type')
        unit_name = factory_conf.get('name')

        if unit_utype != UTYPE_LEDDAR or unit_name is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        new_conf = factory_conf.get('conf') or dict()
        new_connection: dict = new_conf.get('connection') or dict()
        new_connection_type: str = new_connection.get('type', 'usb')
        new_connection_endpoint: str | None = None
        new_connection_address = 1
        if new_connection_type == 'modbus':
            new_connection_address = new_connection.get('address', 1)
            new_connection_endpoint = new_connection.get('endpoint')
            if new_connection_endpoint is None:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

        elif new_connection_type != 'usb':
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        unit_uref, root_uinfo = ph.RootInfoUiconf.add_term_service(
            term_uref, unit_utype, unit_name, [UTYPE_TRAFFIC_RULES, UTYPE_LPR])

        unit_umod = Umod(unit_uref.pid, unit_uref.type, unit_uref.id)
        unit_uconf = ph.build_term_unit_uconf(term_uref, unit_utype, unit_uref.ref, factory_conf)
        unit_umod.install(unit_uconf)

        unit_uconf = unit_umod.load_conf()
        unit_uconf['rw']['conf']['name'] = unit_name
        unit_uconf['ro']['connection'] = {
            'type': new_connection_type,
            'endpoint': new_connection_endpoint,
            'address': new_connection_address
        }

        urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})
        return unit_umod.ref, root_uinfo

    @staticmethod
    def drop_service(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LEDDAR)
        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        root_uinfo, urefs_l1 = rootinfo.drop_unit_r(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l1, forced=True)
        unit_umod.uninstall()

        return {
            'info': [root_uinfo]
        }

    @staticmethod
    def load_service(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LEDDAR)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_e('fail: NotFoundError')
            raise NotFoundError

        unit_uconf_pub = unit_uconf['rw']

        return {
            'type': UTYPE_LEDDAR,
            'conf': unit_uconf_pub['conf'],
            'limits': get_limits()
        }

    @staticmethod
    def save_service(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LEDDAR)

        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')

        def validate_body():
            fs.validate_json({'conf': new_uconf_conf}, unit_umod.schema_path)

        validate_body()

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['conf'] != new_uconf_conf:
            if new_uconf_conf['name'] != unit_uconf['rw']['conf']['name']:
                rootinfo.set_unit_name(unit_umod.ref, new_uconf_conf['name'])

            unit_uconf['rw']['conf'] = new_uconf_conf
            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        return {
            'type': UTYPE_LEDDAR,
            'conf': new_uconf_conf,
            'limits': get_limits()
        }
