import base64
import glob
import os
import subprocess
import ujson

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *

DEMO_KEYREF = 'embedded-demo'

_FPS_TO_TITLE_MAP = {
    25: 'license.lpr.group.25',
    6: 'license.lpr.group.6',
    0: 'license.lpr.group.0'
}


class _Lsvpwc:
    _HASP_FEATURE_HAS_EXPIRED_TEXT = 'Feature has expired'

    _HASP_FEATURE_NOT_FOUND_TEXT = 'Feature not found'
    _HASP_FEATURE_NOT_FOUND_CODE = 'H0031'

    _HASP_KEY_NOT_FOUND_TEXT = 'Sentinel key not found'
    _HASP_KEY_NOT_FOUND_CODE = 'H0007'

    def __init__(self):
        pass

    @classmethod
    def lsvpwc_key(cls, key_ref: str) -> list[dict]:
        """
        if licgroup present but expired, the lsvpwc returns licgroup with
        max_fps==0 and nthreads_max==0 and templates==''
        """

        """
        # /opt/edge/bin/lsvpwc -m /opt/edge/libexec/unity/aorp/modules/ -t json
        [
            {
                "key_id": "1148957907760805623",
                "key_series": "107392",
                "fps_max": 0,
                "nthreads_max": 0,
                "templates": "",
                "flags": "_"
            },
            {
                "key_id": "1087110020766043669",
                "key_series": "107392",
                "fps_max": 25,
                "nthreads_max": 8,
                "templates": "*",
                "flags": "_"
            }
        ]        
        """

        vpw_licgroups = cls._lsvpwc(by_group=False)
        vpw_licgroups = [x for x in vpw_licgroups if x['key_id'] == key_ref]
        return vpw_licgroups

    @classmethod
    def lsvpwc_key_all(cls) -> list[dict]:
        """
        # /opt/edge/bin/lsvpwc -m /opt/edge/libexec/unity/aorp/modules/ -t json -g
        [
            {
                "key_id": "----",
                "key_series": "----",
                "fps_max": 0,
                "nthreads_max": 0,
                "templates": "",
                "flags": "_"
            },
            {
                "key_id": "----",
                "key_series": "----",
                "fps_max": 25,
                "nthreads_max": 8,
                "templates": "*", // space-separated-list or *
                "flags": "_"
            }
        ]
        """

        vpw_licgroups = cls._lsvpwc(by_group=True)
        return vpw_licgroups

    @classmethod
    def _lsvpwc(cls, by_group: bool) -> list[dict]:
        lsvpwc_params = [
            '/opt/edge/bin/lsvpwc',
            '-m', '/opt/edge/libexec/unity/aorp/modules/',
            '-t', 'json'
        ]

        if by_group:
            lsvpwc_params.append('-g')

        lsvpwc_result = subprocess.run(lsvpwc_params, capture_output=True, text=True, encoding='ascii')

        if lsvpwc_result.returncode == 0:
            output = lsvpwc_result.stdout or ''
            try:
                vpw_licgroups = ujson.loads(output)
                return vpw_licgroups

            except ValueError:
                eventlog.log_e('fail: loads <{}>'.format(output))
                raise InternalError

        else:
            std_err = lsvpwc_result.stderr or ''
            if cls._HASP_FEATURE_HAS_EXPIRED_TEXT in std_err:
                eventlog.log_w(cls._HASP_FEATURE_HAS_EXPIRED_TEXT)

            elif cls._HASP_FEATURE_NOT_FOUND_TEXT in std_err or cls._HASP_FEATURE_NOT_FOUND_CODE in std_err:
                eventlog.log_w(cls._HASP_FEATURE_NOT_FOUND_TEXT)

            elif cls._HASP_KEY_NOT_FOUND_TEXT in std_err or cls._HASP_KEY_NOT_FOUND_CODE in std_err:
                eventlog.log_w(cls._HASP_KEY_NOT_FOUND_TEXT)

            else:
                eventlog.log_e('fail: loads <{}>'.format(std_err))
                raise InternalError

        return []

    @staticmethod
    def get_ui_countries_jnode(vpw_license_countries: str) -> list[dict] | None:
        # if lsvpwc returns '' (empty templates), license has expired
        if not vpw_license_countries:
            return []

        lsvpwi_params = [
            '/opt/edge/bin/lsvpwi',
            '-m', '/opt/edge/libexec/unity/aorp/modules/',
            '-h',
            '-t', 'cut',
            '-s', 'id'
        ]

        # if lsvpwc returns * activate all world
        if '*' not in vpw_license_countries:
            lsvpwi_params.append(vpw_license_countries)

        lsvpwi_result = subprocess.run(lsvpwi_params, capture_output=True, text=True, encoding='ascii')
        if lsvpwi_result.stderr:
            eventlog.log_e(f'fail: lsvpwi ({lsvpwi_result.stderr})')
            return None

        template_refs = lsvpwi_result.stdout.split('\n')[:-1]

        from collections import defaultdict
        country_to_region_to_templates_map = defaultdict(lambda: defaultdict(list))
        for template_ref in template_refs:
            items = template_ref.split('.')
            item_count = len(items)
            if item_count == 2:
                country_a2 = items[0]
                region_a2 = country_a2

            elif item_count == 3:
                country_a2 = items[0]
                region_a2 = items[1]

            else:
                eventlog.log_e('fail: unsupported (templater-ref:{})'.format(template_ref))
                raise NotImplementedError

            country_to_region_to_templates_map[country_a2][region_a2].append(template_ref)

        iso3166_params = ['/opt/edge/bin/lpr-iso3166'] + template_refs
        iso3166_result = subprocess.run(iso3166_params, capture_output=True, text=True, encoding='utf-8')
        if not iso3166_result.stdout:
            eventlog.log_e('fail: lpr-iso3166 (stdout:<>)')
            raise InternalError

        country_region_to_titles_map = ujson.loads(iso3166_result.stdout)

        countries_jnode = []
        for country_a2, region_to_templates_map in country_to_region_to_templates_map.items():
            country_title = None
            regions_jnode = []
            for region_a2, template_refs in region_to_templates_map.items():
                region_jnode = {
                    'templates': template_refs
                }

                country_title = country_region_to_titles_map[country_a2][region_a2]['country-title']
                region_title = country_region_to_titles_map[country_a2][region_a2].get('region-title')
                if region_title:
                    region_jnode['title'] = region_title

                regions_jnode.append(region_jnode)

            if 'title' in regions_jnode[0]:
                regions_jnode = sorted(regions_jnode, key=lambda region_item: region_item['title'])

            countries_jnode.append({
                'title': country_title,
                'regions': regions_jnode
            })

        countries_jnode.sort(key=lambda country_item: country_item['title'])
        return countries_jnode


class _Hasplm:
    _LIB_NAME = 'libsntl_adminapi_linux_x86_64.so'

    ATTACHED_KEYS_DIR = '/var/hasplm/installed/107392/'
    DETACHED_KEYS_DIR = '/etc/edge/hasplm/detached/'
    BASE_KEY_NAME_SUFFIX = '_base.v2c'
    UPDATED_KEY_NAME_SUFFIX = '_update*.v2c'

    def __init__(self):
        pass

    @staticmethod
    def _silent_mkdir_r(dirpath: str) -> bool:
        if not fs.path_exists(dirpath):
            fs.silent_mkdir_r(dirpath)
            return True

        return False

    @classmethod
    def test_key_attached(cls, keyref: str) -> bool:
        # mkdir for detached keys when called first time
        cls._silent_mkdir_r(cls.DETACHED_KEYS_DIR)

        path_list = glob.glob(cls.DETACHED_KEYS_DIR + cls.BASE_KEY_NAME_SUFFIX)
        detached_keyrefs = [os.path.basename(x.strip(cls.BASE_KEY_NAME_SUFFIX)) for x in path_list]

        return keyref not in detached_keyrefs

    @classmethod
    def list_detached_keyrefs(cls) -> list[str]:
        if cls._silent_mkdir_r(cls.DETACHED_KEYS_DIR):
            return []

        path_list = glob.glob(cls.DETACHED_KEYS_DIR + '*.v2c')
        detached_keyrefs = [os.path.basename(x.split('_')[0]) for x in path_list]
        detached_keyrefs = list(set(detached_keyrefs))
        return detached_keyrefs

    @staticmethod
    def remove_key(keyref: str) -> None:
        key_to_delete_filepath_set = []
        attached_key = False

        # attached "base" keys
        key_to_delete_filepath = os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_delete_filepath):
            key_to_delete_filepath_set.append(key_to_delete_filepath)
            attached_key = True

        # attached "updated" keys
        updated_key_to_delete_filepath_set = glob.glob(os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)

        for updated_key_to_delete in updated_key_to_delete_filepath_set:
            if os.path.exists(updated_key_to_delete):
                key_to_delete_filepath_set.append(updated_key_to_delete)
                attached_key = True

        # detached "base" key
        key_to_delete_filepath = os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_delete_filepath):
            key_to_delete_filepath_set.append(key_to_delete_filepath)

        # detached "updated" keys
        updated_key_to_delete_filepath_set = glob.glob(os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_delete in updated_key_to_delete_filepath_set:
            if os.path.exists(updated_key_to_delete):
                key_to_delete_filepath_set.append(updated_key_to_delete)

        # delete keys
        for filepath_to_delete in key_to_delete_filepath_set:
            if os.path.exists(filepath_to_delete):
                os.remove(filepath_to_delete)

            else:
                eventlog.log_e('fail: InternalError (file absent)')
                raise InternalError

        if attached_key:
            _Hasplm.restart()

    @classmethod
    def restart(cls) -> None:
        import platform
        if platform.machine() != 'x86_64':
            eventlog.log_w('fail: unsupported (machine is not x86)')
            return

        os.system('service aksusbd restart')

        # wait until license-manager is available
        retry_count = 0
        while retry_count < 4:
            try:
                cls._get_info()
                break
            except RuntimeError:
                retry_count += 1

        if retry_count >= 4:
            raise RuntimeError('fail: HASP license-manager unavailable')

    @classmethod
    def _get_info(cls) -> dict:
        query_text = ''.join([
            '<admin>',
            '<license_manager format="json">',
            '<element name="*"/>',
            '</license_manager>',
            '</admin>'
        ])

        reply_text = cls._call_admin_api_so(query_text)
        hasp_license_manager_info = ujson.loads(reply_text)[:-1]  # strip summary
        return hasp_license_manager_info

    @classmethod
    def _call_admin_api_so(cls, query_text: str, vendor_id: str | None = None) -> str:
        import ctypes

        lib_admin_api = ctypes.CDLL(cls._LIB_NAME)

        session_self = ctypes.POINTER(ctypes.c_void_p)()
        session_scope = '<haspscope><host>localhost</host></haspscope>'
        answer = lib_admin_api.sntl_admin_context_new_scope(ctypes.byref(session_self), session_scope)
        if answer:
            eventlog.log_e(f'fail: sntl_admin_context_new_scope ({answer})')
            raise RuntimeError(answer)

        query_scope = ''
        if vendor_id:
            query_scope = query_scope.join([
                '<haspscope>',
                f'<vendor><id>{vendor_id}</id></vendor>',
                '</haspscope>'
            ])

        try:
            reply_buffer = ctypes.POINTER(ctypes.c_char_p)()
            answer = lib_admin_api.sntl_admin_get(session_self, query_scope, query_text, ctypes.byref(reply_buffer))
            if answer:
                eventlog.log_e(f'fail: sntl_admin_get ({answer})')
                raise RuntimeError(answer)

            try:
                reply_text = ctypes.cast(reply_buffer, ctypes.c_char_p).value
            finally:
                lib_admin_api.sntl_admin_free(reply_buffer)

        finally:
            lib_admin_api.sntl_admin_context_delete(session_self)

        return cls._strip_reply_comments(reply_text)

    @staticmethod
    def _strip_reply_comments(raw_reply_text: str) -> str:
        """
        Strip comments in JSON-string and make string a JSON-array
        """
        begin_comment = 0
        end_comment = 0

        while begin_comment != -1 and end_comment != -1:
            begin_comment = raw_reply_text.find('/*')
            end_comment = raw_reply_text.find('*/')
            if begin_comment == -1 and end_comment == -1:
                continue
            s = raw_reply_text[:begin_comment] + '' + raw_reply_text[end_comment + 2:]
            raw_reply_text = s

        return '[' + raw_reply_text + ']'


class HaspRus:
    _TEMP_V2C_PATH = '/tmp/license'

    _BIN_PATH = '/opt/edge/bin/hasp_update'
    _BIN_CMD_GET_DEMO_KEYREF = 'get_demo'
    _BIN_CMD_GET_KEYS_INFO = 'l'
    _BIN_CMD_GET_HOST_C2V = 'f'
    _BIN_CMD_GET_KEY_C2V = 'i'
    _BIN_CMD_INSTALL_V2C = 'u'

    @classmethod
    def list_attached_keys_info(cls) -> dict:
        """
        # /opt/edge/bin/hasp_update l
        {"1096745228098839639":"HASP-SL","1136738148337368145":"HASP-SL"}
        #
        """
        cli_params = [cls._BIN_PATH, cls._BIN_CMD_GET_KEYS_INFO]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True, encoding='ascii')

        if cli_result.returncode != 0:
            eventlog.log_e('fail: InternalError <{}> ({}, {}, {})'.format(
                cli_result.args, cli_result.returncode, cli_result.stdout, cli_result.stderr))
            raise InternalError

        keyref_to_type_map = ujson.loads(cli_result.stdout)
        return keyref_to_type_map

    @classmethod
    def install_v2c_file(cls, v2c_path: str) -> None:
        cli_params = [cls._BIN_PATH, cls._BIN_CMD_INSTALL_V2C, v2c_path]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True, encoding='ascii')

        if cli_result.returncode != 0 or cli_result.stdout:
            eventlog.log_e('fail: InternalError <{}> ({}, {}, {})'.format(
                cli_result.args, cli_result.returncode, cli_result.stdout, cli_result.stderr))
            raise InternalError

    @classmethod
    def install_v2c_demo(cls) -> None:
        if cls.get_demo_keyref():
            eventlog.log_i('done: already activated')

        else:
            cls.install_v2c_file('/opt/edge/share/edge-embedded-demo.v2c')
            eventlog.log_i('done')

    @classmethod
    def install_v2c_base64(cls, v2c_base64: str) -> None:
        try:
            v2c_text = base64.b64decode(v2c_base64)
        except TypeError:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        with open(cls._TEMP_V2C_PATH, 'wb') as text_file:
            text_file.write(v2c_text)

        cls.install_v2c_file(cls._TEMP_V2C_PATH)

    @classmethod
    def get_host_c2v_base64(cls) -> str:
        cli_params = [cls._BIN_PATH, cls._BIN_CMD_GET_HOST_C2V]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True)

        if cli_result.returncode == 0 and cli_result.stdout:
            return str(base64.b64encode(cli_result.stdout), 'ascii')

        eventlog.log_e('fail: InternalError <{}> ({}, {}, {})'.format(
            cli_result.args, cli_result.returncode, cli_result.stdout, cli_result.stderr))
        raise InternalError

    @classmethod
    def get_key_c2v_base64(cls, key_ref: str) -> bytes:
        cli_params = [cls._BIN_PATH, cls._BIN_CMD_GET_KEY_C2V, key_ref]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True)

        if cli_result.returncode == 0 and cli_result.stdout:
            return base64.b64encode(cli_result.stdout)

        eventlog.log_e('fail: InternalError <{}> ({}, {}, {})'.format(
            cli_result.args, cli_result.returncode, cli_result.stdout, cli_result.stderr))
        raise InternalError

    @classmethod
    def get_demo_keyref(cls) -> str | None:
        """
        Returns haspid of demokey if it is installed, returns None if demo-key not found
        """
        cli_params = [cls._BIN_PATH, cls._BIN_CMD_GET_DEMO_KEYREF]
        cli_result = subprocess.run(cli_params, capture_output=True, text=True, encoding='ascii')
        if cli_result.returncode != 0:
            eventlog.log_e('fail: InternalError <{}> ({}, {}, {})'.format(
                cli_result.args, cli_result.returncode, cli_result.stdout, cli_result.stderr))
            raise InternalError

        demo_keyrefs = ujson.loads(cli_result.stdout)
        if len(demo_keyrefs) > 0:
            return demo_keyrefs[0]


# noinspection PyClassHasNoInit
class LicenseApi:
    @classmethod
    def get_keys_info(cls) -> list[dict]:
        attached_keys_info = cls.get_attached_uikeys_info()
        detached_keys_info = cls.get_detached_uikeys_info()
        return attached_keys_info + detached_keys_info

    @staticmethod
    def get_used_license_count(license_group: str | None = None) -> int:
        lr_uiconf_path_list = glob.glob('/etc/edge/setupstore/admin/actions/pipeline/*/lpr/*.json')

        used_license_count = 0
        for uiconf_path in lr_uiconf_path_list:
            uiconf = fs.load_json3(uiconf_path, default=None)
            if uiconf is None:
                eventlog.log_e('fail: InternalError')
                raise InternalError

            uiconf_pub: dict = uiconf['rw']
            if license_group is None:
                used_license_count += uiconf_pub['license-count']

            elif license_group == uiconf_pub['license-group']:
                used_license_count += uiconf_pub['license-count']

        return used_license_count

    @staticmethod
    def get_total_license_count() -> int:
        total_license_count = 0
        for license_key in _Lsvpwc.lsvpwc_key_all():
            total_license_count += license_key['nthreads_max']

        return total_license_count

    @classmethod
    def test_lpr_license_available(cls) -> None:
        total_count = cls.get_total_license_count()
        used_count = cls.get_used_license_count()
        eventlog.log_d('license total_thread_count: {}, used_thread_count: {}'.format(
            total_count, used_count))

        if used_count >= total_count:
            eventlog.log_d('fail: free lpr licence is not available')
            raise LicServiceError

        eventlog.log_d('done')

    @staticmethod
    def get_attached_uikeys_info() -> list[dict]:
        # TODO removable / detachable
        keyref_to_type_map = HaspRus.list_attached_keys_info()

        attached_uikeys_info = []
        for keyref, keytype in keyref_to_type_map.items():
            sw_key = keytype != 'HASP_HL'
            attached_uikeys_info.append({
                'id': keyref,
                'removable': False,  # sw_key,
                'detachable': False,  # sw_key,
                'attached': _Hasplm.test_key_attached(keyref)
            })

        demo_keyref = HaspRus.get_demo_keyref()
        if demo_keyref:
            # replace demo-key jnode with stub
            first_demo_ui_key_jnode = [x for x in attached_uikeys_info if x['id'] == demo_keyref][0]

            attached_uikeys_info.remove(first_demo_ui_key_jnode)
            attached_uikeys_info.append({
                'id': DEMO_KEYREF,
                'removable': False,
                'detachable': False,
                'attached': True
            })

        else:
            attached_uikeys_info.append({
                'id': DEMO_KEYREF,
                'removable': False,
                'detachable': True,
                'attached': False
            })

        return attached_uikeys_info

    @staticmethod
    def get_detached_uikeys_info() -> list[dict]:
        # TODO removable / detachable
        detached_keyrefs = _Hasplm.list_detached_keyrefs()
        detached_uikeys_info = [{
            'id': keyref,
            'removable': False,  # True,
            'detachable': False,  # True,
            'attached': False
        } for keyref in detached_keyrefs]

        return detached_uikeys_info

    @staticmethod
    def get_host_c2v_base64() -> str:
        return HaspRus.get_host_c2v_base64()

    @staticmethod
    def get_key_c2v_base64(keyref: str) -> str:
        # check if key-id valid
        vpw_licgroups = _Lsvpwc.lsvpwc_key(keyref)
        if not vpw_licgroups:
            eventlog.log_e('fail: NotFoundError (key-ref:{})'.format(keyref))
            raise NotFoundError

        return str(HaspRus.get_key_c2v_base64(keyref), 'ascii')

    @classmethod
    def install_v2c_base64(cls, v2c_base64: str) -> None:
        HaspRus.install_v2c_base64(v2c_base64)
        eventlog.log_d('done')

    @staticmethod
    def attach_key(keyref: str) -> None:
        if keyref == 'embedded-demo':
            HaspRus.install_v2c_demo()
            return

        # try attach other then 'embedded-demo' key
        if keyref not in _Hasplm.list_detached_keyrefs():
            eventlog.log_e('fail: NotFoundError (requested key %s is absent)' % keyref)
            raise NotFoundError

        _silent_mkdir_r(_Hasplm.DETACHED_KEYS_DIR)

        key_to_attach_filepath_set = []

        key_to_attach_filepath = os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_attach_filepath):
            key_to_attach_filepath_set.append(key_to_attach_filepath)

        updated_key_to_attach_filepath_set = glob.glob(os.path.join(_Hasplm.DETACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_attach in updated_key_to_attach_filepath_set:
            if os.path.exists(updated_key_to_attach):
                key_to_attach_filepath_set.append(updated_key_to_attach)

        if key_to_attach_filepath_set:
            fs.move_files(key_to_attach_filepath_set, _Hasplm.ATTACHED_KEYS_DIR)
            _Hasplm.restart()

    @staticmethod
    def detach_key(keyref: str) -> None:
        if keyref not in list(HaspRus.list_attached_keys_info().keys()):
            eventlog.log_e('fail: NotFoundError (key %s is absent)' % keyref)
            raise NotFoundError

        _silent_mkdir_r(_Hasplm.DETACHED_KEYS_DIR)

        key_to_detach_filepath_set = []

        key_to_detach_filepath = os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref + _Hasplm.BASE_KEY_NAME_SUFFIX)
        if os.path.exists(key_to_detach_filepath):
            key_to_detach_filepath_set.append(key_to_detach_filepath)

        updated_key_to_detach_filepath_set = glob.glob(os.path.join(_Hasplm.ATTACHED_KEYS_DIR, keyref)
                                                       + _Hasplm.UPDATED_KEY_NAME_SUFFIX)
        for updated_key_to_detach in updated_key_to_detach_filepath_set:
            if os.path.exists(updated_key_to_detach):
                key_to_detach_filepath_set.append(updated_key_to_detach)

        if key_to_detach_filepath_set:
            fs.move_files(key_to_detach_filepath_set, _Hasplm.DETACHED_KEYS_DIR)
            _Hasplm.restart()
            return

        eventlog.log_e('fail: detach key')
        raise InternalError

    @staticmethod
    def remove_key(keyref: str) -> None:
        _Hasplm.remove_key(keyref)

    @staticmethod
    def get_licinfo_key_keyref(keyref: str) -> dict:
        if keyref == DEMO_KEYREF:
            keyref = HaspRus.get_demo_keyref()

        vpw_licgroups = _Lsvpwc.lsvpwc_key(keyref)
        if not vpw_licgroups:
            eventlog.log_e('fail: NotFoundError (keyref:{})'.format(keyref))
            raise NotFoundError

        from collections import defaultdict
        acc = defaultdict(lambda: defaultdict(set))

        for licgroup in vpw_licgroups:
            group_ref = licgroup['fps_max']
            if acc[group_ref].get('templates', None) is None:
                acc[group_ref]['templates'] = licgroup['templates'].split(' ')

            else:
                acc[group_ref]['templates'] += licgroup['templates'].split(' ')

            if acc[group_ref].get('nthreads_max', None) is None:
                acc[group_ref]['nthreads_max'] = licgroup['nthreads_max']

            else:
                acc[group_ref]['nthreads_max'] += licgroup['nthreads_max']

        ui_groups_info = []
        for licgroup, value in acc.items():
            ui_groups_info.append({
                'id': str(licgroup),
                'title': _FPS_TO_TITLE_MAP[licgroup],
                'total-license-count': value['nthreads_max'],
                'fps-limit': licgroup,
                'countries': _Lsvpwc.get_ui_countries_jnode(' '.join(value['templates']))
            })

        if ui_groups_info:
            return {'lpr': ui_groups_info}

    @staticmethod
    def get_licinfo_key_all(key: str = 'lpr') -> dict:
        vpw_licgroups = _Lsvpwc.lsvpwc_key_all()

        def get_jnode(lic_group):
            """
            :type lic_group: dict
            :rtype: dict
            """
            templates = lic_group['templates'] if lic_group['fps_max'] != 0 else '*'
            used_count = LicenseApi.get_used_license_count(str(lic_group['fps_max']))
            jnode = {
                'id': str(lic_group['fps_max']),
                'title': _FPS_TO_TITLE_MAP[lic_group['fps_max']],
                'total-license-count': lic_group['nthreads_max'],
                'used-license-count': used_count,
                'free-license-count': lic_group['nthreads_max'] - used_count,
                'fps-limit': lic_group['fps_max'],
                'countries': _Lsvpwc.get_ui_countries_jnode(templates)
            }
            return jnode

        ui_groups_info = list(map(get_jnode, vpw_licgroups))

        # drop expired licenses from list
        # expired_license_group_list = filter(lambda x: x['fps-limit'] == 0, ui_groups_info)
        # for expired in expired_license_group_list:
        #     ui_groups_info.remove(expired)

        return {key: ui_groups_info}


def _silent_mkdir_r(dirpath: str) -> bool:
    if not os.path.exists(dirpath):
        fs.silent_mkdir_r(dirpath)
        return True

    return False
