import math
import os
import subprocess
import ujson

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.dgstats.camera import CamDgStats
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def _get_live_image(ui_pid: str) -> str | None:
    # get last overlay imagepath
    or_id = int(ui_pid) * 10
    or_store = f'/var/spool/edge/pipeline-{ui_pid}/or{or_id}-or{or_id}-store/'

    file_names: list[str] = os.listdir(or_store)

    if len(file_names):
        def decode_ts(file_name: str) -> int:
            return int(file_name.split('-')[-1])

        blob_name = max(file_names, key=decode_ts)
        blob_path = os.path.join(or_store, blob_name)
        return blob_path


def _landmark_cli(query_params: list[str], sensor_w_px: float, sensor_h_px: float,
                  focal_dist_px: float) -> dict | None:

    cam_fov_h: float = math.degrees(2 * math.atan2(sensor_w_px, 2 * focal_dist_px))
    cam_fov_v: float = math.degrees(2 * math.atan2(sensor_h_px, 2 * focal_dist_px))

    args: dict = {
        '--fov-vertical': cam_fov_v,
        '--fov-horizontal': cam_fov_h,

        '--position-top': None,

        '--rotation-landmark-image': None,
        '--rotation-landmark-forward': None,
        '--rotation-landmark-left': None
    }

    for name_value in query_params:
        name, value = name_value.split('=', 1)

        if name == 'camera-up':
            args['--position-top'] = float(value)

        elif name == 'landmark-image':
            if value in ['bottom-left', 'bottom-right']:
                args['--rotation-landmark-image'] = value

        elif name == 'landmark-forward':
            args['--rotation-landmark-forward'] = float(value)

        elif name == 'landmark-left':
            args['--rotation-landmark-left'] = float(value)

    invalid_args = [arg_name_arg_value for arg_name_arg_value in iter(args.items()) if arg_name_arg_value[1] is None]
    if invalid_args:
        for invalid_arg_name, _ in invalid_args:
            eventlog.log_d(f'fail: param <{invalid_arg_name}> is absent')
        raise ParamsError

    else:
        cli_params = ['/opt/edge/bin/landmark_v2']
        for name, value in args.items():
            cli_params.append(f'{name}={value}')

        try:
            cli_result = subprocess.run(
                cli_params, capture_output=True, text=True, encoding='ascii',
                check=True, timeout=1.0)

            return ujson.loads(cli_result.stdout)

        except subprocess.TimeoutExpired as x:
            eventlog.log_e(f'fail: <{x.cmd}> (timeout:{x.timeout})')

        except subprocess.CalledProcessError as x:
            eventlog.log_e(f'fail: <{x.cmd}> (returncode:{x.returncode})')

        raise InternalError


# noinspection PyClassHasNoInit
class LocationApi:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LOCATION)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        cam_dgstats = CamDgStats.create(unit_umod.pid)
        cam_stats_jnode = cam_dgstats.get_stats()

        return {
            'type': UTYPE_LOCATION,
            'conf': unit_uconf['rw']['args'],
            'stats': cam_stats_jnode
        }

    @staticmethod
    def save_unit(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LOCATION)

        new_uconf_conf = new_uconf.get('conf')
        fs.validate_json({'args': new_uconf_conf}, unit_umod.schema_path)

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['args'] != new_uconf_conf:
            unit_uconf['rw']['args'] = new_uconf_conf

            from edge.support import hg
            with hg.ConfRepo() as (hg_dir, hg_client):
                hgjpeg_path = fs.join_path(hg_dir, unit_umod.pid, '0.jpeg')
                hgconf_path = fs.join_path(hg_dir, unit_umod.pid, f'{UTYPE_LOCATION}.0.json')

                add_files: bool = not fs.path_exists(hgconf_path)

                jpeg_path: str | None = _get_live_image(unit_umod.pid)
                if jpeg_path is None:
                    eventlog.log_e(f'fail: get snapshot (term-ref:{unit_umod.pid})')
                    raise InternalError

                fs.silent_cp(jpeg_path, hgjpeg_path)
                fs.write_json_file(unit_uconf, hgconf_path, pretty=True)

                if add_files:
                    hg_client.add([hgconf_path.encode('ascii'), hgjpeg_path.encode('ascii')])

                hg_client.commit(message=f'edit({unit_umod.pid}/location.0)', user='admin')

                uref_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
                creator.generate(unit_umod.create_backup_dirpath(), uref_l0_l1, {unit_umod.ref: unit_uconf})

        cam_stats_jnode = CamDgStats.create(unit_umod.pid).get_stats()
        return {
            'type': UTYPE_LOCATION,
            'conf': new_uconf_conf,
            'stats': cam_stats_jnode
        }

    @staticmethod
    def estimate_rotation(uref: str, query_params: list[str]) -> dict:
        umod = Umod.create(uref)
        umod.assert_type(UTYPE_LOCATION)

        def get_sensorwpx_sensorhpx_focalpx(term_pid: str) -> tuple[float, float, float]:
            from edge.webserver.sysconf.uiconf.types.jnode.intrinsics import \
                decode__intrinsic_sensorwpx_sensorhpx_focalpx

            mv_umod = Umod(term_pid, UTYPE_MV_CAMERA, '0')
            mv_uconf = mv_umod.load_conf(default=None)
            if mv_uconf:
                from edge.webserver.sysconf.uiconf.mv_camera import MvCameraUconf
                return decode__intrinsic_sensorwpx_sensorhpx_focalpx(MvCameraUconf(mv_uconf).intrinsics)

            rtsp_umod = Umod(term_pid, UTYPE_RTSP_CAMERA, '0')
            rtsp_uconf = rtsp_umod.load_conf(default=None)
            if rtsp_uconf:
                from edge.webserver.sysconf.uiconf.rtsp_camera import RtspCameraUconf
                return decode__intrinsic_sensorwpx_sensorhpx_focalpx(RtspCameraUconf(rtsp_uconf).intrinsics)

            eventlog.log_e('fail: NotImplementedError')
            raise NotImplementedError

        cli_result: dict = _landmark_cli(query_params, *get_sensorwpx_sensorhpx_focalpx(umod.pid))
        rotation_auto: dict = cli_result['unit-rotation-auto']

        return {
            'camera': {
                'rotation': {
                    'elevation': rotation_auto['elevation'],
                    'azimuth': rotation_auto['azimuth']
                }
            }
        }
