from edge import eventlog
from edge.support import fs
from edge.support.mask import MaskRegion
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.api_license import LicenseApi
from edge.webserver.sysconf.dgstats.camera import CamDgStats
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class LprApi:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LPR)

        cam_dgstats = CamDgStats.create(unit_umod.pid)
        #cam_dgstats.assert_stream_active()
        cam_stats_jnode = cam_dgstats.get_stats()

        unit_uconf = unit_umod.load_conf()
        unit_uconf_pub = unit_uconf['rw']

        lic_limits = LicenseApi.get_licinfo_key_all('license')

        selected_group = [x for x in lic_limits['license'] if x['id'] == unit_uconf_pub['license-group']]
        if not selected_group:
            unit_uconf_pub['license-group'] = '0'
            unit_uconf_pub['license-count'] = 0

        reply_jnode = {
            'type': UTYPE_LPR,
            'conf': unit_uconf_pub,
            'limits': lic_limits,
            'stats': cam_stats_jnode
        }

        return reply_jnode

    @classmethod
    def save_unit(cls, uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_LPR)

        cam_dgstats = CamDgStats.create(unit_umod.pid)
        #cam_dgstats.assert_stream_active()

        new_uconf_conf = new_uconf.get('conf')
        fs.validate_json(new_uconf_conf, unit_umod.schema_path)

        def validate_azone():
            azone_jnode = new_uconf_conf['plate-recognizer']['image-azone']
            try:
                # validate self-intersection of include-mas
                include_jnode = azone_jnode['include-mask']
                mask_region = MaskRegion(include_jnode)

                if not mask_region.test_valid():
                    eventlog.log_d('fail: http400 (invalid include-mask)')
                    raise ParamsError

            except (ValueError, IndexError):
                eventlog.log_d('include-mask contains less then 3 vertexes. using whole image as include-mask')
                azone_jnode['include-mask'] = []

            try:
                # validate self-intersection of exclude-mask
                exclude_jnode = azone_jnode['exclude-mask']

                for exclude_mask in exclude_jnode:
                    mask_region = MaskRegion(exclude_mask)
                    if not mask_region.test_valid():
                        eventlog.log_d('fail: http400 (invalid exclude-mask)')
                        raise ParamsError

            except (ValueError, IndexError):
                eventlog.log_d('at least one of poly-regions in exclude-mask contains less then 3 vertexes.'
                               ' ignoring exclude-mask')
                azone_jnode['exclude-mask'] = []

        lic_limits = LicenseApi.get_licinfo_key_all('license')

        def validate_lic() -> None:
            lic_group = new_uconf_conf['license-group']
            selected_group = [x for x in lic_limits['license'] if x['id'] == lic_group]
            if lic_group == '0' or len(selected_group) == 0:
                eventlog.log_e('fail: http400 (invalid license group)')
                raise ParamsError

        validate_azone()
        validate_lic()

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw'] != new_uconf_conf:
            unit_uconf['rw'] = new_uconf_conf
            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        return {
            'type': UTYPE_LPR,
            'conf': new_uconf_conf,
            'limits': lic_limits,
            'stats': cam_dgstats.get_stats()
        }
