import ujson

from typing import Required, TypedDict

from edge import eventlog
from edge.support import fs
from edge.support import sh
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgstats.camera import CamDgStats
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class MvCameraApi:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_MV_CAMERA)

        unit_uconf = unit_umod.load_conf()
        unit_uconf_pub = unit_uconf['rw']

        stats_jnode = CamDgStats.create(unit_umod.pid).get_stats()
        stats_jnode['vendor'] = unit_uconf_pub.pop('vendor')
        stats_jnode['model'] = unit_uconf_pub.pop('model')
        stats_jnode['serial'] = unit_uconf_pub.pop('serial')
        unit_uconf_pub.pop('type', None)

        return {
            'type': UTYPE_MV_CAMERA,
            'conf': unit_uconf_pub,
            'stats': stats_jnode,
            'limits': unit_umod.schema_limits
        }

    @classmethod
    def save_unit(cls, uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_MV_CAMERA)

        unit_uconf = unit_umod.load_conf()
        unit_uconf_pub = unit_uconf['rw']

        new_uconf_conf = new_uconf.get('conf')
        new_uconf_conf['vendor'] = unit_uconf_pub.get('vendor')
        new_uconf_conf['model'] = unit_uconf_pub.get('model')
        new_uconf_conf['serial'] = unit_uconf_pub.get('serial')

        def validate_uiconf_pub() -> None:
            fs.validate_json(new_uconf_conf, unit_umod.schema_path)

            # ROI checks - (width_offset + ROI_width) < max_width
            from copy import deepcopy
            roi_test = deepcopy(new_uconf_conf)
            roi_test['videostream']['resolution']['w'] += roi_test['videostream']['resolution']['x']
            roi_test['videostream']['resolution']['h'] += roi_test['videostream']['resolution']['y']
            fs.validate_json(roi_test, unit_umod.schema_path)

            control = new_uconf_conf['videostream']['control']
            if control['profile'] == 'auto-70-150':
                min_gain = control.get('gain-auto-min')
                max_gain = control.get('gain-auto-max')

                min_exposure = control.get('exposure-auto-min')
                max_exposure = control.get('exposure-auto-max')

                if None in [min_gain, max_gain, min_exposure, max_exposure]:
                    eventlog.log_d(
                        'fail: ParamsError (gain-auto-min:{}, gain-auto-max:{}, '
                        'exposure-auto-min:{}, exposure-auto-max:{})'.format(
                            min_gain, max_gain, min_exposure, max_exposure))
                    raise ParamsError

                if min_gain > max_gain:
                    # GrabControl works bad when min_value == max_value
                    eventlog.log_d('fail: ParamsError (gain-auto-min:{}, gain-auto-max:{})'.format(
                        min_gain, max_gain))
                    raise ParamsError

                if min_exposure > max_exposure:
                    # GrabControl works bad when min_value == max_value
                    eventlog.log_d('fail: ParamsError (exposure-auto-min:{}, exposure-auto-max:{})'.format(
                        min_exposure, max_exposure))
                    raise ParamsError

            elif control['profile'] == 'manual':
                gain = control.get('gain-manual')
                exposure = control.get('exposure-manual')
                gamma = control.get('gamma-manual')
                if None in [gain, exposure, gamma]:
                    eventlog.log_d(f'fail: ParamsError (gain:{gain}, exposure:{exposure}, gamma:{gamma})')
                    raise ParamsError

        validate_uiconf_pub()

        restart_setupstore = []
        if unit_uconf_pub['videostream'] != new_uconf_conf['videostream']:
            if cls._test_gc_mode_changed(unit_uconf_pub, new_uconf_conf):
                restart_setupstore.append(True)

            unit_uconf['rw'] = new_uconf_conf
            uirefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), uirefs_l0_l1, {unit_umod.ref: unit_uconf})

        if restart_setupstore:
            dg.LifeCycle.sighup_ss_dgmod()

        stats_jnode = CamDgStats.create(unit_umod.pid).get_stats()
        stats_jnode['vendor'] = new_uconf_conf.pop('vendor')
        stats_jnode['model'] = new_uconf_conf.pop('model')
        stats_jnode['serial'] = new_uconf_conf.pop('serial')
        new_uconf_conf.pop('type', None)

        return {
            'type': UTYPE_MV_CAMERA,
            'conf': new_uconf_conf,
            'stats': stats_jnode,
            'limits': unit_umod.schema_limits
        }

    @staticmethod
    def _test_gc_mode_changed(uiconf_pub: dict, new_uiconf_pub: dict) -> bool:
        old_gc_mode = uiconf_pub['videostream']['control']['profile'] == 'auto-70-150'
        new_gc_mode = new_uiconf_pub['videostream']['control']['profile'] == 'auto-70-150'
        return old_gc_mode != new_gc_mode


class CameraInfo(TypedDict, total=False):
    vendor: Required[str]
    model: Required[str]
    serial: Required[str]
    status: Required[str]


# noinspection PyClassHasNoInit
class MvCameraSearchApi:
    CAM_STATUS_USED = 'used'
    CAM_STATUS_UNUSED = 'unused'
    CAM_STATUS_UNDEFINED = 'undefined'

    VENDOR_FLIR = 'flir'
    VENDOR_POINTGRAY = 'pointgrey'
    VENDOR_BASLER = 'basler'
    VENDOR_DALSA = 'dalsa'
    VENDOR_JAI = 'jai'
    VENDOR_GALAXY = 'galaxy'
    VENDOR_LUCID = 'lucid'

    _VENDORS = (
        VENDOR_FLIR,
        VENDOR_POINTGRAY,
        VENDOR_BASLER,
        VENDOR_DALSA,
        VENDOR_JAI,
        VENDOR_GALAXY,
        VENDOR_LUCID)

    @staticmethod
    def _test_cam_used(serial: str) -> bool:
        root_info = rootinfo.decode_info().root
        cam_uimods = unet.map_units_r(
            root_info, lambda x: Umod.create(x['ref']),
            utypes=[UTYPE_MV_CAMERA])

        for uimod in cam_uimods:
            uiconf = uimod.load_conf(default=None)
            if uiconf is None:
                pass

            elif serial == uiconf['rw']['serial']:
                return True

        return False

    @classmethod
    def _test_cam_unused(cls, vendor: str, serial: str) -> bool:
        if vendor not in cls._VENDORS:
            eventlog.log_d('fail: ParamsError (invalid vendor)')
            raise ParamsError

        stdout = sh.exec_cli('/opt/edge/bin/mediagrabber-info',
                             ['list', f'--vendor={vendor}'],
                             tout_sec=60)

        # try:
        #     # workaround for JAI cameras, strip anything before JSON
        #     stdout = stdout[stdout.index('{'):] //TODO: change { to [ ?
        # except ValueError:
        #     pass

        import ujson
        try:
            cams_jnode = ujson.loads(stdout)
            for cam_jnode in cams_jnode:
                if cam_jnode['serial'] == serial:
                    cam_used = cam_jnode.get('used')
                    if cam_used is None:
                        return False
                    else:
                        return not cam_used

            raise InternalError

        except ValueError:
            # camera status is 'undefined' if:
            # - it is not found in mv-camera-uiconf
            # - camera-info is not a valid JSON
            pass

        return False

    @classmethod
    def _get_cam_status(cls, vendor: str, serial: str) -> str:
        if cls._test_cam_used(serial):
            return cls.CAM_STATUS_USED

        elif cls._test_cam_unused(vendor, serial):
            return cls.CAM_STATUS_UNUSED

        else:
            return cls.CAM_STATUS_UNDEFINED

    @classmethod
    def _get_cams_info(cls, vendor: str) -> list[CameraInfo]:
        stdout = sh.exec_cli('/opt/edge/bin/mediagrabber-info',
                             ['list', f'--vendor={vendor}'],
                             tout_sec=60)
        jnode = ujson.loads(stdout)

        cams_info = list[CameraInfo]()
        for cam_jnode in jnode:
            model = cam_jnode['model']
            serial = cam_jnode['serial']
            status = cls._get_cam_status(vendor, serial)

            cams_info.append({
                'vendor': vendor,
                'model': model,
                'serial': serial,
                'status': status
            })

        return cams_info

    @classmethod
    def list_cams(cls) -> list[CameraInfo]:
        cams_info = list[CameraInfo]()
        for vendor in cls._VENDORS:
            cams_info += cls._get_cams_info(vendor)

        return cams_info
