from edge import eventlog
from edge.support import fs, sh
from edge.types.exceptions import *

_UICONF_PATH = '/etc/edge/setupstore/admin/actions/openvpn.json'
_DGCONF_PATH = '/etc/default/openvpn'


def _get_new_ref(uiconf_profiles: list[dict]) -> str:
    used_refs = set([int(x['ref']) for x in uiconf_profiles])

    ref = None
    for candidate in range(0, 32):
        if candidate not in used_refs:
            return str(candidate)

    if ref is None:
        eventlog.log_e('fail: get new-ref')
        raise InternalError


def _openvpn_reload_restart():
    sh.run_silent('service openvpn stop')
    sh.run_silent('systemctl daemon-reload')
    sh.run_silent('service openvpn start')


def _mk_profile_path(ref: str) -> str:
    return f'/etc/openvpn/{ref}.conf'


def _mk_user_path(ref: str) -> str:
    return f'/etc/openvpn/{ref}.user'


def _get_active_profiles(jnode: dict) -> set[str]:
    active_ref_set = set()
    for profile_jnode in jnode['rw']['conf']['profiles']:
        if profile_jnode['stats']['active']:
            active_ref_set.add(profile_jnode['ref'])

    return active_ref_set


def _set_autostart_profile(active_profiles: set[str]) -> None:
    # Edit profile autostart.
    with open(_DGCONF_PATH, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    for old_line in old_defaults_vpn_lines:
        if not old_line.startswith('AUTOSTART='):
            new_defaults_vpn_lines.append(old_line)

    if len(active_profiles) == 0:
        autostart_line = 'AUTOSTART="none"\n'

    else:
        autostart_line = f'AUTOSTART="{" ".join(active_profiles)}"\n'

    new_defaults_vpn_lines.append(autostart_line)

    with open(_DGCONF_PATH, 'w') as new_default_vpn_file:
        new_default_vpn_file.write(''.join(new_defaults_vpn_lines))


def _add_auth_to_profile(conf_path: str, user_path: str,
                         prof_user: str | None = None,
                         prof_pass: str | None = None):
    """
    :raises: BadRequestError
    """
    with open(conf_path, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    auth_user_path = False
    for old_line in old_defaults_vpn_lines:
        if old_line == 'auth-user-pass\n':
            auth_user_path = True
            if None in [prof_user, prof_pass]:
                eventlog.log_e('fail: user/pass not defined')
                raise ParamsError

            auth_conf = f'auth-user-pass {user_path}\n'
            new_defaults_vpn_lines.append(auth_conf)

        else:
            new_defaults_vpn_lines.append(old_line)

    if auth_user_path:
        with open(conf_path, 'w') as new_default_vpn_file:
            new_default_vpn_file.write(''.join(new_defaults_vpn_lines))

        with open(user_path, 'w') as user_vpn_file:
            user_vpn_file.write(f'{prof_user}\n{prof_pass}\n')


# noinspection PyClassHasNoInit,PyUnusedLocal
class OpenVpnApi:
    @staticmethod
    def load_unit(uref: str) -> dict | None:
        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        if unit_uconf is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        return unit_uconf['rw']

    @staticmethod
    def make_profile(uuiref: str, prof_name: str, conf_path: str,
                     prof_user: str | None = None,
                     prof_pass: str | None = None) -> dict:

        # TODO: validate name is unique
        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        unit_uconf_pub: dict = unit_uconf['rw']

        unit_conf: dict = unit_uconf_pub['conf']
        unit_conf_profiles: list = unit_conf['profiles']

        if len([x for x in unit_conf_profiles if x['name'] == prof_name]):
            eventlog.log_d(f'fail: ParamsError (name <{prof_name}> already exist)')
            raise ParamsError

        new_profile_ref = _get_new_ref(unit_conf_profiles)

        user_path = _mk_user_path(new_profile_ref)
        # check if auth-user-path and user + pass not None
        _add_auth_to_profile(conf_path, user_path, prof_user, prof_pass)

        new_profile_path = _mk_profile_path(new_profile_ref)
        fs.silent_cp(conf_path, new_profile_path)

        unit_conf_profiles.append({
            'ref': new_profile_ref,
            'name': prof_name,
            'internal': False,
            'stats': {
                'active': False
            }
        })

        fs.write_json_file(unit_uconf, _UICONF_PATH, pretty=True)
        return unit_uconf_pub

    @staticmethod
    def drop_profile(uref: str, profile_ref: str) -> dict:
        # TODO: check if profile exist
        # TODO: check if profile internal
        profile_path = _mk_profile_path(profile_ref)
        user_path = _mk_user_path(profile_ref)

        fs.silent_rm_r(profile_path)
        fs.silent_rm_r(user_path)

        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        unit_uconf_pub: dict = unit_uconf['rw']
        unit_conf: dict = unit_uconf_pub['conf']

        new_profiles = [x for x in unit_conf['profiles'] if x['ref'] != profile_ref]

        if len(new_profiles) == len(unit_conf['profiles']):
            eventlog.log_e(f'fail: NotFoundError (profile-ref:{profile_ref})')
            raise NotFoundError

        rm_profile_info: dict = [x for x in unit_conf['profiles'] if x['ref'] == profile_ref][0]
        if rm_profile_info['internal']:
            eventlog.log_e(f'fail: ParamsError (profile-ref:{profile_ref}, internal:true')
            raise ParamsError

        unit_conf['profiles'] = new_profiles

        active_profiles = _get_active_profiles(unit_uconf)
        _set_autostart_profile(active_profiles)
        _openvpn_reload_restart()

        fs.write_json_file(unit_uconf, _UICONF_PATH, pretty=True)
        return unit_uconf_pub

    @staticmethod
    def make_profile_active(uref: str, profile_ref: str, active: bool) -> dict:
        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        unit_uconf_pub = unit_uconf['rw']  # type: dict
        unit_conf = unit_uconf_pub['conf']  # type: dict

        profiles_info = [x for x in unit_conf['profiles'] if x['ref'] == profile_ref]
        if len(profiles_info) == 0:
            eventlog.log_e(f'fail: NotFoundError (profile-ref:{profile_ref})')
            raise NotFoundError

        for profile_info in profiles_info:  # type: dict
            profile_info['stats']['active'] = active

            active_profiles = _get_active_profiles(unit_uconf)
            _set_autostart_profile(active_profiles)
            _openvpn_reload_restart()

            fs.write_json_file(unit_uconf, _UICONF_PATH, pretty=True)
            return unit_uconf_pub

        eventlog.log_e(f'fail: NotFoundError (profile-ref:{profile_ref})')
        raise NotFoundError
