from edge import eventlog
from edge.support import db, fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgstats.camera import CameraStats, get_terminal_vs0_stats
from edge.webserver.sysconf.uimod import rootinfo, unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import UTYPE_TERMINAL, UTYPE_HTTP_STREAMING, UTYPE_RTSP_STREAMING


def _get_terminal_stats(term_uref: str) -> dict:
    root_uinfo = rootinfo.decode_info().root
    term_umod = Umod.create(term_uref)

    media_infos = list[CameraStats]()

    term_uinfo = unet.find_unit_r(root_uinfo, uref=term_umod.ref)
    if term_uinfo is None:
        eventlog.log_d(f'fail: NotFoundError ({UTYPE_TERMINAL})')
        raise NotFoundError

    http_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_HTTP_STREAMING)
    if http_uinfo is not None:
        http_umod = Umod.create(http_uinfo['ref'])
        http_uconf = http_umod.load_conf()
        http_uconf_pub = http_uconf['rw']
        http_enabled = http_uconf_pub.get('enabled', False)
        if http_enabled:
            vs0_stats = get_terminal_vs0_stats(term_umod.ref)
            media_infos.append(vs0_stats)

    rtsp_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_RTSP_STREAMING)
    if rtsp_uinfo is not None:
        rtsp_umod = Umod.create(rtsp_uinfo['ref'])
        rtsp_uconf = rtsp_umod.load_conf()
        rtsp_uconf_pub = rtsp_uconf['rw']
        rtsp_enabled = rtsp_uconf_pub.get('enabled', False)
        if rtsp_enabled:
            rtsp_stats = {
                'stream': rtsp_uconf_pub.get('streaming', {}).get('url', 'try again later')  # TODO
            }

            media_infos.append(rtsp_stats)

    return {
        'media': media_infos
    }


class OriginsApi:
    @staticmethod
    def load_unit() -> dict:
        origins = db.get_origins() or list()

        for origin_jnode in origins:
            term_uuid = origin_jnode['uuid']
            term_uref = ph.map_term_uuid_to_uref(term_uuid)

            origin_jnode['services'] = {
                'event': ['feed', 'arch', 'stats']
            }

            if term_uref is None:
                continue

            origin_jnode['services']['media'] = ['live']

            term_stats_jnode = _get_terminal_stats(term_uref)
            origin_jnode['stats'] = term_stats_jnode

            term_pid = Umod.create(term_uref).pid
            sc_path = f'/var/spool/edge/pipeline-{term_pid}/zabbix/sc0'
            sc_stats = fs.load_json3(sc_path, default=None)
            if sc_stats is None:
                continue

            for unit_jnode in sc_stats['units']:
                if unit_jnode.get('type', '') != 'geolocation':
                    continue

                last_geo = unit_jnode.get('stat', dict()).get('last-geoposition')
                if last_geo is None:
                    continue

                origin_jnode['geolocation'] = {
                    'latitude': last_geo.get('latitude'),
                    'longitude': last_geo.get('longitude'),
                    'sunset': last_geo.get('sunset'),
                    'sunrise': last_geo.get('sunrise')
                }

        reply = {
            'origins': origins
        }

        return reply
