from edge import eventlog
from edge.support import fs
from edge.support.template import FormatTemplate
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.rtsp_streaming import RtspStreamingUiconf
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class RtspStreamingApi:
    @classmethod
    def load_unit(cls, uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_RTSP_STREAMING)

        unit_uconf = unit_umod.load_conf(default=None)
        if not unit_uconf:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        #CamDgStats.create(unit_umod.pid).assert_stream_active()

        uiconf_pub = unit_uconf['rw']
        try:
            rtsp_url = RtspStreamingUiconf(uiconf_pub).rtsp_url
        except (ValueError, RuntimeError, IOError):
            eventlog.log_e('fail: InternalError')
            raise InternalError

        uiconf_pub['streaming']['url'] = rtsp_url
        return {
            'type': UTYPE_RTSP_STREAMING,
            'conf': uiconf_pub
        }

    @classmethod
    def save_unit(cls, uref: str, new_uiconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_RTSP_STREAMING)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        #CamDgStats.create(unit_umod.pid).assert_stream_active()

        new_uconf_conf = new_uiconf.get('conf')
        unit_uconf_pub = unit_uconf['rw']

        def validate_uimod_pub() -> None:
            fs.validate_json(new_uconf_conf, unit_umod.schema_path)

            new_streaming: dict = new_uconf_conf['streaming']
            if new_streaming['transport'] == 'rtp-over-udp-or-tcp':
                if new_streaming['min-udp-port'] == 0 and new_streaming['max-udp-port'] != 0:
                    eventlog.log_d('fail: ParamsError')
                    raise ParamsError

                if new_streaming['max-udp-port'] == 0 and new_streaming['min-udp-port'] != 0:
                    eventlog.log_d('fail: ParamsError')
                    raise ParamsError

            # validate 'format' jnodes from json
            footer_rows: list[dict] = new_uconf_conf['live']['footer']['rows']
            if cls._validate_format(footer_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            old_rtsp_url: str = RtspStreamingUiconf(unit_uconf_pub).rtsp_url
            new_rtsp_url: str = new_uconf_conf['streaming']['url']
            if old_rtsp_url != new_rtsp_url:
                eventlog.log_d('fail: ParamsError (rtsp-url is read-only)')
                raise ParamsError

        validate_uimod_pub()

        new_uconf_conf['streaming']['url'] = RtspStreamingUiconf(new_uconf_conf).rtsp_url

        if unit_uconf_pub == new_uconf_conf:
            return new_uconf_conf

        restart_setupstore = []

        if unit_uconf['rw']['enabled'] != new_uconf_conf['enabled']:
            restart_setupstore.append(True)

        unit_uconf['rw'] = new_uconf_conf

        uirefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), uirefs_l0_l1, {unit_umod.ref: unit_uconf})

        if restart_setupstore:
            dg.LifeCycle.sighup_ss_dgmod()

        return {
            'type': UTYPE_RTSP_STREAMING,
            'conf': new_uconf_conf
        }

    @staticmethod
    def _validate_format(jnode: list[dict]) -> bool:
        key_to_true_map = {
            'datetime': True,
            'datetime-ms': True,
            'post-name': True,
            'geolocation': True,
            'vendor': True,
            'model': True,
            'serial': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d('fail: invalid key (%s)' % str(x))
                    return False

        return True
