import uuid

from typing import Callable

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph, edge_creator as creator
from edge.webserver.sysconf.api_seatbelt import SeatbeltApi
from edge.webserver.sysconf.api_tcp_recipient import TcpRecipientApi
from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.types.defs import TERM_UISCHEMA_DIR, TERM_UICONF_DIR
from edge.webserver.sysconf.uiconf.builder import CTRL_DEF_UICONF_DIR
from edge.webserver.sysconf.uiconf.terminal import TermAboutUiconf
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.umod2 import Umod2
from edge.webserver.sysconf.uimod.urole import *
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class TerminalApi:
    @classmethod
    def create_term(cls, ctrl_uref: str, factory_conf: dict) -> dict:
        factory_name = factory_conf['factory']
        factory_info = unet.decode_factory(fs.join_path(CTRL_DEF_UICONF_DIR, factory_name))
        factory_info.root['internal'] = factory_conf.get('internal', False)

        root_uinfo, term_uinfo = ph.RootInfoUiconf.add_term(ctrl_uref, factory_info.root, factory_info.links)

        term_uref = term_uinfo['ref']
        term_uimod = Umod.create(term_uref)
        term_pid = term_uimod.pid
        term_uuid = factory_conf.get('uuid') or str(uuid.uuid4())

        factory_conf['term-pid'] = term_pid
        factory_conf['term-uuid'] = term_uuid
        factory_conf['term-uref'] = term_uref

        def install_unit_fn(unit_uinfo: unet.UnitInfo) -> None:
            unit_utype = unit_uinfo['type']
            if unit_utype in UTYPES_FILELESS:
                return

            unit_uref = unit_uinfo['ref']
            unit_uconf = ph.build_uconf(unit_utype, unit_uref, factory_conf['factory'], factory_conf)
            if unit_utype in UTYPES_UIMOD2:
                unit_umod = Umod2(unit_uref)
                unit_umod.install(UiConf2(unit_uconf))

            else:
                unit_umod = Umod.create(unit_uref)
                unit_umod.install(unit_uconf)

        # noinspection PyBroadException
        try:
            # install uimods
            TimeUiconf.install()
            unet.map_units_r(term_uinfo, install_unit_fn)
            TermAboutUiconf.save_conf(term_pid, term_uuid)
            rootinfo.set_unit_name(term_uref, term_uref)

            # install dgmods
            urefs_l0_l1_ws = unet.find_urefs_r(term_uinfo)
            urefs_l0_l1_ws.append('webserver.0')  # edit ws conf while addind pipeline
            all_used_dgmods = creator.generate(term_uimod.create_backup_dirpath(), urefs_l0_l1_ws, forced=True)

            all_used_dgmods = [
                x for x in all_used_dgmods if x.ref not in (dg.CTRL_WS_DGMOD.ref, dg.CTRL_ES_DGMOD.ref)
            ]

            spool_size = factory_conf.get('spool-size') or 740
            SsDgconf.append_ss_group(term_pid, [x.conf_dir for x in all_used_dgmods], int(spool_size))

            from edge.support import db
            db.edit_term(term_uuid, term_name=term_pid, term_place=term_pid, manual=True)

            dg.LifeCycle.restart_ws_dgmod()
            dg.LifeCycle.sighup_ss_dgmod()
            from edge.webserver.sysconf.api_zabbix import ZabbixApi
            ZabbixApi.generate_web_conf()

            return {
                'info': [root_uinfo],
                'created': [term_uref]
            }

        except NotImplementedError:
            eventlog.log_d('fail: ParamsError (unsupported camera)')
            raise ParamsError

        except BusyError:
            raise

        except Exception:
            eventlog.log_x()
            TerminalApi.drop_terminal2(term_uref)
            raise InternalError

    @staticmethod
    def make_unit(term_uiref: str, factory_conf: dict) -> dict:
        uimod_type = factory_conf.get('type')
        if uimod_type is None:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        from .api_parking import ParkingApi
        from .api_wrong_way import WrongWayApi
        from .api_bus_lane import BusLaneApi
        from .api_red_light import RedLightApi
        from .api_wanted import WantedApi
        from .api_section_speed import SectionSpeedApi
        from .api_spot_speed import SpotSpeedApi
        from .api_railroad_crossing import RailroadCrossingApi

        uitype_to_factory_map = {
            UTYPE_DET_PARKING: ParkingApi.make_detector,
            UTYPE_DET_REDLIGHT: RedLightApi.make_detector,
            UTYPE_DET_SECTION_SPEED: SectionSpeedApi.make_detector,
            UTYPE_DET_SPOT_SPEED: SpotSpeedApi.make_detector,
            UTYPE_DET_WANTED: WantedApi.make_detector,
            UTYPE_DET_WRONG_WAY: WrongWayApi.make_detector,
            UTYPE_DET_BUS_LANE: BusLaneApi.make_detector,
            UTYPE_TCP_RECIPIENT: TcpRecipientApi.make_service,
            #UTYPE_LEDDAR: LeddarApi.make_service,
            UTYPE_DET_RAILROAD: RailroadCrossingApi.make_detector,
            UTYPE_DET_SEATBELT: SeatbeltApi.make_detector
        }

        factory_fn = uitype_to_factory_map.get(uimod_type)
        if not factory_fn:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        term_uimod = Umod.create(term_uiref)
        unit_uiref, terms_info = factory_fn(term_uimod.ref, factory_conf)

        return {
            'info': [terms_info],
            'created': [unit_uiref]
        }

    @staticmethod
    def get_factory_limits(term_uiref: str, roles: list[str]) -> list[dict]:
        uimod = Umod.create(term_uiref)
        uimod.assert_type(UTYPE_TERMINAL)

        service_list = [
            {'role': UROLE_SERVICE, 'type': UTYPE_TCP_RECIPIENT}
            #,{'role': UROLE_SERVICE, 'type': UTYPE_LEDDAR}
        ]

        detector_list = [
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_REDLIGHT},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_RAILROAD},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_WANTED},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_WRONG_WAY},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_BUS_LANE},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_SECTION_SPEED},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_PARKING},
            {'role': UROLE_DETECTOR, 'type': UTYPE_DET_SEATBELT}
        ]

        if ph.RootInfoUiconf.get_raterm_pid() == uimod.pid:
            detector_list.append(
                {'role': UROLE_DETECTOR, 'type': UTYPE_DET_SPOT_SPEED})

        if not roles:
            roles = [UROLE_DETECTOR, UROLE_SERVICE]

        reply = []
        if UROLE_DETECTOR in roles:
            reply.extend(detector_list)

        if UROLE_SERVICE in roles:
            reply.extend(service_list)

        return reply

    @staticmethod
    def load_terminal(term_uiref: str) -> dict:
        unit_umod = Umod.create(term_uiref)
        unit_umod.assert_type(UTYPE_TERMINAL)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_uconf_pri = unit_uconf['ro']

        ext_uiconf_pub = unit_uconf['rw']
        ext_uiconf_pub['origin-guid'] = unit_uconf_pri['uuid']
        return {
            'type': UTYPE_TERMINAL,
            'conf': ext_uiconf_pub
        }

    @staticmethod
    def save_terminal(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_TERMINAL)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')
        fs.validate_json(new_uconf_conf, unit_umod.schema_path)

        unit_uconf_pri: dict = unit_uconf['ro']
        unit_uconf_pub: dict = unit_uconf['rw']

        if unit_uconf_pri['uuid'] != new_uconf_conf['origin-guid']:
            eventlog.log_d('fail: ParamsError (#/origin-guid is read-only)')
            raise ParamsError

        new_origin_name = new_uconf_conf['origin-name']
        new_origin_place = new_uconf_conf['origin-place']

        if unit_uconf_pub['origin-name'] != new_origin_name or unit_uconf_pub['origin-place'] != new_origin_place:
            term_uuid = TerminalUiconf(unit_umod.conf_path).get_uuid()
            rootinfo.set_unit_name(unit_umod.ref, new_origin_name)

            from edge.support import db
            db.edit_term(term_uuid, new_origin_name, new_origin_place, manual=True)

            TermAboutUiconf.save_conf(unit_umod.pid, term_uuid, term_place=new_origin_place, term_name=new_origin_name)

            unit_uconf_pub['origin-name'] = new_origin_name
            unit_uconf_pub['origin-place'] = new_origin_place

            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        return {
            'type': UTYPE_TERMINAL,
            'conf': new_uconf_conf
        }

    @classmethod
    def drop_terminal(cls, uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_TERMINAL)

        root_uinfo = rootinfo.decode_info().root
        unit_uinfo = unet.find_unit_r(root_uinfo, uref=uref)
        if unit_uinfo is None:
            eventlog.log_e('fail: ParamsError')
            raise ParamsError

        elif unit_uinfo.get('internal', True):
            eventlog.log_d('fail: ParamsError (terminal is internal)')
            raise ParamsError

        reply = cls.drop_terminal2(uref)

        # restarting edgeserver with PCD
        eventlog.log_i('exec: restarting modules')
        dg.LifeCycle.sighup_ss_dgmod()
        dg.LifeCycle.restart_ws_dgmod()
        from edge.webserver.sysconf.api_zabbix import ZabbixApi
        ZabbixApi.generate_web_conf()

        return reply

    @classmethod
    def drop_terminal2(cls, term_uref: str) -> dict:
        from edge.webserver.sysconf.dgmod.es64.es64_module import es64_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.gc.gc_module import gc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.lr.lr_module import lr_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.mc.mc_module import mc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.mg.mg_module import mg_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ms.ms_module import ms_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.msw.msw_module import msw_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ra.ra_module import ra_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rc.rc_module import rc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rg.rg_module import rg_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rn.rn_module import rn_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rs.rs_module import rs_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.sc.sc_module import sc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.tl.tl_module import tl_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.tr.tr_module import tr_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ws.ws_module import ws_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.psu.psu_module import psu_create_dgconf_deleter

        dgconf_deleter_factory_fn_list: list[Callable[[IUiConfProvider], list[IDgConfDeleter]]] = [
            es64_create_dgconf_deleter,
            gc_create_dgconf_deleter,
            lr_create_dgconf_deleter,
            mc_create_dgconf_deleter,
            mg_create_dgconf_deleter,
            ms_create_dgconf_deleter,
            msw_create_dgconf_deleter,
            ra_create_dgconf_deleter,
            rc_create_dgconf_deleter,
            rg_create_dgconf_deleter,
            rn_create_dgconf_deleter,
            rs_create_dgconf_deleter,
            sc_create_dgconf_deleter,
            tl_create_dgconf_deleter,
            tr_create_dgconf_deleter,
            ws_create_dgconf_deleter,
            psu_create_dgconf_deleter
        ]

        root_uinfo = rootinfo.decode_info().root
        term_uinfo = unet.find_unit_r(root_uinfo, uref=term_uref)
        if term_uinfo is None:
            eventlog.log_e(f'fail: find_unit_r (uref:{term_uref})')
            raise NotImplementedError

        term_umod = Umod.create(term_uref)
        all_used_deleters = list[IDgConfDeleter]()
        for factory_fn in dgconf_deleter_factory_fn_list:
            dgconf_deleter = factory_fn(term_umod)
            all_used_deleters.extend(dgconf_deleter)

        term_urefs = unet.find_urefs_r(term_uinfo)
        for unit_uref in term_urefs:
            unit_umod = ph.RootInfoUiconf.get_uimods(root_uinfo, {unit_uref}).pop()  # TODO
            for factory_fn in dgconf_deleter_factory_fn_list:
                dgconf_deleter = factory_fn(unit_umod)
                all_used_deleters.extend(dgconf_deleter)

        list(map(lambda deleter: deleter.run(), all_used_deleters))

        shemas_dirpath = fs.join_path(TERM_UISCHEMA_DIR, term_umod.pid)
        ui_dirpath = fs.join_path(TERM_UICONF_DIR, term_umod.pid)
        list(map(fs.silent_rm_r, [shemas_dirpath, ui_dirpath]))

        terms_info, urefs_l1 = rootinfo.drop_unit_r(term_uref)

        from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
        SsDgconf.remove_ss_group(term_umod.pid)

        return {
            'info': [terms_info]
        }

    @staticmethod
    def get_ra_type(term_pid: str) -> str | None:
        location_uimod = Umod(term_pid, UTYPE_LOCATION, '0')
        try:
            conf = location_uimod.load_conf()
            conf_ro = conf.get('ro') or dict()
            return conf_ro.get('type')

        except FsError:
            pass
