import os
from typing import Final, NotRequired, TypedDict, cast

from edge import eventlog
from edge.support import fs, net, sh
from edge.support.cgi import CgiBackup
from edge.types.exceptions import InternalError, ParamsError
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.utype import *


class _NtpServerConf(TypedDict):
    address: str
    prefered: bool


class _SourceConf(TypedDict):
    type: str
    autosync: NotRequired[bool]


class _GnssSourceConf(_SourceConf):
    pass


class _NtpSourceConf(_SourceConf):
    servers: list[_NtpServerConf]


def _map_enpoint_gnss_to_pps(gnss_endpoint: str) -> str | None:
    """
    :param gnss_endpoint:
        /dev/ttyS3
    :return:
        /dev/pps1
    """
    if not gnss_endpoint:
        eventlog.log_e_param('gnss_endpoint')
        raise NotImplementedError

    pps_root_dir = '/sys/devices/virtual/pps'
    if not os.path.exists(pps_root_dir):
        return

    #
    # cat /sys/devices/virtual/pps/pps1/path
    # /dev/ttyS3
    subdir_names = os.listdir(pps_root_dir)
    for subdir_name in subdir_names:
        file_path = fs.join_path(pps_root_dir, subdir_name, 'path')
        with open(file_path, 'r') as text_file:
            lines = text_file.readlines()

        for line in lines:
            if line.strip() == gnss_endpoint:
                return f'/dev/{subdir_name}'


class _NtpdateConfTweaker:
    NTPSEC_NTPDATE_DEB__CONF_PATH = '/etc/default/ntpsec-ntpdate'
    NTPDATE_DEB__CONF_PATH = '/etc/default/ntpdate'

    @classmethod
    def _get_conf_path(cls):
        if fs.path_exists(cls.NTPSEC_NTPDATE_DEB__CONF_PATH):
            # bookworm
            return cls.NTPSEC_NTPDATE_DEB__CONF_PATH
        elif fs.path_exists(cls.NTPDATE_DEB__CONF_PATH):
            # buster
            return cls.NTPDATE_DEB__CONF_PATH
        else:
            eventlog.log_e('fail: unsupported')
            raise NotImplementedError

    @classmethod
    def tweak(cls, ntp_autosync: bool) -> None:
        conf_path = cls._get_conf_path()

        with open(conf_path, 'r') as text_file:
            curr_lines = text_file.readlines()

        # Remove lines starting with 'exit 0'
        next_lines = [line for line in curr_lines if not line.startswith('exit 0')]

        if ntp_autosync:
            # do sync time using ntpdate on system startup
            pass
        else:
            # do not sync time using ntpdate on system startup
            next_lines.insert(0, 'exit 0\n')

        with open(conf_path, 'w') as text_file:
            text_file.write(''.join(next_lines))


class _NtpConfBuider:
    CONF_PATH = '/etc/ntp.conf'

    def __init__(self, pri_uconf: dict) -> None:
        self._pri: Final[dict] = pri_uconf

        self._footer: Final[list[str]] = [
            "#",
            "driftfile /var/lib/ntp/ntp.drift",
            "#statsdir /var/log/ntpstats/",
            "statistics loopstats peerstats clockstats",
            "filegen loopstats file loopstats type day enable",
            "filegen peerstats file peerstats type day enable",
            "filegen clockstats file clockstats type day enable",
            "#",
            "restrict -4 default kod notrap nomodify nopeer noquery",
            "restrict -6 default kod notrap nomodify nopeer noquery",
            "restrict 127.0.0.1",
            "restrict ::1",
            "#",
            "tinker panic 0"
        ]

    def _encode_ntp_source(self, lines: list[str], ntp_source: _NtpSourceConf) -> bool:
        if not ntp_source.get('autosync', False):
            return False

        for server_conf in ntp_source['servers']:
            address = server_conf['address']
            if server_conf['prefered']:
                lines.append(f'server {address} prefer')
            else:
                lines.append(f'server {address}')

        for server_conf in ntp_source['servers']:
            address = server_conf['address']
            lines.append(f'restrict {address} noquery notrap')

        return True

    def _encode_gnss_source(self, lines: list[str], gnss_source: _GnssSourceConf) -> bool:
        if not gnss_source.get('autosync', False):
            return False

        gnss_jnode: dict | None = self._pri.get('gnss')
        if gnss_jnode is None:
            eventlog.log_e_param('#/gnss')
            return False

        gnss_endpoint: str = gnss_jnode['endpoint']
        gnss_offset: float = gnss_jnode.get('gnss-offset') or 0.0

        lines.extend([
            f'server 127.127.28.0 minpoll 4 maxpoll 4 prefer',
            f'fudge 127.127.28.0 time1 {gnss_offset} flag1 1 refid GPS'
        ])

        pps_endpoint = _map_enpoint_gnss_to_pps(gnss_endpoint)
        if pps_endpoint is not None:
            pps_port = pps_endpoint[-1]
            lines.extend([
                f'server 127.127.22.{pps_port}',
                f'fudge 127.127.22.{pps_port} flag2 1 flag3 1 refid PPS',
                'tos mindist 0.01'
            ])

        return True

    def build(self, ntp_source: _NtpSourceConf, gnss_source: _GnssSourceConf) -> bool:
        lines: list[str] = []

        autosync = self._encode_ntp_source(lines, ntp_source)
        autosync = self._encode_gnss_source(lines, gnss_source) or autosync

        lines.extend(self._footer)

        with open(self.CONF_PATH, 'w') as text_file:
            text_file.write('\n'.join(lines))

        return autosync


class _ChronyConfBuilder:
    CONF_DIR = '/etc/chrony/conf.d'
    SOURCES_DIR = '/etc/chrony/sources.d'

    def __init__(self, pri_uconf: dict) -> None:
        self._pri: Final[dict] = pri_uconf

    def _encode_ntp_source(self, ntp_source: _NtpSourceConf) -> bool:
        conf_lines = list[str]()

        if ntp_source.get('autosync', False):
            for server_conf in ntp_source['servers']:
                address = server_conf['address']
                if server_conf['prefered']:
                    conf_lines.append(f'server {address} iburst prefer')
                else:
                    conf_lines.append(f'server {address} iburst')

        conf_path = fs.join_path(self.SOURCES_DIR, 'edge.sources')
        with open(conf_path, 'w') as text_file:
            text_file.write('\n'.join(conf_lines))

        return bool(conf_lines)

    def _encode_gnss_source(self, gnss_source: _GnssSourceConf) -> bool:
        conf_lines = list[str]()

        if gnss_source.get('autosync', False):
            gnss_jnode = self._pri.get('gnss') or dict()

            gnss_endpoint = gnss_jnode.get('endpoint').split('/')[-1]
            gnss_offset = gnss_jnode.get('gnss-offset') or 0.0
            pps_offset = gnss_jnode.get('pps-offset') or 0.0

            conf_lines.extend([
                f'refclock SHM 0 offset {gnss_offset} delay 0.05 refid GNSS noselect',
                f'refclock SOCK /var/run/chrony.{gnss_endpoint}.sock offset {pps_offset} refid PPS'
            ])

        conf_path = fs.join_path(self.CONF_DIR, 'edge.conf')
        with open(conf_path, 'w') as text_file:
            text_file.write('\n'.join(conf_lines))

        return bool(conf_lines)

    def build(self, ntp_source: _NtpSourceConf, gnss_source: _GnssSourceConf) -> bool:
        autosync = self._encode_ntp_source(ntp_source)
        autosync = self._encode_gnss_source(gnss_source) or autosync

        return autosync


def setup_ntpd(pri_uconf: dict, ntp_source: _NtpSourceConf, gnss_source: _GnssSourceConf) -> None:
    _NtpdateConfTweaker.tweak(ntp_source.get('autosync', False))
    autosync = _NtpConfBuider(pri_uconf).build(ntp_source, gnss_source)

    sh.run_silent('systemctl stop ntp')
    sh.run_silent('systemctl disable ntp')
    if autosync:
        sh.run_silent('systemctl enable ntp')
        sh.run_silent('systemctl start ntp')


def setup_chronyd(pri_uconf: dict, ntp_source: _NtpSourceConf, gnss_source: _GnssSourceConf) -> None:
    autosync = _ChronyConfBuilder(pri_uconf).build(ntp_source, gnss_source)

    sh.run_silent('systemctl stop chrony')
    sh.run_silent('systemctl disable chrony')
    if autosync:
        sh.run_silent('systemctl enable chrony')
        sh.run_silent('systemctl start chrony')
        sh.run_silent('systemctl restart gpsd')

def get_gnss_sync_source(sync_sources: list[_SourceConf]) -> _GnssSourceConf:
    gnss_sources = [source for source in sync_sources if source['type'] == 'gnss']
    gnss_sources_count = len(gnss_sources)
    if gnss_sources_count == 1:
        return gnss_sources[0]

    elif gnss_sources_count == 0:
        return {
            'autosync': False,
            'type': 'gnss'
        }

    else:
        eventlog.log_d('fail: ParamsError (#/args/sources)')
        raise ParamsError

def get_ntp_sync_source(sync_sources: list[_SourceConf]) -> _NtpSourceConf:
    ntp_sources: list[_SourceConf] = [source for source in sync_sources if source['type'] == 'ntp']
    ntp_sources_count = len(ntp_sources)
    if ntp_sources_count == 1:
        return cast(_NtpSourceConf, ntp_sources[0])

    elif ntp_sources_count == 0:
        return {
            'autosync': False,
            'type': 'ntp',
            'servers': []
        }

    else:
        eventlog.log_d('fail: ParamsError (#/args/sources)')
        raise ParamsError


# noinspection PyClassHasNoInit
class TimeApi:
    _GNSS_SOURCE_LIMITS = {
        'type': 'gnss',
        'sync-auto': True,
        'sync-manual': False
    }

    _GNSS_SOURCE_CONF: _GnssSourceConf = {
        'type': 'gnss',
        'autosync': False
    }

    @classmethod
    def load_unit(cls) -> dict:
        TimeUiconf.install()

        uiconf = TimeUiconf.load_conf()
        uiconf_pub: dict = uiconf['rw']

        ext_uiconf_pub = uiconf_pub
        ext_sources = ext_uiconf_pub['args']['sources']

        # patch config if GNSS device detected
        if len([x for x in ext_sources if x['type'] == 'gnss']) == 0:
            if net.query_gps_count() > 0:
                ext_sources.append(cls._GNSS_SOURCE_CONF)

        if len([x for x in ext_uiconf_pub['limits']['sources'] if x['type'] == 'gnss']) == 0:
            ext_uiconf_pub['limits']['sources'].append(cls._GNSS_SOURCE_LIMITS)

        TimeUiconf.save_conf(uiconf)

        unit_ext_jnode = {
            'type': UTYPE_TIME,
            'conf': ext_uiconf_pub['args'],
            'limits': ext_uiconf_pub['limits']
        }

        return unit_ext_jnode

    @classmethod
    @fs.Lock('/var/run/edge/datetime', 2)
    @CgiBackup('/etc/edge/setupstore/backup/datetime', {
        'time.json': TimeUiconf.CONF_PATH
    })
    def save_unit(cls, new_uconf: dict) -> dict:
        fs.validate_json({'args': new_uconf['conf']}, TimeUiconf.SCHEMA_PATH)  # TODO
        new_uconf_conf = new_uconf['conf']

        unit_uconf = TimeUiconf.load_conf()
        unit_uconf_pub: dict = unit_uconf['rw']
        if new_uconf_conf['tz'] not in unit_uconf_pub['limits']['tz']:
            eventlog.log_d('fail: ParamsError (#/args/tz)')
            raise ParamsError

        next_gnss_source: _GnssSourceConf = get_gnss_sync_source(new_uconf_conf.get('sources', []))
        next_ntp_source: _NtpSourceConf = get_ntp_sync_source(new_uconf_conf.get('sources', []))

        timekeeper_type = os.environ.get('EDGE_TIMEKEEPER__TYPE') or 'ntp'
        if timekeeper_type == 'ntp':
            setup_ntpd(unit_uconf['ro'], next_ntp_source, next_gnss_source)
        elif timekeeper_type == 'chrony':
            setup_chronyd(unit_uconf['ro'], next_ntp_source, next_gnss_source)
        else:
            eventlog.log_d('fail: EDGE_TIMEKEEPER__TYPE')
            raise ParamsError

        # if tz changed restart
        if unit_uconf_pub['args']['tz'] != new_uconf_conf['tz']:
            eventlog.log_i('exec: restart overlayrender, speedcontrol, webserver, tracker')
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_RENDERER)
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_SPEEDCONTROL)
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_TRACKER)
            dg.LifeCycle.restart_ws_dgmod()
            # TODO: restart edgeserver if next_time < current_time

        # write new jnode to disk
        unit_uconf_pub['args'] = new_uconf_conf
        TimeUiconf.save_conf(unit_uconf)

        return cls.load_unit()  # TODO: strip limits from uiconf

    @classmethod
    def sync_manual(cls, sync_source: dict) -> dict:
        import datetime

        fs.validate_json(sync_source, TimeUiconf.SCHEMA_MANUAL_PATH)
        source_type = sync_source.get('type')
        if source_type == 'ntp':
            cls._sync_manual_ntp(sync_source)
            return {
                'type': 'ntp',
                'time': datetime.datetime.now().isoformat()
            }

        elif source_type == 'user':
            cls._sync_manual_user(sync_source)
            return {
                'type': 'user',
                'time': datetime.datetime.now().isoformat()
            }

        else:
            eventlog.log_i('fail: ParamsError (#/type)')
            raise ParamsError

    @staticmethod
    def _sync_manual_ntp(sync_source: dict) -> None:
        # set OS time
        servers = [sh.quote(server['address']) for server in sync_source.get('servers', [])]
        if not servers:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        timekeeper_type = os.environ.get('EDGE_TIMEKEEPER__TYPE') or 'ntp'
        if timekeeper_type == 'ntp':
            sh_cmd = f'ntpdate -u {" ".join(servers)}'
        elif timekeeper_type == 'chrony':
            sh_cmd = f'chronyc makestep'
        else:
            eventlog.log_d('fail: EDGE_TIMEKEEPER__TYPE')
            raise ParamsError

        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

    @staticmethod
    def _sync_manual_user(sync_source: dict) -> None:
        import iso8601
        tz_datetime = iso8601.parse_date(sync_source['time'])
        utc_datetime = tz_datetime - tz_datetime.utcoffset()

        # set OS time
        sh_cmd = 'date --set="{:%Y-%m-%d %H:%M:%S}"'.format(utc_datetime)
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: http500. sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: http500. sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError
