from edge import eventlog
from edge.support import fs
from edge.support.template import FormatTemplate
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyClassHasNoInit
class TransitApi:
    @staticmethod
    def load_detector(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_DET_TRANSIT)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        reply = {
            'type': UTYPE_DET_TRANSIT,
            'conf': unit_uconf['rw']['args']
        }

        reply.update(ph.get_detector_limits_stats(uref))
        return reply

    @classmethod
    def save_detector(cls, uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_DET_TRANSIT)

        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')

        def validate_uiconf_pub() -> None:
            fs.validate_json({'args': new_uconf_conf}, unit_umod.schema_path)

            images_jnode = new_uconf_conf['images']
            v_rows = images_jnode['vehicle']['footer']['rows']
            s_rows = images_jnode['scene']['footer']['rows']

            if cls._validate_format(v_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            elif cls._validate_format(s_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

        validate_uiconf_pub()

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['args'] != new_uconf_conf:
            if new_uconf_conf['name'] != unit_uconf['rw']['args']['name']:
                rootinfo.set_unit_name(unit_umod.ref, new_uconf_conf['name'])

            unit_uconf['rw']['args'] = new_uconf_conf

            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        reply = {
            'type': UTYPE_DET_TRANSIT,
            'conf': new_uconf_conf
        }

        reply.update(ph.get_detector_limits_stats(uref))
        return reply

    @staticmethod
    def _validate_format(jnode: list[dict]) -> bool:
        key_to_true_map = {
            'speed': True,
            'datetime': True,
            'datetime-ms': True,
            'datetime-ms-gnss': True,
            'datetime-ms-sync': True,
            'datetime-control': True,
            'datetime-control-ms': True,
            'time-control-ms': True,
            'post-name': True,
            'geolocation': True,
            'geolocation-sync': True,
            'plate': True,
            'lane': True,
            'dir': True,
            'dir-avg': True,
            'video-speed': True,
            'video-speed2': True,
            'radar-speed': True,
            'radar-speed2': True,
            'vendor': True,
            'model': True,
            'serial': True,
            'wim-speed': True,
            'main-speed': True,
            'main-speed2': True,
            'control-speed': True,
            'position-f': True,
            'mmr-class': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d(f'fail: invalid key ({str(x)})')
                    return False

        return True
