from abc import ABCMeta, abstractmethod
from typing import Final

from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.types.conf import IUiConfProvider

UrefToConfMap = dict[str, dict]
UrefToUmodMap = dict[str, IUiConfProvider]
UtypeToUrefToUmodMap = dict[str, UrefToUmodMap]


class Udeps:
    def __init__(self, linked_uimods: UtypeToUrefToUmodMap) -> None:
        self._type_to_ref_uimod_map: Final[UtypeToUrefToUmodMap] = linked_uimods

    def __str__(self) -> str:
        type_to_refs_map = dict[str, list[str]]()

        for uitype, ref_uimod_map in self._type_to_ref_uimod_map.items():
            type_to_refs_map[uitype] = list(ref_uimod_map.keys())

        return f'{type_to_refs_map}'

    def test_contains(self, uiref: str, uitypes: set[str]) -> bool:
        for uitype in uitypes:
            if uiref in self._type_to_ref_uimod_map[uitype]:
                return True

        return False

    def try_find_any(self, uitypes: set[str]) -> IUiConfProvider | None:
        for uitype in uitypes:
            for uimod in self._type_to_ref_uimod_map[uitype].values():
                return uimod

    def find_any(self, uitypes: set[str]) -> IUiConfProvider:
        result = self.try_find_any(uitypes)
        if result is None:
            raise NotFoundError(uitypes)
        else:
            return result

    def try_find_all(self, uitype: str) -> list[IUiConfProvider]:
        return list(self._type_to_ref_uimod_map[uitype].values())

    def find_all(self, uitype: str) -> list[IUiConfProvider]:
        result = self.try_find_all(uitype)
        if not result:
            raise NotFoundError(uitype)
        else:
            return result


class IDgConfBuilder(metaclass=ABCMeta):
    @property
    @abstractmethod
    def dgmod(self) -> dg.DgMod | None:
        """
        :returns: None if
            - dgmod has nothing to backup
            - dgmod is not registered in the setupstore group.
        """
        pass

    @abstractmethod
    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        pass

    @abstractmethod
    def build(self) -> None:
        pass

    @abstractmethod
    def notify(self) -> None:
        pass


class IDgConfDeleter(metaclass=ABCMeta):
    @abstractmethod
    def run(self) -> None:
        pass
