from typing import Final

from edge.support import net
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, UrefToConfMap, Udeps, IDgConfDeleter
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.controller import CtrlUiconf
from edge.webserver.sysconf.uiconf.db import DbUiconf
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def _reconfigure_es(endpoints: list[str]) -> bool:
    for _ in [1, 2, 3]:
        try:
            net.query_es_config(endpoints, tout_sec=1)
            return True
        except net.NetError:
            import time
            time.sleep(0.300)

    return False


def _mk_react_path(rule_uimod: IUiConfProvider) -> str:
    return f'/var/spool/edge/system/sc{rule_uimod.pid}-es65535-fact'


# noinspection PyUnusedLocal
def es64_create_dgconf_builder(
        target_uref: str,
        target_udeps: Udeps,
        uref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    """
    :param target_uref:
        The configuration files for this unit must be regenerated.
    :param target_udeps:
        The direct dependencies (linked units) of the target unit.
    :param uref_to_newconf_map:
    :param forced:
        The value is true when
            - the terminal is being created.
            - the terminal is being regenerated.
            - the detector is being removed.
            - the service is being removed.
            - the systest activation is being changed.
    """

    activate = forced

    if not activate:
        return []

    root_info = rootinfo.decode_info().root
    tr_uirefs = unet.find_urefs_r(root_info, UTYPE_TRAFFIC_RULES)
    rules_uimods = list(map(Umod.create, tr_uirefs))
    return [_EsCreator64(rules_uimods)]


class _EsCreator64(IDgConfBuilder):
    def __init__(self, rules_uimods: list[IUiConfProvider]) -> None:
        self._rules_uimods: Final[list[IUiConfProvider]] = rules_uimods

        self._main_conf = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return dg.CTRL_ES_DGMOD

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        es_new_conf = {
            'api': net.get_es_uds_endpoint(),
            'db': DbUiconf.CONF_PATH,
            'monitoring': {
                'stat-path': '/var/spool/edge/system/zabbix/es65535',
                'stat-tout': 60
            },
            'fact-react': list(map(_mk_react_path, self._rules_uimods)),
            'systest': CtrlUiconf.get_systest()
        }

        self._main_conf = es_new_conf

    def build(self) -> None:
        es_dgmod = dg.CTRL_ES_DGMOD
        es_dgmod.install()
        es_dgmod.save_main_conf(self._main_conf)
        es_dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        if not _reconfigure_es(self._main_conf['fact-react']):
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_EVENTSTORE)


class _EsDeleter64(IDgConfDeleter):
    def __init__(self, uimod_fa: IUiConfProvider) -> None:
        self._uimod_fa = uimod_fa

    def run(self) -> None:
        dgmod = dg.CTRL_ES_DGMOD
        main_conf = dgmod.load_main_conf()

        endpoint = _mk_react_path(self._uimod_fa)
        main_conf['fact-react'] = [x for x in main_conf['fact-react'] if x != endpoint]

        dgmod.save_main_conf(main_conf)

        if not _reconfigure_es(main_conf['fact-react']):
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_EVENTSTORE)


def es64_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_TERMINAL:
        return [_EsDeleter64(target_uimod)]

    return []
