from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def le_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_LEDDAR}):
        return []

    try:
        uimod_ledar = target_uideps.find_any({UTYPE_LEDDAR})
        uimod_rules = target_uideps.find_any({UTYPE_TRAFFIC_RULES})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_ledar.ref)
    ])

    if not activate:
        return []

    return [_LeCreator(uimod_ledar, uimod_rules)]


class _LeCreator(IDgConfBuilder):
    def __init__(self, uimod_ledar: IUiConfProvider, uimod_rules: IUiConfProvider) -> None:
        self._dgmod_leddar: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_LEDDAR, uimod_ledar.id)

        self._uimod_ledar: Final[IUiConfProvider] = uimod_ledar
        self._uimod_rules: Final[IUiConfProvider] = uimod_rules

        self._main_active = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_leddar

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        ui_pid = self._uimod_ledar.pid
        spool_dir = ph.get_spool_dir(ui_pid)

        #
        main_conf = {
            'driver': {
                'connection': {
                    'type': 'usb',
                    'address': 1,
                    'endpoint': '/dev/ttyUSB0'
                },
                'dist-thr': 5.1
            },
            'bg-detector': {
                'beam-angle': 45.1,
                'beam-seg-count': 16,
                'beam-depth-map': [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
            },
            'loop-detector': {
                'changelog': {
                    'entry-count': 50,
                    'entry-tout': 300,
                    'entry-age': 300
                },
                'loops': [{
                    'ref': 'loop0',
                    'dist': [0.1, 0.5],
                    'amp': [0.0, 1000.0],
                    'seg': [7, 8]
                }]
            },
            'classifier': {
                'bike': [80, 220],
                'car': [220, 560]
            },
            'scene-sync': '/var/spool/edge/mg0-lg0-pipe',
            'scene-event': {
                'endpoints': [
                    '/tmp/lg0-tr0-scene'
                ]
            },
            'track-event': {
                'endpoints': [
                    fs.join_path(spool_dir, f'leddar{self._uimod_ledar.id}-sc{self._uimod_rules.id}-track')
                ]
            }
        }

        self._main_active = main_conf, True

    def build(self) -> None:
        main_conf, ss_active = self._main_active  # type: dict, bool

        edge_leddar: dg.DgMod = self._dgmod_leddar
        edge_leddar.install()
        edge_leddar.save_main_conf(main_conf)
        edge_leddar.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _LeDeleter(IDgConfDeleter):
    def __init__(self, uimod_ledar: IUiConfProvider) -> None:
        self._dgmod_ledar: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_LEDDAR, uimod_ledar.id)

    def run(self) -> None:
        self._dgmod_ledar.uninstall()


def le_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_LEDDAR:
        return [_LeDeleter(target_uimod)]
    return []
