from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.lr.dgconf.unity import bus, essentials, recipient, speedcam, anpr, leddar, wiegand, waga, sbr, mmr
from edge.webserver.sysconf.dgmod.unity import UnityConfBuilder
from edge.webserver.sysconf.dgstats.camera import CamDgStats
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def lr_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_LPR}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_lpr = target_uideps.find_any({UTYPE_LPR})
        uimod_roadway = target_uideps.find_any({UTYPE_ROADWAY})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimods_ledar = target_uideps.try_find_all(UTYPE_LEDDAR)
    uimods_location = target_uideps.try_find_all(UTYPE_LOCATION)
    uimods_tcp_recipient = target_uideps.try_find_all(UTYPE_TCP_RECIPIENT)
    uimods_waga = target_uideps.try_find_all(UTYPE_IPCOM_WAGA)

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref),
        uiref_to_newconf_map.get(uimod_lpr.ref),
        uiref_to_newconf_map.get(uimod_roadway.ref),

        any([uiref_to_newconf_map.get(x.ref) for x in uimods_ledar]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_location]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_tcp_recipient]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_waga])
    ])

    if not activate:
        return []

    return [_LrCreator(uimod_lpr, uimod_cam, uimod_fusion, uimod_roadway,
                       uimods_tcp_recipient, uimods_ledar, uimods_location, uimods_waga)]


UNIT__ESSENTIALS0__API0 = 'essentials.0~api.0'
UNIT__ESSENTIALS0__COUNTERS0 = 'essentials.1~counters.0'
UNIT__MEDIA_SERVER0__MEDIA0 = 'media-server.0~media.0'
UNIT__EVENT_EMITTER0__TX_TRACK_BUS0 = 'event-emitter.0~tx-track-bus.0'
UNIT__EVENT_EMITTER0__TX_SCENE_BUS0 = 'event-emitter.1~tx-scene-bus.0'
UNIT__ANPR0__ANPR0 = 'anpr.0~anpr.0'
UNIT__MMR0__MMR0 = 'mmr.0~mmr.0'
UNIT__SBR0__SBR0 = 'sbr.0~sbr.0'
UNIT__SPEEDCAM0__SPEEDCAM0 = 'speedcam.0~bracket.0'
UNIT__IPCOM0_WAGA0 = 'ipcom.0~waga.0'
UNIT__LEDDAR0__LEDDAR0 = 'leddar.0~leddar.0'
UNIT__WIEGAND_PCB0__RECIPIENT0 = 'wiegand-pcb.0~recipient.0'
UNIT__WIEGAND_UDP0__RECIPIENT0 = 'wiegand-udp.0~recipient.0'
UNIT__RULE_PCB0__RULE_PCB0 = 'rule.pcb0~rule.pcb0'
UNIT__RULE_UDP0__RULE_UDP0 = 'rule.udp0~rule.udp0'
UNIT__RULE_TCPx__RULE_TCPx = 'rule.tcp{}~rule.tcp{}'


class _LrCreator(IDgConfBuilder):
    def __init__(self,
                 uimod_lpr: IUiConfProvider,
                 uimod_cam: IUiConfProvider,
                 uimod_fusion: IUiConfProvider,
                 uimod_roadway: IUiConfProvider,
                 uimod_tcp_recipient: list[IUiConfProvider],
                 uimod_ledar: list[IUiConfProvider],
                 uimod_location: list[IUiConfProvider],
                 uimod_ipcom_waga: list[IUiConfProvider]):

        self._dgmod_lpr: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_LPR, uimod_lpr.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam
        self._uimod_fusion: Final[IUiConfProvider] = uimod_fusion
        self._uimod_lpr: Final[IUiConfProvider] = uimod_lpr
        self._uimod_roadway: Final[IUiConfProvider] = uimod_roadway

        self._uimods_ipcom_waga: Final[list[IUiConfProvider]] = uimod_ipcom_waga
        self._uimods_ledar: Final[list[IUiConfProvider]] = uimod_ledar
        self._uimods_location: Final[list[IUiConfProvider]] = uimod_location
        self._uimods_tcp_recipient: Final[list[IUiConfProvider]] = uimod_tcp_recipient

        self._builder = UnityConfBuilder()

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_lpr

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        spool_path = ph.get_spool_dir(self._uimod_lpr.pid)

        lpr_uiconf = uiref_to_conf_map.get(self._uimod_lpr.ref) or self._uimod_lpr.load_conf()
        lpr_pri = lpr_uiconf['ro']
        lpr_pub = lpr_uiconf['rw']
        cam_uiconf = uiref_to_conf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()
        cam_pri = cam_uiconf['ro']

        if self._uimod_cam.type == UTYPE_MV_CAMERA:
            if cam_pri['raw-format'] == 'MONO8T':
                add_scene_react_dir = fs.join_path(
                    spool_path, f'mg{self._uimod_cam.id}-lr{self._uimod_lpr.id}-scene')
            else:
                add_scene_react_dir = fs.join_path(
                    spool_path, f'rc{self._uimod_cam.id}-lr{self._uimod_lpr.id}-scene')

        elif self._uimod_cam.type == UTYPE_RTSP_CAMERA:
            add_scene_react_dir = fs.join_path(
                spool_path, f'rg{self._uimod_cam.id}-lr{self._uimod_lpr.id}-scene')

        else:
            raise NotImplementedError

        image_wh = CamDgStats.create(self._uimod_lpr.pid).stream_wh or (640, 480)
        scene_event_jnode = [
            fs.join_path(spool_path, f'lr{self._uimod_lpr.id}-tr{self._uimod_fusion.id}-scene')
        ]

        self._builder.unit(*essentials.create_unit__essentials__api(
            UNIT__ESSENTIALS0__API0, self._uimod_lpr.pid))
        self._builder.unit(*essentials.create_unit__essentials__counters(
            UNIT__ESSENTIALS0__COUNTERS0, spool_path))
        self._builder.unit(*bus.create_unit__media_server__media(
            UNIT__MEDIA_SERVER0__MEDIA0, add_scene_react_dir))
        self._builder.unit(*bus.create_unit__event_emitter__tx_track_bus(
            UNIT__EVENT_EMITTER0__TX_TRACK_BUS0, lpr_pri.get('translate-plate') or True,
            fs.join_path(spool_path, f'lr{self._uimod_lpr.id}-lr{self._uimod_lpr.id}-store'),
            fs.join_path(spool_path, f'lr{self._uimod_lpr.id}-tr{self._uimod_fusion.id}-track')))
        self._builder.unit(*bus.create_unit__event_emitter__tx_scene_bus(
            UNIT__EVENT_EMITTER0__TX_SCENE_BUS0, scene_event_jnode))
        self._builder.unit(*anpr.create_unit__anpr__anpr(
            UNIT__ANPR0__ANPR0, lpr_pri, lpr_pub, image_wh))

        self._builder.link_dist({UNIT__MEDIA_SERVER0__MEDIA0, UNIT__ESSENTIALS0__API0}, {UNIT__ANPR0__ANPR0})
        self._builder.link_dist({UNIT__ANPR0__ANPR0}, {UNIT__ESSENTIALS0__API0, UNIT__EVENT_EMITTER0__TX_SCENE_BUS0})
        self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__EVENT_EMITTER0__TX_SCENE_BUS0)

        units_seq = [UNIT__ANPR0__ANPR0]
        mmr_conf = lpr_pri.get(UNIT__MMR0__MMR0, None)
        if mmr_conf is not None:
            self._builder.unit(*mmr.create_unit__mmr___mmr(UNIT__MMR0__MMR0, mmr_conf))
            units_seq.append(UNIT__MMR0__MMR0)

        sbr_conf = lpr_pri.get(UNIT__SBR0__SBR0, None)
        if sbr_conf is not None:
            self._builder.unit(*sbr.create_unit__sbr___sbr(UNIT__SBR0__SBR0, sbr_conf))
            units_seq.append(UNIT__SBR0__SBR0)

        location_uimod = self._uimods_location[0] if len(self._uimods_location) > 0 else None

        if location_uimod:
            location_uiconf = uiref_to_conf_map.get(location_uimod.ref) or location_uimod.load_conf()
            speedcam_jnode = location_uiconf.get('ro', dict()).get('speedcam')
            if speedcam_jnode is not None:
                add_speedcam = speedcam_jnode.get('with-cam-n', False)
                if add_speedcam:
                    self._builder.unit(*speedcam.create_unit__speedcam___speedcam(
                        UNIT__SPEEDCAM0__SPEEDCAM0, location_uiconf, self._uimod_cam.conf_path, location_uimod.conf_path))
                    units_seq.append(UNIT__SPEEDCAM0__SPEEDCAM0)

        if len(self._uimods_ipcom_waga) == 1:
            waga_uimod = self._uimods_ipcom_waga[0]
            waga_conf = uiref_to_conf_map.get(waga_uimod.ref) or waga_uimod.load_conf()
            self._builder.unit(*waga.create_unit__ipcom___waga(UNIT__IPCOM0_WAGA0, waga_conf))
            units_seq.append(UNIT__IPCOM0_WAGA0)

        units_seq.append(UNIT__EVENT_EMITTER0__TX_TRACK_BUS0)
        self._builder.link_seq(units_seq)
        self._builder.link_dist([UNIT__ESSENTIALS0__COUNTERS0], units_seq)

        # create wiegand unit
        wiegand_conf = lpr_pri.get('wiegand-conf', None)
        if wiegand_conf:
            self._builder.unit(*wiegand.create_unit__wiegand___pcb(
                UNIT__WIEGAND_PCB0__RECIPIENT0, wiegand_conf))
            self._builder.unit(*recipient.create_unit__rule___rule2(
                UNIT__RULE_PCB0__RULE_PCB0, wiegand_conf['trigger']))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, UNIT__RULE_PCB0__RULE_PCB0, UNIT__WIEGAND_PCB0__RECIPIENT0])
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__WIEGAND_PCB0__RECIPIENT0)

        # create wiegand-udp unit
        wiegand_udp_conf = lpr_pri.get('wiegand-udp-conf', None)
        if wiegand_udp_conf:
            self._builder.unit(*wiegand.create_unit__wiegand___udp(UNIT__WIEGAND_UDP0__RECIPIENT0, wiegand_udp_conf))

            self._builder.unit(*recipient.create_unit__rule___rule2(
                UNIT__RULE_UDP0__RULE_UDP0, wiegand_udp_conf['trigger']))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, UNIT__RULE_UDP0__RULE_UDP0, UNIT__WIEGAND_UDP0__RECIPIENT0])
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__WIEGAND_UDP0__RECIPIENT0)

        dir_a = dir_r = dir_u = term_uuid = None
        if len(self._uimods_tcp_recipient):
            roadway_uiconf = uiref_to_conf_map.get(self._uimod_roadway.ref) or self._uimod_roadway.load_conf()
            ui_road_jnode = roadway_uiconf['rw']['args']
            ui_dir_jnode = ui_road_jnode.get('dir') or dict()
            dir_a = ui_dir_jnode.get('approaching')
            dir_r = ui_dir_jnode.get('receding')
            dir_u = ui_dir_jnode.get('unknown')
            term_uimod = Umod.create_term(self._uimod_lpr.pid)
            term_uuid = ph.TerminalUiconf(term_uimod.conf_path).get_uuid()

        tcp_index = 0
        for tcp_recipient_uimod in self._uimods_tcp_recipient:
            tcp_index += 1
            tcp_recipient_conf = uiref_to_conf_map.get(tcp_recipient_uimod.ref) or tcp_recipient_uimod.load_conf()

            self._builder.unit(*recipient.create_unit__recipient___vast2(tcp_recipient_uimod.ref, tcp_recipient_conf,
                                                                         dir_a, dir_r, dir_u, term_uuid))

            rule_ref = UNIT__RULE_TCPx__RULE_TCPx.format(tcp_index, tcp_index)
            self._builder.unit(*recipient.create_unit__rule___rule(rule_ref, tcp_recipient_conf))

            self._builder.link_seq([UNIT__ANPR0__ANPR0, rule_ref, tcp_recipient_uimod.ref])
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, tcp_recipient_uimod.ref)

        if len(self._uimods_ledar) == 1:
            leddar_uimod = self._uimods_ledar[0]
            leddar_conf = uiref_to_conf_map.get(leddar_uimod.ref) or leddar_uimod.load_conf()
            leddar_endpoint = fs.join_path(spool_path, 'lg{0}-sc{0}-track'.format(self._uimod_lpr.id))

            self._builder.unit(*leddar.create_unit__leddar___leddar(
                UNIT__LEDDAR0__LEDDAR0, leddar_conf, leddar_endpoint))

            self._builder.link(UNIT__MEDIA_SERVER0__MEDIA0, UNIT__LEDDAR0__LEDDAR0)
            self._builder.link(UNIT__ESSENTIALS0__COUNTERS0, UNIT__LEDDAR0__LEDDAR0)

    def build(self) -> None:
        lr_dgmod = self._dgmod_lpr
        self._builder.build(lr_dgmod.conf_dir)
        lr_dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _LrDeleter(IDgConfDeleter):
    def __init__(self, uimod_lr: IUiConfProvider) -> None:
        self._dgmod_lr: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_LPR, uimod_lr.id)

    def run(self) -> None:
        self._dgmod_lr.uninstall()


def lr_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_LPR:
        return [_LrDeleter(target_uimod)]
    return []
