from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.mg.dgconf.mgd_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.mg.dgconf.mgd_unit import UnitConfBuilder
from edge.webserver.sysconf.dgmod.mg.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.mg.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.lpr import LprUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.mv_camera import MvCameraUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.radar import RadarUiConfProvider
from edge.webserver.sysconf.dgmod.mg.uiconf.redlight import RedlightUiConfProvider, RailroadUiConfProvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def mg_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_MV_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_lpr = target_uideps.find_any({UTYPE_LPR})
        uimod_ra = target_uideps.try_find_any({UTYPE_RADAR})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_psuctl = target_uideps.try_find_any({UTYPE_PSUCTL})
    uimods_railroad = target_uideps.try_find_all(UTYPE_DET_RAILROAD)
    uimods_redlight = target_uideps.try_find_all(UTYPE_DET_REDLIGHT)

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_redlight]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_railroad])
    ])

    if not activate:
        return []

    return [_MgCreator(uimod_cam, uimod_lpr, uimod_fusion, uimods_railroad, uimods_redlight, uimod_ra, uimod_psuctl)]


class _MgCreator(IDgConfBuilder):
    def __init__(self,
                 uimod_cam: IUiConfProvider,
                 uimod_lpr: IUiConfProvider,
                 uimod_fusion: IUiConfProvider,
                 uimods_railroad: list[IUiConfProvider],
                 uimods_redlight: list[IUiConfProvider],
                 uimod_ra: IUiConfProvider | None,
                 uimod_psuctl: IUiConfProvider | None) -> None:

        self._dgmod_mg: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_MEDIAGRABBER, uimod_cam.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam
        self._uimod_fusion: Final[IUiConfProvider] = uimod_fusion
        self._uimod_lpr: Final[IUiConfProvider] = uimod_lpr
        self._uimod_ra: Final[IUiConfProvider | None] = uimod_ra
        self._uimod_psuctl: Final[IUiConfProvider | None] = uimod_psuctl

        self._uimods_railroad: Final[list[IUiConfProvider]] = uimods_railroad
        self._uimods_redlight: Final[list[IUiConfProvider]] = uimods_redlight

        self._unit_builder = UnitConfBuilder(self._uimod_cam.id, ph.get_spool_dir(self._uimod_cam.pid))
        self._halt_builder = HaltConfBuilder(self._uimod_psuctl)

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_mg

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        providers = list[IUiConfProvider2]()

        uimod = self._uimod_cam
        providers.extend([
            MvCameraUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf()),
            FusionUiConfProvider(self._uimod_fusion.id),
            LprUiConfProvider(self._uimod_lpr.id)
        ])

        if self._uimod_ra:
            providers.extend([
                RadarUiConfProvider(self._uimod_ra.id)
            ])

        if len(self._uimods_redlight) == 1:
            uimod = self._uimods_redlight[0]
            providers.extend([
                RedlightUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        elif len(self._uimods_railroad) == 1:
            uimod = self._uimods_railroad[0]
            providers.extend([
                RailroadUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        for provider in providers:
            provider.setup_builders(self._unit_builder, self._halt_builder)

    def build(self) -> None:
        dgmod = self._dgmod_mg
        dgmod.install()

        self._unit_builder.build(self._dgmod_mg.conf_dir)
        dgmod.save_haltaction_conf(self._halt_builder.build())
        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        self.dgmod.sigint()


class _MgDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_mg: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_MEDIAGRABBER, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_mg.uninstall()


def mg_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_MV_CAMERA:
        return [_MgDeleter(target_uimod)]
    return []
