from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.unity import UnityConfBuilder
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def ms_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_MEDIASTORE}):
        return []

    try:
        uimod_ms = target_uideps.find_any({UTYPE_MEDIASTORE})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced

    if not activate:
        return []

    return [_MsCreator(uimod_ms)]


class _MsCreator(IDgConfBuilder):
    def __init__(self, uimod_ms: IUiConfProvider) -> None:
        self._dgmod_bs: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_BLOBSTORE, uimod_ms.id)

        self._uimod_ms: Final[IUiConfProvider] = uimod_ms
        self._unit_builder = UnityConfBuilder()
        self._main_conf: dict | None = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_bs

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_ms.pid)

        uiconf = uiref_to_conf_map.get(self._uimod_ms.ref) or self._uimod_ms.load_conf()
        pri_jnode = uiconf['ro']

        cleanup_jnode = pri_jnode.get('cleanup') or dict()

        limits_jnode = dict()

        if cleanup_jnode['max-du-mb'][0]:
            limits_jnode['max-du'] = cleanup_jnode['max-du-mb'][1]

        if cleanup_jnode['min-df-mb'][0]:
            limits_jnode['min-df'] = cleanup_jnode['min-df-mb'][1]

        if cleanup_jnode['max-abs-time-sec'][0]:
            limits_jnode['max-abs-time'] = cleanup_jnode['max-abs-time-sec'][1]

        if cleanup_jnode['max-rel-time-sec'][0]:
            limits_jnode['max-rel-time'] = cleanup_jnode['max-rel-time-sec'][1]

        # filesize = pri_jnode.get('gopsize-mb') or 16
        #
        main_conf = {
            'links': [
                ['blobstore-server.0', 'blobstore-manager.0'],
                ['blobstore-stats.0', 'blobstore-server.0'],
                ['blobstore-stats.0', 'blobstore-manager.0~chan.0'],
                ['blobstore-stats.0', 'blobstore-manager.0~chan.0~video-cache.0'],
                ['blobstore-stats.0', 'blobstore-manager.0~chan.1'],
                ['blobstore-stats.0', 'blobstore-manager.0~chan.1~video-cache.0']
            ],
            'nodes': [
                {
                    'bundle': ['libunity_essentials', '1.0.0'],
                    'type': 'essentials',
                    'name': 'essentials.0',
                    'units': [
                        {
                            'type': 'counters',
                            'name': 'blobstore-stats.0',
                            'conf': {
                                'stat-path': fs.join_path(spool_dir, 'zabbix', 'bs0'),
                                'stat-tout': 60
                            }
                        }
                    ]
                },
                {
                    'bundle': ['libunity_blobstore', '1.0.0'],
                    'type': 'blobstore-server',
                    'name': 'blobstore-server.0',
                    'conf': {
                        'endpoint': f'/run/edge/blobstore.{self._uimod_ms.id}-uds',
                        'thread-count': 4
                    }
                },
                {
                    'bundle': ['libunity_blobstore', '1.0.0'],
                    'type': 'blobstore-manager',
                    'name': 'blobstore-manager.0',
                    'units': [
                        {
                            'type': 'blobstore-manager~chan',
                            'name': 'blobstore-manager.0~chan.0',
                            'conf': {
                                'ref': '0',
                                'logs-name': 'bs-chan0',
                                'temp-dir': fs.join_path(spool_dir, 'bs0-bs0-tmp')
                            },
                            'units': [
                                {
                                    'type': 'blobstore-manager~chan~video-cache',
                                    'name': 'blobstore-manager.0~chan.0~video-cache.0',
                                    'conf': {
                                        'directory': fs.join_path(spool_dir, 'bs0-bs0-archive'),
                                        'logs-name': 'bs-chan0-archive',
                                        'limits': limits_jnode
                                    }
                                }
                            ]
                        },
                        {
                            'type': 'blobstore-manager~chan',
                            'name': 'blobstore-manager.0~chan.1',
                            'conf': {
                                'ref': '1',
                                'logs-name': 'bs-chan1',
                                'temp-dir': fs.join_path(spool_dir, 'bs0-bs1-tmp')
                            },
                            'units': [
                                {
                                    'type': 'blobstore-manager~chan~video-cache',
                                    'name': 'blobstore-manager.0~chan.1~video-cache.0',
                                    'conf': {
                                        'directory': fs.join_path(spool_dir, 'bs0-bs1-archive'),
                                        'logs-name': 'bs-chan1-archive',
                                        'limits': limits_jnode
                                    }
                                }
                            ]
                        }
                    ]
                }
            ]
        }

        # create setupstore.conf
        self._main_conf = main_conf

    def build(self) -> None:
        dgmod = self.dgmod
        dgmod.install()

        unity_conf_dir = fs.join_path(dgmod.conf_dir, 'unit.conf.d')
        fs.silent_rm_r(unity_conf_dir)
        fs.silent_mkdir_r(unity_conf_dir)

        conf_path = fs.join_path(unity_conf_dir, 'blobstore.conf')
        fs.write_json_file(self._main_conf, conf_path, pretty=True)

        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        self.dgmod.sighup()


class _MsDeleter(IDgConfDeleter):
    def __init__(self, uimod_ms: IUiConfProvider) -> None:
        self._dgmod_bs: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_BLOBSTORE, uimod_ms.id)
        self._uimod_ms: Final[IUiConfProvider] = uimod_ms

    def run(self) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_ms.pid)
        archive_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-archive'.format(self._uimod_ms.pid))
        tmp_dir = fs.join_path(spool_dir, 'bs{0}-bs{0}-tmp'.format(self._uimod_ms.pid))

        fs.silent_rm_r(archive_dir)
        fs.silent_rm_r(tmp_dir)
        self._dgmod_bs.uninstall()


def ms_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_MEDIASTORE:
        return [_MsDeleter(target_uimod)]
    return []
