from typing import Final
import os

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, IDgConfDeleter, Udeps, UrefToConfMap
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


class _PsuCreator(IDgConfBuilder):
    def __init__(self, uimod_psu: IUiConfProvider) -> None:
        self._uiref: Final[str] = uimod_psu.ref

    @property
    def dgmod(self) -> dg.DgMod | None:
        return None

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        pass

    def build(self) -> None:
        preup_conf = '\n'.join([
            '#!/bin/sh -e',
            '',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} camera-on',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} radar-on',

            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-ethernet',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-camera',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-radar',
        ])

        postdown_conf = '\n'.join([
            '#!/bin/sh',
            '',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-camera',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-radar',
            f'/opt/edge/bin/psuctl run-uref {self._uiref} led-ethernet',
        ])

        # def run(self) -> None:
        preup_path = f'/etc/edge/controller.0/preup.d/edge-{self._uiref}'
        postdown_path = f'/etc/edge/controller.0/postdown.d/edge-{self._uiref}'

        fs.write_text_file([preup_conf], preup_path)
        fs.write_text_file([postdown_conf], postdown_path)
        os.chmod(preup_path, 0o755)
        os.chmod(postdown_path, 0o755)

    def notify(self) -> None:
        pass


class _PsuDeleter(IDgConfDeleter):
    def __init__(self, uimod_psu: IUiConfProvider) -> None:
        self._filename = uimod_psu.ref

    def run(self) -> None:
        preup_path = f'/etc/edge/controller.0/preup.d/{self._filename}.sh'
        postdown_path = f'/etc/edge/controller.0/postdown.d/{self._filename}.sh'

        fs.silent_rm_r(preup_path)
        fs.silent_rm_r(postdown_path)


def psu_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_PSUCTL}):
        return []

    try:
        uimod_psu = target_uideps.find_any({UTYPE_PSUCTL})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_psu.ref)
    ])

    if not activate:
        return []

    return [_PsuCreator(uimod_psu)]


def psu_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_PSUCTL:
        return [_PsuDeleter(target_uimod)]

    return []
