from abc import ABCMeta, abstractmethod

import ujson

from typing import Any, Final

from twisted.internet import reactor
from twisted.internet.interfaces import IReactorCore
from twisted.internet.protocol import Protocol, ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

from edge import eventlog
from edge.support import fs

_QUERY_DONE = 1
_QUERY_INVALID_STATE = 2
_QUERY_INVALID_ARG = 3


def _get_reactor():
    """
    :rtype: IReactorCore
    """
    return reactor


# noinspection PyMethodMayBeStatic
class _IRadar(metaclass=ABCMeta):
    @abstractmethod
    def query_set_alignment(self, mode: int, speed: int, query_arg: Any) -> int:
        pass

    @abstractmethod
    def query_set_location(
            self,
            camera_f: float,
            camera_l: float,
            camera_t: float,
            radar_f: float,
            radar_l: float,
            radar_u: float,
            radar_fl: float,
            radar_fu: float,
            radar_lu: float,
            camera_fl: float,
            camera_fu: float,
            camera_lu: float,
            sensor_wh_mm: tuple[float, float],
            sensor_wh_px: tuple[int, int],
            focal_dist: int,
            query_arg: Any) -> int:
        pass


class _IRadarHandler(metaclass=ABCMeta):
    @abstractmethod
    def connection_made(self, radar: _IRadar) -> None:
        pass

    @abstractmethod
    def connection_lost(self, query_arg: Any) -> None:
        pass

    @abstractmethod
    def handle_set_alignment(self, status: int, query_arg: Any) -> None:
        pass

    @abstractmethod
    def handle_set_location(self, status: int, query_arg: Any) -> None:
        pass


class _UdsTransport(LineOnlyReceiver):
    _UTF8_ENCODING = 'utf-8'
    delimiter = b'\0'

    def connectionMade(self):
        raise NotImplementedError

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        raise NotImplementedError

    def lineReceived(self, raw_response):
        unicode_reply = raw_response.decode(_UdsTransport._UTF8_ENCODING)
        eventlog.log_d(f'raw-reply: {unicode_reply}')

        try:
            reply_jnode = ujson.loads(unicode_reply)
            self.rx_reply(reply_jnode)
        except ValueError:
            eventlog.log_x()

    def rx_reply(self, reply_jnode: dict) -> None:
        raise NotImplementedError

    def tx_query(self, query_jnode: dict) -> int:
        if not isinstance(query_jnode, dict):
            eventlog.log_d('fail: invalid query')
            raise ValueError

        unicode_query = ujson.dumps(query_jnode)
        eventlog.log_d(f'raw-query: {unicode_query}')

        utf8_query = unicode_query.encode(_UdsTransport._UTF8_ENCODING)
        self.sendLine(utf8_query)
        return _QUERY_DONE


class _Radar(_UdsTransport, _IRadar):
    _QUERY_SET_LOCATION = 1,
    _QUERY_SET_ALIGNMENT = 2

    def __init__(self, radar_handler: _IRadarHandler) -> None:
        _UdsTransport.__init__(self)

        if not isinstance(radar_handler, _IRadarHandler):
            eventlog.log_w('fail: invalid value of radar-handler')
            raise ValueError

        self._radar_handler = radar_handler
        self._query_action = None
        self._query_arg = None

    def connectionMade(self):
        eventlog.log_w('done: open ra-socket')
        self._query_arg = None
        self._radar_handler.connection_made(self)

    def connectionLost(self, reason=connectionDone):
        eventlog.log_d('done: close ra-socket {}'.format(reason))

        query_arg = self._query_arg
        self._query_arg = None

        self._radar_handler.connection_lost(query_arg)

    def rx_reply(self, reply_jnode: dict) -> None:
        if self._query_action == _Radar._QUERY_SET_LOCATION:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_location(status, query_arg)

        elif self._query_action == _Radar._QUERY_SET_ALIGNMENT:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_alignment(status, query_arg)

        else:
            eventlog.log_w('fail: invalid query-action')

    def query_set_alignment(self, mode: str, speed: int, query_arg: Any) -> int:
        if self._query_arg:
            eventlog.log_w('fail: another query in progress')
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_w('fail: invalid value of query-arg')
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_ALIGNMENT
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            'action': 'alignment',
            'args': {
                'sensor': {
                    'mode': mode,
                    'mode-alignment': {
                        'speed-filter': speed
                    }
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE

    def query_set_location(
            self,
            camera_f: float,
            camera_l: float,
            camera_u: float,
            radar_f: float,
            radar_l: float,
            radar_u: float,
            radar_fl: float,
            radar_fu: float,
            radar_lu: float,
            camera_fl: float,
            camera_fu: float,
            camera_lu: float,
            sensor_wh_mm: tuple[float, float],
            sensor_wh_px: tuple[int, int],
            focal_dist: int,
            query_arg: Any) -> int:

        if self._query_arg:
            eventlog.log_d('fail: another query in progress')
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_d('fail: invalid value of query-arg')
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_LOCATION
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            'action': 'location',
            'args': {
                'sensor': {
                    'position': {
                        'forward': radar_f,
                        'left': radar_l,
                        'up': radar_u
                    },
                    'rotation': {
                        'elevation': radar_fu,
                        'azimuth': radar_fl,
                        'pitch': radar_lu
                    }
                },
                'camera': {
                    'position': {
                        'forward': camera_f,
                        'left': camera_l,
                        'up': camera_u
                    },
                    'rotation': {
                        'elevation': camera_fu,
                        'azimuth': camera_fl,
                        'pitch': camera_lu
                    },
                    'frustum': {
                        'sensor-wh': sensor_wh_mm,
                        'focal-dist': focal_dist,
                        'forward': [5, 100]
                    },
                    'image-wh': sensor_wh_px
                },
                'image-mapper': {
                    'rhc-to-image-map': [
                        [1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [0, 0, 0, 1]]
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE


class _RadarClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, radar_handler):
        self._radar_action_handler = radar_handler

    def buildProtocol(self, addr) -> Protocol:
        self.resetDelay()
        protocol = _Radar(self._radar_action_handler)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason) -> None:
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason) -> None:
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


class ApiConfBuilder(_IRadarHandler):
    def __init__(self, uds_filepath: str, lock_path: str) -> None:
        self._uds_path: Final[str] = uds_filepath
        self._lock_path: Final[str] = lock_path

        self._camera_f: float | None = None
        self._camera_l: float | None = None
        self._camera_u: float | None = None

        self._radar_f: float | None = None
        self._radar_l: float | None = None
        self._radar_u: float | None = None

        self._radar_fl: float | None = None  # forward-to-left angle (azimuth)
        self._radar_fu: float | None = None  # forward-to-up angle (elevation)
        self._radar_lu: float | None = None  # left-to-up angle (roll)

        self._camera_fl: float | None = None  # forward-to-left angle (azimuth)
        self._camera_fu: float | None = None  # forward-to-up angle (elevation)
        self._camera_lu: float | None = None  # left-to-up angle (roll)

        self._sensor_wh_mm: tuple[float, float] | None = None
        self._sensor_wh_px: tuple[int, int] | None = None
        self._focal_dist: int | None = None

        self._uds_reply_stat = 2

    def setup_unit_position(self, camera_f: float, camera_l: float, camera_u: float,
                            radar_f: float, radar_l: float, radar_u: float) -> None:
        """
        :param camera_f:
            values are in range [-10.0; 10.0], m
        :param camera_l:
            values are in range [-10.0; 10.0], m
        :param camera_u:
            values are in range [0; 10.0], m
        :param radar_f:
            values are in range [-10.0; 10.0], m
        :param radar_l:
            values are in range [-10.0; 10.0], m
        :param radar_u:
            values are in range [0; 10.0], m
        """
        self._camera_f = camera_f
        self._camera_l = camera_l
        self._camera_u = camera_u
        self._radar_f = radar_f
        self._radar_l = radar_l
        self._radar_u = radar_u

    def setup_radar_orientation(self, radar_fl: float, radar_fu: float, radar_lu: float) -> None:
        """
        :param radar_fl:
            values are in range [-15.0; 15.0], degrees
        :param radar_fu:
            values are in range [0; 9.0], degrees
        :param radar_lu:
            values are in range [-5.0; 5.0], degrees
        """
        self._radar_fl = radar_fl
        self._radar_fu = radar_fu
        self._radar_lu = radar_lu

    def setup_camera_orientation(self, camera_fl: float, camera_fu: float, camera_lu: float) -> None:
        """
        :param camera_fl:
            values are in range [-15.0; 15.0], degrees
        :param camera_fu:
            values are in range [0; 9.0], degrees
        :param camera_lu:
            values arein range [-5.0; 5.0], degrees
        """
        self._camera_fl = camera_fl
        self._camera_fu = camera_fu
        self._camera_lu = camera_lu

    def setup_camera(self, sensor_wh_px: tuple[int, int], sensor_wh_mm: tuple[float, float], focal_dist: int) -> None:
        """
        :param sensor_wh_px:
            size of camera sensor in px
        :param sensor_wh_mm:
            size of camera sensor in mm
        :param focal_dist:
            focal distance of camera, mm
        """
        self._sensor_wh_mm = sensor_wh_mm
        self._sensor_wh_px = sensor_wh_px
        self._focal_dist = focal_dist

    def connection_made(self, radar: _IRadar) -> None:
        radar.query_set_location(
            self._camera_f, self._camera_l, self._camera_u,
            self._radar_f, self._radar_l, self._radar_u,
            self._radar_fl, self._radar_fu, self._radar_lu,
            self._camera_fl, self._camera_fu, self._camera_lu,
            self._sensor_wh_mm, self._sensor_wh_px, self._focal_dist,
            'str')
        eventlog.log_d('')

    def connection_lost(self, query_arg) -> None:
        eventlog.log_d('')

    def handle_set_location(self, status: int, query_arg: Any) -> None:
        eventlog.log_d(f'status: {status}')
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    def handle_set_alignment(self, status: int, query_arg: Any) -> None:
        eventlog.log_d(f'status: {status}')
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    def build(self) -> None:
        @fs.Lock(self._lock_path, try_count=2)
        def fn() -> None:
            client_factory = _RadarClientFactory(radar_handler=self)
            twisted_reactor = _get_reactor()
            twisted_reactor.connectUNIX(self._uds_path, client_factory)

            twisted_reactor.callLater(15, twisted_reactor.stop)
            twisted_reactor.addSystemEventTrigger('before', 'shutdown', eventlog.log_d, 'Shutdown...')

            twisted_reactor.run()

            if self._uds_reply_stat != 1:
                raise RuntimeError(f'fail: invalid uds-reply-stat ({self._uds_reply_stat})')

        fn()
