from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.rg.rg_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.rg.rg_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.rg.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.rg.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.lpr import LprUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.radar import RadarUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.redlight import RedlightUiConfProvider, RailroadUiConfProvider
from edge.webserver.sysconf.dgmod.rg.uiconf.rtsp_camera import RtspCameraUiConfProvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def rg_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_RTSP_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_RTSP_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_lpr = target_uideps.find_any({UTYPE_LPR})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_ra = target_uideps.try_find_any({UTYPE_RADAR})
    uimod_psuctl = target_uideps.try_find_any({UTYPE_PSUCTL})

    uimods_rl = target_uideps.try_find_all(UTYPE_DET_REDLIGHT)
    uimods_rc = target_uideps.try_find_all(UTYPE_DET_RAILROAD)

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_rl]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_rc])
    ])

    if not activate:
        return []

    return [_RgCreator(uimod_cam, uimod_lpr, uimod_fusion, uimods_rl, uimods_rc, uimod_ra, uimod_psuctl)]


class _RgCreator(IDgConfBuilder):
    def __init__(self, cam_uimod: IUiConfProvider,
                 lpr_uimod: IUiConfProvider,
                 fusion_uimod: IUiConfProvider,
                 rl_uimods: list[IUiConfProvider],
                 railroad_crossing_uimods: list[IUiConfProvider],
                 ra_uimod: IUiConfProvider | None,
                 uimod_psuctl: IUiConfProvider | None) -> None:

        self._dgmod_rg: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RTSP_GRABBER, cam_uimod.id)

        self._uimod_cam: Final[IUiConfProvider] = cam_uimod
        self._uimod_fusion: Final[IUiConfProvider] = fusion_uimod
        self._uimod_lpr: Final[IUiConfProvider] = lpr_uimod
        self._uimod_ra: Final[IUiConfProvider | None] = ra_uimod

        self._uimods_rl: Final[list[IUiConfProvider]] = railroad_crossing_uimods
        self._uimods_rc: Final[list[IUiConfProvider]] = rl_uimods

        self._uimod_psuctl: Final[IUiConfProvider | None] = uimod_psuctl

        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_rg

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        main_builder = MainConfBuilder(self._uimod_cam.id, ph.get_spool_dir(self._uimod_cam.pid))
        halt_builder = HaltConfBuilder(self._uimod_psuctl)

        providers = list[IUiConfProvider2]()

        uimod = self._uimod_cam
        providers.extend([
            RtspCameraUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf()),
            FusionUiConfProvider(self._uimod_fusion.id),
            LprUiConfProvider(self._uimod_lpr.id)
        ])

        if self._uimod_ra:
            providers.extend([
                RadarUiConfProvider(self._uimod_ra.id)
            ])

        if len(self._uimods_rc) == 1:
            uimod = self._uimods_rc[0]
            providers.extend([
                RedlightUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        elif len(self._uimods_rl) == 1:
            uimod = self._uimods_rl[0]
            providers.extend([
                RailroadUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        for provider in providers:
            provider.setup_builders(main_builder, halt_builder)

        main_conf = main_builder.build()
        halt_conf = halt_builder.build()

        self._main_halt = main_conf, halt_conf

    def build(self) -> None:
        main_conf, halt_conf = self._main_halt  # type: dict, str

        dgmod = self._dgmod_rg
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_haltaction_conf(halt_conf)
        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _RgDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_cam: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RTSP_GRABBER, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_cam.uninstall()


def rg_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_RTSP_CAMERA:
        return [_RgDeleter(target_uimod)]

    return []
