from copy import deepcopy
from typing import Final

from edge.support import fs


class _CamServiceConfBuilder:
    def __init__(self, name: str) -> None:
        self._name: str = name
        self._intrinsics_path: str | None = None
        self._extrinsics_path: str | None = None

    def setup_intrinsics(self, intrinsics_path: str) -> None:
        self._intrinsics_path = intrinsics_path

    def setup_extrinsics(self, extrinsics_path: str) -> None:
        self._extrinsics_path = extrinsics_path

    def build(self) -> dict:
        if self._intrinsics_path is None or self._extrinsics_path is None:
            return dict()

        return {
            self._name: {
                'intrinsics': self._intrinsics_path,
                'extrinsics': self._extrinsics_path
            }
        }


class _RoadMaskBuilder:
    def __init__(self, name: str) -> None:
        self._name: str = name
        self._enabled: bool = False
        self._conf: dict | None = None

    def setup_enabled(self, enabled: bool) -> None:
        self._enabled = enabled

    def setup_conf(self, conf: dict) -> None:
        self._conf = conf

    def build(self) -> dict:
        if not self._enabled or self._conf is None or not self._conf['lanes']:
            return dict()

        return {
            self._name: {
                'lanes': self._conf['lanes'],
                'scene-wh': self._conf['scene-wh']
            }
        }


class _VehicleAnnoBuilder:
    def __init__(self) -> None:
        self._lr: dict | None = None
        self._ra: dict | None = None
        self._ra_enabled: bool = False

    def setup_lr(self, lr: dict) -> None:
        self._lr = lr

    def setup_ra(self, ra: dict) -> None:
        self._ra = ra

    def setup_ra_enabled(self, ra_enabled: bool) -> None:
        self._ra_enabled = ra_enabled

    def build(self) -> dict:
        if None in (self._lr, self._ra):
            return dict()

        if self._ra_enabled:
            return {
                'lr': self._lr,
                'ra': self._ra
            }

        else:
            return {
                'lr': self._lr
            }


class _MarkingsBuilder:
    def __init__(self) -> None:
        self._with_line_f: bool = False
        self._with_line_l: bool = False
        self._markings: dict | None = None

    def setup_enabled(self, with_line_f: bool, with_line_l: bool) -> None:
        self._with_line_f = with_line_f
        self._with_line_l = with_line_l

    def setup_markings_conf(self, markings: dict) -> None:
        self._markings = markings

    def build(self) -> dict:
        if self._markings is None:
            return dict()

        markings_jnode = deepcopy(self._markings)

        if not self._with_line_f:
            markings_jnode.pop('line-f', None)

        if not self._with_line_l:
            markings_jnode.pop('line-l', None)

        return markings_jnode


class MainConfBuilder:
    def __init__(self, uimod_id: str, spool_dir: str) -> None:
        self.uimod_id: Final[str] = uimod_id
        self._spool_dir: Final[str] = spool_dir

        self._cam_builder = _CamServiceConfBuilder('cam')
        self._road_mask_builder = _RoadMaskBuilder('road-mask')
        self._vehicle_anno_builder = _VehicleAnnoBuilder()
        self._markings_builder = _MarkingsBuilder()

        self._jnode = {
            'services': dict(),
            'scene-event': {
                'max-file-count': 8,
                'endpoints': []
            },
            'Renderer': {
                'profile': {
                    'annotation': {
                        'vehicle': dict(),
                        'trafficlight': [],
                        'markings': dict()
                    }
                }
            }
        }

    def setup_tracker_react(self, react_dir: str) -> None:
        self._jnode.update({
            'Source': {
                'Url': fs.join_path(self._spool_dir, react_dir)
            }
        })

    def setup_trafficlight_react(self, react_dir: str, tl_id: int) -> None:
        self._jnode['services'].update({
                'trafficlight': {
                    'endpoint': fs.join_path(self._spool_dir, react_dir),
                    'ref': tl_id,
                    'red-only': False,
                    'precision-ms': 0
                }
            })

    def setup_event_blobs(self, blobs_dir: str, codec: str) -> None:
        self._jnode.update({
            'image-store': {
                'format': codec,
                'endpoint': fs.join_path(self._spool_dir, blobs_dir),
                'max-count': 8
            }
        })

    def setup_event_args(self, args_dirs: list[str]) -> None:
        jnode: dict = self._jnode['scene-event']
        endpoints_jnode: list[str] = jnode['endpoints']

        for args_dir in args_dirs:
            endpoints_jnode.append(fs.join_path(self._spool_dir, args_dir))

    def setup_service_cam__intrinsics(self, intrinsics_path: str) -> None:
        self._cam_builder.setup_intrinsics(intrinsics_path)

    def setup_service_cam__extrinsics(self, extrinsics_path: str) -> None:
        self._cam_builder.setup_extrinsics(extrinsics_path)

    def setup_service_origin(self, conf_path: str) -> None:
        self._jnode['services'].update({
            'origin': conf_path
        })

    def setup_service_datetime(self, conf_path: str) -> None:
        self._jnode['services'].update({
            'datetime': conf_path
        })

    def setup_service_gnss(self, conf_path: str) -> None:
        self._jnode['services'].update({
            'geolocation': conf_path
        })

    def setup_service_stats(self, stats_path: str) -> None:
        self._jnode['services'].update({
            'monitoring': {
                'path': fs.join_path(self._spool_dir, stats_path),
                'tout': 30
            }
        })

    def setup_image_footer(self, image_w: int, footer: dict) -> None:
        self._jnode['Renderer']['profile'].update({
            'width': image_w,
            'footer': footer
        })

    def setup_image_anno__conf(self, anno_conf: dict) -> None:
        """
        :param anno_conf:
            http-streaming/0.json#/rw/live/renderer/vehicle
            rtsp-streaming/0.json#/rw/live/renderer/vehicle
        """
        self._vehicle_anno_builder.setup_lr({
            'plate': anno_conf.get('video-plate', 'never'),
            'dir': anno_conf.get('video-dir', 'never'),
            'lane': anno_conf.get('video-lane', 'never'),
            'trace': anno_conf.get('video-plate-trace', 'never'),
            'center': anno_conf.get('video-plate-center', 'never'),
            'vehicle-bbox': anno_conf.get('video-vehicle-bbox', 'never'),
            'vehicle-cuboid': anno_conf.get('video-vehicle-cuboid', 'never')
        })

        self._vehicle_anno_builder.setup_ra({
            'target': anno_conf.get('radar-target', 'never'),
            'vehicle': anno_conf.get('radar-vehicle', 'never'),
            'speed': anno_conf.get('radar-speed', 'never'),
            'len': anno_conf.get('radar-length', 'never'),
            'dir': anno_conf.get('radar-dir', 'never')
        })

    def setup_image_tl_anno__conf(self, tl_anno_conf: list) -> None:
        self._jnode['Renderer']['profile']['annotation']['trafficlight'] = tl_anno_conf

    def setup_image_anno__ra_enabled(self, enabled: bool) -> None:
        self._vehicle_anno_builder.setup_ra_enabled(enabled)

    def setup_image_markings__enabled(self, with_line_f: bool, with_line_l: bool) -> None:
        self._markings_builder.setup_enabled(with_line_f, with_line_l)

    def setup_image_markings__conf(self, markings: dict) -> None:
        self._markings_builder.setup_markings_conf(markings)

    def setup_spotspeed(self, speed_limit: int, max_overspeed: int) -> None:
        self._jnode['Renderer'].update({
            'YellowSpeedThreshold': speed_limit,
            'RedSpeedThreshold': speed_limit + max_overspeed
        })

    def setup_roadway__enabled(self, enabled: bool) -> None:
        self._road_mask_builder.setup_enabled(enabled)

    def setup_roadway__conf(self, conf: dict) -> None:
        """
        :param conf:
            /etc/edge/setupstore/admin/actions/pipeline/0/lane-enforcement/0.json#/rw/args
        """
        self._road_mask_builder.setup_conf(conf)

    def build(self) -> dict:
        self._jnode['services'].update(self._cam_builder.build())
        self._jnode.update(self._road_mask_builder.build())

        self._jnode['Renderer']['profile']['annotation']['vehicle'].update(
            self._vehicle_anno_builder.build())

        self._jnode['Renderer']['profile']['annotation']['markings'].update(
            self._markings_builder.build())

        return self._jnode
