from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.rs.rs_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.rs.rs_setupstore import SetupStoreConfBuilder
from edge.webserver.sysconf.dgmod.rs.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.rs.uiconf.camera import MvCameragUiConfProvider
from edge.webserver.sysconf.dgmod.rs.uiconf.streaming import RtspStreamingUiConfProvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def rs_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_RTSP_STREAMING}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA})
        uimod_rstreaming = target_uideps.find_any({UTYPE_RTSP_STREAMING})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref),
        uiref_to_newconf_map.get(uimod_rstreaming.ref)
    ])

    if not activate:
        return []

    return [_RsCreator(uimod_cam, uimod_rstreaming)]


class _RsCreator(IDgConfBuilder):
    def __init__(self, cam: IUiConfProvider, rs: IUiConfProvider) -> None:
        self._dgmod_rs: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RTSP_SERVER, rs.id)

        self._uimod_cam: Final[IUiConfProvider] = cam
        self._uimod_rs: Final[IUiConfProvider] = rs

        self._info_main_ss = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_rs

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        ss_builder = SetupStoreConfBuilder()
        main_builder = MainConfBuilder(self._uimod_rs.id, ph.get_spool_dir(self._uimod_rs.pid))

        providers = list[IUiConfProvider2]()

        uimod = self._uimod_rs
        providers.extend([
            RtspStreamingUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
        ])

        uimod = self._uimod_cam
        providers.extend([
            MvCameragUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
        ])

        for provider in providers:
            provider.setup_builders(main_builder, ss_builder)

        main_conf = main_builder.build()
        ss_active = ss_builder.build()

        self._info_main_ss = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._info_main_ss  # type: dict, bool

        dgmod = self._dgmod_rs
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _RsDeleter(IDgConfDeleter):
    def __init__(self, uimod_rstreaming: IUiConfProvider) -> None:
        self._dgmod_rstreaming: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RTSP_SERVER, uimod_rstreaming.id)

    def run(self) -> None:
        self._dgmod_rstreaming.uninstall()


def rs_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_RTSP_STREAMING:
        return [_RsDeleter(target_uimod)]

    return []
