from copy import deepcopy
from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.db import DbUiconf
from edge.webserver.sysconf.uiconf.terminal import TermAboutUiconf
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uimod.systest import SystestUimod
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


def _mk_eventstore_react_path(ui_pid: str) -> str:
    return f'/var/spool/edge/system/sc{ui_pid}-es65535-fact'


# noinspection PyUnusedLocal
def sc_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_TRAFFIC_RULES}):
        return []

    try:
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_gnss = target_uideps.find_any({UTYPE_GNSS})
        uimod_roadway = target_uideps.find_any({UTYPE_ROADWAY})
        uimod_rules = target_uideps.find_any({UTYPE_TRAFFIC_RULES})
        uimod_term = target_uideps.find_any({UTYPE_TERMINAL})
        uimod_transit = target_uideps.find_any({UTYPE_DET_TRANSIT})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_cam = target_uideps.try_find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
    uimod_iit_ca = target_uideps.try_find_any({UTYPE_IIT_CA})
    uimod_iit_ca_enduser = target_uideps.try_find_any({UTYPE_IIT_CA_END_USER})
    uimod_leddar = target_uideps.try_find_any({UTYPE_LEDDAR})
    uimod_location = target_uideps.try_find_any({UTYPE_LOCATION})

    uimods_buslane = target_uideps.try_find_all(UTYPE_DET_BUS_LANE)
    uimods_parking = target_uideps.try_find_all(UTYPE_DET_PARKING)
    uimods_railroad = target_uideps.try_find_all(UTYPE_DET_RAILROAD)
    uimods_redlight = target_uideps.try_find_all(UTYPE_DET_REDLIGHT)
    uimods_seatbelt = target_uideps.try_find_all(UTYPE_DET_SEATBELT)
    uimods_sectionspeed = target_uideps.try_find_all(UTYPE_DET_SECTION_SPEED)
    uimods_spotspeed = target_uideps.try_find_all(UTYPE_DET_SPOT_SPEED)
    uimods_spotspeed_waga = target_uideps.try_find_all(UTYPE_DET_SPOT_SPEED_WAGA)
    uimods_wanted = target_uideps.try_find_all(UTYPE_DET_WANTED)
    uimods_wrongway = target_uideps.try_find_all(UTYPE_DET_WRONG_WAY)

    activate = forced or any([
        uimod_cam and uiref_to_newconf_map.get(uimod_cam.ref),
        uimod_iit_ca and uiref_to_newconf_map.get(uimod_iit_ca.ref),
        uimod_iit_ca_enduser and uiref_to_newconf_map.get(uimod_iit_ca_enduser.ref),
        uimod_leddar and uiref_to_newconf_map.get(uimod_leddar.ref),
        uimod_location and uiref_to_newconf_map.get(uimod_location.ref),
        uiref_to_newconf_map.get(uimod_gnss.ref),
        uiref_to_newconf_map.get(uimod_roadway.ref),
        uiref_to_newconf_map.get(uimod_rules.ref),
        uiref_to_newconf_map.get(uimod_term.ref),
        uiref_to_newconf_map.get(uimod_transit.ref),

        any([uiref_to_newconf_map.get(x.ref) for x in uimods_buslane]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_parking]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_railroad]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_redlight]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_seatbelt]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_sectionspeed]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_spotspeed]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_spotspeed_waga]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_wanted]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_wrongway])
    ])

    if not activate:
        return []

    creator = _ScCreator(
        uimod_rules, uimod_roadway, uimod_fusion, uimod_transit, uimod_location, uimod_gnss, uimod_cam,
        uimod_leddar, uimod_term, uimods_wrongway, uimods_wanted, uimods_spotspeed, uimods_sectionspeed,
        uimods_redlight, uimods_railroad, uimods_parking, uimods_buslane, uimods_spotspeed_waga, uimods_seatbelt,
        uimod_iit_ca_enduser)

    return [creator]


class _MarkingsBuilder:
    def __init__(self) -> None:
        self._with_line_f: bool = False
        self._with_line_l: bool = False
        self._with_vehicle: bool = False
        self._markings: dict | None = None

    def setup_enabled(self, enabled: dict[str, bool]) -> None:
        self._with_line_f = enabled['line-f']
        self._with_line_l = enabled['line-l']
        self._with_vehicle = enabled['vehicle']

    def setup_markings_conf(self, markings: dict | None) -> None:
        self._markings = markings

    def build(self) -> dict:
        if self._markings is None:
            return {
                'line-f': [],
                'line-l': [],
                'vehicle': self._with_vehicle
            }

        markings_jnode: dict = deepcopy(self._markings)
        markings_jnode['vehicle'] = self._with_vehicle

        if not self._with_line_f:
            markings_jnode['line-f'] = []

        if not self._with_line_l:
            markings_jnode['line-l'] = []

        markings_jnode['line-f'] = markings_jnode.get('line-f') or []
        markings_jnode['line-l'] = markings_jnode.get('line-l') or []

        return markings_jnode


class _ScCreator(IDgConfBuilder):
    def __init__(self, rules_uimod: IUiConfProvider,
                 roadway_uimod: IUiConfProvider,
                 fusion_uimod: IUiConfProvider,
                 transit_uimod: IUiConfProvider,
                 location_uimod: IUiConfProvider,
                 gnss_uimod: IUiConfProvider,
                 cam_uimod: IUiConfProvider | None,
                 leddar_uimod: IUiConfProvider | None,
                 term_uimod: IUiConfProvider,

                 wrongway_uimods: list[IUiConfProvider],
                 wanted_uimods: list[IUiConfProvider],
                 spotspeed_uimods: list[IUiConfProvider],
                 sectionspeed_uimods: list[IUiConfProvider],
                 redlight_uimods: list[IUiConfProvider],
                 rc_uimods: list[IUiConfProvider],
                 parking_uimods: list[IUiConfProvider],
                 buslane_uimods: list[IUiConfProvider],
                 spotspeed_waga_uimods: list[IUiConfProvider],
                 seatbelt_uimods: list[IUiConfProvider],
                 iit_ca_enduser_uimod2: IUiConfProvider | None) -> None:

        self._edge_sc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_SPEEDCONTROL, rules_uimod.id)

        self._uimod_cam: Final[IUiConfProvider | None] = cam_uimod
        self._uimod_fusion: Final[IUiConfProvider] = fusion_uimod
        self._uimod_gnss: Final[IUiConfProvider] = gnss_uimod
        self._uimod_iit_ca_enduser: Final[IUiConfProvider | None] = iit_ca_enduser_uimod2
        self._uimod_leddar: Final[IUiConfProvider | None] = leddar_uimod
        self._uimod_location: Final[IUiConfProvider] = location_uimod
        self._uimod_roadway: Final[IUiConfProvider] = roadway_uimod
        self._uimod_rules: Final[IUiConfProvider] = rules_uimod
        self._uimod_term: Final[IUiConfProvider] = term_uimod
        self._uimod_transit: Final[IUiConfProvider] = transit_uimod

        self._uimods_buslane: Final[list[IUiConfProvider]] = buslane_uimods
        self._uimods_parking: Final[list[IUiConfProvider]] = parking_uimods
        self._uimods_rc: Final[list[IUiConfProvider]] = rc_uimods
        self._uimods_rl: Final[list[IUiConfProvider]] = redlight_uimods
        self._uimods_seatbelt: Final[list[IUiConfProvider]] = seatbelt_uimods
        self._uimods_sectionspeed: Final[list[IUiConfProvider]] = sectionspeed_uimods
        self._uimods_spotspeed: Final[list[IUiConfProvider]] = spotspeed_uimods
        self._uimods_spotspeed_waga: Final[list[IUiConfProvider]] = spotspeed_waga_uimods
        self._uimods_wanted: Final[list[IUiConfProvider]] = wanted_uimods
        self._uimods_wrongway: Final[list[IUiConfProvider]] = wrongway_uimods

        self._path_to_conf_map = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._edge_sc

    def setup(self, uiref_to_newconf_map: UrefToConfMap) -> None:
        ui_pid = self._uimod_rules.pid

        edge_sc: dg.DgMod = self._edge_sc
        roadway_uiconf = uiref_to_newconf_map.get(self._uimod_roadway.ref) or self._uimod_roadway.load_conf()
        transit_uiconf = uiref_to_newconf_map.get(self._uimod_transit.ref) or self._uimod_transit.load_conf()
        transit_pri = transit_uiconf['ro']
        main_conf = deepcopy(transit_pri.get('main', {}))

        spool_dir = ph.get_spool_dir(ui_pid)

        main_conf['services'] = {
            'stat': {
                'tout': 60,
                'path': fs.join_path(spool_dir, 'zabbix', 'sc0')
            },
            'gpsd': self._uimod_gnss.conf_path,
            'db.conf': DbUiconf.CONF_PATH,
            'tz.conf': TimeUiconf.CONF_PATH,
            'origin': TermAboutUiconf.get_conf_path(ui_pid),
            'event': {
                'endpoints': [_mk_eventstore_react_path(ui_pid)],
                'args-ua': transit_pri.get('require-args-ua', False),
                'local-arch-base': transit_pri.get('local-arch-base')
            },
            'react': {
                'endpoint': fs.join_path(spool_dir, f'tr{self._uimod_fusion.id}-sc{self._uimod_rules.id}-track')
            },
            'systest': SystestUimod.CONF_PATH,
            'plate-info': '/opt/edge/share/lpr/plate-info.json',
            'vs': transit_pri.get('use-vs', True)
        }

        if len(self._uimods_rl) == 1:
            rl_uimod = self._uimods_rl[0]
            red_light_jnode = uiref_to_newconf_map.get(rl_uimod.ref) or rl_uimod.load_conf()
            pri_jnode = red_light_jnode['ro']
            params_jnode = pri_jnode.get('params', {})
            tl_id = params_jnode.get('controller', [{}])[0].get('id', 1)
            red_light_only = params_jnode.get('red-light-only', False)
            tl_pressison = pri_jnode.get('tl-precision-ms', 0)
            main_conf['services']['trafficlight'] = {
                'endpoint': fs.join_path(spool_dir, f'tl0-sc{self._uimod_rules.id}-track'),
                'ref': tl_id,
                'red-only': red_light_only,
                'precision-ms': tl_pressison
            }

        elif len(self._uimods_rc) == 1:
            rc_uimod = self._uimods_rc[0]
            railroad_crossing_jnode = uiref_to_newconf_map.get(rc_uimod.ref) or rc_uimod.load_conf()
            pri_jnode = railroad_crossing_jnode['ro']
            params_jnode = pri_jnode.get('params', {})
            tl_id = params_jnode.get('controller', [{}])[0].get('id', 1)
            red_light_only = params_jnode.get('red-light-only', True)
            tl_pressison = pri_jnode.get('tl-precision-ms', 0)
            main_conf['services']['trafficlight'] = {
                'endpoint': fs.join_path(spool_dir, f'tl0-sc{self._uimod_rules.id}-track'),
                'ref': tl_id,
                'red-only': red_light_only,
                'precision-ms': tl_pressison
            }

        terminal_uiconf = uiref_to_newconf_map.get(self._uimod_term.ref) or self._uimod_term.load_conf()
        trainer_jnode = terminal_uiconf['ro'].get('trainer')
        if trainer_jnode:
            trainer_jnode2 = deepcopy(trainer_jnode)
            trainer_jnode2['directory'] = trainer_jnode2['directory'] + f'-{self._uimod_term.pid}'
            main_conf['services'].update({'trainer': trainer_jnode2})

        images_jnode = deepcopy(transit_uiconf['rw']['args']['images'])

        markings_builder = _MarkingsBuilder()
        markings_builder.setup_enabled(images_jnode['scene']['markings'])

        transit_conf = {
            'plate-detected-mode': transit_pri.get('plate-detected-mode', False),
            'with-video': transit_pri.get('with-video', False),
            'name': '',
            'enabled': True,
            'enable-overview': transit_pri.get('enable-overview', False),
            'images': {
                'best': images_jnode
            },
            'rules': [{
                'name': '',
                'enabled': True,
                'limits': {
                    'require-plate': transit_pri.get('require-plate', True)
                }
            }],
        }

        if self._uimod_location and self._uimod_cam:
            location_jnode = uiref_to_newconf_map.get(self._uimod_location.ref) or self._uimod_location.load_conf()
            location_pri_jnode = location_jnode['ro']
            location_pri_speedcam_jnode = location_pri_jnode.get('speedcam') or dict()

            transit_conf['images']['best']['scene']['markings']['line-f'] = []
            transit_conf['images']['best']['scene']['markings']['line-l'] = []

            markings_builder.setup_markings_conf(location_pri_jnode.get('markings', {}))

            main_conf['services']['cam'] = {
                'intrinsics': self._uimod_cam.conf_path,
                'extrinsics': self._uimod_location.conf_path
            }

        transit_conf['images']['best']['scene']['markings'] = markings_builder.build()

        path_to_conf_map = [
            (edge_sc.get_conf_path('main.conf'), main_conf),
            (edge_sc.get_conf_path('transit.conf'), transit_conf)
        ]

        if self._uimod_leddar:
            transit_conf['le-loop-ref'] = 'loop0'
            transit_conf['rules'][0]['limits']['require-plate'] = False
            main_conf['services']['leddar'] = fs.join_path(spool_dir, 'lg{}-sc{}-track'.format(
                self._uimod_leddar.id, self._uimod_rules.id))

        # add parking rules
        parking_path_to_conf_map = []
        for parking_uimod in self._uimods_parking:
            parking_rule_jnode = uiref_to_newconf_map.get(parking_uimod.ref) or parking_uimod.load_conf()
            k = edge_sc.get_conf_path(f'parking_{parking_uimod.id}.conf')
            v = parking_rule_jnode['rw']['args']
            parking_path_to_conf_map.append((k, v))

        for file_path, content in parking_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # # add spot-speed rules
        spotspeed_path_to_conf_map = []
        for spotspeed_uimod in self._uimods_spotspeed:
            spotspeed_rule_jnode = uiref_to_newconf_map.get(spotspeed_uimod.ref) or spotspeed_uimod.load_conf()
            h = spotspeed_rule_jnode['ro'].get('plate-over-ground')
            log = spotspeed_rule_jnode['ro'].get('log-cam-speed', False)

            jnode = deepcopy(spotspeed_rule_jnode['rw']['args'])
            if h is not None:
                jnode['plate-over-ground'] = h
                jnode['log-cam-speed'] = log

            k = edge_sc.get_conf_path(f'spotspeed_{spotspeed_uimod.id}.conf')
            spotspeed_path_to_conf_map.append((k, jnode))

        for file_path, content in spotspeed_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # # add spot-speed rules
        spotspeed_waga_path_to_conf_map = []
        for spotspeed_waga_uimod in self._uimods_spotspeed_waga:
            spotspeed_waga_rule_jnode = uiref_to_newconf_map.get(spotspeed_waga_uimod.ref)\
                                        or spotspeed_waga_uimod.load_conf()
            h = spotspeed_waga_rule_jnode['ro'].get('plate-over-ground')
            log = spotspeed_waga_rule_jnode['ro'].get('log-cam-speed', False)

            jnode = deepcopy(spotspeed_waga_rule_jnode['rw']['args'])
            if h is not None:
                jnode['plate-over-ground'] = h
                jnode['log-cam-speed'] = log

            k = edge_sc.get_conf_path(f'spotspeed_waga_{spotspeed_waga_uimod.id}.conf')
            spotspeed_waga_path_to_conf_map.append((k, jnode))

        for file_path, content in spotspeed_waga_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # # add section-speed rules
        sectionspeed_path_to_conf_map = []
        for sectionspeed_uimod in self._uimods_sectionspeed:
            sectionspeed_rule_jnode = uiref_to_newconf_map.get(sectionspeed_uimod.ref) or sectionspeed_uimod.load_conf()
            k = edge_sc.get_conf_path(f'sectionspeed_{sectionspeed_uimod.id}.conf')
            jnode = sectionspeed_rule_jnode['rw']['args']
            jnode['ignore-ra'] = sectionspeed_rule_jnode['ro'].get('ignore-ra', False)
            jnode['section']['assert-dist'] = sectionspeed_rule_jnode['ro'].get('assert-dist', 1000)
            sectionspeed_path_to_conf_map.append((k, jnode))

        for file_path, content in sectionspeed_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add wrong-way rules
        wrongway_path_to_conf_map = []
        for wrongway_uimod in self._uimods_wrongway:
            wrongway_rule_jnode = uiref_to_newconf_map.get(wrongway_uimod.ref) or wrongway_uimod.load_conf()
            k = edge_sc.get_conf_path(f'wrongway_{wrongway_uimod.id}.conf')
            v = wrongway_rule_jnode['rw']['args']
            wrongway_path_to_conf_map.append((k, v))

        for file_path, content in wrongway_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add red-light rules
        rl_path_to_conf_map = []
        for rl_uimod in self._uimods_rl:
            red_light_rule_jnode = uiref_to_newconf_map.get(rl_uimod.ref) or rl_uimod.load_conf()
            rl_args_jnode = deepcopy(red_light_rule_jnode['rw']['args'])
            rl_args_jnode['direction'] = red_light_rule_jnode['ro'].get('direction', 'receding')
            rl_args_jnode['on-stopbar'] = red_light_rule_jnode['ro'].get('on-stopbar', 'required')

            k = edge_sc.get_conf_path(f'redlight_{rl_uimod.id}.conf')
            rl_path_to_conf_map.append((k, rl_args_jnode))

        for file_path, content in rl_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add railroadcrossing rules
        rc_path_to_conf_map = []
        for rc_uimod in self._uimods_rc:
            rc_rule_jnode = uiref_to_newconf_map.get(rc_uimod.ref) or rc_uimod.load_conf()
            rc_args_jnode = deepcopy(rc_rule_jnode['rw']['args'])
            rc_args_jnode['direction'] = rc_rule_jnode['ro'].get('direction', 'receding')
            rc_args_jnode['on-stopbar'] = rc_rule_jnode['ro'].get('on-stopbar', 'required')

            k = edge_sc.get_conf_path(f'railroad-crossing_{rc_uimod.id}.conf')
            rc_path_to_conf_map.append((k, rc_args_jnode))

        for file_path, content in rc_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add wanted rules
        wanted_path_to_conf_map = []
        for wanted_uimod in self._uimods_wanted:
            wanted_rule_jnode = uiref_to_newconf_map.get(wanted_uimod.ref) or wanted_uimod.load_conf()
            k = edge_sc.get_conf_path(f'wanted_{wanted_uimod.id}.conf')
            v = wanted_rule_jnode['rw']['args']
            wanted_path_to_conf_map.append((k, v))

        for file_path, content in wanted_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add bus-lane rules
        buslane_path_to_conf_map = []
        for buslane_uimod in self._uimods_buslane:
            buslane_rule_jnode = uiref_to_newconf_map.get(buslane_uimod.ref) or buslane_uimod.load_conf()

            buslane_args_jnode = deepcopy(buslane_rule_jnode['rw']['args'])
            buslane_args_jnode['min-image-dist'] = buslane_rule_jnode['ro'].get('min-image-dist', 1e3)
            k = edge_sc.get_conf_path(f'buslane_{buslane_uimod.id}.conf')
            buslane_path_to_conf_map.append((k, buslane_args_jnode))

        for file_path, content in buslane_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        # add seatbelt rules
        seatbelt_path_to_conf_map = []
        for seatbelt_uimod in self._uimods_seatbelt:
            seatbelt_rule_jnode = uiref_to_newconf_map.get(seatbelt_uimod.ref) or seatbelt_uimod.load_conf()
            k = edge_sc.get_conf_path(f'seatbelt_{seatbelt_uimod.id}.conf')

            min_w = seatbelt_rule_jnode['ro'].get('min-weight', 0)

            jnode = deepcopy(seatbelt_rule_jnode['rw']['args'])
            if min_w is not None:
                for seat_jnode in jnode['seats']:
                    seat_jnode['min-weight'] = min_w

            seatbelt_path_to_conf_map.append((k, jnode))

        for file_path, content in seatbelt_path_to_conf_map:
            path_to_conf_map.append((file_path, content))

        def get_path_fn(name_conf: tuple[str, str]):
            return name_conf[0]

        # TODO: separate red-light and railroad-crossing unit
        main_conf['detectors'] = {
            'transit.conf': [edge_sc.get_conf_path('transit.conf')],
            'parking.conf': list(map(get_path_fn, parking_path_to_conf_map)),
            'spot-speed.conf': list(map(get_path_fn, spotspeed_path_to_conf_map)),
            'spot-speed-waga.conf': list(map(get_path_fn, spotspeed_waga_path_to_conf_map)),
            'wrong-way.conf': list(map(get_path_fn, wrongway_path_to_conf_map)),
            'seatbelt.conf': list(map(get_path_fn, seatbelt_path_to_conf_map)),
            'bus-lane.conf': list(map(get_path_fn, buslane_path_to_conf_map)),
            'red-light.conf': list(map(get_path_fn, rl_path_to_conf_map + rc_path_to_conf_map)),
            'railroad-crossing.conf': [],
            # 'red-light.conf': map(get_path_fn, redlight_path_to_conf_map),
            # 'railroad-crossing.conf': map(get_path_fn, railroad_crossing_path_to_conf_map),
            'wanted.conf': list(map(get_path_fn, wanted_path_to_conf_map)),
            'section-speed.conf': list(map(get_path_fn, sectionspeed_path_to_conf_map))
        }

        ui_road_jnode = roadway_uiconf['rw']['args']
        ui_dir_jnode = ui_road_jnode.get('dir') or dict()
        road_conf_jnode = {
            'road': {
                'dir': {
                    'approaching': ui_dir_jnode.get('approaching') or 'approaching',
                    'receding': ui_dir_jnode.get('receding') or 'receding',
                    'unknown': ui_dir_jnode.get('unknown') or 'unknown'
                }
            },
            'scene-wh': ui_road_jnode['scene-wh']}

        if len(ui_road_jnode['lanes']) > 0:
            road_conf_jnode['road']['lanes'] = ui_road_jnode['lanes']

        main_conf['services'].update({'road.conf': edge_sc.get_conf_path('road.conf')})

        k = edge_sc.get_conf_path('road.conf')
        path_to_conf_map.append((k, road_conf_jnode))

        # add sign
        if self._uimod_iit_ca_enduser:
            iit_ca_enduser_jnode = uiref_to_newconf_map.get(self._uimod_iit_ca_enduser.ref)
            iit_ca_enduser_uiconf = UiConf2(iit_ca_enduser_jnode or self._uimod_iit_ca_enduser.load_conf())

            iit_ca_enduser_uiconf_conf = iit_ca_enduser_uiconf.pub_conf
            active = iit_ca_enduser_uiconf_conf.get('enabled', False)
            if active:
                iit_conf_filepath = edge_sc.get_conf_path('sign.iit.conf')
                path_to_conf_map.append((iit_conf_filepath, {
                    'key-model': iit_ca_enduser_uiconf.pri['model'],
                    'key-serial': iit_ca_enduser_uiconf.pri['serial'],
                    'key-sign-pass': iit_ca_enduser_uiconf.pri['password'],
                    'append-cert': iit_ca_enduser_uiconf.pri.get('append-cert', False),
                    'hash-algo': 'gost34311',
                    'sign-algo': 'dstu4145'
                }))

                main_conf['services']['sign'] = {
                    'provider': {
                        'iit': iit_conf_filepath,
                        'uiconf': self._uimod_iit_ca_enduser.conf_path,
                        'uiref': self._uimod_iit_ca_enduser.ref,
                    }
                }

        self._path_to_conf_map = path_to_conf_map

    def build(self) -> None:
        path_to_conf_map: dict = self._path_to_conf_map

        edge_sc: dg.DgMod = self._edge_sc
        edge_sc.install()
        edge_sc.save_setupstore_conf(active=True)

        for conf_path, content in path_to_conf_map:
            fs.write_json_file(content, conf_path, pretty=True)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _ScDeleter(IDgConfDeleter):
    def __init__(self, uimod_rules: IUiConfProvider) -> None:
        self._dgmod_sc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_SPEEDCONTROL, uimod_rules.id)

    def run(self) -> None:
        self._dgmod_sc.uninstall()


def sc_create_dgconf_deleter(target_uimod: Umod) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_TRAFFIC_RULES:
        return [_ScDeleter(target_uimod)]

    return []
