from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.tl.tl_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.tl.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.tl.uiconf.camera import CameraUiConfProvider
from edge.webserver.sysconf.dgmod.tl.uiconf.railroad import RailroadUiConfProvider
from edge.webserver.sysconf.dgmod.tl.uiconf.redlight import RedlightUiConfProvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def tl_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_TRAFFIC_LIGHT}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
        uimod_rules = target_uideps.find_any({UTYPE_TRAFFIC_RULES})
        uimod_tl = target_uideps.find_any({UTYPE_TRAFFIC_LIGHT})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimods_railroad = target_uideps.try_find_all(UTYPE_DET_RAILROAD)
    uimods_redlight = target_uideps.try_find_all(UTYPE_DET_REDLIGHT)

    activate = forced or any([
        uiref_to_newconf_map.get(x.ref) for x in uimods_redlight]) or len(uimods_redlight) == 0 or \
            (any([uiref_to_newconf_map.get(x.ref) for x in uimods_railroad]) or len(uimods_railroad) == 0)

    if not activate:
        return []

    return [_TlCreator(uimod_cam, uimod_tl, uimod_rules, uimods_redlight, uimods_railroad)]


class _TlCreator(IDgConfBuilder):
    def __init__(self,
                 cam: IUiConfProvider,
                 tl: IUiConfProvider,
                 rules: IUiConfProvider,
                 redlight: list[IUiConfProvider],
                 railroad: list[IUiConfProvider]) -> None:

        self._dgmod_tl: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_TRAFFIC_LIGHT, tl.id)

        self._uimod_cam: Final[IUiConfProvider] = cam
        self._uimod_rules: Final[IUiConfProvider] = rules
        self._uimod_tl: Final[IUiConfProvider] = tl

        self._uimods_railroad: Final[list[IUiConfProvider]] = railroad
        self._uimods_redlight: Final[list[IUiConfProvider]] = redlight

        self._main_ss = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_tl

    def setup(self, uiref_to_newconf_map: UrefToConfMap) -> None:
        main_builder = MainConfBuilder(self._uimod_tl.id, ph.get_spool_dir(self._uimod_tl.pid))
        main_builder.setup_track_event_args([
            f'tl{main_builder.uimod_id}-sc{self._uimod_rules.id}-track'
            # , f'tl{main_builder.uimod_id}-or{self._rules_uimod.id}-track'
        ])

        providers = list[IUiConfProvider2]()

        if len(self._uimods_redlight) == 1:
            rl_uimod = self._uimods_redlight[0]
            rl_uiconf = uiref_to_newconf_map.get(rl_uimod.ref) or rl_uimod.load_conf()
            rl_uiconf_pri = rl_uiconf['ro']

            with_trafficlight_65535 = rl_uiconf_pri.get('with-trafficlight.65535', False)
            with_trafficlight = not with_trafficlight_65535

            providers.extend([
                CameraUiConfProvider(self._uimod_cam.id),
                RedlightUiConfProvider(rl_uimod.id, rl_uiconf)
            ])

        elif len(self._uimods_railroad) == 1:
            rc_uimod = self._uimods_railroad[0]
            rc_uiconf = uiref_to_newconf_map.get(rc_uimod.ref) or rc_uimod.load_conf()
            rc_uiconf_pri = rc_uiconf['ro']

            with_trafficlight_65535 = rc_uiconf_pri.get('with-trafficlight.65535', False)
            with_trafficlight = not with_trafficlight_65535

            providers.extend([
                CameraUiConfProvider(self._uimod_cam.id),
                RailroadUiConfProvider(rc_uimod.id, rc_uiconf)
            ])

        else:
            with_trafficlight = False

        for provider in providers:
            provider.setup_builders(main_builder)

        main_conf = main_builder.build()

        self._main_ss = main_conf, with_trafficlight

    def build(self) -> None:
        main_conf, ss_active = self._main_ss  # type: dict, bool

        dgmod = self._dgmod_tl
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        dg.LifeCycle.sighup_ss_dgmod()


class _TlDeleter(IDgConfDeleter):
    def __init__(self, uimod_tl: IUiConfProvider) -> None:
        self._dgmod_tl: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_TRAFFIC_LIGHT, uimod_tl.id)

    def run(self) -> None:
        self._dgmod_tl.uninstall()


def tl_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_TRAFFIC_LIGHT:
        return [_TlDeleter(target_uimod)]

    return []
