import os

from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import Final, Type

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import InternalError
from edge.webserver.sysconf.uiconf.rtsp_streaming import RtspStreamingUiconf
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uimod.utype import *

CTRL_DEF_UICONF_DIR = '/opt/edge/share/setupstore/ui/'
TERM_DEF_UICONF_DIR = fs.join_path(CTRL_DEF_UICONF_DIR, 'pipeline')


class _IUnitUconfBuilder(metaclass=ABCMeta):
    @abstractmethod
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        pass


class _UnitUconfBuilder(_IUnitUconfBuilder):
    def __init__(self, utype: str, defconf: dict | str) -> None:
        if isinstance(defconf, str):
            defconf_path = fs.join_path(TERM_DEF_UICONF_DIR, utype, defconf)
            self._defconf: dict = fs.load_json3(defconf_path)

        elif isinstance(defconf, dict):
            import copy
            self._defconf: dict = copy.deepcopy(defconf)

        else:
            eventlog.log_e(f'fail: utype:{utype}')
            raise InternalError

    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        return self._defconf


class _TerminalUconfBuilder(_UnitUconfBuilder):
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        if factory_conf is None:
            eventlog.log_e(f'fail: factory-conf:None')
            raise InternalError

        uconf = super().build(uref, factory_conf)

        term_uref = factory_conf['term-uref']

        uconf['ro']['uuid'] = factory_conf['term-uuid']
        uconf['ro']['ctime'] = datetime.now().isoformat()
        uconf['ro']['factory'] = factory_conf['factory']

        uconf['rw']['origin-place'] = factory_conf['term-pid']
        uconf['rw']['origin-name'] = term_uref

        return uconf


class _RtspCameraUconfBuilder(_UnitUconfBuilder):
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        uconf = super().build(uref, factory_conf)

        if factory_conf.get('with-intrinsic', False):
            uconf['rw']['conf']['intrinsic'] = {
                'active': False,
                'frustum': {
                    'pixel-size': 3.45,
                    'focal-dist': 50,
                    'sensor-size': [
                        1920, 1200
                    ]
                }
            }

        cam_jnode = factory_conf.get('rtsp-camera')
        if cam_jnode is not None:
            uconf['rw']['conf']['url'] = cam_jnode.get('url') or ''

        return uconf


class _RtspStreamingUconfBuilder(_UnitUconfBuilder):
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        if factory_conf is None:
            eventlog.log_e(f'fail: factory-conf:None')
            raise InternalError

        uconf = super().build(uref, factory_conf)

        uconf['rw']['streaming']['tcp-port'] += int(factory_conf['term-pid'])
        uconf['rw']['streaming']['url'] = RtspStreamingUiconf(uconf['rw']).rtsp_url
        return uconf


class _LprUconfBuilder(_UnitUconfBuilder):
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        uconf = super().build(uref, factory_conf)

        nn_detector = os.environ.get('EDGE_TARGET_AUTOSDK__WITH_NN_PLATE_DETECTOR') or 'None'
        if nn_detector == '1':
            uconf['rw']['plate-detector']['image-scale'] = 25
            uconf['rw']['plate-recognizer']['width']['min'] = 70
            uconf['rw']['plate-recognizer']['width']['max'] = 200

        return uconf


class _MvCameraUconfBuilder(_IUnitUconfBuilder):
    # noinspection PyUnusedLocal
    def __init__(self, utype: str, defconf: dict | str) -> None:
        pass

    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        if factory_conf is None:
            eventlog.log_e(f'fail: factory-conf:None')
            raise InternalError

        cam_vendor = factory_conf['vendor'].lower()
        cam_model = str.replace(str(factory_conf['model']), ' ', '-')

        import fnmatch
        model_files = os.listdir(fs.join_path(TERM_DEF_UICONF_DIR, UTYPE_MV_CAMERA))
        model_files2 = fnmatch.filter(model_files, f'{cam_vendor}_{cam_model}*jnode.json')
        for model_file in model_files2:
            # TODO: what to do if there are jnodes for one camera with different focal-lengths?
            uconf_path = fs.join_path(TERM_DEF_UICONF_DIR, UTYPE_MV_CAMERA, model_file)
            uconf = fs.load_json3(uconf_path)
            uconf['rw']['serial'] = str(factory_conf['serial'])
            return uconf

        eventlog.log_e(f'fail: unsupported (vendor:{cam_vendor}, model:{cam_model})')
        raise NotImplementedError


class _PsuCtlUconfBuilder(_IUnitUconfBuilder):
    # noinspection PyUnusedLocal
    def __init__(self, utype: str, defconf: dict | str) -> None:
        pass

    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        if factory_conf is None:
            eventlog.log_e(f'fail: factory-conf:None')
            raise InternalError

        uiconf = UiConf2(dict())
        uiconf.set_pri({
            'lock-path': f'/var/run/edge/{uref}.lock',
            'endpoint': factory_conf['psuctl_endpoint'],
            'scripts': factory_conf['psuctl_scripts'],
        })

        return uiconf.raw


class _LocationUconfBuilder(_UnitUconfBuilder):
    def build(self, uref: str, factory_conf: dict | None = None) -> dict:
        uconf = super().build(uref, factory_conf)

        if factory_conf is None:
            return uconf

        location: dict | None = factory_conf.get('location')
        if location is None:
            return uconf

        uconf_ro = uconf['ro']
        uconf_ro['driver'] = location['driver']
        uconf_ro['type'] = location['device']
        uconf_ro['uri'] = location.get('endpoint')

        return uconf


class UconfBuilder:
    def __init__(self, utype_to_builder_conf_map: dict[str, dict | str]) -> None:
        """
        :param utype_to_builder_conf_map:
            mv-terminal-info.json#/defconf
            radar-terminal-info.json#/defconf
            rtsp-terminal-info.json#/defconf
            rtsp-waga-terminal-info.json#/defconf
            wspases-flex-r3d-terminal-info.json#/defconf
        """
        self._utype_to_builder_conf_map: Final[dict[str, dict | str]] = utype_to_builder_conf_map
        self._utype_to_builder_type_map: Final[dict[str, Type[_IUnitUconfBuilder]]] = {
            UTYPE_TERMINAL: _TerminalUconfBuilder,
            UTYPE_MV_CAMERA: _MvCameraUconfBuilder,
            UTYPE_RTSP_CAMERA: _RtspCameraUconfBuilder,
            UTYPE_RTSP_STREAMING: _RtspStreamingUconfBuilder,
            UTYPE_PSUCTL: _PsuCtlUconfBuilder,
            UTYPE_LPR: _LprUconfBuilder,
            UTYPE_LOCATION: _LocationUconfBuilder
        }

    def build(self, utype: str, uref: str, factory_conf: dict | None = None) -> dict:
        builder_type = self._utype_to_builder_type_map.get(utype) or _UnitUconfBuilder
        builder_conf = self._utype_to_builder_conf_map.get(utype) or 'def-jnode.json'
        builder = builder_type(utype, builder_conf)
        return builder.build(uref, factory_conf)
