import sys

from functools import reduce
from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.uref import Uref, Uref2
from edge.webserver.sysconf.uimod.utype import *

_ROOT_DEFCONF_PATH = '/opt/edge/share/setupstore/ui/root-info.json'
_ROOT_CONF_PATH = '/etc/edge/setupstore/root-info.json'


def decode_info() -> unet.UnetInfo:
    uiconf = fs.load_json3(_ROOT_CONF_PATH, default=None)
    if uiconf is None:
        raise NotImplementedError

    uiconf_pri: dict = uiconf['pri']
    uiconf_pub: dict = uiconf['pub']

    root: unet.UnitInfo = uiconf_pub['info']
    links: list[unet.UnetLink] = uiconf_pri['links']
    ver: str | None = uiconf.get('ver')

    return unet.UnetInfo(root, links, ver)


def encode_info(unet_info: unet.UnetInfo) -> None:
    uiconf = {
        'pri': {
            'links': unet_info.links
        },
        'pub': {
            'info': unet_info.root
        }
    }

    if unet_info.ver:
        uiconf['ver'] = unet_info.ver

    fs.write_json_file(uiconf, _ROOT_CONF_PATH, pretty=True)


def drop_unit_r(unit_uref: str) -> tuple[unet.UnitInfo, set[str]]:
    """
    Unlinks and removes a unit and its sub-units from the root_info tree.

    :param unit_uref:
        The ref of the unit to be removed.

    :returns: (root_info, unlinked_units_urefs)
        [0] root_info without the dropped unit and its sub-units.
        [1] urefs of the units that were linked to the dropped unit and/or its sub-units.
    """
    unet_info = decode_info()

    # find the unit to be removed and its (the most nested) parent.
    unit_uinfo, parent_uinfo = unet.find_unit_parent_r(unet_info.root, unit_uref=unit_uref)
    if parent_uinfo is None:
        eventlog.log_d('fail: NotFoundError')
        raise NotFoundError

    # define the refs of the unit and its sub-units to be removed.
    drop_units_urefs = set[str](unet.find_urefs_r(unit_uinfo))
    assert len(drop_units_urefs) > 0

    # define the links to be removed.
    drop_links_urefs = set[str]()
    for uref in drop_units_urefs:
        for link in unet_info.links:
            if uref in link:
                drop_links_urefs.update(link)

    # strip units
    parent_uinfo['units'] = [x for x in parent_uinfo['units'] if x['ref'] != unit_uref]

    # strip links.
    for uref in drop_units_urefs:
        unet_info.links = [link for link in unet_info.links if uref not in link]

    encode_info(unet_info)

    # define the urefs of the units that were linked to the dropped unit and/or its sub-units.
    unlinked_units_urefs = drop_links_urefs - drop_units_urefs

    return unet_info.root, unlinked_units_urefs


def set_unit_name(unit_uref: str, unit_name: str) -> None:
    unet_info = decode_info()
    unit_uinfo = unet.find_unit_r(unet_info.root, uref=unit_uref)
    if unit_uinfo is None:
        eventlog.log_w('fail: NotFoundError')
        raise NotFoundError

    unit_uinfo['name'] = unit_name
    encode_info(unet_info)


def get_linked_urefs(uiref: str, links: list[unet.UnetLink] | None = None) -> set[str]:
    """
    :param uiref:
        The endpoint that must to be a part of the matched link.
    :param links:
        The pre-loaded (cached) list of links. The list will be loaded if not provided.
    :return:
        A set of the unique endpoints of the matched links.
        The uiref endpoint is included in the result if at least one link matches.
    """
    if links is None:
        links = decode_info().links

    used_urefs = set[str]()
    for link in links:
        if uiref in link:
            used_urefs.update(link)

    return used_urefs


class UrefGenerator:
    def __init__(self, root: unet.UnitInfo) -> None:
        """
        :param root:
            The root unit for the operation.
            Defines the scope of the uniqueness.
        """
        self._root: Final[unet.UnitInfo] = root

    def create_term_uref(self) -> Uref:
        used_urefs: set[str] = set(unet.find_urefs_r(self._root, utypes=[UTYPE_TERMINAL]))
        for pid in range(sys.maxsize):
            candidate = Uref.create_term_uref(pid)
            if candidate.ref not in used_urefs:
                eventlog.log_d(f'done: ref:{candidate.ref}')
                return candidate

        raise NotImplementedError

    def create_unit_uref(self, term_uref: str, unit_utype: str) -> Uref:
        """
        :param term_uref:
            The uiref of the target terminal unit.
        :param unit_utype:
            The type of the unit to be created.

        The part of term_uiref becomes the part of the new unit uiref.
        This is to ensure uniqueness.
        """
        assert self._root['type'] in {
            UTYPE_ROOT,
            UTYPE_SYSTEM,
            UTYPE_CONTROLLER,
            UTYPE_TERMINAL,
            UTYPE_TRAFFIC_RULES,
            UTYPE_MV_CAMERA,
            UTYPE_RTSP_CAMERA,
            UTYPE_LPR
        }

        term_pid = Uref.create_uref(term_uref).pid
        used_refs: set[str] = set(unet.find_urefs_r(self._root, utypes=[unit_utype]))
        for unit_uid in range(sys.maxsize):
            if unit_utype == UTYPE_HTTP_STREAMING and unit_uid % 2 != 0:
                continue
            elif unit_utype == UTYPE_RTSP_STREAMING and unit_uid % 2 != 1:
                continue

            candidate = Uref.create_unit_uref(term_pid, unit_utype, unit_uid)
            if candidate.ref not in used_refs:
                eventlog.log_d(f'done: ref:{candidate.ref}')
                return candidate

        raise NotImplementedError

    def create_unit_uref2(self, unit_utype: str) -> Uref2:
        assert self._root['type'] in {
            UTYPE_ROOT,
            UTYPE_SYSTEM,
            UTYPE_CONTROLLER,
            UTYPE_WIEGAND_PCB,
            UTYPE_IIT_CA,
            UTYPE_TERMINAL
        }

        used_urefs: set[str] = set(unet.find_urefs_r(self._root, utypes=[unit_utype]))
        for chan in range(sys.maxsize):
            candidate = Uref2.create_uref(unit_utype, chan)
            if candidate.ref not in used_urefs:
                eventlog.log_d(f'done: ref:{candidate.ref}')
                return candidate

        raise NotImplementedError

    def create_uref_map(self, factory: unet.UnitInfo, term_uref: str) -> dict[str, str]:
        """
        :param factory:
            The set of urefs to remap.
        :param term_uref:
            The uiref of the target terminal unit.
        :returns:
            old_to_new_uref_map
        """
        def reduce_fn(acc: dict[str, str], factory_uinfo: unet.UnitInfo) -> dict[str, str]:
            factory_uref: str = factory_uinfo['ref']
            factory_utype: str = factory_uinfo['type']

            if factory_utype in UTYPES_UIMOD2:
                new_unit_ref = self.create_unit_uref2(factory_utype).ref
            else:
                new_unit_ref = self.create_unit_uref(term_uref, factory_utype).ref

            acc[factory_uref] = new_unit_ref
            return acc

        factory_uinfos: list[unet.UnitInfo] = unet.map_units_r(factory)
        return reduce(reduce_fn, factory_uinfos, dict())
