from typing import Self, Final

from edge.support import fs
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.types.defs import TERM_UICONF_DIR, TERM_UISCHEMA_DIR, CTRL_BACKUP_DIR
from edge.webserver.sysconf.uimod.uref import Uref
from edge.webserver.sysconf.uimod.utype import UTYPES_FILELESS


class Umod(Uref, IUiConfProvider):
    def __init__(self, pid: str, utype: str, uid: str) -> None:
        super().__init__(pid, utype, uid)
        self._conf: dict | None = None

    def __hash__(self) -> int:
        return hash(self.type + self.id)

    def __str__(self) -> str:
        return self.ref

    def __eq__(self, other: Self) -> bool:
        return (isinstance(other, self.__class__) and
                getattr(other, 'pid', None) == self.pid and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    @classmethod
    def create(cls, uref: str) -> Self:
        """
        :param uref:
            'wrong-way.1_14'
        """
        uiref: Uref = cls.create_uref(uref)
        return cls(uiref.pid, uiref.type, uiref.id)

    @classmethod
    def create_term(cls, term_pid: str) -> Self:
        uiref: Uref = cls.create_term_uref(term_pid)
        return cls(uiref.pid, uiref.type, uiref.id)

    @property
    def conf_dir(self) -> str:
        return fs.join_path(TERM_UICONF_DIR, self.pid, self.type)

    @property
    def conf_path(self) -> str:
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(self.conf_dir, f'{old_id}.json')
        # return fs.join_path(self.conf_dir, '{}.json'.format(self.id))

    def load_conf(self, **kwargs) -> dict:
        """
        :raise FsError:
        """
        if self._conf is None:
            self._conf = fs.load_json3(self.conf_path, **kwargs)

        return self._conf

    def save_conf(self, conf_jnode: dict) -> None:
        fs.write_json_file(conf_jnode, self.conf_path, pretty=True)

    @property
    def schema_path(self) -> str:
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(TERM_UISCHEMA_DIR, self.pid, self.type, f'{old_id}.json')

    @property
    def schema_limits(self) -> dict | None:
        schema_jnode = fs.load_json3(self.schema_path)
        return schema_jnode.get('xLimits')

    @property
    def lock_path(self) -> str:
        # TODO: migrate to /run/lock/edge/
        return fs.join_path('/var/run/edge', '{}.lock').format(self.ref)

    def create_backup_dirpath(self) -> str:
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def install_schema(self) -> None:
        conf_path: Final[str] = self.conf_path
        if fs.path_exists(conf_path):
            fs.ln_schema(self.schema_path, conf_path)

    def install(self, uconf: dict) -> None:
        assert self.type not in UTYPES_FILELESS

        fs.silent_mkdir_r(self.conf_dir)
        fs.write_json_file(uconf, self.conf_path, pretty=True)
        self.install_schema()

    def uninstall(self) -> None:
        if self.type not in UTYPES_FILELESS:
            fs.silent_rm_r(self.conf_path)
            fs.silent_rm_r(self.schema_path)
