import os

from typing import Dict, List, Self, Tuple

from edge import eventlog
from edge.support.db import DbArgs
from edge.types.exceptions import JnodeError


class ApiArgs:
    def __init__(self, endpoint: str):
        self.endpoint = endpoint

    @classmethod
    def unpack(cls, jnode: str) -> Self:
        """
        :param jnode:
            main.conf#/api
        """
        return cls(jnode)


class StoreArgs:
    def __init__(self, db: DbArgs, systest: bool):
        self.db = db
        self.systest = systest

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            main.conf#/

        :raise JnodeError:
        """

        # unpack #/systest
        systest = jnode.get('systest', False)
        if type(systest) != bool:
            eventlog.log_e_param('#/systest')
            raise JnodeError

        # unpack #/db
        db_path = jnode.get('db')
        if not (type(db_path) in [str] and len(db_path)):
            eventlog.log_e_param('#/db')
            raise JnodeError

        try:
            db = DbArgs.deserialize2(db_path)
        except (KeyError, ValueError):
            eventlog.log_e_param('#/db')
            raise JnodeError

        return cls(db, systest)


class ReactArgs:
    def __init__(self, endpoints: List[str]):
        self.endpoints = endpoints

    @classmethod
    def unpack(cls, jnode: List[str]) -> Self:
        """
        :param jnode:
            main.conf#/fact-react
        """
        return cls(jnode)


class EventArgs:
    MODE_DEFAULT = 1
    MODE_WITH_JSON = 2

    def __init__(self, mode: int, udp_peer_host: str, udp_peer_port: int):
        self.mode = mode
        self.udp_peer_host = udp_peer_host
        self.udp_peer_port = udp_peer_port

    @classmethod
    def unpack(cls, jnode: Tuple[int, str, int]) -> Self:
        """
        :param jnode:
            main.conf#/fact-event

        :raise JnodeError:
        """
        if len(jnode) != 3:
            eventlog.log_e_param('#/')
            raise JnodeError

        mode, udp_peer_host, udp_peer_port = jnode
        if mode not in [EventArgs.MODE_DEFAULT, EventArgs.MODE_WITH_JSON]:
            eventlog.log_e_param('#/0')
            raise JnodeError

        return cls(mode, udp_peer_host, udp_peer_port)


class StatsArgs:
    def __init__(self, stat_filepath: str, stat_tout_sec: int):
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            main.conf#/monitoring

        :raise JnodeError:
        """

        # unpack #/stat-path
        stat_path = jnode.get('stat-path')
        if not (str == type(stat_path) and len(stat_path) < 64 and os.path.isabs(stat_path)):
            eventlog.log_e_param('#/stat-path')
            raise JnodeError

        # unpack #/stat-tout
        stat_tout_sec = jnode.get('stat-tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_e_param('#/stat-tout')
            raise JnodeError

        return cls(stat_path, stat_tout_sec)
