from typing import Dict, Optional, Tuple, Final

from edge import eventlog
from edge.image.types.footer import ImFooterConf
from edge.image.types.jnode.footer import decode__im_footer_conf
from edge.services.types.about import AboutServiceConf
from edge.services.types.cam import CamServiceConf
from edge.services.types.gnss import GnssServiceConf
from edge.services.types.jnode.cam import decode__cam_service_conf
from edge.services.types.jnode.gnss import decode__gnss_service_conf_file
from edge.services.types.jnode.tl import decode__tl_service_conf
from edge.services.types.jnode.tz import decode__tz_service_conf_file
from edge.services.types.lane import LaneServiceConf
from edge.services.types.tl import TlServiceConf
from edge.services.types.tz import TzServiceConf
from edge.types.exceptions import JnodeError

from renderer.services.types.event import EventEmitterConf
from renderer.services.types.jnode.event import decode__im_store_conf, decode__json_store_conf
from renderer.services.types.jnode.lane import decode__lane_service_conf
from renderer.services.types.jnode.react import decode__react_service_conf
from renderer.services.types.jnode.renderer import decode__renderer_conf
from renderer.services.types.jnode.stats import decode__stats_service_conf
from renderer.services.types.react import ReactServiceConf
from renderer.services.types.renderer import RendererConf
from renderer.services.types.stats import StatsServiceConf


class ImLoaderArgs:
    def __init__(self, dh: int, image_wh: Tuple[int, int]):
        self.dh = dh
        self.image_wh = image_wh


class KernelConf:
    def __init__(self,
                 react: ReactServiceConf,
                 ee_conf: EventEmitterConf,
                 renderer: RendererConf,
                 lane_service: LaneServiceConf | None,
                 tz: TzServiceConf,
                 gpsd: GnssServiceConf,
                 footer: ImFooterConf,
                 monitoring: StatsServiceConf,
                 im_loader: ImLoaderArgs,
                 about: AboutServiceConf,
                 tl_service_args: Optional[TlServiceConf],
                 landmark_args: Optional[CamServiceConf] = None):

        self.lane: Final[LaneServiceConf | None] = lane_service
        self.react = react
        self.ee_conf = ee_conf
        self.renderer = renderer
        self.tz = tz
        self.gpsd = gpsd
        self.footer = footer
        self.stats = monitoring
        self.im_loader = im_loader
        self.about = about
        self.tl = tl_service_args
        self.cam = landmark_args


def decode__kernel_conf(jnode: Dict, log_level: str) -> KernelConf:
    """
    :param jnode:
        main.conf#/
    :param log_level:

    :raise JnodeArgs:
    """
    lane_jnode: dict | None = jnode.get('lane')
    lane_service_conf = decode__lane_service_conf(lane_jnode) if lane_jnode else None

    rn_jnode = jnode['Renderer']
    profile_jnode = rn_jnode['profile']
    footer = decode__im_footer_conf(profile_jnode['footer'])

    react = decode__react_service_conf(jnode['Source'])
    event = decode__json_store_conf(jnode['scene-event'], log_level)

    im_loader = ImLoaderArgs(footer.height, (profile_jnode['width'], 65535))
    im_store = decode__im_store_conf(jnode['image-store'])

    annotaion_jnode = profile_jnode.get('annotation', {})

    yellow = rn_jnode.get('YellowSpeedThreshold', 65535)
    red = rn_jnode.get('RedSpeedThreshold', 65535)

    # define LrStoreTout
    def_lr_store_tout = 10
    lr_store_tout = rn_jnode.get('LrStoreTout', def_lr_store_tout)
    if not (1 <= lr_store_tout <= 60):

        eventlog.log_w('The param LrStoreTout has invalid value %d. The default value %d is used.'
                       % (lr_store_tout, def_lr_store_tout))
        lr_store_tout = def_lr_store_tout

    renderer = decode__renderer_conf(annotaion_jnode, yellow, red, lr_store_tout)

    # deserialize services_jnode
    services_jnode = jnode['services']

    tl_service_args = decode__tl_service_conf(services_jnode.get('trafficlight'))

    conf_path = services_jnode['origin']
    about = AboutServiceConf(conf_path)

    # datetime service
    conf_path = services_jnode.get('datetime')
    if not conf_path:
        eventlog.log_e_param('#/services/datetime')
        raise JnodeError

    tz = decode__tz_service_conf_file(conf_path)

    # geolocation service
    conf_path = services_jnode.get('geolocation')
    if not conf_path:
        eventlog.log_e_param('#/services/geolocation')
        raise JnodeError

    gpsd = decode__gnss_service_conf_file(conf_path)

    # stat service
    conf_jnode = services_jnode['monitoring']
    stats = decode__stats_service_conf(conf_jnode)

    landmark_args = None
    cam_jnode = services_jnode.get('cam')
    if cam_jnode is not None:
        landmark_args = decode__cam_service_conf(cam_jnode)

    if renderer.ra_anno.radar_present or renderer.lr_anno.show_vehicle_cuboid:
        if landmark_args is None:
            eventlog.log_e_param('#/services/cam')
            raise JnodeError

    ee_conf = EventEmitterConf(event, im_store)
    return KernelConf(react, ee_conf, renderer, lane_service_conf, tz, gpsd, footer,
                      stats, im_loader, about, tl_service_args, landmark_args)
