from typing import Final, Self

from edge.image.builder import LiveImBuilder, LiveImConf
from edge.services import lane, cam, tz, gnss, about, tl
from edge.services.types.about import IAboutService
from edge.services.types.cam import ICamService
from edge.services.types.gnss import IGnssService
from edge.services.types.lane import ILaneService
from edge.services.types.tl import ITlService
from edge.services.types.tz import ITzService
from edge.support.service import ServicePack
from edge.types.image import ImRgba

from renderer._args import KernelConf
from renderer.services.event import init_event_emitter
from renderer.services.react_tr import init_react_service, IReactService
from renderer.services.renderer import init_renderer
from renderer.services.stats import init_stats_service, IStatService
from renderer.services.types.event import IEventEmitter
from renderer.services.types.renderer import IRenderer
from renderer.types.tr import TrScene


class _KernelDeps:
    def __init__(self,
                 stat_service: IStatService,
                 react_service: IReactService,
                 event_consumer: IEventEmitter,
                 renderer: IRenderer) -> None:

        self.stat = stat_service
        self.react = react_service
        self.event = event_consumer
        self.renderer = renderer

    @classmethod
    def create(cls, services: ServicePack) -> Self:
        return cls(
            services.get(IStatService),
            services.get(IReactService),
            services.get(IEventEmitter),
            services.get(IRenderer))


class BaseKernel:
    def __init__(self, deps: _KernelDeps) -> None:
        self._deps: Final[_KernelDeps] = deps
        self._deps.react.set_handler(self._handle_react)

    def _handle_react(self, tr_scene: TrScene, im_rgba: ImRgba):
        self._deps.renderer.render_tr_scene(tr_scene, im_rgba)
        self._deps.event.emit(tr_scene.im_ref.ts, im_rgba)
        self._deps.stat.wh.im_size(im_rgba.w, im_rgba.h, im_rgba.dh)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, conf: KernelConf) -> None:
        services = ServicePack()

        if conf.cam is not None:
            services.add(ICamService, cam.init_service(conf.cam))

        services.add(IStatService, init_stats_service(twisted_reactor, conf.stats))
        services.add(ILaneService, lane.init_service(conf.lane))
        services.add(IReactService, init_react_service(twisted_reactor, conf.react, conf.im_loader))
        services.add(ITzService, tz.init_service(conf.tz))
        services.add(IGnssService, gnss.init_service(twisted_reactor, conf.gpsd, services.get(ITzService)))

        services.add(IAboutService, about.init_service(conf.about, services.get(IGnssService)))
        services.add(ITlService, tl.init_service(twisted_reactor, conf.tl))

        im_builder = LiveImBuilder(LiveImConf(conf.footer))

        services.add(IRenderer, init_renderer(twisted_reactor, services, conf.renderer, im_builder))
        services.add(IEventEmitter, init_event_emitter(twisted_reactor, conf.ee_conf))

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', lambda: self.destroy())

        super().__init__(_KernelDeps.create(services))

        self._services: Final[ServicePack] = services
        self._reactor = twisted_reactor

    def destroy(self) -> None:
        self._services.destroy()

    def run(self) -> None:
        self._reactor.run()
