from typing import List, Optional, Tuple

from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF
from edge.types.track import LrTrackItem, RaTrackItem, TrTrackItem

_FAKE_PLATE_SIZE = 0.01


def check_quad_dist_from_img_borders(plate_quad: QuadI, lr_scene_w: int, lr_scene_h: int) -> bool:
    bbox: RectF = plate_quad.get_rel_bbox_recf(lr_scene_w, lr_scene_h)

    plate_w = bbox.fr - bbox.fl
    plate_h = bbox.fb - bbox.ft

    return (bbox.ft > plate_h and
            bbox.fb < (1.0 - 2 * plate_h) and
            bbox.fl > plate_w and
            bbox.fr < (1.0 - plate_w))


def get_lr_item_dist(lr_item: LrTrackItem, lr_scene_w: int, lr_scene_h: int) -> int:
    bbox: RectF = lr_item.plate_quad.get_rel_bbox_recf(lr_scene_w, lr_scene_h)

    # vehicle's plate is fully visible on the image (1 bit).
    dist1 = 1 if bbox.get_visible_100() else 0

    # the weight of plate's recognition (7 bits).
    dist2 = lr_item.plate_weight

    # vehicle's body is more visible (1 bit).
    dist3 = 1 if check_quad_dist_from_img_borders(lr_item.plate_quad, lr_scene_w, lr_scene_h) else 0

    # plate's height [0, 1000] (10 bits).
    dist4 = int((bbox.fb - bbox.ft) * 1000)

    dist = (dist1 << 18) + (dist2 << 11) + (dist3 << 10) + dist4
    return dist


def get_ra_item_dist(ra_item: RaTrackItem,
                     lr_item: Optional[LrTrackItem] = None,
                     lr_scene_w: Optional[int] = None,
                     lr_scene_h: Optional[int] = None
                     ) -> int:

    # if None in [lr_item, lr_scene_w, lr_scene_h]:
    #     best_line_y = 0.5
    # else:
    #     l, t, r, b = lr_item.plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h)
    #     best_line_y = (t + b) / 2.0

    # ra-item is fully visible on the image (1 bit).
    def check_borders(coord):
        return (0.0 + _FAKE_PLATE_SIZE) <= coord <= (1.0 - _FAKE_PLATE_SIZE)

    dist1 = 1 if check_borders(ra_item.frame_l_rel) and check_borders(ra_item.frame_t_rel) else 0

    # radar item is closer to the center of number plate (7 bit)

    #   (doesn't work because the farther from camera the car is locate,
    #   the less is the distance between ra-point and center of number plate

    # dist2 = 100 - abs(int(100 * (ra_item.y - best_line_y)))

    dist2 = 0

    dist = (dist1 << 7) + dist2
    return dist


def get_tr_item_dist(tr_item: TrTrackItem, lr_scene_w: Optional[int], lr_scene_h: Optional[int]) -> int:
    lr_dist = 0
    if tr_item.lr_item:
        # ? to 468 968
        lr_dist = get_lr_item_dist(tr_item.lr_item, lr_scene_w, lr_scene_h)

    ra_dist = 0
    if tr_item.ra_item:
        # 1 to 228 when ra item on image
        # can be negative when ra item out off image
        ra_dist = get_ra_item_dist(tr_item.ra_item, tr_item.lr_item, lr_scene_w, lr_scene_h)

    return lr_dist + ra_dist * 10


def get_best_ra_ts_and_rect(
        track_items: List[RaTrackItem],
        search_image_list: List[int]
        ) -> Optional[Tuple[int, RectF]]:

    best_line_y = 0.5
    best_item_dist = 2.0
    best_item_index = None
    for index, item in enumerate(track_items):
        dist = abs(item.frame_t_rel - best_line_y)
        if dist < best_item_dist:
            if item.frame_ts in search_image_list:
                # noinspection PyChainedComparisons
                if item.frame_l_rel < 1.0 - _FAKE_PLATE_SIZE and item.frame_l_rel > _FAKE_PLATE_SIZE \
                        and item.frame_t_rel < 1.0 - _FAKE_PLATE_SIZE and item.frame_t_rel > _FAKE_PLATE_SIZE:
                    best_item_dist = dist
                    best_item_index = index

    if best_item_index is None:
        return None

    radar_point = track_items[best_item_index]

    plate_rect = RectF(
        radar_point.frame_l_rel - _FAKE_PLATE_SIZE, radar_point.frame_t_rel - _FAKE_PLATE_SIZE,
        radar_point.frame_l_rel + _FAKE_PLATE_SIZE, radar_point.frame_t_rel + _FAKE_PLATE_SIZE)

    return track_items[best_item_index].frame_ts, plate_rect
