from typing import Callable, Final

from edge import eventlog
from edge.image.builder import EventMeasCamImBuilder, EventOverCamImBuilder
from edge.services.types.vs import VsImageInfo
from edge.support.imloader import RgbaImLoader, JpegImLoader, IJpegImLoader
from edge.support.service import ServicePack
from edge.types.blob import MemBlob
from edge.types.exceptions import GenericError, ParamsError
from edge.types.image import ImJpeg, ImRgba
from edge.types.jnode.dir import encode_dir_text
from edge.types.location import Location
from edge.types.origin import Origin
from edge.types.shape.point import PointI
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectI
from edge.types.track import TrTrackItem
from edge.types.trackset import Trackset
from edge.types.trackset.vrp import TrackVrp

from speedcontrol.detectors.base.best import get_tr_item_dist
from speedcontrol.detectors.base.types.conf import DetectorConf
from speedcontrol.detectors.base.types.deps import DetectorDeps
from speedcontrol.detectors.base.types.footer import BestImFooterVars
from speedcontrol.detectors.base.types.jnode.conf import decode__detector_conf
from speedcontrol.detectors.base.types.jnode.stats import encode__detector_stats
from speedcontrol.detectors.base.types.stats import DetectorStats
from speedcontrol.detectors.types.detector import IBaseDetector, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.measurement import Measurement
from speedcontrol.events.position import Position
from speedcontrol.events.vehicle import Vehicle
from speedcontrol.events.vrp import VrpType, Vrp
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.measurement import BestWaypoint


class _TrItemImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem, weight: int) -> None:
        super().__init__(tr_item.ts)
        self.tr_item: Final[TrTrackItem] = tr_item
        self.weight: Final[weight] = weight


class _Detector(IBaseDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf) -> None:
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf
        self._stats = DetectorStats()

        self._meas_loader = RgbaImLoader(self._stats.meas_im, self._deps.vs, conf.meas_builder.scene.footer.height)
        self._meas_builder = EventMeasCamImBuilder(conf.meas_builder, deps.cam)
        self._meas_rgba: ImRgba = ImRgba()

        self._over_loader: IJpegImLoader = JpegImLoader(self._stats.over_im,
                                                        self._deps.vs if conf.enable_overview else None)
        self._over_builder = EventOverCamImBuilder()
        self._over_jpeg: ImJpeg = ImJpeg()

        self._handler: Callable[[Trackset, BaseEvent], None] | None = None

    def destroy(self) -> None:
        self._meas_loader.destroy()
        self._over_loader.destroy()

    def set_handler(self, event_handler: Callable[[Trackset, BaseEvent], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__detector_stats('base.0', self._stats))

    # TODO: fix support of tracksets with only LeTrack
    def analyze(self, trackset: Trackset) -> None:
        if trackset.lr_track:
            lr_scene_w = trackset.lr_track.imref.width
            lr_scene_h = trackset.lr_track.imref.height
        else:
            lr_scene_w = None
            lr_scene_h = None

        im_infos: list[_TrItemImageInfo] = [
            _TrItemImageInfo(tr_item, get_tr_item_dist(tr_item, lr_scene_w, lr_scene_h))
            for tr_item in trackset.tr_track.values()
        ]

        im_infos.sort(key=lambda ref: ref.weight, reverse=True)

        # try to load an image from the measurement camera
        try:
            if trackset.lr_track is None:
                lr_imref = None
            elif trackset.lr_track.imref.url is None:
                lr_imref = None
            else:
                lr_imref = trackset.lr_track.imref

            meas_iminfo: _TrItemImageInfo | None = self._meas_loader.load(self._meas_rgba, im_infos, lr_imref)
            if meas_iminfo is None:
                meas_tr_item: TrTrackItem = trackset.tr_track[lr_imref.ts]
            else:
                meas_tr_item: TrTrackItem = meas_iminfo.tr_item

        except ParamsError:
            eventlog.log_i('fail: load measurement image')
            raise

        # try to load an image from the overview camera
        try:
            if trackset.wa_track is None:
                over_imref = None
            elif trackset.wa_track.overview_ref.url is None:
                over_imref = None
            else:
                over_imref = trackset.wa_track.overview_ref

            # TODO: use the same image timestamp as for the measurement camera
            self._over_loader.load(self._over_jpeg, im_infos, over_imref)

        except ParamsError:
            pass

        # Optimize plate recognition results

        def upscale_raas_plate_rect(bbox: RectI) -> RectI:
            return RectI(
                bbox.il - bbox.ih,
                bbox.it - bbox.ih,
                bbox.ir + bbox.ih,
                bbox.ib + bbox.ih)

        if meas_tr_item.lr_item is not None:
            vrp_text_quad: QuadI = meas_tr_item.lr_item.plate_quad
            raas_roi: RectI = upscale_raas_plate_rect(vrp_text_quad.bbox_recti)

        else:
            abs_image_x: float = meas_tr_item.ra_item.frame_l_rel * self._meas_rgba.w
            abs_image_y: float = meas_tr_item.ra_item.frame_t_rel * self._meas_rgba.h
            padding: float = 0.05 * self._meas_rgba.w

            vrp_text_bbox = RectI(
                abs_image_x - padding,
                abs_image_y - padding,
                abs_image_x + padding,
                abs_image_y + padding)

            vrp_text_quad = QuadI.create(vrp_text_bbox)
            raas_roi: RectI = upscale_raas_plate_rect(vrp_text_bbox)

        def raas_vrp_handler(raas_vrp: TrackVrp | None) -> None:
            if raas_vrp is None and trackset.lr_track is not None:
                raas_vrp: TrackVrp = TrackVrp.from_proto(trackset.lr_track.vrp, vrp_text_quad)

            self._mk_base_event(trackset, meas_tr_item, raas_vrp, vrp_text_quad)

        self._deps.raas.analyze(self._meas_rgba, raas_roi, raas_vrp_handler)

    def _mk_track_item_pos(self, tr_item: TrTrackItem, vrp_center: PointI) -> Position | None:
        # Radar data is available
        if tr_item.ra_item is not None:
            return Position(tr_item.ra_item.scene_f_cm, tr_item.ra_item.scene_l_cm)

        # Vehicle bbox is available
        if self._deps.cam is not None:
            if tr_item.lr_item is not None:
                if tr_item.lr_item.vehicle_rect is not None:
                    vr_item: RectI = tr_item.lr_item.vehicle_rect
                    vr_cx: int = (vr_item.il + vr_item.ir) // 2
                    vr_b: int = vr_item.ib

                    world_f_m, world_l_m = self._deps.cam.mk_world_point(vr_cx, vr_b, world_t=0.0)
                    return Position(world_f_m * 100, world_l_m * 100)  # map to cm

        # Vehicle registration plate is available
        if self._deps.cam is not None:
            world_f_m, world_l_m = self._deps.cam.mk_world_point(vrp_center.il, vrp_center.it, world_t=0.5)
            return Position(world_f_m * 100, world_l_m * 100)  # map to cm

    def mk_vehicle(self,
                   trackset: Trackset,
                   tr_item: TrTrackItem,
                   vrp: TrackVrp | None,
                   vrp_center: PointI) -> Vehicle:
        """
        :param trackset:
        :param tr_item:
            The LrRaTrackItem which corresponds to the loaded image from the measurement camera.
        :param vrp:
            The final VRP-result. Value provided by RaaS or AutoSDK
        :param vrp_center:
        """

        item_dir = trackset.tr_track.get_item_dir(tr_item.ts)
        item_dir_alias = self._conf.footer_defvars.get(encode_dir_text(item_dir))

        # global vehicle direction
        avg_dir = trackset.get_direction()
        avg_dir_alias = self._conf.footer_defvars.get(encode_dir_text(avg_dir))

        if vrp:
            country_num: int = self._deps.plate_info.get_country_code(vrp.type.template_ref)
            plate_attrs: str | None = self._deps.plate_info.get_plate_attrs(vrp.type.template_ref)
            plate_templ: str | None = self._deps.plate_info.get_plate_template(vrp.type.template_ref)

            plate_type = VrpType(vrp.type.country_a2, country_num, plate_attrs, plate_templ)
            plates = [Vrp(plate_type, vrp.text_quad, vrp.text_ucode, vrp.text_conf, trackset.wa_track)]

        else:
            plates = []

        length = trackset.ra_track.len_cm if trackset.ra_track else None
        try:
            position = self._mk_track_item_pos(tr_item, vrp_center)
        except ParamsError:
            position = None

        return Vehicle(
            plates, length, item_dir, item_dir_alias, avg_dir, avg_dir_alias, position,
            trackset.lr_track and trackset.lr_track.type_jnode,
            trackset.lr_track and trackset.lr_track.seats_jnode,
            trackset.wa_track,
            trackset.spotspeeds)

    def _mk_base_event(self,
                       trackset: Trackset,
                       meas_tr_item: TrTrackItem,
                       final_vrp: TrackVrp | None,
                       vrp_quad: QuadI) -> None:
        """
        :param trackset:
        :param meas_tr_item:
        :param final_vrp:
        :param vrp_quad:
        """

        vehicle = self.mk_vehicle(trackset, meas_tr_item, final_vrp, PointI(*vrp_quad.center))

        main_ts = self._meas_rgba.ts
        control_ts = self._deps.clock.get_control_ts(main_ts)
        origin: Origin = self._deps.about.get_origin(main_ts)

        def mk_best_location() -> Location:
            if meas_tr_item.lane_item is not None:
                return Location(origin.location.place, meas_tr_item.lane_item.lane)

            return Location(origin.location.place)

        best_location = mk_best_location()

        def mk_best_images() -> list[MemBlob]:
            footer_vars = BestImFooterVars(
                self._deps.tz, self._deps.about,
                main_ts,
                time_sync=trackset.time_sync,
                control_ts=control_ts,
                defvars=self._conf.footer_defvars,
                location=best_location,
                geolocation=origin.geolocation,
                plate_text=vehicle.vrp_text_ucode,
                vehicle=vehicle)

            scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(main_ts)
            blobs = self._meas_builder.build(
                self._meas_rgba, scene_dt_utc, origin.geolocation,
                footer_vars, trackset.lr_track, vrp_quad,
                True, True, True, True, vehicle_pos=vehicle.position)

            return blobs

        best_images: list[MemBlob] = mk_best_images()

        if self._over_jpeg.ts > 14:  # TODO: 14?
            ov_blob = self._over_builder.build_scene(self._over_jpeg)
            best_images.append(ov_blob)

        # define #/measurement/0
        measurement = Measurement(trackset.uuid, None, [
            Waypoint(Waypoint.NAME_CATCH, trackset.tr_track.min_ts),
            Waypoint(Waypoint.NAME_LOST, trackset.tr_track.max_ts),
            BestWaypoint(main_ts, control_ts, best_location, best_images, vehicle,
                         direction=encode_dir_text(vehicle.dir))
        ])

        base_event = BaseEvent([measurement], trackset.lr_track and trackset.lr_track.alarms_jnode)

        self._deps.trainer.collect(base_event, trackset, self._meas_rgba)

        self._handler(trackset, base_event)


def create(services: ServicePack, info: DetectorUnitInfo) -> IBaseDetector:
    try:
        conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
        return _Detector(DetectorDeps.create(services), conf)
    except GenericError:
        eventlog.log_e(f'fail: {info.identity.ref}')
        raise
