from typing import Dict

from edge import eventlog
from edge.image.types.jnode.builder import decode__im_builder_conf
from edge.types.exceptions import JnodeError

from speedcontrol.detectors.types.jnode.conf import decode__base_rule_conf
from speedcontrol.detectors.parking.types.conf import DetectorConf, RuleConf, RuleLimits
from speedcontrol.conf.jnode.mask import decode__mask_conf
from edge.types.identity import Identity


def decode__rule_limits(jnode: Dict) -> RuleLimits:
    """
    :raise JnodeError:
    """
    # deserialize #/max-standing-time
    max_standing_time = jnode.get('max-standing-time')
    if not (int == type(max_standing_time) and max_standing_time > 0):
        eventlog.log_e_param('#/max-standing-time')
        raise JnodeError

    return RuleLimits(max_standing_time)


def decode__rule_conf(jnode: Dict, footer_defvars: Dict) -> RuleConf:
    """
    :raise JnodeError:
    """
    if type(jnode) != dict:
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/name
    # deserialize #/enabled
    # deserialize #/schedule
    # deserialize #/ticket
    rule = decode__base_rule_conf(jnode)

    # deserialize #/limits
    limits_jnode = jnode.get('limits')
    if type(jnode) != dict:
        eventlog.log_e_param('#/limits')
        raise JnodeError

    limits = decode__rule_limits(limits_jnode)

    return RuleConf(rule.name, rule.enabled, rule.schedule, rule.ticket, limits, footer_defvars)


def decode__detector_conf(identity: Identity, jnode: Dict, footer_defvars: Dict) -> DetectorConf:
    """
    :raise JnodeError:
    """
    # deserialize #/name
    name = jnode.get('name')
    if not type(name) in [str] or len(name) > 32:
        eventlog.log_e_param('#/name')
        raise JnodeError

    # deserialize #/store
    store = jnode.get('store')
    if not (type(store) in [str] and 1 <= len(store) <= 2):
        eventlog.log_e_param('#/store')
        raise JnodeError

    # deserialize #/scene
    scene_jnode = jnode.get('scene')
    if type(scene_jnode) != dict:
        eventlog.log_e_param('#/scene')
        raise JnodeError
    scene = decode__mask_conf(scene_jnode)

    # deserialize #/rules
    rules_jnode = jnode.get('rules')
    if type(rules_jnode) != list:
        eventlog.log_e_param('#/rules')
        raise JnodeError

    rules = []
    for rule_jnode in rules_jnode:
        rule = decode__rule_conf(rule_jnode, footer_defvars)
        if rule.enabled:
            rules.append(rule)

    # deserialize #/images
    images_jnode = jnode.get('images')
    if type(images_jnode) != dict:
        eventlog.log_e_param('#/images')
        raise JnodeError

    init_im_engine = decode__im_builder_conf(images_jnode['violation'])
    fini_im_engine = init_im_engine

    # deserialize range tout
    range_tout = jnode.get('range-tout')
    if type(range_tout) not in [int]:
        eventlog.log_e_param('#/range_tout')
        raise JnodeError

    # deserialize lost tout
    lost_tout = jnode.get('lost-tout')
    if type(lost_tout) not in [int]:
        eventlog.log_e_param('#/lost_tout')
        raise JnodeError

    return DetectorConf(
        identity, name, store, scene, rules,
        init_im_engine, fini_im_engine, range_tout, lost_tout)
