import functools

from typing import Callable, Final, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.services.types.vs import VsSessionAbsentError, VsError
from edge.support.mask import IntersectionMask
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem
from edge.types.trackset import Trackset

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.redlight.rl_rule import Rule
from speedcontrol.detectors.redlight.types.deps import DetectorDeps
from speedcontrol.detectors.redlight.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.redlight.types.stats import DetectorStats
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.buslane import Evidence



class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf
        self._stats: Final[DetectorStats] = DetectorStats()
        self._scene_mask: Final[IntersectionMask] = IntersectionMask(conf.scene)

        imbuilder = EventMeasCamImBuilder(conf.imbuilder)

        self._ime_prebar = ImBuilderEnv(imbuilder)
        self._ime_postbar_init = ImBuilderEnv(imbuilder)
        self._ime_postbar_fini = ImBuilderEnv(imbuilder)

        if self._conf.video:
            self._im_fitter = WhScaler((self._conf.video.max_w, self._conf.video.max_h))

        self._rules = list[Rule]()

        rule_index = 0
        for rule_conf in conf.rules:
            rule_name = conf.identity.child_ref(IDetector.TYPE_REDLIGHT_RULE, rule_index)
            rule = Rule(deps, rule_conf, conf.name, rule_name,
                        self._ime_prebar, self._ime_postbar_init, self._ime_postbar_fini)

            self._rules.append(rule)
            rule_index += 1

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = handler

    def _split_track(self, trackset: Trackset, stopbar_mask: IntersectionMask
                     ) -> Tuple[list[TrTrackItem], list[TrTrackItem], list[TrTrackItem]]:

        lr_imref = trackset.lr_track.imref
        tr_items = [
            x for x in trackset.tr_track.values()
            if x.lr_item and check_quad_dist_from_img_borders(x.lr_item.plate_quad, lr_imref.width, lr_imref.height)
        ]

        stopbar_pre = list[TrTrackItem]()
        stopbar_on = list[TrTrackItem]()
        stopbar_post = list[TrTrackItem]()
        for tr_item in tr_items:
            answer = stopbar_mask.check_point_abs(tr_item.lr_item.plate_quad.center, self._conf.det_dir)

            if answer == stopbar_mask.STOPBAR_PRE:
                stopbar_pre.append(tr_item)

            elif answer == stopbar_mask.STOPBAR_ON:
                stopbar_on.append(tr_item)

            elif answer == stopbar_mask.STOPBAR_POST:
                stopbar_post.append(tr_item)

        return stopbar_pre, stopbar_on, stopbar_post

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        self._stats.total += 1

        if '*' in base_event.plate_ucode:
            eventlog.log_i(f'done: skip trackset ({base_event.plate_utf8}, reason:asterisk)')
            self._stats.lr_invalid += 1
            return

        if self._conf.det_dir != trackset.get_direction():
            eventlog.log_i(f'done: skip trackset ({base_event.plate_utf8}, reason:invalid-dir)')
            self._stats.zone_invalid += 1
            return

        pre_stopbar, on_stopbar, post_stop_bar = self._split_track(trackset, self._scene_mask)
        if not pre_stopbar:
            eventlog.log_i('done: skip trackset ({}, pre-stopbar:0)'.format(base_event.plate_utf8))
            self._stats.zone_invalid += 1
            return

        elif self._conf.on_stopbar_required and len(on_stopbar) == 0:
            eventlog.log_i('done: skip trackset ({}, pre-stopbar:{}, on-stopbar:0)'.format(
                base_event.plate_utf8, len(pre_stopbar)))
            self._stats.zone_invalid += 1
            return

        elif not post_stop_bar:
            eventlog.log_i('done: skip trackset ({}, pre-stopbar:{}, on-stopbar:{}, post-stopbar:0)'.format(
                base_event.plate_utf8, len(pre_stopbar), len(on_stopbar)))
            self._stats.zone_invalid += 1
            return

        for rule in self._rules:
            try:
                rule_answer: Tuple[Evidence, TimeSpan] | None = rule.check(
                    trackset, base_event, pre_stopbar, post_stop_bar)

                if rule_answer is None:
                    continue

            except GenericError:
                eventlog.log_i(f'fail: skip trackset (<{base_event.plate_utf8}>)')
                continue

            evidence, time_span = rule_answer
            handler = functools.partial(self._handle_evidence_with_vs, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_prebar.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                    eventlog.log_i(f'done: <{base_event.plate_utf8}>, with-video-session:true')
                    return

                except VsSessionAbsentError:
                    eventlog.log_w(f'fail: VsSessionAbsentError (<{base_event.plate_utf8}>)')

                except VsError:
                    eventlog.log_w(f'fail: VsError (<{base_event.plate_utf8}>)')

                except GenericError:
                    eventlog.log_w(f'fail: GenericError (<{base_event.plate_utf8}>)')

                except Exception:
                    eventlog.log_x()

            eventlog.log_i(f'done: <{base_event.plate_utf8}>, with-video-session:false')
            self._handle_evidence_wout_vs(base_event, evidence)
            return

        eventlog.log_i(
            'done: skip trackset ({}, pre-stopbar:{}, on-stopbar:{}, post-stopbar:{}, reason:violation-absent'
            .format(base_event.plate_utf8, len(pre_stopbar), len(on_stopbar), len(post_stop_bar)))

    def _handle_evidence_with_vs(self, base: BaseEvent, evidence: Evidence, video_blob: FileBlob | None) -> None:
        eventlog.log_i(f'exec: {base.plate_utf8}, with-video:{video_blob is not None}')
        self._handle_evidence(base, evidence, video_blob)

    def _handle_evidence_wout_vs(self, base: BaseEvent, evidence: Evidence) -> None:
        eventlog.log_i(f'exec: {base.plate_utf8}')
        self._handle_evidence(base, evidence, None)

    def _handle_evidence(self, base: BaseEvent, evidence: Evidence, video_blob: FileBlob | None) -> None:
        if video_blob:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._conf.identity, base, evidence)):
            self._stats.emit_fail += 1

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        total_fails = self._stats.lr_invalid + self._stats.zone_invalid
        jnode = {
            'type': 'red-light',
            'name': f'{self._conf.identity.ref}',
            'stat': {
                'assert-track': {
                    'count': [self._stats.total, self._stats.total - total_fails, total_fails],
                    'zone-unmatch': self._stats.zone_invalid,
                    'lr-invalid': self._stats.lr_invalid
                },
                'emit-event': {
                    'count': [
                        self._stats.emit_total,
                        self._stats.emit_total - self._stats.emit_fail,
                        self._stats.emit_fail
                    ]
                }
            }
        }

        json_handler(jnode)
        list(map(lambda x: x.collect_stats(json_handler), self._rules))


def create(detectors: DetectorGroups, services: ServicePack, infos: list[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            if conf.rules:
                detector = _Detector(DetectorDeps.create(services), conf)
                detectors.plate_lost.add(detector)

        except GenericError:
            eventlog.log_e('fail')
            raise
