from typing import Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.services.types.tl import TlHaltSpanInfo
from edge.services.types.vs import VsImageInfo
from edge.types.timespan import TimeSpan
from edge.types.trackset.le import LeTrackItem
from edge.types.trackset import Trackset

from speedcontrol.detectors.redlight_leddar.types.conf import RuleConf
from speedcontrol.detectors.redlight_leddar.types.deps import RuleDeps
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.redlight import Evidence
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.redlight import RedLightViolationWaypoint


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    ticket_info = TicketInfo(conf.ticket.name)
    rule_info = RuleInfo(conf.name, None, ticket_info)
    return DetectorInfo(detector_name, rule_info)


class _SceneImageInfo(VsImageInfo):
    def __init__(self, le_item: LeTrackItem):
        super().__init__(le_item.ts)
        self.le_item = le_item


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 detector_name: str,
                 ime_postbar: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)

        self._detector_name = detector_name
        self._ime_postbar = ime_postbar

    def check(self, trackset: Trackset, base_event: BaseEvent) -> Optional[Tuple[Evidence, TimeSpan]]:
        tl_result: List[TimeSpan] = self._deps.tl.get_halt_spans(trackset.le_track.min_ts, trackset.le_track.max_ts)
        le_result: List[TimeSpan] = self._deps.leddar.check(trackset.le_track.min_ts, trackset.le_track.max_ts)

        # TODO:
        for tl_violation in tl_result:
            for le_violation in le_result:
                # prepare im-environments
                post_im_infos = list(map(_SceneImageInfo, trackset.le_track))
                post_im_info1 = self._deps.vs.export_image_rgba(
                    self._ime_postbar.scene_rgba, post_im_infos[0:5], self._ime_postbar.scene_dh)

                # prepare evidence
                evidence = self._mk_evidence(base_event, trackset, post_im_info1.le_item)
                if evidence:
                    ref_ts = trackset.le_track[0].ts
                    return evidence, TimeSpan(ref_ts, ref_ts)

    def _mk_evidence(
            self, base_event: BaseEvent, trackset: Trackset, post_le_item1: LeTrackItem) -> Optional[Evidence]:

        def mk_waypoint(deps: RuleDeps, im_env: ImBuilderEnv, le_item: LeTrackItem, waypoint_name: str) -> Waypoint:
            origin = deps.about.get_origin(im_env.scene_rgba.ts)
            wp_direction = trackset.le_track.get_item_direction(le_item)

            halt_span_info: TlHaltSpanInfo = deps.tl.get_halt_span_info(im_env.scene_rgba.ts)

            footer_vars = ImFooterVars(deps.tz, self._deps.about, im_env.scene_rgba.ts,
                                       defvars=self._conf.footer_defvars,
                                       location=origin.location,
                                       geolocation=origin.geolocation,
                                       plate_text=base_event.plate_ucode,
                                       direction=wp_direction,
                                       halt_time=halt_span_info.halt_time)

            vrp_text_quad = le_item.get_bbox_quad_abs(*im_env.scene_rgba.wh)
            scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(im_env.scene_rgba.ts)
            im_blobs = im_env.builder.build(
                im_env.scene_rgba, scene_dt_utc, origin.geolocation,
                footer_vars, trackset.lr_track, vrp_text_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_i('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, im_blobs,
                ViolationVehicle(plate_quads=[vrp_text_quad]),
                halt_span_info)

            return waypoint

        try:
            post_waypoint1 = mk_waypoint(self._deps, self._ime_postbar, post_le_item1, Waypoint.NAME_POST_STOPBAR)
        except RuntimeError:
            eventlog.log_x()
            return None

        event_ts = self._ime_postbar.scene_rgba.ts

        evidence = Evidence(
            self._deps.about.get_origin(event_ts),
            event_ts, self._info)
        evidence.add_waypoints([post_waypoint1])
        return evidence
