import functools

from typing import Callable, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.services.types.vs import VsSessionAbsentError, VsError
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.timespan import TimeSpan
from edge.types.trackset import Trackset

from speedcontrol.detectors.spotspeed_waga.spsw_rule import Rule
from speedcontrol.detectors.spotspeed_waga.types.deps import DetectorDeps
from speedcontrol.detectors.spotspeed_waga.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.spotspeed_waga.types.jnode.stats import encode__detector_stats
from speedcontrol.detectors.spotspeed_waga.types.stats import DetectorStats
from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.spotspeed import Evidence
from speedcontrol.events.evidence.spotspeed_waga import SpotSpeedWagaEvidence


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf
        self._stats: Final[DetectorStats] = DetectorStats()

        if self._conf.video:
            self._im_fitter = WhScaler(self._conf.video.max_wh)

        imbuilder = EventMeasCamImBuilder(conf.im_engine)
        self._ime_init = ImBuilderEnv(imbuilder)
        self._ime_fini = ImBuilderEnv(imbuilder)
        self._ime_best = ImBuilderEnv(imbuilder)

        self._rules: List[Rule] = []
        rule_index = 0
        for rule_conf in conf.rules:
            if rule_conf.enabled:  # TODO: strip disabled rules when application starts
                rule_stats_name = conf.identity.child_ref(IDetector.TYPE_SPOTSPEED_WAGA_RULE, rule_index)
                rule_index += 1
                rule = Rule(
                    deps, rule_conf, rule_stats_name,
                    self._ime_init, self._ime_fini, self._ime_best)

                self._rules.append(rule)

        self._handler: Callable[[Event], bool] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], bool]) -> None:
        assert self._handler is None
        self._handler = event_handler

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__detector_stats(self._conf.identity, self._stats))
        list(map(lambda x: x.collect_stats(json_handler), self._rules))

    def analyze(self, track: Trackset, base_event: BaseEvent) -> None:
        self._stats.total += 1
        if track.lr_track is None:
            self._stats.lr_absent += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format('-', 'lr-absent'))
            return

        if track.wa_track is None:
            self._stats.wa_absent += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(base_event.plate_utf8, 'wa-absent'))
            return

        # best_vehicle = base.best_wp.vehicle  # type: # Vehicle
        # base_racam_speed = best_vehicle.ra_cam_speed
        #
        # if base_racam_speed is None:
        #     self._stats.ra_cam_speed += 1
        #     eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
        #         plate_text_utf8, 'racam-speed-invalid'))
        #     return

        if '*' in base_event.plate_ucode:
            self._stats.lr_invalid += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(base_event.plate_utf8, 'lr-invalid'))
            return

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(track, base_event)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_w('fail')
                continue

            evidence, time_span = reply
            handler = functools.partial(self.handle_evidence, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_init.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w('fail: VsSessionAbsentError')

                except VsError:
                    eventlog.log_w('fail: VsError')

                except GenericError:
                    eventlog.log_w('fail')

                except Exception:
                    eventlog.log_x()

            handler(None)
            break

    def handle_evidence(self, base: BaseEvent, evidence: SpotSpeedWagaEvidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._conf.identity, base, evidence)):
            self._stats.emit_fail += 1


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detector = _Detector(DetectorDeps.create(services), conf)
            detectors.plate_lost.add(detector)
        except GenericError:
            eventlog.log_e(f'fail: {info.identity.ref}')
            raise
