from typing import Callable, Final, List

from edge import eventlog
from edge.support.service import ServicePack
from edge.types.blob import BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.trackset import Trackset

from speedcontrol.detectors.types.detector import IDetector, DetectorGroups, DetectorUnitInfo
from speedcontrol.detectors.wanted.wn_rule import Rule
from speedcontrol.detectors.wanted.types.deps import DetectorDeps
from speedcontrol.detectors.wanted.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.events.event import BaseEvent, Event


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._conf: Final[DetectorConf] = conf

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            if rule_conf.enabled:
                self._rules.append(Rule(deps, rule_conf, conf.name))

        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track: Trackset, base: BaseEvent) -> None:
        for rule in self._rules:
            evidence = rule.check(base)
            if evidence:
                self._handler(Event(self._conf.identity, base, evidence))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detector = _Detector(DetectorDeps.create(services), conf)
            detectors.plate_lost.add(detector)
        except GenericError:
            eventlog.log_e(f'fail: {info.identity.ref}')
            raise
