from typing import Final, Optional

from edge import eventlog

from speedcontrol.detectors.wanted.types.conf import RuleConf
from speedcontrol.detectors.wanted.types.deps import RuleDeps
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.wanted import Evidence
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.waypoint.measurement import BestWaypoint


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    ticket_info = TicketInfo(conf.ticket.name)
    rule_info = RuleInfo(conf.name, None, ticket_info, {
        'plate-dist': conf.limits.plate_dist,
        'plate-list': {
            'owner': conf.limits.list_owner,
            'name': conf.limits.list_name
        }
    })

    return DetectorInfo(detector_name, rule_info)


class Rule:
    def __init__(self, deps: RuleDeps, conf: RuleConf, detector_name: str):
        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)

    def check(self, base_event: BaseEvent) -> Optional[Evidence]:
        all_points = base_event.measurements[0].waypoints
        best_points = [waypoint for waypoint in all_points if waypoint.name == waypoint.NAME_BEST]
        if len(best_points) != 1:
            eventlog.log_e('fail: find best-waypoint')
            return

        best_point: BestWaypoint = best_points[0]
        with self._deps.wanted as api:
            matches = api.check_plate(
                best_point.vehicle.vrp_text_ucode,
                self._conf.limits.plate_dist)

        if len(matches) == 0:
            return

        # filter matches
        rule_limits = self._conf.limits

        rule_matches = []
        for match in matches:
            owner = match.list.owner
            name = match.list.ident

            if rule_limits.list_owner == '*' or rule_limits.list_owner == owner:
                if rule_limits.list_name == '*' or rule_limits.list_name == name:
                    rule_matches.append(match)

        if len(rule_matches) == 0:
            return

        # eventlog.log_i('get wanted plate %s' % best_waypoint.vehicle.plate_text)
        origin = self._deps.about.get_origin(best_point.main_ts)
        evidence = Evidence(origin, best_point.main_ts, self._info, rule_matches)
        return evidence
