from typing import Dict, List

from edge import eventlog
from edge.image.types.jnode.builder import decode__im_builder_conf
from edge.types.dir import Dir
from edge.types.exceptions import JnodeError
from edge.types.identity import Identity

from speedcontrol.detectors.types.jnode.conf import decode__base_rule_conf, decode__video_conf
from speedcontrol.detectors.wrongway.types.conf import DetectorConf, RuleConf, RuleLimits, DirLimit
from speedcontrol.conf.jnode.mask import decode__mask_conf


def encode_dir_limit_text(value_code: DirLimit) -> str:
    code_to_text_map = {
        DirLimit.CODE_U: 'unknown',
        DirLimit.CODE_A: 'approaching',
        DirLimit.CODE_R: 'receding',
        DirLimit.CODE_NOWAY: 'any'
    }

    return code_to_text_map[value_code]


def decode__dir_limit_text(value_text: str) -> DirLimit:
    text_to_code_map: Dict[str, Dir] = {
        'unknown': DirLimit.CODE_U,
        'approaching': DirLimit.CODE_A,
        'receding': DirLimit.CODE_R,
        'any': DirLimit.CODE_NOWAY
    }

    value_code = text_to_code_map.get(value_text)
    if value_code is None:
        raise NotImplementedError()

    return DirLimit(value_code)


def decode__rule_limits(jnode: Dict) -> RuleLimits:
    """
    :raise JsonError:
    """
    # deserialize #/dir
    dir_limit_jnode = jnode.get('dir')
    if type(dir_limit_jnode) not in [str] or (dir_limit_jnode not in ['approaching', 'receding', 'any']):
        eventlog.log_e_param('#/dir')
        raise JnodeError

    dir_limit = decode__dir_limit_text(dir_limit_jnode)

    # deserialize #/lanes
    lanes_jnode = jnode.get('lanes', [])

    if type(lanes_jnode) != list:
        eventlog.log_e_param('#/lanes')
        raise JnodeError

    for lane in lanes_jnode:
        if not type(lane) in [str]:
            eventlog.log_e_param('#/lanes/%s' % lane)
            raise JnodeError

    return RuleLimits(lanes_jnode, dir_limit)


def decode__rule_conf(rule_jnode: Dict, footer_defvars: Dict) -> RuleConf:
    """
    :raise JsonError:
    """
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/name
    # deserialize #/enabled
    # deserialize #/schedule
    # deserialize #/ticket
    rule = decode__base_rule_conf(rule_jnode)

    # deserialize #/limits
    limits_jnode = rule_jnode.get('limits')
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/limits')
        raise JnodeError

    limits = decode__rule_limits(limits_jnode)

    return RuleConf(
        rule.name, rule.enabled, rule.schedule, rule.ticket,
        limits, footer_defvars)


def decode__detector_conf(identity: Identity, jnode: Dict, footer_defvars: Dict) -> DetectorConf:
    """
    :raise JsonError:
    """
    # deserialize #/name
    name = jnode.get('name')
    if not type(name) in [str] or len(name) > 32:
        eventlog.log_e_param('#/name')
        raise JnodeError

    # deserialize #/store
    store = jnode.get('store')
    if not (type(store) in [str] and 1 <= len(store) <= 2):
        eventlog.log_e_param('#/store')
        raise JnodeError

    # deserialize #/rules
    rules_jnode = jnode.get('rules')
    if type(rules_jnode) != list:
        eventlog.log_e_param('#/rules')
        raise JnodeError

    # deserialize #/scene
    scene_jnode = jnode.get('scene')
    if type(scene_jnode) != dict:
        eventlog.log_e_param('#/scene')
        raise JnodeError
    scene = decode__mask_conf(scene_jnode)

    rules: List[RuleConf] = []
    for rule_jnode in rules_jnode:
        rule_conf = decode__rule_conf(rule_jnode, footer_defvars)
        if rule_conf.enabled:
            rules.append(rule_conf)

    # deserialize #/images
    images_jnode = jnode.get('images')
    if type(images_jnode) != dict:
        eventlog.log_e_param('#/images')
        raise JnodeError

    im_engine = decode__im_builder_conf(images_jnode['violation'])

    # deserialize #/videos/measuments
    video_jnode = jnode.get('videos', {}).get('violation')
    if video_jnode is not None:
        if video_jnode.get('enabled', False):
            video_args = decode__video_conf(video_jnode)
        else:
            video_args = None
    else:
        video_args = None

    return DetectorConf(identity, name, store, rules, scene, im_engine, video_args)
