from datetime import datetime
from typing import Final, Optional, List, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.image.types.footer import ImFooterVars
from edge.services.types.vs import VsImageInfo
from edge.support.schedule import Schedule
from edge.types.dir import Dir
from edge.types.jnode.dir import encode_dir_text
from edge.types.origin import Origin
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem
from edge.types.trackset import Trackset

from speedcontrol.detectors.wrongway.types.conf import RuleConf, DirLimit
from speedcontrol.detectors.wrongway.types.deps import RuleDeps
from speedcontrol.detectors.wrongway.types.jnode.conf import encode_dir_limit_text
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.wrongway import Evidence
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle
from speedcontrol.events.waypoint.violation import ViolationWaypoint


class _TrImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.ts)
        self.tr_item: Final[TrTrackItem] = tr_item


def _create_detector_info(detector_name: str, conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.name, schedule_info, ticket_info, {
        'lanes': conf.limits.lanes,
        'dir': encode_dir_limit_text(conf.limits.dir)
    })

    return DetectorInfo(detector_name, rule_info)


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 detector_name: str,
                 init_im_env: ImBuilderEnv,
                 fini_im_env: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(detector_name, conf)

        self._ime_init: Final[ImBuilderEnv] = init_im_env
        self._ime_fini: Final[ImBuilderEnv] = fini_im_env

        self._scheduler = Schedule(conf.schedule, self._deps.tz)

    def check(self,
              trackset: Trackset,
              base_event: BaseEvent,
              lane_to_tr_item: List[Tuple[str, TrTrackItem]]) -> Optional[Tuple[Evidence, TimeSpan]]:

        min_ts: int = trackset.lr_track.min_ts
        max_ts: int = trackset.lr_track.max_ts

        if not (self._scheduler.check_ts_usec(max_ts) or self._scheduler.check_ts_usec(min_ts)):
            return

        lane_match_list: List[TrTrackItem] = []
        for lane_ref, tr_item in lane_to_tr_item:
            if len(self._conf.limits.lanes) == 0 or lane_ref in self._conf.limits.lanes:
                lane_match_list.append(tr_item)

        if len(lane_match_list) < 3:
            return

        # lane_match_list.sort(key=lambda ref: ref.ts, reverse=False)

        violation_items: List[TrTrackItem] = []

        for tr_item in lane_match_list:
            tr_item_dir: Dir = trackset.tr_track.get_item_dir(tr_item.ts)
            if tr_item_dir in (Dir.CODE_A, Dir.CODE_R):
                if self._conf.limits.dir in (DirLimit.CODE_NOWAY, tr_item_dir):
                    violation_items.append(tr_item)

        if len(violation_items) < 2:
            return

        scene_ms_refs = [_TrImageInfo(x) for x in violation_items]
        init_scene_ms_refs = scene_ms_refs[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)

        fini_scene_ms_refs = scene_ms_refs[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        init_winner_img = self._deps.vs.export_image_rgba(
            self._ime_init.scene_rgba, init_scene_ms_refs, self._ime_init.scene_dh)

        fini_winner_img = self._deps.vs.export_image_rgba(
            self._ime_fini.scene_rgba, fini_scene_ms_refs, self._ime_fini.scene_dh)

        if None in [init_winner_img, fini_winner_img]:
            eventlog.log_i('fail: get init/fini image from vs')
            return

        return self._mk_evidence(trackset, base_event, init_winner_img.tr_item, fini_winner_img.tr_item)

    def _mk_evidence_waypoint(self,
                              trackset: Trackset,
                              vehicle: Vehicle,
                              tr_item: TrTrackItem,
                              tr_item_ime: ImBuilderEnv) -> ViolationWaypoint:
        """
        :raise RuntimeError:
        """
        lane_ref: Optional[str] = tr_item.lane_item.lane if tr_item.lane_item is not None else ''
        origin: Origin = self._deps.about.get_origin(tr_item.ts, lane_ref)

        item_dir = trackset.tr_track.get_item_dir(tr_item.ts)
        item_dir_alias = self._conf.footer_defvars.get(encode_dir_text(item_dir))

        footer_vars = ImFooterVars(
            self._deps.tz, self._deps.about, tr_item.ts,
            defvars=self._conf.footer_defvars,
            location=origin.location,
            geolocation=origin.geolocation,
            plate_text=vehicle.vrp_text_ucode,
            direction=item_dir,
            spot_speed=tr_item.ra_item and tr_item.ra_item.speed_kmh,
            spot_speed_ra_n=vehicle.spotspeeds.ra_n,
            spot_speed_cam_2=vehicle.spotspeeds.cam_2,
            spot_speed_cam_n=vehicle.spotspeeds.cam_n,
            spot_speed_wim=vehicle.spotspeeds.wim,
            spot_speed_main=vehicle.spotspeeds.main,
            spot_speed_control=vehicle.spotspeeds.control,
            mmr_class=vehicle.mmr_class)

        item_dt_utc: datetime = self._deps.tz.map_ts_usec_to_datetime_utc(tr_item.ts)
        item_images = tr_item_ime.builder.build(
            tr_item_ime.scene_rgba, item_dt_utc, origin.geolocation,
            footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
            vehicle=False, scene=True, plate=True)

        if len(item_images) != 2:
            eventlog.log_i('fail: get all blobs for evidence')
            raise RuntimeError

        vehicle = ViolationVehicle(
            position=None,
            direction=item_dir,
            direction_alias=item_dir_alias,
            plate_quads=[tr_item.lr_item.plate_quad])

        waypoint = ViolationWaypoint(
            ViolationWaypoint.NAME_VIOLATION,
            tr_item.ts, None, origin.location, item_images,
            vehicle)

        return waypoint

    def _mk_evidence(self,
                     trackset: Trackset,
                     base_event: BaseEvent,
                     init_tr_item: TrTrackItem,
                     fini_tr_item: TrTrackItem) -> Optional[Tuple[Evidence, TimeSpan]]:

        assert init_tr_item.ts == self._ime_init.scene_rgba.ts
        assert fini_tr_item.ts == self._ime_fini.scene_rgba.ts

        vehicle: Vehicle = base_event.best_wp.vehicle
        try:
            init_wp = self._mk_evidence_waypoint(trackset, vehicle, init_tr_item, self._ime_init)
            fini_wp = self._mk_evidence_waypoint(trackset, vehicle, fini_tr_item, self._ime_fini)
        except RuntimeError:
            return

        duration_sec: int = (fini_tr_item.ts - init_tr_item.ts) // 1000000

        event_ts = fini_tr_item.ts
        origin = self._deps.about.get_origin(event_ts)
        evidence = Evidence(origin, event_ts, duration_sec, self._info)
        evidence.add_waypoints([init_wp, fini_wp])

        return evidence, TimeSpan(trackset.lr_track.min_ts, trackset.lr_track.max_ts)
