from typing import Dict, Optional

from edge.services.types.tz import ITzService
from edge.types.dir import Dir
from edge.types.jnode.dir import encode_dir_text
from edge.types.origin import Origin

from speedcontrol.events.detector import DetectorInfo
from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.jnode.schedule import encode__schedule_info
from speedcontrol.events.waypoint.violation import ViolationWaypoint


class SpotSpeedWagaEvidence(BaseEvidence):
    def __init__(self,
                 origin: Origin,
                 event_ts: int,
                 wim_speed: int,
                 cam_speed: int,
                 detector_info: DetectorInfo,
                 direction: Dir,
                 speed_limit: int):

        super().__init__(BaseEvidence.EVENT_TYPE__SPOTSPEED, origin, event_ts)

        self.detector = detector_info
        self.wim_speed = wim_speed
        self.cam_speed = cam_speed
        self.dir = direction
        self.speed_limit = speed_limit

    @property
    def best_waypoint(self) -> Optional[ViolationWaypoint]:
        bw = [x for x in self.waypoints if x.name == ViolationWaypoint.NAME_VIOLATION_BEST]
        return bw[0] if len(bw) else None

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = BaseEvidence.serialize(self, tz_service)
        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector.name,
                'rule': {
                    'name': self.detector.rule.name,
                    'schedule': encode__schedule_info(self.detector.rule.schedule),
                    'ticket': {
                        'name': self.detector.rule.ticket.name
                    },
                    'limits': self.detector.rule.limits
                }
            },
            'measurements': {
                'speed': {
                    'main': {
                        'measured': self.wim_speed,
                        'source': 'wim'
                    },
                    'control': [{
                        'measured': self.cam_speed,
                        'source': 'camera'
                    }],
                    'wim': self.wim_speed,
                    'camera': self.cam_speed
                },
                'overspeed': {
                    'main': {
                        'measured': self.wim_speed - self.speed_limit,
                        'source': 'wim'
                    },
                    'control': [{
                        'measured': self.cam_speed - self.speed_limit,
                        'source': 'camera'
                    }],
                    'wim': self.wim_speed - self.speed_limit,
                    'camera': self.cam_speed - self.speed_limit
                },
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if self.dir != Dir.CODE_U:
            jnode['evidence']['measurements']['dir'] = encode_dir_text(self.dir)

        if len(self._videos):
            videos_jnode = [x.serialize(tz_service) for x in self._videos]
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode
