from typing import Dict, Optional

from edge.support.speed import map_speed_uint
from edge.types.dir import Dir
from edge.types.jnode.dir import encode_dir_text
from edge.types.jnode.shape.quad import encode__quadi
from edge.types.jnode.shape.rect import encode__recti
from edge.types.trackset.spotspeed import SpotSpeeds

from speedcontrol.events.jnode.vrp import encode__vrp
from speedcontrol.events.jnode.position import encode__position
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle


def encode__violation_vehicle(self: ViolationVehicle) -> Optional[Dict]:
    jnode = dict()

    if self.dir and self.dir != Dir.CODE_U:
        jnode['dir'] = encode_dir_text(self.dir)

    if self.dir_alias:
        jnode['dir-alias'] = self.dir_alias

    if self.position:
        jnode['position'] = encode__position(self.position)

    if len(self.plate_quads):
        jnode['licence-plates'] = [
            {
                'bbox4': encode__quadi(quad),
                'bbox2': encode__recti(quad.bbox_recti)
            }
            for quad in self.plate_quads
        ]

    elif len(self.plate_rects):
        jnode['licence-plates'] = [
            {'bbox2': encode__recti(rect)} for rect in self.plate_rects
        ]

    # Jnode contains at least one key
    return jnode if jnode else None


def encode__vehicle_spotspeeds(spotspeeds: SpotSpeeds) -> Dict:
    main_jnode = dict()
    if spotspeeds.main is not None:
        main_jnode.update({
            'source': spotspeeds.main.source,
            'measured': map_speed_uint(spotspeeds.main.measured)
        })

    control_jnode = dict()
    if spotspeeds.control is not None:
        control_jnode.update({
            'source': spotspeeds.control.source,
            'measured': map_speed_uint(spotspeeds.control.measured)
        })

    radar_jnode = dict()
    if spotspeeds.ra_n is not None:
        ra_n_uint = map_speed_uint(spotspeeds.ra_n)
        if ra_n_uint > 240:
            ra_n_str = '>240'

        elif ra_n_uint < 20:
            ra_n_str = '<20'

        else:
            ra_n_str = str(ra_n_uint)

        radar_jnode.update({
            'measured': ra_n_uint,
            'measured2': ra_n_str
        })

    camera_jnode = dict()

    if spotspeeds.cam_2 is not None:
        camera_jnode.update({
            'measured': map_speed_uint(spotspeeds.cam_2)
        })

    if spotspeeds.cam_n is not None:
        camera_jnode.update({
            'measured2': map_speed_uint(spotspeeds.cam_n)
        })

    wim_jnode = dict()
    if spotspeeds.wim is not None:
        wim_jnode.update({
            'measured': map_speed_uint(spotspeeds.wim)
        })

    jnode = dict()

    if main_jnode:
        jnode['main'] = main_jnode

    if control_jnode:
        jnode['control'] = control_jnode

    if radar_jnode:
        jnode['radar'] = radar_jnode

    if camera_jnode:
        jnode['camera'] = camera_jnode

    if wim_jnode:
        jnode['wim'] = wim_jnode

    return jnode


def encode__vehicle(self: Vehicle) -> Dict:
    jnode = {
        'license-plates': [encode__vrp(plate) for plate in self.plates]
    }

    if self.wa_track:
        jnode['info'] = [{
            'provider': self.wa_track.provider,
            'road-temperature': self.wa_track.road_temperature,
            'lane': self.wa_track.lane,
            'class': self.wa_track.vehicle_class,
            'event-id': self.wa_track.event_id,
            'event-ts': self.wa_track.event_ts,
            'event-dur': self.wa_track.duration
        }]

        jnode['weight'] = {
            'measured': self.wa_track.weight,
            'axle-count': self.wa_track.axle_count,
            'axle-layout': self.wa_track.axle_layout
        }

    if self.type_jnode is not None:
        jnode['type'] = self.type_jnode

    if self.seats_jnode is not None:
        jnode['seats'] = self.seats_jnode

    if self.position is not None:
        jnode['position'] = encode__position(self.position)

    if self.dir != Dir.CODE_U:
        jnode['dir'] = encode_dir_text(self.dir)

    if self.dir_alias is not None:
        jnode['dir-alias'] = self.dir_alias

    if self.avg_dir != Dir.CODE_U:
        jnode['dir-avg'] = encode_dir_text(self.avg_dir)

    if self.avg_dir_alias is not None:
        jnode['dir-alias-avg'] = self.avg_dir_alias

    spotspeed_jnode = encode__vehicle_spotspeeds(self.spotspeeds)
    if spotspeed_jnode:
        jnode['spot-speed'] = spotspeed_jnode

    size_jnode = dict()

    if self.length is not None:
        size_jnode.update({
            'l': self.length
        })

    if self.wa_track is not None:
        size_jnode.update({
            'w': self.wa_track.width,
            'h': self.wa_track.height,
            'l': self.wa_track.length
        })

    if size_jnode:
        jnode['size'] = size_jnode

    return jnode
