from edge import eventlog
from edge.types.blob import StructBlobMeta, TBlob
from edge.support import bs
from edge.types.identity import Identity

from speedcontrol.services.types.bs import IBsService, BsServiceConf, BsSenderToChanMap


class _BsService(IBsService):
    def __init__(self, twisted_reactor, bs_conf: BsServiceConf, counters: bs.IBsCounters):
        IBsService.__init__(self)
        self._reactor = twisted_reactor
        self._session_args = (
            f'speedcontrol.{bs_conf.module_id}',
            counters,
            65535,
            bs.LOG_LEVEL_NAME_TO_CODE_MAP.get(bs_conf.log_level, 'info'))

        self._session = bs.Session(counters=counters, closed=True)

        self._counters = counters
        self._sender_to_chan_map: BsSenderToChanMap = dict()

    def destroy(self) -> None:
        self._session.close()

    def setup(self, sender_to_chan_map: BsSenderToChanMap) -> None:
        self._sender_to_chan_map = sender_to_chan_map

    def get_size(self, sender: Identity, blob_id: str, profile: str) -> int:
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[profile]
        else:
            chan = chan_or_profiles

        reply = self._session.get_blob_size(chan, blob_id)
        return reply

    def save(self, sender: Identity, blob: TBlob) -> None:
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[blob.profile]
        else:
            chan = chan_or_profiles

        self._counters.inc_blob(chan)
        self._save(chan, blob)

    @bs.reopen_session
    def _save(self, chan: str, blob: TBlob) -> None:
        """
        :raise bs.BsError:
        :raise bs.BsSessionError:
        """
        try:
            blob_ts: int = blob.min_ts // 1000
            if blob.file_path is not None:
                ident = self._session.save_blob2(chan, blob_ts, blob.file_path, blob.file_meta)
            elif isinstance(blob.meta, StructBlobMeta):
                ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.meta.struct)
            else:
                ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.file_meta)

        except bs.BsError:
            eventlog.log_e(f'fail: save (chan={chan}, profile={blob.profile})')
            raise

        blob.name = f'{chan}_{ident}.{blob.file_ext}'


def init_service(twisted_reactor, conf: BsServiceConf, counters: bs.IBsCounters) -> IBsService:
    return _BsService(twisted_reactor, conf, counters)
