import ujson
import psycopg2
import psycopg2.pool
import psycopg2.extras

from typing import Dict, List, Optional, Tuple

from edge import eventlog
from edge.support.db import DbJsonArg, DbArgs
from speedcontrol.services.types.db import IDbService, DbServiceConf

psycopg2.extras.register_default_json(globally=True, loads=ujson.loads)
psycopg2.extras.register_default_jsonb(globally=True, loads=ujson.loads)


class _DbService(IDbService):
    def __init__(self, origin_ref: int, db_args: DbArgs):
        super().__init__()

        self._origin_ref = origin_ref
        self._link = psycopg2.connect(
            host=db_args.db_host,
            port=db_args.db_port,
            user=db_args.db_user,
            password=db_args.db_pass,
            database=db_args.db_name)

    def destroy(self) -> None:
        self._link.close()

    def set_log(self, detector_ref: int, plate_text: str,
                log_min_ts: int, log_max_ts: int, log_jnode: Optional[Dict]) -> bool:

        # TODO: del in db, if log_jnode is None
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_set(%s,%s,%s,%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text,
                     DbJsonArg(log_jnode), log_min_ts, log_max_ts])

                row_count = db_cursor.fetchone()
                return row_count == 1

        except psycopg2.Error as x:
            eventlog.log_e(f'edge.plate_motion_set <{str(x)}>')

        return False

    def get_log(self, detector_ref: int, plate_text: str, plate_dist: int = 0
                ) -> Optional[List[Tuple[str, int, Dict]]]:
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT plate,distance,log FROM edge.plate_motion_get(%s,%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text, plate_dist])

                return db_cursor.fetchall()

        except psycopg2.Error as x:
            eventlog.log_e(f'edge.plate_motion_get <{str(x)}>')

    def del_log(self, detector_ref: int, plate_text: str) -> bool:
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_del(%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text])

                row_count = db_cursor.fetchone()
                return row_count == 1

        except psycopg2.Error as x:
            eventlog.log_e(f'edge.plate_motion_del <{str(x)}>')

        return False

    def clear_log(self, detector: int, max_ts: int) -> None:
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_clear(%s,%s,%s)',
                    [self._origin_ref, detector, max_ts])

                row_count = db_cursor.fetchone()
                if row_count:
                    eventlog.log_d(f'plate_motion_clear: deleted <{row_count}> records')

        except psycopg2.Error as x:
            eventlog.log_e(f'edge.plate_motion_clear <{str(x)}>')

    def get_last_parking(self, plate_text: str, origin_uuid: str, min_search_time: str) -> bool:
        api_query_jnode = {
            'event': {
                'type': ['parking'],
                'origin': [origin_uuid],
                'plate-pattern': plate_text,
                'fromtime': min_search_time
            }
        }

        with self._link, self._link.cursor() as db_cursor:
            db_cursor.execute(
                'SELECT errcode,errmsg,result FROM api_event_filter(%s)',
                [ujson.dumps(api_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()

        return errcode == 0


def init_service(args: DbServiceConf) -> IDbService:
    return _DbService(args.origin_id, args.db_args)

