import ctypes
import ujson

from edge.support import fs, sh
from typing import Callable, Dict, Optional, Self

from edge import eventlog
from edge.types.exceptions import BufferSizeError, ToutError, InternalError

from speedcontrol.services.types.sign import SignServiceConf, SignError, ISignService


class _StubService(ISignService):
    @property
    def cert(self) -> bytes:
        return b''

    @property
    def hash_type(self) -> str:
        import hashlib
        return hashlib.md5().name

    def hash_blob(self, blob_buffer: bytes, hash_buffer: ctypes.Array) -> int:
        import hashlib

        md5 = hashlib.md5(blob_buffer)
        md5_data: str = md5.hexdigest()
        md5_size = len(md5_data)

        if md5_size <= len(hash_buffer):
            hash_buffer.value = md5_data.encode('ascii')
            return md5_size

        else:
            raise BufferSizeError(md5_size)

    def sign_blob(self, blob_buffer: str, sign_handler: Callable[[Optional[str], Optional[str]], None]) -> None:
        sign_handler(None, None)

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        pass


class _CryptoApi:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    LOG_FLAGS_NONE = 0
    LOG_FLAGS_STDERR = 1

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    # typedef void (*sign_handler_fn)(void *aHandlerArg, int aStat,
    # const uint8_t *aBlobData, uint32_t aBlobSize, const uint8_t *aSignData, uint32_t aSignSize);
    SignHandler = ctypes.CFUNCTYPE(
        None,
        ctypes.c_void_p, ctypes.c_int,
        ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
        ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32)

    # typedef void (*crypto_stats_handler_fn)(void *aHandlerArg, const char *aJson);
    StatsHandler = ctypes.CFUNCTYPE(
        None,
        ctypes.c_void_p, ctypes.POINTER(ctypes.c_char))

    _lib = None
    _libpath = '/opt/edge/lib/libcrypto-provider.so'

    if fs.path_exists(_libpath):
        _lib = ctypes.CDLL(_libpath)

        # void CryptoLogOpen(const char *aName, int aMaxLevel, int aFlags);
        logopen = _lib.CryptoLogOpen
        logopen.restype = ctypes.c_void_p
        logopen.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)

        # int
        # CryptoProviderOpen(const char *aConfPath)
        open = _lib.CryptoProviderOpen
        open.restype = ctypes.c_int
        open.argtypes = (ctypes.c_char_p,)

        # int
        # CryptoProviderClose()
        close = _lib.CryptoProviderClose
        close.argtypes = ()

        # int
        # CryptoProviderStart()
        start = _lib.CryptoProviderStart
        start.argtypes = ()

        # int
        # CryptoProviderStop()
        stop = _lib.CryptoProviderStop
        stop.argtypes = ()

        # int
        # CryptoProviderGetCrt(uint8_t *aCrtData, uint32_t *aCrtSize);
        getcrt = _lib.CryptoProviderGetCrt
        getcrt.restype = ctypes.c_int
        getcrt.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # void
        # CryptoProviderCollectStats(crypto_stats_handler_fn aHandler, void *aHandlerArg);
        collect_stats = _lib.CryptoProviderCollectStats
        collect_stats.restype = ctypes.c_void_p
        collect_stats.argtypes = (StatsHandler, ctypes.c_void_p)

        # int
        # CryptoProviderSign(const uint8_t *aBlobData, uint32_t aBlobSize,
        #     sign_handler_fn aHandler, void *aHandlerArg)
        sign = _lib.CryptoProviderSign
        sign.restype = ctypes.c_int
        sign.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32, SignHandler, ctypes.c_void_p)

        # int
        # CryptoProviderHash (const uint8_t *aBlobData, uint32_t aBlobSize,
        #     uint8_t *aHashData, uint32_t *aHashSize);
        hash = _lib.CryptoProviderHash
        hash.restype = ctypes.c_int
        hash.argtypes = (ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                         ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    hash_type = 'gost34311'

    @classmethod
    def log_level_name_to_code(cls, name: str) -> int:
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }

        return name_to_code.get(name, cls.LOG_I)

    @classmethod
    def test_available(cls) -> bool:
        return cls._lib is not None


class _SignWork:
    """
    :type _service: _SignService | None
    """
    _ref_to_self_map = dict()

    def __init__(self, twisted_reactor,
                 service,
                 blob_buffer: bytes,
                 handler: Callable[[Optional[bytes], Optional[bytes]], None]):
        """
        :type service: _SignService
        """
        self._ref = id(self)

        self._reactor = twisted_reactor
        self._service = service
        self._handler = handler

        def c_call_handler_py(
                work_ref: int, sign_stat: int,
                blob_data: ctypes.POINTER(ctypes.c_uint8),
                blob_size: int,
                sign_data: ctypes.POINTER(ctypes.c_uint8),
                sign_size: int
                ) -> None:
            """
            The handler is called from crypto-provider's thread. Do synchronize with the twisted reactor
            """
            if sign_stat == _CryptoApi.RC_DONE:
                eventlog.log_d(
                    f'done: CryptoProviderSign (stat:{sign_stat}, blob-size:{blob_size}, sign-size:{sign_size})')
                blob_data_bytes: bytes = ctypes.string_at(blob_data, blob_size)
                sign_data_bytes: bytes = ctypes.string_at(sign_data, sign_size)

            else:
                eventlog.log_w(f'fail: CryptoProviderSign (stat:{sign_stat})')
                blob_data_bytes: bytes | None = None
                sign_data_bytes: bytes | None = None

            self._reactor.callFromThread(_SignWork._twisted_handler, work_ref, blob_data_bytes, sign_data_bytes)

        self._c_call_handler = _CryptoApi.SignHandler(c_call_handler_py)
        self._c_call_args = (blob_buffer, self._c_call_handler, self._ref)

    def __del__(self) -> None:
        if _SignWork._ref_to_self_map.pop(self._ref, None):
            eventlog.log_i(f'done: fini sign-work (ref:{self._ref})')

    @classmethod
    def _twisted_handler(cls, work_ref: int, blob_data: Optional[bytes], sign_data: Optional[bytes]) -> None:
        eventlog.log_d('exec: fini sign-work (ref:{}, ref-to-self-map:{})'.format(
            work_ref, list(_SignWork._ref_to_self_map.keys())))

        self: Optional[Self] = cls._ref_to_self_map.pop(work_ref, None)
        if self is None:
            eventlog.log_w(f'fail: fini sign-work (ref:{work_ref})')

        else:
            eventlog.log_d(f'done: fini sign-work (ref:{work_ref})')

            # noinspection PyBroadException
            try:
                self._c_call_handler = None
                self._c_call_args = None
                self._handler(blob_data, sign_data)
                eventlog.log_d(f'done: ref:{self._ref}')
            except Exception:
                eventlog.log_x()

    def run_async(self) -> None:
        """
        :raise SignError:
        """
        eventlog.log_d(f'exec: init sign-work (ref:{self._ref})')

        _SignWork._ref_to_self_map[self._ref] = self

        # noinspection PyBroadException
        try:
            self._service.sign_buffer(*self._c_call_args)
            eventlog.log_d(f'done: init sign-work (ref:{self._ref})')
            return
        except SignError:
            eventlog.log_e(f'fail: init sign-work (ref:{self._ref}, session error')
        except Exception:
            eventlog.log_x()

        self._c_call_handler = None
        self._c_call_args = None
        del _SignWork._ref_to_self_map[self._ref]
        eventlog.log_d(f'done: fini sign-work (ref:{self._ref})')

        raise SignError


class _SignService(ISignService):
    def __init__(self, twisted_reactor, crypto_api: type(_CryptoApi)):
        super().__init__()
        self._reactor = twisted_reactor
        self._api = crypto_api
        self._cert = None

    def destroy(self) -> None:
        self._api.stop()
        self._api.close()

    @property
    def cert(self) -> bytes:
        if self._cert is None:
            buffer_data = ctypes.create_string_buffer(b'', size=1)
            for _ in range(2):
                try:
                    buffer_data2 = ctypes.cast(buffer_data, ctypes.POINTER(ctypes.c_uint8))

                    buffer_size = len(buffer_data)
                    buffer_size_used = ctypes.c_uint32(buffer_size)
                    buffer_size_used2 = ctypes.pointer(buffer_size_used)

                    rc = self._api.getcrt(buffer_data2, buffer_size_used2)
                    if rc == self._api.RC_DENY:
                        eventlog.log_e('fail: GetCrt <RC_DENY>')
                        raise SignError

                    elif rc != self._api.RC_DONE:
                        eventlog.log_e(f'fail: GetCrt <{rc}>')
                        raise SignError

                    buffer_used = buffer_size_used.value
                    if buffer_used > buffer_size:
                        eventlog.log_e(f'fail: crt-buffer too small (min-size: {buffer_used})')
                        raise BufferSizeError(buffer_used)

                    self._cert = buffer_data.raw
                    break

                except BufferSizeError as x:
                    buffer_data = ctypes.create_string_buffer(b'', size=x.min_size)

        return self._cert

    @property
    def hash_type(self) -> str:
        return self._api.hash_type

    def hash_blob(self, blob_buffer: bytes, hash_buffer: ctypes.Array) -> int:
        """
        :raise SignError:
        """

        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        buffer_data = ctypes.cast(hash_buffer, ctypes.POINTER(ctypes.c_uint8))

        buffer_size = len(hash_buffer)
        buffer_size_used = ctypes.c_uint32(buffer_size)
        buffer_size_used2 = ctypes.pointer(buffer_size_used)

        rc = self._api.hash(blob_data, blob_size, buffer_data, buffer_size_used2)
        if rc == self._api.RC_DENY:
            eventlog.log_e('fail: Hash <RC_DENY>')
            raise SignError

        elif rc != self._api.RC_DONE:
            eventlog.log_e('fail: Hash (%s)' % rc)
            raise SignError

        buffer_used = buffer_size_used.value
        if buffer_used > buffer_size:
            eventlog.log_e(f'fail: hash-buffer too small (min-size: {buffer_used})')
            raise BufferSizeError(buffer_used)

        return buffer_used

    def sign_blob(self, blob_buffer: bytes, sign_handler: Callable[[Optional[bytes], Optional[bytes]], None]) -> None:
        try:
            sign_work = _SignWork(self._reactor, self, blob_buffer, sign_handler)
            sign_work.run_async()
        except SignError:
            eventlog.log_e('fail: session error')
            raise
        except Exception:
            eventlog.log_x()
            raise SignError

    def sign_buffer(self, blob_buffer: bytes, cb: _CryptoApi.SignHandler, cb_arg: int) -> None:
        """
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        rc = self._api.sign(blob_data, blob_size, cb, cb_arg)
        if rc == self._api.RC_DENY:
            eventlog.log_e('fail: Sign <RC_DENY>')
            raise SignError

        elif rc != self._api.RC_DONE:
            eventlog.log_e(f'fail: Sign <answer:{rc}>')
            raise SignError

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        def stat_handler_py(_, stat_data):
            # noinspection PyBroadException
            try:
                stat_str = ctypes.string_at(stat_data)
                stat_jnode = ujson.loads(stat_str)
                json_handler(stat_jnode)
            except Exception:
                eventlog.log_x()

        _sign_handler_c = _CryptoApi.StatsHandler(stat_handler_py)
        self._api.collect_stats(_sign_handler_c, 14)


def _test_access(uiconf_path: str, uiref: str) -> bool:
    """
    :param uiconf_path:
        '/etc/edge/setupstore/admin/actions/iit-ca.0.json'
    :param uiref:
        'iit-ca.0~end-user.0'
    :return:
    """
    iit_ca_jnode = fs.load_json3(uiconf_path, default=None)

    if iit_ca_jnode is None:
        eventlog.log_e(f'fail: iit-ca-enduser not loaded ({uiconf_path})')
        return False

    iit_ca_enduser_jnode = iit_ca_jnode.get(uiref)

    if iit_ca_enduser_jnode is None:
        eventlog.log_e(f'fail: iit-ca-enduser not found ({uiref})')
        return False

    iit_ca_enduser_pri = iit_ca_enduser_jnode['pri']
    key_serial = iit_ca_enduser_pri['serial']
    key_pass = iit_ca_enduser_pri['password']

    if key_serial and key_pass and fs.path_exists('/opt/edge/bin/crypto-iit-cli'):
        try:
            reply = sh.exec_cli('/opt/edge/bin/crypto-iit-cli',
                                ['key', 'test-access', key_serial, key_pass],
                                tout_sec=10, encoding='utf-8')

            if 'PASSED' in reply:
                return True

        except ToutError:
            eventlog.log_w("fail: crypto-iit-cli test-access (timeout)")
        except InternalError:
            eventlog.log_w("fail: crypto-iit-cli test-access (internal error)")

    iit_ca_enduser_jnode['pub']['enabled'] = False
    fs.write_json_file(iit_ca_jnode, uiconf_path, pretty=True)
    return False


def init_service(twisted_reactor, conf: Optional[SignServiceConf]) -> ISignService:
    if conf is None:
        eventlog.log_i('using _StubService')
        return _StubService()

    elif not _CryptoApi.test_available():
        eventlog.log_w('using _StubService')
        return _StubService()

    elif not _test_access(conf.ui_conf_path, conf.ui_ref):
        eventlog.log_w('using _StubService')
        return _StubService()

    else:
        eventlog.log_i('using libsign-iit')

    log_level = _CryptoApi.log_level_name_to_code(conf.log_level)
    #_CryptoApi.logopen(bytes('speedcontrol.iit', 'ascii'), log_level, _CryptoApi.LOG_FLAGS_NONE)

    rc = _CryptoApi.open(bytes(conf.iit_conf_path, 'ascii'))
    _CryptoApi.start()
    if rc == _CryptoApi.RC_DENY:
        eventlog.log_e('fail: Open <RC_DENY>')
        raise SignError

    elif rc != _CryptoApi.RC_DONE:
        eventlog.log_e(f'fail: Open <{rc}>')
        raise SignError

    return _SignService(twisted_reactor, _CryptoApi)
