import time
import os

from collections import defaultdict
from typing import Final

from edge import eventlog
from edge.services.types.stats import IStatsProvider, StatsServiceConf
from edge.support import bs, fs

from speedcontrol.services.types.stats import IEventCounters, IScStatsService


class BsCounters(bs.IBsCounters):
    def __init__(self):
        bs.IBsCounters.__init__(self)

        def _dict_factory():
            return {
                'blob-count': 0,
                'done-count': 0,
                'fail-count': 0
            }
        self._counters = defaultdict(_dict_factory)

    def inc_blob(self, chan: str) -> None:
        self._counters[chan]['blob-count'] += 1

    def inc_done(self, chan: str) -> None:
        self._counters[chan]['done-count'] += 1

    def inc_fail(self, chan: str) -> None:
        self._counters[chan]['fail-count'] += 1

    def serialize(self) -> dict:
        return self._counters


class _EventCounters(IEventCounters):
    def __init__(self):
        self._fact_count = int(0)

    def inc_fact(self) -> None:
        self._fact_count += 1

    @property
    def fact_count(self) -> int:
        return self._fact_count


class _StatService(IScStatsService):
    def __init__(self, twisted_reactor, conf: StatsServiceConf) -> None:
        self._providers: Final[list[IStatsProvider]] = []
        self._conf: Final[StatsServiceConf] = conf

        self._event = _EventCounters()
        self._bs = BsCounters()

        # reset on start
        twisted_reactor.callLater(0, self._collect2)

        if conf.interval_sec > 0:
            twisted_reactor.callLater(conf.interval_sec, self._collect, twisted_reactor)

    @property
    def event(self) -> IEventCounters:
        return self._event

    @property
    def bs(self) -> bs.IBsCounters:
        return self._bs

    def add_providers(self, providers: list[IStatsProvider]) -> None:
        self._providers.extend(providers)

    def _collect(self, twisted_reactor) -> None:
        # noinspection PyBroadException
        try:
            self._collect2()
        except Exception:
            eventlog.log_x()

        twisted_reactor.callLater(self._conf.interval_sec, self._collect, twisted_reactor)

    def _collect2(self) -> None:
        sys_utc_time = time.time()

        units = [
            {
                'type': 'stats',
                'name': 'stats',
                'stat': {
                    'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                    'tout-sec': self._conf.interval_sec
                }
            }
        ]

        list(map(lambda provider: provider.collect_stats(units.append), self._providers))
        stat = {
            'units': units,
            'stat': {
                'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                'tout-sec': self._conf.interval_sec
            },
            'event': {
                'count': self.event.fact_count
            }
        }

        bs_stats = self.bs.serialize()
        if bs_stats:
            stat['bs'] = bs_stats

        try:
            fs.write_json_file(stat, self._conf.filepath_tmp, True)
            os.rename(self._conf.filepath_tmp, self._conf.filepath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{self._conf.filepath}> ({x})')


def init_service(twisted_reactor, args: StatsServiceConf) -> IScStatsService:
    return _StatService(twisted_reactor, args)
