import ctypes
import os

from glob import glob
from tempfile import NamedTemporaryFile
from typing import Callable, List, Optional, Self, Tuple

from edge import eventlog
from edge.image.helpers import ImHelpers
from edge.support import fs
from edge.types.blob import FileBlob, JsonBlobMeta
from edge.types.image import ImRgba, ImJpeg

from edge.services.types.vs import VsServiceConf, TVsImageInfo, VsError, VsSessionAbsentError, \
    VsSessionExpiredError, IVsService

_IMAGE_POOL_MAX = 3
_VIDEO_POOL_MAX = 3


class _StubService(IVsService):
    def __init__(self):
        IVsService.__init__(self)

    def export_image_rgba(self, im_rgba: ImRgba, im_infos: List[TVsImageInfo], im_footer_h: int = 0,
                          bs_chan: Optional[str] = None) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        return None

    def export_image_jpeg(self,
                          im_buffer: ImJpeg,
                          im_infos: List[TVsImageInfo],
                          bs_chan: Optional[str] = None
                          ) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        return None

    def export_video(self, min_ts: int, max_ts: int, max_w: int, max_h: int, rec_quality: int,
                     handler: Callable[[Optional[FileBlob]], None]) -> None:
        """
        :param min_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_w:
            valid values in range [16; 32752], multiple 16.
        :param max_h:
            valid values in range [16; 32752], multiple 16.
        :param rec_quality:
            Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].

        :param handler:

        :raise VsSessionAbsentError:
        :raise VsError:
        """
        return None


class _Api:
    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2
    RC_CONN = -3

    LOG_FLAGS_NONE = 0
    LOG_FLAGS_STDERR = 1

    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    IMAGE_GRAY = 1
    IMAGE_RGBA = 2
    IMAGE_JPEG = 3

    VIDEO_H264_AVI = 1
    VIDEO_H264_MKV = 2
    VIDEO_H264_MP4 = 3

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libblobstore-bsv*.so')[0])

    # int
    # BsvOpen(uint16_t aStore, uint8_t aLogLevel, void **aSession);

    open = _lib.BsvOpen
    open.argtypes = (ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_void_p))
    open.restype = ctypes.c_int

    # int
    # BsvLoadImage(void *aSession, const char *aChannel, uint64_t anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize,uint16_t *anImWidth, uint16_t *anImHeight);

    load = _lib.BsvLoadImage
    load.argtypes = (
        ctypes.c_void_p, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint16), ctypes.POINTER(ctypes.c_uint16))

    load.restype = ctypes.c_int

    # int
    # BsvLoadImageCoded(void *aSession, const char *aChannel, uint64_t *anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize);

    loadCoded = _lib.BsvLoadImageCoded
    loadCoded.argtypes = (
        ctypes.c_void_p,
        ctypes.c_char_p, ctypes.POINTER(ctypes.c_uint64),
        ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32))

    loadCoded.restype = ctypes.c_int

    # void
    # typedef void (*bsv_load_record_handler_fn)(int aStat, uint64_t aRecMinTs,
    #               uint64_t aRecMaxTs, const char *aRecPath,
    #               void *aHandlerArg);

    LoadRecordHandler = ctypes.CFUNCTYPE(
        None, ctypes.c_int, ctypes.c_uint64,
        ctypes.c_uint64, ctypes.c_char_p,
        ctypes.c_void_p)

    # int
    # BsvLoadRecordAsync(void *aSession, const char *aChannel,
    #     uint64_t aRecMinTs, uint64_t aRecMaxTs,
    #     uint8_t aRecFormat, uint32_t aRecQuality, uint16_t aRecSkipRate,
    #     uint16_t aRecMaxWidth, uint16_t aRecMaxHeight, const char *aRecPath,
    #     bsv_load_record_handler_fn aHandlerFn, void *aHandlerArg);

    save = _lib.BsvLoadRecordAsync
    save.argtypes = (
        ctypes.c_void_p, ctypes.c_char_p,
        ctypes.c_uint64, ctypes.c_uint64,
        ctypes.c_uint8, ctypes.c_uint32, ctypes.c_uint16,
        ctypes.c_uint16, ctypes.c_uint16, ctypes.c_char_p,
        LoadRecordHandler, ctypes.c_void_p)

    save.restype = ctypes.c_int

    # void
    # BsvClose(void *aSession);
    close = _lib.BsvClose
    close.argtype = ctypes.c_void_p

    # void
    # BsvLogOpen(const char *aName, int aMaxLevel, int aFlags);
    logOpen = _lib.BsvLogOpen
    logOpen.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)

    # void
    # BsvLogClose();
    logClose = _lib.BsvLogClose


class _VsSession:
    def __init__(self, handle: Optional[ctypes.c_void_p], chan: str):
        self._chan = chan
        self._handle = handle
        self.closed = handle is None

    @classmethod
    def open(cls, store: int, chan: str, log_level: int = _Api.LOG_I) -> Self:
        """
        :raise VsSessionExpiredError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(store, log_level, handle)
        #_Api.logOpen(bytes('speedcontrol.{}'.format(store), 'ascii'), log_level, _Api.LOG_FLAGS_NONE)

        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvOpen (rc:%s)' % rc)
            raise VsSessionExpiredError

        return cls(handle=handle, chan=chan)

    def export_image_rgba(self,
                          im_ts: int,
                          im_buffer: ctypes.Array,
                          bs_chan: Optional[str] = None
                          ) -> Tuple[int, int]:
        """
        :param im_ts:
            value in msec
        :param bs_chan:
        :param im_buffer:

        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_w, im_h = ctypes.c_uint16(), ctypes.c_uint16()

        chan = self._chan if bs_chan is None else bs_chan
        rc = _Api.load(
            self._handle, bytes(chan, 'ascii'), ctypes.c_uint64(im_ts), ctypes.c_uint8(_Api.IMAGE_RGBA),
            im_data, im_size, im_w, im_h)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_w('fail: buffer too small')
            return im_w.value, im_h.value

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e(f'fail: BsvLoadImage (rc:{rc})')
            raise VsError

        return im_w.value, im_h.value

    def export_image_jpeg(self, im_ts: int, im_buffer: ctypes.Array, bs_chan: Optional[str] = None) -> Tuple[int, int]:
        """
        :param im_ts:
            value in msec
        :param im_buffer:
        :param bs_chan:

        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_ts_c = ctypes.c_uint64(im_ts)

        chan = self._chan if bs_chan is None else bs_chan
        rc = _Api.loadCoded(
            self._handle, bytes(chan, 'ascii'), im_ts_c, ctypes.c_uint8(_Api.IMAGE_JPEG),
            im_data, im_size)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_e(f'fail: buffer too small {im_size.value}')
            return im_size.value, im_ts_c.value

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e(f'fail: BsvLoadImage (rc:{rc}) {im_size.value}')
            raise VsError

        return im_size.value, im_ts_c.value

    def export_video(self, min_ts: int, max_ts: int, max_w: int, max_h: int, rec_q: int,
                     file_path: str, cb, cb_arg: int) -> None:
        """
        :param min_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_ts:
            ms since 01.01.1970 00:00 UTC
        :param max_w:
            valid values in range [16; 32752], multiple 16.
        :param max_h:
            valid values in range [16; 32752], multiple 16.
        :param rec_q:
            Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].
        :param file_path:
        :param cb:
            _Api.LoadRecordHandler
        :param cb_arg:

        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        rc = _Api.save(
            self._handle, ctypes.c_char_p(bytes(self._chan, 'ascii')),
            ctypes.c_uint64(min_ts),
            ctypes.c_uint64(max_ts),
            ctypes.c_uint8(_Api.VIDEO_H264_MP4),
            ctypes.c_uint32(rec_q),
            ctypes.c_uint16(1),  # skip-rate
            ctypes.c_uint16(max_w),
            ctypes.c_uint16(max_h),
            ctypes.c_char_p(bytes(file_path, 'ascii')),
            cb, ctypes.c_void_p(cb_arg))

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadRecordAsync (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e(f'fail: BsvLoadRecordAsync (rc:{rc})')
            raise VsError

    def close(self) -> None:
        if not self.closed:
            _Api.logClose()
            _Api.close(self._handle)
            self.closed = True


class _VsSessionPool:
    _LOG_NAME_TO_CODE_MAP = {
        'none': _Api.LOG_N,
        'error': _Api.LOG_E,
        'warn': _Api.LOG_W,
        'info': _Api.LOG_I,
        'debug': _Api.LOG_D,
        'trace': _Api.LOG_T
    }

    def __init__(self, bs_store: int, bs_chan: str, bs_log: str, max_count: int):
        log_code = _VsSessionPool._LOG_NAME_TO_CODE_MAP.get(bs_log, 'info')
        self._conf = bs_store, bs_chan, log_code
        self._sessions = [_VsSession(handle=None, chan=bs_chan) for _ in range(max_count)]

    def __del__(self) -> None:
        list(map(lambda x: x.close(), self._sessions))

    def get_session(self) -> _VsSession:
        """
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        if not self._sessions:
            eventlog.log_e('fail: session absent')
            raise VsSessionAbsentError

        session = self._sessions.pop()
        try:
            attempt = 0
            while attempt < 3:
                attempt += 1
                try:
                    if session.closed:
                        session = _VsSession.open(*self._conf)

                    eventlog.log_d(f'done: session:{id(session)}')
                    return session

                except VsSessionExpiredError:
                    session.close()

                # todo: consider to sleep a little bit
                eventlog.log_e(f'fail: open session (attempt:{attempt}/3)')

            raise VsError

        except VsError:
            self._sessions.append(session)
            raise

        except Exception:
            self._sessions.append(session)
            raise

    def put_session(self, session: _VsSession) -> None:
        assert session not in self._sessions
        self._sessions.append(session)
        eventlog.log_d(f'done: session:{id(session)}')


class _VsImageSessionPool(_VsSessionPool):
    pass


class _VsVideoSessionPool(_VsSessionPool):
    pass


class _VsVideoWork:
    """
    :type _session: _VsSession | None
    """
    _ref_to_self_map = dict()

    def __init__(self, twisted_reactor,
                 tmp_path: str,
                 mk_session_fn: Callable[[], _VsSession],
                 rm_session_fn: Callable[[_VsSession], None],
                 min_ts: int,
                 max_ts: int,
                 max_w: int,
                 max_h: int,
                 req_q: int,
                 handler: Callable[[Optional[FileBlob]], None]):

        self._ref = id(self)

        self._reactor = twisted_reactor
        self._handler = handler
        self._mk_session_fn = mk_session_fn
        self._rm_session_fn = rm_session_fn
        self._session = None

        #
        self._tmp_file = NamedTemporaryFile(dir=tmp_path, prefix='bsVid')
        eventlog.log_d('done: mk tmpfile (ref:{}, name:{})'.format(self._ref, self._tmp_file.name))

        #
        def c_call_handler_py(rec_stat: int, rec_min_ts: int, rec_max_ts: int, rec_path: str, work_ref: int) -> None:
            """
            The handler is called from video-provider's thread. Do synchronize with the twisted reactor
            """
            eventlog.log_d(f'exec: ref:{self._ref}')
            self._reactor.callFromThread(_VsVideoWork._twisted_handler,
                                         rec_stat, rec_min_ts, rec_max_ts, rec_path, work_ref)

        self._c_call_handler = _Api.LoadRecordHandler(c_call_handler_py)
        self._c_call_args = (min_ts, max_ts, max_w, max_h, req_q, self._tmp_file.name,
                             self._c_call_handler, self._ref)

    def __del__(self) -> None:
        self._tmp_file = None
        if _VsVideoWork._ref_to_self_map.pop(self._ref, None):
            eventlog.log_w(f'done: fini video-work (ref:{self._ref})')

    @staticmethod
    def _twisted_handler(rec_stat: int, rec_min_ts: int, rec_max_ts: int, rec_path: str, work_ref: int) -> None:
        eventlog.log_d(f'exec: fini video-work (ref:{work_ref})')
        self: _VsVideoWork = _VsVideoWork._ref_to_self_map.pop(work_ref, None)
        if self is None:
            eventlog.log_w(f'fail: fini video-work (ref:{work_ref})')
            return

        else:
            eventlog.log_d(f'done: fini video-work (ref:{work_ref})')

        #
        if rec_stat == _Api.RC_CONN:
            eventlog.log_e('fail: BsvLoadRecordAsync (kBS_CONN)')
            self._session.close()
            video_blob = None

        elif rec_stat != _Api.RC_DONE:
            eventlog.log_e(f'fail: BsvLoadRecordAsync (stat:{rec_stat})')
            video_blob = None

        else:
            eventlog.log_d(f'done: BsvLoadRecordAsync (stat:{rec_stat})')
            max_w = self._c_call_args[2]
            max_h = self._c_call_args[3]
            video_blob_meta = JsonBlobMeta({'size': [max_w, max_h, os.path.getsize(rec_path)]})
            video_blob = FileBlob(rec_min_ts * 1000, rec_max_ts * 1000, self._tmp_file, meta=video_blob_meta)

        self._rm_session_fn(self._session)
        self._session = None

        # noinspection PyBroadException
        try:
            self._c_call_handler = None
            self._c_call_args = None
            self._tmp_file = None
            self._handler(video_blob)
            eventlog.log_d('done: ref:{}, blob:{}'.format(
                self._ref, video_blob.file_path if video_blob else 'none'))
        except Exception:
            eventlog.log_x()

    def run_async(self) -> None:
        """
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        self._session = self._mk_session_fn()

        _VsVideoWork._ref_to_self_map[self._ref] = self
        eventlog.log_d(f'exec: init video-work (ref:{self._ref})')

        # noinspection PyBroadException
        try:
            self._session.export_video(*self._c_call_args)
            eventlog.log_d(f'done: init video-work (ref:{self._ref})')
            return
        except VsError:
            eventlog.log_e(f'fail: init video-work (ref:{self._ref}, session error')
        except Exception:
            eventlog.log_x()

        self._c_call_handler = None
        self._c_call_args = None
        self._tmp_file = None
        del _VsVideoWork._ref_to_self_map[self._ref]
        eventlog.log_d(f'done: fini video-work (ref:{self._ref})')

        self._rm_session_fn(self._session)
        self._session = None
        raise VsError


class _VsService(IVsService):
    def __init__(self, twisted_reactor, conf: VsServiceConf):
        IVsService.__init__(self)
        self._reactor = twisted_reactor

        fs.silent_rm_r(conf.tmp_path)
        fs.silent_mkdir_r(conf.tmp_path)
        self._tmp_path = conf.tmp_path

        self._image_pool = _VsImageSessionPool(conf.store, conf.chan, conf.log_level, _IMAGE_POOL_MAX)
        self._video_pool = _VsVideoSessionPool(conf.store, conf.chan, conf.log_level, _VIDEO_POOL_MAX)

    def destroy(self) -> None:
        del self._image_pool
        del self._video_pool

    def export_image_rgba(self, im_rgba: ImRgba, im_infos: List[TVsImageInfo], im_footer_h: int = 0,
                          bs_chan: Optional[str] = None) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        session = self._image_pool.get_session()
        try:
            for im_info in im_infos:
                for _ in (0, 1):
                    try:
                        image_w, image_h = session.export_image_rgba(im_info.ts // 1000, im_rgba.buffer, bs_chan)
                    except VsSessionExpiredError:
                        eventlog.log_w('fail: VsSessionExpiredError')
                        raise

                    except VsError:
                        eventlog.log_i(f'fail: export_image_rgba (image-ts:{im_info.ts}, trying the next one)')
                        break

                    if image_w > im_rgba.w or image_h > im_rgba.h:
                        im_rgba.reset(image_w, image_h, im_footer_h, im_info.ts)

                    elif image_w > 0 and image_h > 0:
                        im_rgba.reset(image_w, image_h, im_footer_h, im_info.ts)
                        return im_info

                    else:
                        eventlog.log_e(f'fail: export_image_rgba (image-w:{image_w}, image-h:{image_h})')
                        raise VsError
        finally:
            self._image_pool.put_session(session)

    def export_image_jpeg(self,
                          im_buffer: ImJpeg,
                          im_infos: List[TVsImageInfo],
                          bs_chan: Optional[str] = None) -> Optional[TVsImageInfo]:
        """
        :raise VsError:
        """
        session = self._image_pool.get_session()
        try:
            for im_info in im_infos:
                for _ in (0, 1):
                    try:
                        image_size, image_ts = session.export_image_jpeg(im_info.ts // 1000, im_buffer.buffer, bs_chan)
                    except VsSessionExpiredError:
                        eventlog.log_w('fail: VsSessionExpiredError')
                        raise

                    except VsError:
                        eventlog.log_i(f'fail: export_image_jpeg (image-ts:{im_info.ts}, trying the next one)')
                        break

                    if image_size > im_buffer.buffer_size:
                        im_buffer.reset(0, 0, image_size, image_ts * 1000)

                    elif image_size > 0:
                        image_w, image_h = ImHelpers.get_jpeg_size2(im_buffer.buffer)
                        im_buffer.reset(image_w, image_h, image_size, image_ts * 1000)
                        return im_info

                    else:
                        eventlog.log_e(f'fail: export_image_jpeg (image-size:{image_size}')
                        raise VsError
        finally:
            self._image_pool.put_session(session)

    def export_video(self, min_ts: int, max_ts: int, max_w: int, max_h: int, rec_quality: int,
                     blob_handler: Callable[[Optional[FileBlob]], None]) -> None:
        """
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        video_work = None
        try:
            video_work = _VsVideoWork(
                self._reactor,
                self._tmp_path,
                lambda: self._video_pool.get_session(),
                lambda session: self._video_pool.put_session(session),
                min_ts, max_ts, max_w, max_h, rec_quality, blob_handler)

            video_work.run_async()

        except VsSessionAbsentError:
            eventlog.log_w('fail: session absent')
            if video_work:
                video_work.__del__()
            raise

        except VsError:
            eventlog.log_e('fail: session error')
            if video_work:
                video_work.__del__()
            raise
        except Exception:
            eventlog.log_x()
            if video_work:
                video_work.__del__()
            raise VsError


def init_service(twisted_reactor, conf: Optional[VsServiceConf]) -> IVsService:
    return _StubService() if conf is None else _VsService(twisted_reactor, conf)
