import os

from time import gmtime, strftime

from edge import eventlog
from edge.services.stats import StatsService
from edge.support import fs

from tracker.types.lrra import LrRaTrackset


class SceneReactCounters:
    def __init__(self):
        self.scene_count = int(0)
        self.lr_scene_count = int(0)
        self.ra_scene_count = int(0)
        self.lr_scene_item_count = int(0)
        self.ra_scene_item_count = int(0)

    def inc_scene_count(self) -> None:
        self.scene_count += 1

    def inc_lr_scene_item_count(self, lr_scene_item_count) -> None:
        self.lr_scene_count += 1
        self.lr_scene_item_count += lr_scene_item_count

    def inc_ra_scene_item_count(self, ra_scene_item_count) -> None:
        self.ra_scene_count += 1
        self.ra_scene_item_count += ra_scene_item_count


class TrackEventCounters:
    def __init__(self):
        self.track_count = int(0)

        self.lrra_track_count = int(0)
        self.lr_track_count = int(0)
        self.ra_track_count = int(0)

    def inc_track_count(self, tr_track: LrRaTrackset) -> None:
        self.track_count += 1

        if bool(tr_track.lr_trackset) and bool(tr_track.lr_trackset.lr_jnode) and bool(tr_track.ra_track):
            self.lrra_track_count += 1

        if bool(tr_track.lr_trackset) and bool(tr_track.lr_trackset.lr_jnode):
            self.lr_track_count += 1

        if bool(tr_track.ra_track):
            self.ra_track_count += 1


class SceneEventCounters:
    def __init__(self, scene_consumer_count):
        self.scene_count_list = [int(0)] * scene_consumer_count

    def inc_scene_count(self, consumer_index: int) -> None:
        self.scene_count_list[consumer_index] += 1


class TrStatService(StatsService):
    def __init__(self, twisted_reactor, conf, scene_consumer_count) -> None:
        super().__init__(twisted_reactor, conf)
        self.scene_provider = SceneReactCounters()

        self.track_emitter = TrackEventCounters()
        self.scene_emitter = SceneEventCounters(scene_consumer_count)

    def _collect(self, unixtime_utc: float) -> None:
        stat_file_lines: list[str] = [
            # header
            'stat.time-str: %s\n' % strftime('%d.%m.%Y %H:%M:%S', gmtime(unixtime_utc)),
            f'stat.time-sec: {int(round(unixtime_utc))}\n',
            f'stat.tout-sec: {self._conf.interval_sec}\n\n',

            # react-counters
            f'react.scene-count: {self.scene_provider.scene_count}\n',
            f'react.scene.lr-scene-count: {self.scene_provider.lr_scene_count}\n',
            f'react.scene.ra-scene-count: {self.scene_provider.ra_scene_count}\n',
            f'react.scene.lr-scene-item-count: {self.scene_provider.lr_scene_item_count}\n',
            f'react.scene.ra-scene-item-count: {self.scene_provider.ra_scene_item_count}\n\n',

            # event-counters
            f'event.track-count: {self.track_emitter.track_count}\n',
            f'event.track.lr-track-count: {self.track_emitter.lr_track_count}\n',
            f'event.track.ra-track-count: {self.track_emitter.ra_track_count}\n',
            f'event.track.lrra-track-count: {self.track_emitter.lrra_track_count}\n',
        ]

        jnode = {
            'units': self._collect_units(unixtime_utc),

            'react': {
                'scene': {
                    'count': self.scene_provider.scene_count,
                    'lr-scene-count': self.scene_provider.lr_scene_count,
                    'ra-scene-count': self.scene_provider.ra_scene_count,
                    'lr-scene-item-count': self.scene_provider.lr_scene_item_count,
                    'ra-scene-item-count': self.scene_provider.ra_scene_item_count,
                }
            },

            'event': {
                'track': {
                    'count': {
                        'total-count': self.track_emitter.track_count,
                        'lr-track-count': self.track_emitter.lr_track_count,
                        'ra-track-count': self.track_emitter.ra_track_count,
                        'lrra-track-count': self.track_emitter.lrra_track_count,
                    },
                }
            }
        }

        eventlog.log_d(f'tx-scene-count-list: {self.scene_emitter.scene_count_list}')

        try:
            tmp_filepath = self._conf.filepath_tmp
            fs.write_json_file(jnode, tmp_filepath)
            os.rename(tmp_filepath, self._conf.filepath + '.json')
        except IOError as x:
            eventlog.log_w(f'fail: write <{stat_file_lines}> ({x})')

        try:
            tmp_filepath = self._conf.filepath_tmp
            fs.write_text_file(stat_file_lines, tmp_filepath)
            os.rename(tmp_filepath, self._conf.filepath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{stat_file_lines}> ({x})')
