import collections
from collections.abc import Generator
from typing import Any, Final, Callable

from edge.types.dir import Dir
from tracker.services.types.jnode.track_matcher import encode__track_matcher_stats
from tracker.services.types.track_matcher import ILrRaMatcher, LrRaMatcherConf, TrackMatcherStats

type _RaOidToMatchCountMap = dict[int, int]
type _LrOidToRaOidToMatchCountMap = dict[int, _RaOidToMatchCountMap]


class _LrRaMatcher(ILrRaMatcher):
    def __init__(self, twisted_reactor, conf: LrRaMatcherConf) -> None:
        self._reactor = twisted_reactor
        self._conf: Final[LrRaMatcherConf] = conf
        self._stats = TrackMatcherStats()

        self._lr_oid_to_delayed_call_map: dict[int, Any] = dict()
        self._lr_oid_to_ra_oid_to_match_count_map: _LrOidToRaOidToMatchCountMap = collections.defaultdict(
            collections.Counter)

        self._lr_sender: tuple[int, int] | None = None

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__track_matcher_stats(self._stats))

    def _handle_lr_oid_tout(self, lr_oid: int) -> None:
        self._lr_oid_to_ra_oid_to_match_count_map.pop(lr_oid, None)
        self._lr_oid_to_delayed_call_map.pop(lr_oid, None)

    def add_matches(self, lr_sender: tuple[int, int], lrra_matches: Generator[tuple[int, int], None, None]) -> None:
        if self._lr_sender == lr_sender:
            for lr_oid, ra_oid in lrra_matches:
                rm_delayed_call = self._lr_oid_to_delayed_call_map.get(lr_oid, None)
                if rm_delayed_call is None:
                    self._lr_oid_to_delayed_call_map[lr_oid] = self._reactor.callLater(
                        self._conf.match_tout_sec, self._handle_lr_oid_tout, lr_oid)
                else:
                    rm_delayed_call.reset(self._conf.match_tout_sec)

                self._lr_oid_to_ra_oid_to_match_count_map[lr_oid][ra_oid] += 1

        else:
            self._lr_oid_to_ra_oid_to_match_count_map.clear()
            self._lr_oid_to_delayed_call_map.clear()
            self._lr_sender = lr_sender

    def get_ra_oids_desc(self, lr_sender: tuple[int, int], lr_dir: Dir, lr_oids: list[int]) -> list[int]:
        if self._lr_sender != lr_sender:
            self._stats.invalid_sender += 1
            return []

        final_ra_oid_to_match_count_map: _RaOidToMatchCountMap = collections.Counter()
        for lr_oid in lr_oids:
            ra_oid_to_match_count_map = self._lr_oid_to_ra_oid_to_match_count_map.get(lr_oid, None)
            if ra_oid_to_match_count_map:
                final_ra_oid_to_match_count_map += ra_oid_to_match_count_map

        # sort entries by count in desc order
        ra_oids_desc = sorted(final_ra_oid_to_match_count_map,
                              key=final_ra_oid_to_match_count_map.get,
                              reverse=True)

        if len(ra_oids_desc) == 0:
            self._stats.count.unmatch += 1
        else:
            self._stats.count.match += 1

        if lr_dir == Dir.CODE_A:
            self._stats.match_hist.dir_a[5 if len(ra_oids_desc) >= 5 else len(ra_oids_desc)] += 1

        elif lr_dir == Dir.CODE_R:
            self._stats.match_hist.dir_r[5 if len(ra_oids_desc) >= 5 else len(ra_oids_desc)] += 1

        elif lr_dir == Dir.CODE_U:
            self._stats.match_hist.dir_u[5 if len(ra_oids_desc) >= 5 else len(ra_oids_desc)] += 1

        else:
            raise NotImplementedError

        return ra_oids_desc

def init_service(twisted_reactor, conf: LrRaMatcherConf) -> ILrRaMatcher:
    return _LrRaMatcher(twisted_reactor, conf)
