import sys
from typing import Self, Final

from edge.support.speed import map_speed_int
from edge.types.dir import Dir


class RaSceneItemImg:
    def __init__(self, x: float, y: float) -> None:
        self.x = x
        self.y = y


class RaSceneItem:
    def __init__(self,
                 oid: int,
                 oid_stat: int,
                 frame_ts: int,
                 frame_l_rel: float,
                 frame_t_rel: float,
                 scene_ts: int,
                 scene_f_cm: int,
                 scene_l_cm: int,
                 speed_kmh: int,
                 len_cm: int) -> None:
        """
        :param oid:
        :param oid_stat:
        :param frame_ts:
        :param frame_l_rel:
        :param frame_t_rel:
        :param scene_ts:
        :param scene_f_cm:
        :param scene_l_cm:
        :param speed_kmh:
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        :param len_cm:
        """

        self.oid: Final[int] = oid
        self.stat: Final[int] = oid_stat

        self.frame_ts: Final[int] = frame_ts
        self.frame_l_rel: Final[float] = frame_l_rel
        self.frame_t_rel: Final[float] = frame_t_rel

        self.scene_ts: Final[int] = scene_ts
        self.scene_f_cm: int = scene_f_cm  # TODO: final. See _debouce_ra_items_pos()
        self.scene_l_cm: int = scene_l_cm  # TODO: final. See _debouce_ra_items_pos()

        self.speed_kmh: Final[int] = speed_kmh
        self.len_cm: Final[int] = len_cm

    @property
    def speed_dir(self):
        if self.speed_kmh > 0:
            return Dir.CODE_R
        elif self.speed_kmh < 0:
            return Dir.CODE_A
        else:
            return Dir.CODE_U


class RaScene:
    STATUS_VALID = 1
    STATUS_INVALID = 2

    def __init__(self, items: list[RaSceneItem], status: int) -> None:
        self.items: Final[list[RaSceneItem]] = items
        self.status: Final[int] = status


class RaTrack:
    def __init__(self, oid: int, init_item: RaSceneItem) -> None:
        self.oid: Final[int] = oid

        self._max_length: int = 0
        self._min_ts: int = sys.maxsize
        self._max_ts: int = 0

        self.trace: list[RaSceneItem] = []
        self.append(init_item)

    @property
    def max_length(self) -> int:
        return self._max_length

    @property
    def min_ts(self) -> int:
        return self._min_ts

    @property
    def max_ts(self) -> int:
        return self._max_ts

    @classmethod
    def create(cls, items_asc: list[RaSceneItem]) -> Self:
        ra_track = RaTrack(items_asc[0].oid, items_asc[0])
        for index in range(1, len(items_asc), 1):
            ra_track.append(items_asc[index])

        return ra_track

    def __len__(self):
        return len(self.trace)

    def append(self, ra_item: RaSceneItem) -> None:
        if self._min_ts > ra_item.frame_ts:
            self._min_ts = ra_item.frame_ts

        if self._max_ts < ra_item.frame_ts:
            self._max_ts = ra_item.frame_ts

        if self._max_length < ra_item.len_cm:
            self._max_length = ra_item.len_cm

        self.trace.append(ra_item)

    @property
    def speed_n_kmh(self) -> int:
        """
        :returns:
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        """
        speeds: list[int] = [item.speed_kmh for item in self.trace]
        speeds.sort()

        index = len(speeds) // 2
        return int(speeds[index])

    @property
    def speed_2_kmh(self) -> int | None:
        """
        :returns:
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        """
        if len(self.trace) < 2:
            return

        min_item: RaSceneItem = self.trace[0]
        max_item: RaSceneItem = self.trace[-1]

        # Positive value of df_m means that the vehicle is moving in receeding direction relative to the radar
        df_m: float = (max_item.scene_f_cm - min_item.scene_f_cm) / 100.0
        dl_m: float = (max_item.scene_l_cm - min_item.scene_l_cm) / 100.0

        # Positive value of dr_m means that the vehicle is moving in receeding direction relative to the radar
        dr_m: float = pow(df_m ** 2 + dl_m ** 2, 0.5)
        dr_m = +dr_m if df_m >= 0 else -dr_m

        dt_sec: float = (max_item.scene_ts - min_item.scene_ts) / 1e6
        return map_speed_int(dr_m / dt_sec * 3.6)

    @property
    def dir_2(self) -> Dir:
        newest_f: int = self.trace[-1].scene_f_cm
        oldest_f: int = self.trace[0].scene_f_cm

        df = newest_f - oldest_f
        if df > 0:
            return Dir.CODE_R
        elif df < 0:
            return Dir.CODE_A
        else:
            return Dir.CODE_U
